#include "Log_Manager.h"
#include "handlers.h"

#include "ace/Time_Value.h"
#include <iostream>

#define DEFAULT_HIGH_WATERMARK 131072

Log_Callback::Log_Callback(Log_Manager *manager) : ACE_Log_Msg_Callback() {
	//ACE_TRACE(ACE_TEXT("Log_Callback::Log_Callback"));
	manager_ = manager;
}

Log_Callback::~Log_Callback() {
	ACE_TRACE(ACE_TEXT("Log_Callback::~Log_Callback"));
	manager_ = NULL;
}

void Log_Callback::log(ACE_Log_Record &log_record) {
	//first we disable the callback to avoid an infinite loop...
	if ( manager_->forwardLog() )
		ACE_LOG_MSG->clr_flags (ACE_Log_Msg::MSG_CALLBACK);	
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Callback::log"));
	//we need to send the log to the server here

	putLogInQueue(log_record);
	sendLogToServer();
	
	//job is done, we can enable the callback again...
	if ( manager_->forwardLog() )
		ACE_LOG_MSG->set_flags (ACE_Log_Msg::MSG_CALLBACK);
}

void Log_Callback::putLogInQueue(ACE_Log_Record &log_record) {
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Callback::putLogInQueue"));
	
	//we get the size of the message
	size_t size = log_record.length();

	//first, we create a message block to store the log
	ACE_Message_Block *mb = new ACE_Message_Block(size);
	if ( mb )
		//we copy the log in the message block
		ACE_OS::memcpy( mb->wr_ptr(), log_record.msg_data(), size);
	else 
		return;
	
	//we tell the message block how much data we put in it
	mb->wr_ptr(size);

	ACE_Time_Value timeout(ACE_OS::gettimeofday());
	ACE_Message_Queue<ACE_SYNCH> *tmp_queue = manager_->getQueue();
	if ( tmp_queue ) {
		//we enqueue the message, if the queue is full or blocked,
		//this call will return after one seconde.
		if ( manager_->getQueue()->enqueue_tail(mb, &timeout) == -1 ) {
			//it failed, we forget about this message and delete the message block		
			delete mb;
		}
	}

	//ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Callback::putLogInQueue ] queue size= %d\n"),tmp_queue->message_count()));
	return;
}

int Log_Callback::sendLogToServer() {
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Callback::sendLogToServer"));

	//here we need to format the message properly and notify the reactor to invoke the callback that sends the log
	ACE_Message_Queue<ACE_SYNCH> *tmp_queue = manager_->getQueue();
	if ( !tmp_queue )
		return -1;
	else {
		//we don't want to notify the reactor all the time, only every 5 message or 3 secondes
		if ( ( tmp_queue->message_count() >= 5 ) && ( !manager_->isNotificationSent() ) ) {
			ACE_Reactor::instance()->notify(manager_->getLogHandler(), ACE_Event_Handler::READ_MASK);
			manager_->setNotificationSent(1);
			//ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Callback::sendLogToServer ] Notification Sent\n")));
		}
	}
	return 0;
}

/*******************************************************************/

Log_Manager::Log_Manager(DUP_Log *logHandler) {
	ACE_TRACE(ACE_TEXT("Log_Manager::Log_Manager"));
	cb = NULL;
	file = NULL;
	log_handler = logHandler;
	notificationSent = 0;
	logForwardEnabled = false;
}

Log_Manager::~Log_Manager() {
	ACE_TRACE(ACE_TEXT("Log_Manager::~Log_Manager"));
	if ( cb ) {
		ACE_LOG_MSG->clr_flags (ACE_Log_Msg::MSG_CALLBACK);
		delete cb;
	}
	cb = NULL;
	if ( file ) {
		ACE_LOG_MSG->clr_flags (ACE_Log_Msg::OSTREAM);
		delete file;
	}
	file = NULL;
	log_handler = NULL;
	logQueue.deactivate();
	logQueue.flush();
}

int Log_Manager::init(ACE_TString flowID, ACE_TString severity, int isSyslogEnabled, int isForwardEnabled, ACE_TString logFilePath) {
	ACE_TRACE(ACE_TEXT("Log_Manager::init"));

	//init the message queue
	logQueue.high_water_mark(DEFAULT_HIGH_WATERMARK);
	logQueue.low_water_mark(0);
	logQueue.flush();

	//we need to set up our debug level here
	//we need to get the parameter from the command line to get 
	//the debug level (or log severity)
	if ( severity == "debug" ) {
		//the log level is highest, we get the errors, warnings and debug info;
		ACE_LOG_MSG->priority_mask( LM_TRACE | LM_DEBUG | LM_WARNING | LM_ERROR, ACE_Log_Msg::PROCESS);
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Severity set = debug\n")));
	}
	else if ( severity == "warning" ) {
		//the log leve is medium, we get ther errors and warnings
		ACE_LOG_MSG->priority_mask( LM_TRACE | LM_WARNING | LM_ERROR, ACE_Log_Msg::PROCESS);
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Severity set = warning\n")));
	}
	else {
		//the log level is not specified or is the default one, we get the errors only
		ACE_LOG_MSG->priority_mask( LM_TRACE | LM_ERROR, ACE_Log_Msg::PROCESS);
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Severity set = error\n")));
	}

	//used to invoke a callback for logging operations
	cb = new Log_Callback(this);

	ACE_TString name = "dup_";
	name += flowID;
	if ( isSyslogEnabled == 1 ) {
		//We are logging to the syslog of the system and stderr
		ACE_LOG_MSG->open(name.c_str(), ACE_Log_Msg::SYSLOG, name.c_str());
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Logging to Syslog On\n")));
	}
	else {
		ACE_LOG_MSG->open(name.c_str());
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Logging to Syslog Off\n")));
	}

		
	//the log is going to stderr
	ACE_LOG_MSG->set_flags (ACE_Log_Msg::STDERR);
	
	if ( isForwardEnabled == 1 ) {
		//enable the logging from the callback
		ACE_LOG_MSG->set_flags (ACE_Log_Msg::MSG_CALLBACK);
		//setting the callback object
		ACE_LOG_MSG->msg_callback(cb);
		logForwardEnabled = true;
		log_handler->enableTimer();
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Log Propagation On\n")));
	}
	else {
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Log Propagation Off\n")));
	}

	//we need to enable the log info to a file
	//we need to get the parameter from the command line to see if we need to send
	//the log in a file
	
	ACE_TString log_file = logFilePath;//ACE_TEXT_CHAR_TO_TCHAR("c:\\logdup.txt");
	if ( log_file != "NULL" ) {
		log_file += "dup_";
		
		//the flow name contains ':', which is a forbidden character for a file name
		//So we replace ':' by '-'
		char *fid = flowID.rep();
		ACE::strrepl(fid,':','-');
		log_file += fid;
		free(fid);
		
		log_file += ".log";
		
		file = new std::ofstream(log_file.c_str());
		if ( !file ) 
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Log_Manager_ACE::init ] Log file couldn't be opened")));
		else {
			//the file has successfuly been opened
			//setting the output strean

			//need to check again what 0 means
			ACE_LOG_MSG->msg_ostream(file);
			//enable the logging to the output stream
			ACE_LOG_MSG->set_flags (ACE_Log_Msg::OSTREAM);
			ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Logging to file: %s\n"), log_file.c_str()));
		}
	}

	return 0;
}

