#include "duplicator.h"

#define buffer_length 512


/*********************************************/
/***** Error management methods  *************/
/*********************************************/
const class Duplicator_Error duplicator_error[] =
{
	{ NO_ERRORS, "No error" }, // NO_ERROR is a word reserved under Windows
	{ INVALID_NAME, "Invalid name" },
	{ INVALID_BLOCK_SIZE, "Invalid block size" },
	{ INVALID_BLOCK_COUNT, "Invalid block count" },
	{ INVALID_PEER_ADDRESS, "Invalid peer address" },
	{ PROBLEM_FILENAME, "Problem to get the mmap filename"},
	{ PROBLEM_MMAP_ALLOC, "Problem when allocating mmap memory"},
	{ PROBLEM_LOCAL_SOCKET_OPEN, "Could not open local socket for listen"},
	{ PROBLEM_REMOTE_SOCKET_OPEN, "Could not open remote socket for listen"},
	{ PROBLEM_REMOTE_SOCKET_CONNECT, "Could not open remote socket for connect"},
	{ PEER_UNREACHABLE, "Remote peer unreachable" },
	{ INVALID_NB_CONSUMERS_TO_CONNECT, "Invalide number of consumers expected to connect" },

	{ UNKNOWN_ERROR, "Unknown error" } /* TO BE KEPT AS THE LAST ONE */
};


const ACE_TString &Duplicator_Error::find_error_message(int error_nb)
{ 
	ACE_TRACE(ACE_TEXT("Duplicator_Error::find_error_message"));

	int i;
	for (i = 0; duplicator_error[i].code != UNKNOWN_ERROR; i++)
	{
		if (duplicator_error[i].code == error_nb)
		{
			break;
		}
	}
	return duplicator_error[i].message; 
}



/* *********************************************/
/********* Duplicator methods ******************/
/***********************************************/

void Duplicator::ctrl_c_handler(int sig)
{
	ACE_TRACE(ACE_TEXT("Duplicator::ctrl_c_handler"));

	ACE_UNUSED_ARG(sig);
	ACE_Reactor::instance()->end_reactor_event_loop ();
}



Duplicator::Duplicator() : sigpipe_catch(SIG_IGN,SIGPIPE,0,SA_RESTART),
    sigint_catch(ctrl_c_handler,SIGINT,0,SA_RESTART),
    is_provider_(false), block_size_(0), block_count_(0),
    mem_pool(NULL), provider(0), isSynchronized_(0), 
	serverRequestDataTransfer(0), nbRemoteConsumers(0), logManager(NULL)
{
	ACE_TRACE(ACE_TEXT("Duplicator::Duplicator"));
}


Duplicator::~Duplicator()
{
	ACE_TRACE(ACE_TEXT("Duplicator::~Duplicator"));

	/* deallocate stuff */
	if (mem_pool)
	{
		delete mem_pool;
		mem_pool = NULL;
	}

	if (flow_mmap_filename != "")
	{
		ACE_OS::unlink(flow_mmap_filename.c_str());
		flow_mmap_filename.clear();
	}

#if (!(defined (ACE_WIN32)))
		if (flow_socket_filename != "")
		{
			// remove the ACE_FILE from the file system without closing the ACE_FILE handle
			ACE_OS::unlink(flow_socket_filename.c_str());
			flow_socket_filename.clear();
		}
	
		if (flow_provider_socket_filename != "")
		{
			ACE_OS::unlink(flow_provider_socket_filename.c_str());
			flow_provider_socket_filename.clear();
		}
#endif
	if ( logManager )
		delete logManager;
	logManager = NULL;
}

int Duplicator::init(ACE_TString flowID_, bool is_provider, unsigned int block_size, unsigned int block_count, ACE_TString peer, ACE_TString logSeverity, int syslogEnabled, int forwardTheLog, ACE_TString logFilePath)
{
	ACE_TRACE(ACE_TEXT("Duplicator::init"));

	ACE_TCHAR tmp_filename[DUP_MAX_FILENAME_LEN + 1];
	ACE_TString flow_mmap_filename;
#if (defined (ACE_WIN32))
	int port_consumers;
	port_provider = 1;
#endif

	socket_server = NULL;

	int result;
	int err = 0;

	/* Check arguments correctness */
	if (flowID_ == "") return INVALID_NAME;
	flowID = flowID_;
		
	if (block_size == 0 ) return INVALID_BLOCK_SIZE;
	block_size_ = block_size + sizeof(ACE_UINT32);

	if ( block_count == 0 )     return INVALID_BLOCK_COUNT;

	block_count_ = 2; // ? Why ?	

	if (peer == "") return INVALID_PEER_ADDRESS;	

	is_provider_ = is_provider;

	isSynchronized_ = 0;
	serverRequestDataTransfer = 0;
	nbRemoteConsumers = 0;
	isLocalProviderConnected_ =0;

	// just make sure the lists are empty at the beginning
	consumers_connected.reset();


	current_provider_block = 0; 
	current_consumer_block = 0; 
	memset(blocks, 0, sizeof(blocks));

	stats.blocks_read = 0;
	stats.blocks_written = 0;

	/* Initialize working variables */
	requested_size = block_size_ * block_count_;
	err = ACE::get_temp_dir(tmp_filename, DUP_MAX_FILENAME_LEN);
	if (err == -1) return PROBLEM_FILENAME;

#if (defined (ACE_WIN32))
		// Some chracters are not allowed in a filename under windows
		// We replace them by '-'

		/* Find flow filename base */
		flow_mmap_filename = tmp_filename;
		flow_mmap_filename += DUP_FLOW_FILENAME_PREFIX;
		ACE_TString flowIDModified = flowID;		

		for (unsigned int i = 0; i < flowIDModified.length(); i++) {
			if ((flowIDModified[i] == ':') ||(flowIDModified[i] == '/') ||(flowIDModified[i] == '?') ||(flowIDModified[i] == '"') ||(flowIDModified[i] == '*') ||(flowIDModified[i] == '<') ||(flowIDModified[i] == '>') ||(flowIDModified[i] == '|'))
				flowIDModified[i] = '-';
		}
		
		flow_mmap_filename += flowIDModified;
		flow_mmap_filename += DUP_FLOW_FILENAME_MMAP;
	
		if (flow_mmap_filename.length() > DUP_MAX_FILENAME_LEN)
		{
			return PROBLEM_FILENAME;
		}
#else
		/* Find flow filename base */
		flow_mmap_filename = tmp_filename;
		flow_mmap_filename += DUP_FLOW_FILENAME_PREFIX;
		flow_mmap_filename += flowID;
		flow_mmap_filename += DUP_FLOW_FILENAME_MMAP;

		if (flow_mmap_filename.length() > DUP_MAX_FILENAME_LEN)
		{
			return PROBLEM_FILENAME;
		}
#endif

#if (!(defined (ACE_WIN32)))
		flow_socket_filename += tmp_filename;
		flow_socket_filename += DUP_FLOW_FILENAME_PREFIX;
		flow_socket_filename += flowID;
		flow_socket_filename += DUP_FLOW_FILENAME_SOCKET;

		if (flow_socket_filename.length() > DUP_MAX_FILENAME_LEN)
		{
			return PROBLEM_FILENAME;
		}
#else
		//Assign ports for the consumer and the provider from flowID 
		if ( get_provider_and_consumer_ports(port_consumers) == false )
			return INVALID_NAME;
#endif

	/* open MMAP file */
	mem_pool = new ACE_Lite_MMAP_Memory_Pool(flow_mmap_filename.c_str(), 0);
	if (!mem_pool)
	{
		return PROBLEM_MMAP_ALLOC;
	}

	ACE_OS::unlink(flow_mmap_filename.c_str());
	if ( mem_pool->init_acquire(requested_size,given_size,err) == NULL )
			return PROBLEM_MMAP_ALLOC;

	if ( err == 0 ) //It means the file already exists -> the file cannot be
			           //removed but can be opened -> ownership pb or hack attempt
		return PROBLEM_MMAP_ALLOC;

	ACE_OS::memset(mem_pool->base_addr(),given_size,0);
	
	/* Open sockets */
#if (!(defined (ACE_WIN32)))
	ACE_UNIX_Addr local_addr(flow_socket_filename.c_str());
	ACE_OS::unlink(flow_socket_filename.c_str());
#else
	ACE_INET_Addr local_addr(port_consumers, "localhost", AF_INET);
#endif
	
	//ACE_MEM_Addr local_addr;
	err = local_consumer.open(local_addr);
	if (err == -1)
	{
		ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::init ] Error opening socket for local consumers. Error code : %d\n"), ACE_OS::last_error()), PROBLEM_LOCAL_SOCKET_OPEN);
	}

	DUP_Server *hdl = NULL;
	ACE_INET_Addr server_addr(SF2D_DEFAULT_DUPLI_PORT, "localhost", AF_INET);

	// initiate the connection on the server socket
	result = server.connect(hdl, server_addr);
	if (result == -1)
	{
		ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::init ] ERROR: Error connecting to the server.\n")), PROBLEM_REMOTE_SOCKET_CONNECT);
	}

	setServerSocket(&(hdl->peer()));
	

	//send it using the sendXMLToServer method
	if ( sendConnectionMessageToServer() < 0 ) {
		//ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::init::init ] ERROR: send FlowID to server not successfull\n")), -1);
		ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Duplicator::init::init ] ERROR: send FlowID to server not successfull\n")));
		return -1;
	}

	//we instantiate the Dup_Log
	DUP_Log *lg = new DUP_Log();

	//we initiate the logManager
	logManager = new Log_Manager(lg);
	logManager->init(flowID, logSeverity, syslogEnabled, forwardTheLog, logFilePath);

	/*
	//HERE WE NEED TO USE THE PROPER XML MESSAGE. WE DONT SEND THE FLOWID IN A 512B BUFFER!!!!!
	int buffer_size = buffer_length * sizeof(ACE_TCHAR);
	ACE_TCHAR buffer[buffer_length];

	// put the ID of the flow into the buffer
	result = ACE_OS::snprintf(buffer, buffer_length, ACE_TEXT("%s"), flowID.c_str());
	if (result >= buffer_length)
	{
		ACE_ERROR_RETURN((LM_DEBUG, ACE_TEXT("[ Duplicator ] ERROR: snprintf truncated the ID of the flow (length > %d)\n"), buffer_length), -1);
	}
	else if (result < 0)
	{
		ACE_ERROR_RETURN((LM_DEBUG, ACE_TEXT("[ Duplicator ] ERROR: snprintf output error\n")), -1);
	}
	
	// send the buffer to the server
	result = hdl->peer().send(buffer, buffer_size);
	if (result < buffer_size)
	{
		ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator ] ERROR: send not successfull\n")), -1);
	}
	*/

	
	/* if provider, peer contain the port number to accept() to
	   if consumer, peer contant machine:port to connect() to
	 */


    if (is_provider_)
	{
		return open_provider_part(peer);
	}
	else
	{
		if (peer != "noprovider" )
			return connect_to_distant_duplicator(peer);
		else 
			return 0;
	}
}


int Duplicator::connect_to_distant_duplicator(ACE_TString peer)
{
	ACE_TRACE(ACE_TEXT("Duplicator::connect_to_distant_duplicator"));

	peer_ = peer;
	DUP_Remote_Provider *hdl = NULL;
	
	ACE_INET_Addr remote_addr(peer_.c_str());
	int err = remote_provider.connect(hdl, remote_addr);
	while (err == -1)
	{
		ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Duplicator::connect_to_distant_duplicator ] ERROR: Cannot connect to remote socket '%s', trying again... Error code:%d\n"), peer_.c_str(), ACE_OS::last_error()));
		ACE_Time_Value tmp(0, 1000);
		ACE_OS::sleep(tmp);
		hdl = NULL;
		err = remote_provider.connect(hdl, remote_addr);
	}
	return 0;
}

int Duplicator::open_provider_part(ACE_TString peer)
{
	ACE_TRACE(ACE_TEXT("Duplicator::open_provider_part"));

	is_provider_ = true;
#if (!(defined (ACE_WIN32)))
	ACE_TCHAR tmp_filename[DUP_MAX_FILENAME_LEN + 1];
	ACE::get_temp_dir(tmp_filename, DUP_MAX_FILENAME_LEN);
	flow_provider_socket_filename = "";
	flow_provider_socket_filename += tmp_filename;
	flow_provider_socket_filename += DUP_FLOW_FILENAME_PREFIX;
	flow_provider_socket_filename += flowID;
	flow_provider_socket_filename += DUP_FLOW_FILENAME_PSOCKET;
	if (flow_provider_socket_filename.length() > DUP_MAX_FILENAME_LEN)
	{
		return PROBLEM_FILENAME;
	}
    // Not portable on non UNIX systems (eg Windows) !!!
   	ACE_UNIX_Addr local_addr_p(flow_provider_socket_filename.c_str());
	ACE_OS::unlink(flow_provider_socket_filename.c_str());
#else
	ACE_INET_Addr local_addr_p((u_short) port_provider, "localhost", AF_INET);
#endif
	
	int err = local_provider.open(local_addr_p);
	int count = 0;
	while (err == -1)
	{
		local_provider.close();
		ACE_DEBUG((ERROR_PREFIX, ACE_TEXT("[ Duplicator::open_provider_part] Error code : %d. "), ACE_OS::last_error()));
		count++;
		if (count <= 3) {
#if (!(defined (ACE_WIN32)))
			ACE_DEBUG((ERROR_PREFIX, ACE_TEXT("[ Duplicator::open_provider_part] Attempt %d: Cannot open local socket %s. Retrying in one second\n"), count, flow_provider_socket_filename.c_str()));
#else
			ACE_DEBUG((ERROR_PREFIX, ACE_TEXT("[ Duplicator::open_provider_part] Attempt %d: Cannot open socket for the local provider on port %d. Retrying in one second\n"), count, port_provider));
#endif
			local_provider.close();
			ACE_Time_Value tmp(0, 1000);
			ACE_OS::sleep(tmp);
			err = local_provider.open(local_addr_p);
		}
		else {
#if (!(defined (ACE_WIN32)))
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[  Duplicator::open_provider_part ] Too many attempts, giving up! ERROR: Cannot open local socket %s.\n"), flow_provider_socket_filename.c_str()));
			return PROBLEM_LOCAL_SOCKET_OPEN;
			//ACE_ERROR_RETURN(ERROR_PREFIX, ACE_TEXT("Too many attempts, giving up! ERROR: Cannot open local socket %s.\n"), flow_provider_socket_filename.c_str()), PROBLEM_LOCAL_SOCKET_OPEN);
#else
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[  Duplicator::open_provider_part ] Too many attempts, giving up! ERROR: Cannot open socket for the local provider on port %d.\n"), port_provider));
			return PROBLEM_LOCAL_SOCKET_OPEN;
			//ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("Too many attempts, giving up! ERROR: Cannot open socket for the local provider on port %d.\n"), port_provider), PROBLEM_LOCAL_SOCKET_OPEN);
#endif
		}
	}

	peer_ = peer;
	int port = ACE_OS::atoi(peer_.c_str());
	ACE_INET_Addr remote_addr(peer_.c_str());
	err = remote_consumer.open(remote_addr);
	while (err == -1)
	{
		ACE_DEBUG((ERROR_PREFIX, ACE_TEXT("[ Duplicator::open_provider_part ] ERROR: Cannot open remote socket on port %d. Trying port %d\n"), port, port + 1));
		remote_consumer.close();
		port++;
		ACE_TCHAR dpeer[6] = {0};
		ACE_OS::snprintf(dpeer, sizeof(dpeer), ACE_TEXT("%d"), port);
		peer_ = dpeer;
		remote_addr.set(peer_.c_str());
		err = remote_consumer.open(remote_addr);
	}

	//We inform the server of the port number of the socket we just opened
	//We need to allocate a string of a proper space to contain the message
	char tmp_str[11];
	size_t size_needed = sizeof(SOCKET_FOR_REMOTE_CONSUMERS_OPENED_MESSAGE) + flowID.length() + ACE_OS::sprintf(tmp_str, "%d", port) + 1; 
	ACE_TCHAR *socketOpenedMessage = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
	ACE_OS::memset(socketOpenedMessage, 0, size_needed);
	//char socketOpenedMessage[MAX_MESSAGE_SIZE]={0};
	ACE_OS::sprintf(socketOpenedMessage, SOCKET_FOR_REMOTE_CONSUMERS_OPENED_MESSAGE, flowID.c_str(), port);

	//then we send it
	int result = sendXMLMessageToServer(socketOpenedMessage);
	
	ACE_OS::free(socketOpenedMessage);
	
	return result;
}

int Duplicator::close_provider_part()
{
	ACE_TRACE(ACE_TEXT("Duplicator::close_provider_part"));

	is_provider_ = false;

	// close the socket with each consumer while keeping the iterator in a stable mode
	int remoteConsumerLeft = true;
	while (remoteConsumerLeft) {
		DUP_Remote_Consumer *c = NULL;
		remoteConsumerLeft = false;
		for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++)
		{
			ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[Duplicator::close_provider_part]iterator\n")));
			if ((*i)->is_remote_consumer == true) {
				remoteConsumerLeft = true;
				c = (DUP_Remote_Consumer *) (*i);
				break;
			}
		} // for
		if (remoteConsumerLeft) {
			ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[Duplicator::close_provider_part]after close\n")));
			consumers_connected.remove(c);
			c->close();
		}
	}

	current_provider_block = 0; 
	current_consumer_block = 0; 
	memset(blocks, 0, sizeof(blocks));

	setLocalProviderConnected(0);

	int err = local_provider.close();
	
	if (err == -1)
	{
#if (!(defined (ACE_WIN32)))
		ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::init ] ERROR: Cannot close local socket %s\n"), flow_provider_socket_filename.c_str()), PROBLEM_LOCAL_SOCKET_CLOSE);
#else
		ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::init ] ERROR: Cannot close the socket reserved for the local provider on port %d\n"), port_provider), PROBLEM_LOCAL_SOCKET_CLOSE);
#endif
	}

	err = remote_consumer.close();
	if (err == -1)
	{
		ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::init ] ERROR: Cannot close the socket reserved for remote connections\n")), PROBLEM_REMOTE_SOCKET_CLOSE);
	}
	return 0;
}

int Duplicator::run_event_loop(void)
{
	ACE_TRACE(ACE_TEXT("Duplicator::run_event_loop"));
	// run the event loop of the reactor
	return ACE_Reactor::instance()->run_reactor_event_loop();
}

int Duplicator::stop_event_loop(void)
{
	ACE_TRACE(ACE_TEXT("Duplicator::stop_event_loop"));
	// end the event loop of the reactor
	return ACE_Reactor::instance()->end_reactor_event_loop();
}

char *Duplicator::mem_base(unsigned char block_number)
{ 
	ACE_TRACE(ACE_TEXT("Duplicator::mem_base"));

	if (mem_pool == NULL)
	{
		return NULL;
	}

	if (block_number > block_count_ )
	{
		return NULL;
	}

	return (char *) mem_pool->base_addr() + (block_size_ * block_number);
}



char *Duplicator::data_base(unsigned char block_number)
{ 
	ACE_TRACE(ACE_TEXT("Duplicator::data_base"));

	char *data = mem_base(block_number);
	if (data == NULL)
	{
		return NULL;
	}
	return data + sizeof(ACE_UINT32);
}

int Duplicator::consumer_block_release(long consumer_block)
{
	ACE_TRACE(ACE_TEXT("Duplicator::consumer_block_release"));

	ACE_UNUSED_ARG(consumer_block);

	int result;

	// check that no consumers are reading this block
	// and that there is a provider
	bool consumers_reading = false;
	for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++)
	{
		bool read = (*i)->has_read_block(0);
		
		if (read != true)
		{
			consumers_reading = true;
			break;
		}
	} // for

	if ((consumers_reading == false) && (get_provider() != NULL))
	{
		// notify the provider that it can write this block
		result = get_provider()->buffer_action(0);
		// check that the notification succeeded
		if (result != 0)
		{
			// notification failure
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Duplicator::consumer_block_release ] ERROR: Notification to the provider failed\n")));
			
		}
	}
	return 0;
}

int Duplicator::block_release(DUP_Consumer *c)
{
	ACE_TRACE(ACE_TEXT("Duplicator::block_release(DUP_Consumer)"));

	int result;

	if (!c)
	{
		return -1;
	}
	
	// check that 'c' is actually connected
	if (consumers_connected.find(c) == -1)
	{
		// if not, error
		ACE_DEBUG((ERROR_PREFIX, ACE_TEXT("[ Duplicator::block_release(DUP_Consumer) ] ERROR: Consumer %x is not in the list of connected consumers\n"), c));
		return -1;
		//ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::block_release(DUP_Consumer) ] ERROR: Consumer %x is not in the list of connected consumers\n"), c), -1);
	}
    // update the number of blocks read for the stats
    stats.blocks_read++;
	    
    // release the block
	if (c->blocks_read[0] == false) {
		c->blocks_read[0] = true;
		result = consumer_block_release(0);
		return result;
	}	    

	return 0;
}

int Duplicator::block_release(DUP_Provider *p, unsigned int size, int block_number)
{
	ACE_TRACE(ACE_TEXT("Duplicator::block_release(DUP_Provider)"));
	ACE_UNUSED_ARG(block_number);

	int result;
	// check that the pointer to the provider is not null
	if (!p)
	{
		return -1;
	}

	// check that 'p' is actually the one writing this block
	if (p != get_provider())
	{
		ACE_DEBUG((ERROR_PREFIX, ACE_TEXT("[ Duplicator::block_release(DUP_Provider) ] ERROR: The provider is not the right one\n")));
		return -1;
		//ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::block_release(DUP_Provider) ] ERROR: The provider is not the right one\n")), -1);
	}
	// update the number of blocks written for the stats
	stats.blocks_written++;

	// check if there are consumers switched on
	if (consumers_connected.size() == 0)
	{
		// no consumers are connected
		// the block is ready to be written again by the provider
		// send the notification
		return consumer_block_release(0);
	}
	else
	{
		// there are some consumers switched on
		// notify all the switched on consumers that the block is ready to be read
		bool has_consumers_switched_on = false;
		for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++)
		{
			//there is a new block and nobody read it yet, so set that in all the consumers currently consuming
			if ((*i)->is_paused == false) {
				(*i)->blocks_read[0] = false;
				has_consumers_switched_on = true;
			}
		}
		if (has_consumers_switched_on == false)
			return consumer_block_release(0);

		blocks[0].size = size;

		// send the number and size of the block to all the consumers
		// to notify them that they can read the block
		for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++)
		{
			// send the notification
			if ((*i)->is_paused == false) {
				result = (*i)->buffer_action(0, size);
			
				// check that the notification succeeded
				if (result != 0)
				{
					// notification failure
					ACE_DEBUG((ERROR_PREFIX, ACE_TEXT("[ Duplicator::block_release(DUP_Provider) ] ERROR: Notification to consumer %x failed\n"), (*i)));
				}
			}
		} // for
	} // if
	
	return 0;
}

int Duplicator::connect(DUP_Consumer *c)
{
	ACE_TRACE(ACE_TEXT("Duplicator::connect(DUP_Consumer)"));

	// check that this client is not already connected
	if (consumers_connected.find(c) == 0)
	{
		// this consumer is already connected
		ACE_DEBUG((ERROR_PREFIX, ACE_TEXT("[ Duplicator::connect ] ERROR: Consumer %x is already connected\n"), c));
		return -1;
		//ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::connect ] ERROR: Consumer %x is already connected\n"), c), -1);
	}

	// add the consumer to the list of connected consumers
	consumers_connected.insert(c);

	ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Duplicator::connect ] Consumer (%x) connected\n"), c));
	ACE_DEBUG((DEBUG_PREFIX, ACE_TEXT("[ Duplicator::connect ] Number of consumers connected : %d\n"), consumers_connected.size()));

	displayAllConsumersConnected();
	displayAllConsumersSwitchedOn();

	return 0;
}

int Duplicator::displayAllConsumersConnected() {

  ACE_TRACE(ACE_TEXT("Duplicator::displayAllConsumersConnected"));

  ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::displayAllConsumersConnected ] (%d) \n"),consumers_connected.size()));

  for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++)
    if ((*i)->is_remote_consumer == false)
      ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("(%s)\n"), (*i)->client_id));
  ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("\n")));

  return 0;
}

int Duplicator::displayAllConsumersSwitchedOn() {
	ACE_TRACE(ACE_TEXT("Duplicator::displayAllConsumersSwitchedOn"));
	
	int consumers_switched_on_count = 0;
	for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++)
	{
			if ((*i)->is_paused == false) 
				consumers_switched_on_count++;
	}

  ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::displayAllConsumersSwitched_on ] (%d) \n"), consumers_switched_on_count ));

  for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++) {
    if ((*i)->is_remote_consumer == false && (*i)->is_paused == false)
      ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("(%s)"), (*i)->client_id));
  }

  ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("\n")));
  return 0;
}

int Duplicator::pause(ACE_TString clientID) {
  ACE_TRACE(ACE_TEXT("Duplicator::pause"));

  ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Duplicator::pause for %s] \n"),clientID.c_str() ));

  ACE_TString current_id;
  int consumers_switched_on_count = 0;
  for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++)
  {
			if ((*i)->is_paused == false) 
				consumers_switched_on_count++;
  }
  ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::pause ] Number of consumers switched_on : %d\n"), consumers_switched_on_count));

  for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++) {
    if ((*i)->is_remote_consumer == false) {
      current_id = (*i)->client_id;
      if (clientID == current_id ) {
		ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::pause ... OK] \n"), (*i)->client_id));
		// release the block the consumer was reading
		DUP_Local_Consumer *consumer_to_pause = (DUP_Local_Consumer *) (*i);
		consumer_to_pause->is_paused = true;
		break;
      }
    }
  }
  return 0;
}

int Duplicator::pause_dupProvider_dupConsumer(ACE_TString host_to_pause) {
  ACE_TRACE(ACE_TEXT("Duplicator::pause_dupProvider_dupConsumer"));

  ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::pause_dupProvider_dupConsumer ] (%s) \n"), host_to_pause.c_str()  ));

  for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++) {
    if ((*i)->is_remote_consumer == true) {
      if (((*i)->host) == host_to_pause) {
		ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::pause_dupProvider_dupConsumer ... OK]\n")  ));
		(*i)->is_paused = true;
		break;
      }
    }
  }
  return 0;
}

int Duplicator::wake_up_host(ACE_TString host_to_wake_up){
	ACE_TRACE(ACE_TEXT("Duplicator::wake_up_host"));

	ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::wake_up_host ... ] (%s) \n"), host_to_wake_up.c_str()  ));

	for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++) {
		if ((*i)->is_remote_consumer == true) {
			if (((*i)->host) == host_to_wake_up) {
				if ((*i)->is_paused == false) {
					ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::wake_up_host ]  Consumer %x is already switched_on \n"),*i));
				}
				else {
					ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::wake_up_host ... OK] \n") ));
					(*i)->is_paused = false;
					break;
				}
			}
		}
	}
	return 0;
}

int Duplicator::restart(ACE_TString clientID) {
  ACE_TRACE(ACE_TEXT("Duplicator::restart"));

  ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Duplicator::restart ... ] (%s) \n"), clientID.c_str()  ));

  ACE_TString current_clientID;
  for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++) {
    if ((*i)->is_remote_consumer == false) {
      current_clientID = (*i)->client_id;
      if (current_clientID == clientID ) {
	ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::restart ... OK] \n") ));
	(*i)->is_paused = false;
	break;
      }
    }
  }
  return 0;
}

int Duplicator::disconnect(DUP_Consumer *c)
{
	ACE_TRACE(ACE_TEXT("Duplicator::disconnect(DUP_Consumer)"));

	if (!c)
	{
		return -1;
	}

	// check that the consumer is connected
	if (consumers_connected.find(c) == -1)
	{
		// this consumer is not connected
		ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::disconnect ] ERROR: Consumer %x is not connected\n"), c), -1);
	}

	// release the block the consumer was reading
	if (c->blocks_read[0] == false) {
		c->blocks_read[0] = true;
		consumer_block_release(0);
	}

	// remove the client of the list of connected consumers
	consumers_connected.remove(c);
	
	ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Duplicator::disconnect ] The consumer %x disconnected\n"), c));
	ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Duplicator::disconnect ] Number of consumers connected : %d\n"), consumers_connected.size()));

	return 0;
}

int Duplicator::connect(DUP_Provider *p)
{
	ACE_TRACE(ACE_TEXT("Duplicator::connect(DUP_Provider)"));

	int result;

	// check if there is already one provider
	if (provider != 0)
	{
		// error, there is already one provider
		ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Duplicator::connect ] ERROR: There is already one provider\n")));
		return -1;
		//ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::connect ] ERROR: There is already one provider\n")), -1);
	}

	ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Duplicator::connect ] The provider %x connected\n"), p));

	provider = p;

	// check if consumers are reading this block
	bool consumers_reading = false;
	//If the flow is synchronized, this loop is irrelevant because consumers have no block to read anyway
	//CAREFULL: This test may have some side effect in this case:
	//1) A consumer connects
	//2) A provider with a flow not synchronized connects
	//3) The provider quits
	//4) A provider with a flow synchronized connects.
	
	//This loop is important when the provider connects, it ensures that there is no consumer reading the SHM
	//written by a former provider. If consumers are reading, the provider doesn't do anything and wait that the consumer finishes reading.
	//But if a provider connects, and no consumer are reading, we are then ready to write a bock, as a result we set bock read to false 
	//for the future block (in "if (consumers_reading == false)" ).
	//If the provider disconnects before writting the block, the consumer have block_read set to true but no block to read. Then the provider comes back and don't do
	//anything because some consumer are reading a block that doesn't exist... => interlock
	if ( isSynchronized() == 0 ) { //there is no synchronized flow
		for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++)
		{
			if ((*i)->is_paused == false) {
				bool read = (*i)->has_read_block(0);

				if (read != true)
				{
					consumers_reading = true;
					break;
				}
			}
		} // for
	}//if
	
	if (consumers_reading == false)
	{
		//I remember Stephane put that here. So far, it doesn't seem necessary to have that here and will
		//in some cases creates interlock problems. However, putting this code here in the first place has an explanation.
		//So far I cannot find any case where that could be usefull, so it is commented to avoid the bug.
		//For the full explanation, read the previous comment in this method.
		/*for (CONSUMER_ITERATOR i = consumers_connected.begin(); i != consumers_connected.end(); i++)
		{
			//there is a new block and nobody read it yet, so set that in all the consumers currently consuming
			if ((*i)->is_paused == false) {
				(*i)->blocks_read[0] = false;
				cout << "------> Duplicator::connect(DUP_Provider) set BLOCK_READ = FALSE" << endl;
			}
		}*/

		// no consumers are reading this block
		// notify the provider that the block can be written
		result = p->buffer_action(0);

		// check that the notification succeeded
		if (result != 0)
		{
			// notification failure
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Duplicator::connect ] ERROR: There is already one provider\n")));
			return -1;
			//ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::connect ] ERROR: There is already one provider\n")), -1);
		}
	}
	
	return 0;
}

int Duplicator::disconnect(DUP_Provider *p)
{
	ACE_TRACE(ACE_TEXT("Duplicator::disconnect(DUP_Provider)"));
	
	if (provider == p)
	{
		provider = NULL;
	}

	ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Duplicator::disconnect ] The provider %x disconnected\n"), p));
	return 0;
}



int Duplicator::connect(DUP_Server *s)
{
	ACE_TRACE(ACE_TEXT("Duplicator::connect(DUP_Server)"));
	ACE_UNUSED_ARG(s);
	ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Duplicator::connect ] The server connected\n")));
	return 0;
}

int Duplicator::disconnect(DUP_Server *s)
{
    ACE_TRACE(ACE_TEXT("Duplicator::disconnect(DUP_Server)"));

	ACE_UNUSED_ARG(s);

	ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Duplicator::disconnect ] The server disconnected, killing myself\n")));

	return 0;
}

int Duplicator::sendConnectionMessageToServer() {
	ACE_TRACE(ACE_TEXT("Duplicator::sendConnectionMessageToServer"));

    //first we create the message
	
	//We need to allocate a string of a proper space to contain the message
	size_t size_needed = sizeof(CONNECT_MESSAGE) + flowID.length() + 1;
	ACE_TCHAR *theMessage = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
	ACE_OS::memset(theMessage, 0, size_needed);
	
	//char theMessage[MAX_MESSAGE_SIZE]={0};
	ACE_OS::sprintf(theMessage,CONNECT_MESSAGE, flowID.c_str());

	//then we send it
	int result = sendXMLMessageToServer(theMessage);
	
	ACE_OS::free(theMessage);
	
	return result;
}

int Duplicator::sendXMLMessageToServer(ACE_TString xml)
{	
	//ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Duplicator::sendXMLMessageToServer ]\n")));

	ACE_TRACE(ACE_TEXT("Duplicator::sendXMLMessageToServer"));
	/* As always in the Smartflow2 system, we first then the size then the data */
	unsigned int result;
	char size[512] = {0};
	ACE_OS::snprintf(size, sizeof(size), "%d", xml.length());
	
	ACE_SOCK_STREAM *socket = getServerSocket();
	if ( socket == NULL ) {
		ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Duplicator::sendXMLMessageToServer ] Socket Server NULL \n")));
		return -1;
	}
	else {
		//first we send the message size
		result = socket->send_n(size, sizeof(size));
		if ( result < sizeof(size) ) {
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Duplicator::sendXMLMessageToServer ] ERROR: send size of message not successfull\n")));
			return -1;
		   //ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::sendXMLMessageToServer ] ERROR: send size of message not successfull\n")), -1);
		}

		//then the message itself
		const char *xml_message = ACE_TEXT_ALWAYS_CHAR(xml.c_str());
		result = socket->send_n(xml_message, ACE_OS::strlen(xml_message));
		if ( result < ACE_OS::strlen(xml_message) ) {
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Duplicator::sendXMLMessageToServer ] ERROR: send message not successfull")));
			return -1;
		   //ACE_ERROR_RETURN((ERROR_PREFIX, ACE_TEXT("[ Duplicator::sendXMLMessageToServer ] ERROR: send message not successfull ")), -1);
		}

		return 0;
	}
}

void Duplicator::setServerSocket(ACE_SOCK_Stream* socket_) {
	ACE_TRACE(ACE_TEXT("Duplicator::setServerSocket"));

	socket_server = socket_;
}

ACE_SOCK_Stream* Duplicator::getServerSocket() {
	ACE_TRACE(ACE_TEXT("Duplicator::getServerSocket"));

	return socket_server;
}

int Duplicator::setSynchronized(int val) {
	ACE_TRACE(ACE_TEXT("Duplicator::setSynchronized"));

	if ( (val != 0) && (val != 1) )
		return -1;
	else
		isSynchronized_ = val;
	return 0;
}

int Duplicator::setLocalProviderConnected(int val) {
	ACE_TRACE(ACE_TEXT("Duplicator::setLocalProviderConnected"));

	if ( (val != 0) && (val != 1) )
		return -1;
	else
		isLocalProviderConnected_ = val;
	return 0;

}

int Duplicator::setRequestDataTransfer(int val) {
	ACE_TRACE(ACE_TEXT("Duplicator::setRequestDataTransfer"));

	if ( (val != 0) && (val != 1) )
		return -1;
	else {
		serverRequestDataTransfer = val;
#if (defined (ACE_WIN32))
		//this code is only required on Windows. We need to invoke the DUP_Local_Provider::handle_input()
		//It is not required on Linux or Mac because the reactor invoke the handle_input() method 
		//until the data on the socket has been read
		DUP_Local_Provider *pro = (DUP_Local_Provider*) provider;
		if ( pro != NULL )
			pro->handle_input();
#endif
	}
	return 0;
}

int Duplicator::sendToServerTheNumberOfRemoteDuplicatorsConnected() {
	ACE_TRACE(ACE_TEXT("Duplicator::sendToServerTheNumberOfRemoteDuplicatorsConnected"));
	
	//first we create the message
	//We need to allocate a string of a proper space to contain the message
	char tmp_str[11];
	size_t size_needed = sizeof(DUPLICATOR_CONNECTED_MESSAGE) + flowID.length() + ACE_OS::sprintf(tmp_str, "%d", nbRemoteConsumers) + 1; 
	ACE_TCHAR *theMessage = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
	ACE_OS::memset(theMessage, 0, size_needed);

	//char theMessage[MAX_MESSAGE_SIZE]={0};
	ACE_OS::sprintf(theMessage, DUPLICATOR_CONNECTED_MESSAGE, flowID.c_str(), nbRemoteConsumers);

	//then we send it
	int result = sendXMLMessageToServer(theMessage);
	
	ACE_OS::free(theMessage);
	
	return result;
}

int Duplicator::sendLogToServer(const std::string &logMessage) {
	ACE_TRACE(ACE_TEXT("Duplicator::sendLogToServer"));
	
	int size_needed = sizeof(LOG_MESSAGE) + flowID.length() + logMessage.length() + 1;
	ACE_TCHAR *theMessage = (ACE_TCHAR *) ACE_OS::malloc(size_needed); 
	ACE_OS::memset(theMessage, 0, size_needed);
	
	//char theMessage[MAX_MESSAGE_SIZE]={0};
	ACE_OS::sprintf(theMessage, LOG_MESSAGE, flowID.c_str(), logMessage.c_str());

	//then we send it
	int result = sendXMLMessageToServer(theMessage);
	
	ACE_OS::free(theMessage);
	
	return result;
}

void Duplicator::dump0()
{
	/*
	std::cout << "*** DUPLICATOR *********" << std::endl;
	std::cout << "  Name          : " << flowID <<std::endl;
	std::cout << "  Provider ?    : " << is_provider_ <<std::endl;
	std::cout << "  Bloc size     : " << block_size_ <<std::endl;
	std::cout << "  Bloc count    : " << block_count_ <<std::endl;
	std::cout << "  Peer value    : " << peer_ <<std::endl;
	std::cout << std::endl;
	std::cout << "  Mem_base      : " << mem_base(0) <<std::endl;
	std::cout << "  MemPool base  : " << mem_pool->base_addr() <<std::endl;
	std::cout << std::endl;
	*/
}

#ifndef DEBUG
  #define dump1(N)    
#else
void Duplicator::dump1(char *head)
{
	ACE_UNUSED_ARG(head);
	/*
	std::cout << "*** " << head << " *********" << std::endl;
	std::cout << "  consumers size        : " << consumers_connected.size() <<std::endl;
	std::cout << "  provider addr         : " << provider <<std::endl;
	std::cout << "  current_provider_block : " << (short) current_provider_block <<std::endl;
	std::cout << "  current_consumer_block : " << (short) current_consumer_block <<std::endl;
	std::cout << "  blocks[0].size         : " << blocks[0].size <<std::endl;
	std::cout << "  blocks[1].size         : " << blocks[1].size <<std::endl;
	std::cout << "  blocks_seen            : " << block_seen <<std::endl;
	std::cout << std::endl;
	std::cout << "************************" << std::endl;
	std::cout << std::endl;
	*/
}
#endif

#if (defined (ACE_WIN32))
bool Duplicator::get_provider_and_consumer_ports(int &port_consumers_) {
	ACE_TRACE(ACE_TEXT("Duplicator::get_provider_and_consumer_ports"));

	//check if the flow name is correct
	if (flowID == "") return false;
	
	for (unsigned int i = 0; i < flowID.length(); i++) {
		int valeur;
		switch (flowID[i]) {
			case 'a' :
			case 'A' :
				valeur = 1;
				break;
			case 'b' :
			case 'B' :
				valeur = 2;
				break;
			case 'c' :
			case 'C' :
				valeur = 3;
			break;
			case 'd' :
			case 'D' :
				valeur = 4;
				break;
			case 'e' :
			case 'E' :
				valeur = 5;
				break;
			case 'f' :
			case 'F' :
				valeur = 6;
				break;
			case 'g' :
			case 'G' :
				valeur = 7;
				break;
			case 'h' :
			case 'H' :
				valeur = 8;
				break;
			case 'i' :
			case 'I' :
				valeur = 9;
				break;
			case 'j' :
			case 'J' :
				valeur = 10;
				break;
			case 'k' :
			case 'K' :
				valeur = 11;
				break;
			case 'l' :
			case 'L' :
				valeur = 12;
				break;
			case 'm' :
			case 'M' :
				valeur = 13;
				break;
			case 'n' :
			case 'N' :
				valeur = 14;
				break;
			case 'o' :
			case 'O' :
				valeur = 15;
				break;
			case 'p' :
			case 'P' :
				valeur = 16;
				break;
			case 'q' :
			case 'Q' :
				valeur = 17;
				break;
			case 'r' :
			case 'R' :
				valeur = 18;
				break;
			case 's' :
			case 'S' :
				valeur = 19;
				break;
			case 't' :
			case 'T' :
				valeur = 20;
				break;
			case 'u' :
			case 'U' :
				valeur = 21;
				break;
			case 'v' :
			case 'V' :
				valeur = 22;
				break;
			case 'w' :
			case 'W' :
				valeur = 23;
				break;
			case 'x' :
			case 'X' :
				valeur = 24;
				break;
			case 'y' :
			case 'Y' :
				valeur = 25;
				break;
			case 'z' :
			case 'Z' :
				valeur = 26;
				break;
			default :
				valeur = atoi(&flowID[i]);
				if (valeur <=0 )
					valeur = 1;
				break;
		}
		port_provider *= valeur;
		port_provider += i;
	}
	port_provider %= 62357;
	port_provider += 1217;
	port_consumers_ = (int) port_provider + 1;
	return true;
}
#endif

