#ifndef DUPLICATOR_H
#define DUPLICATOR_H

#include "ace/ACE.h"
#include "ace/FILE.h"
#include "ace/FILE_Addr.h"
#include "ace/Log_Msg.h"
#include "ace/Signal.h"
#include "ace/Singleton.h"
#include "ace/Acceptor.h"
#include "ace/Connector.h"
#include "ace/SOCK_Acceptor.h"
#include "ace/SOCK_Connector.h"

#include <iostream>

#if (!(defined (ACE_WIN32)))
	#include "ace/LSOCK_Acceptor.h"
	#include "ace/LSOCK_Connector.h"
#endif

//#include "ace/MEM_Acceptor.h"
//#include "ace/MEM_Connector.h"
#include "ace/MMAP_Memory_Pool.h"
#include "ace/Unbounded_Set.h"

#include "sf2d_common_ACE.h"
#include "duplicator_common_ACE.h"
#include "handlers.h"
#include "Log_Manager.h"

/*********************************************/
/***** XML Message to send to the server *****/
/*********************************************/

#define CONNECT_MESSAGE ACE_TEXT("<?xml version='1.0'?>\n<smartspace>\n <type>3</type>\n <target/> <flowID>%s</flowID>\n</smartspace>")
#define CONSUMER_CONNECTED_MESSAGE ACE_TEXT("<?xml version='1.0'?>\n<smartspace>\n <type>5</type>\n <target>31</target>\n <flowID>%s</flowID>\n <clientID>%s</clientID>\n </smartspace>")
#define DUPLICATOR_CONNECTED_MESSAGE ACE_TEXT("<?xml version='1.0'?>\n<smartspace>\n <type>5</type>\n <target>32</target>\n <flowID>%s</flowID>\n <nbRemoteConsumers>%d</nbRemoteConsumers>\n</smartspace>")
#define LOG_MESSAGE ACE_TEXT("<?xml version='1.0'?>\n<smartspace>\n <type>5</type>\n <target>36</target>\n <flowID>%s</flowID>\n <servername>%s</servername>\n <messages>%s</messages>\n </smartspace>")
#define SOCKET_FOR_REMOTE_CONSUMERS_OPENED_MESSAGE ACE_TEXT("<?xml version='1.0'?>\n<smartspace>\n <type>5</type>\n <target>34</target> <flowID>%s</flowID> <portnumber>%d</portnumber>\n</smartspace>")
#define MAX_MESSAGE_SIZE 131072

/* **** Error management definitions *************/
/************************************************/
/************************************************/
enum duplicator_error_enum {
	UNKNOWN_ERROR = -1,
	NO_ERRORS = 0, // NO_ERROR is reserved on Windows
	INVALID_NAME = 1,
	INVALID_BLOCK_SIZE = 2,
	INVALID_BLOCK_COUNT = 3,
	INVALID_PEER_ADDRESS = 4,
	PROBLEM_FILENAME = 5,
	PROBLEM_MMAP_ALLOC = 6,
	PROBLEM_LOCAL_SOCKET_OPEN = 7,
	PROBLEM_REMOTE_SOCKET_OPEN = 8,
	PROBLEM_REMOTE_SOCKET_CONNECT = 9,
	PEER_UNREACHABLE = 10,
	INVALID_NB_CONSUMERS_TO_CONNECT = 11,
	PROBLEM_LOCAL_SOCKET_CLOSE = 12,
	PROBLEM_REMOTE_SOCKET_CLOSE = 13

};


class Duplicator_Error {
	public :
		duplicator_error_enum code;
		const ACE_TString message;
		static const ACE_TString &find_error_message(int error_nb);
};

extern const class Duplicator_Error duplicator_error[];

typedef ACE_Unbounded_Set<DUP_Consumer *> CONSUMER_LIST;
typedef ACE_Unbounded_Set_Iterator<DUP_Consumer *> CONSUMER_ITERATOR;


/* ********************************************/
/********* Duplicator class ******************/
/*********************************************/
/** Duplicator's main class.
\internal
This class is a Singleton, there can only be one per application.
It is constructed before the main and destroyed after it (that's why the
constructor does nothing and the init() method exists).
It holds the main loop via the ACE reactor, and manages shared memory blocks and
opening/closing of synchronisation sockets.
The principle is to open a mmapped file, whose name is dependant of the flow
name, and one local pipe for consumers. Also, depending if the provider is
local or remote, it opens either a local unix socket or a remote connection. 
Then blocks are made available to the provider for filling, and once the
provider did it, the block availability is signaled to consumers (for remote
consumers, the buffer is sent by the socket). Each consumer handler then
signals that they have released the buffer, and when all of them have released
it, the block is made available again for the provider.
*/

class Duplicator {
public:
	Duplicator();
	~Duplicator();

/** The init method.
\internal
It is called from main and prepares the duplicator for operation.
\param flowID_ Name of the flow this duplicator must handle
\param is_provider True if the flow provider is on the local machine
\param block_size Maximum size of one data block to transfer
\param block_count Number of blocks to allocate to allow simultaneous read/write
\param peer if (is_provider) then the port number to open. else, the address to connect to to get the flow data input. 
\return 0 if the initialisation finished correctly, otherwise an error number. (See class Duplicator_Error)
*/
	int init(ACE_TString flowID_, bool is_provider, unsigned int block_size, unsigned int block_count, ACE_TString peer, ACE_TString logSeverity, int syslogEnabled, int forwardTheLog, ACE_TString logFilePath);
	int connect_to_distant_duplicator(ACE_TString peer);
	int open_provider_part(ACE_TString peer);
	int close_provider_part();
	void dump0(void);
	void dump1(char *);
	int run_event_loop(void);
	int stop_event_loop(void);

	/* Can be called from handlers */
	char *mem_base(unsigned char block_number);
	char *data_base(unsigned char block_number);   
	unsigned int block_size() {return block_size_;}

 private:
	ACE_Sig_Action sigpipe_catch;
	ACE_Sig_Action sigint_catch;
	static void ctrl_c_handler(int sig);
	ACE_TString flowID;
	bool is_provider_;
	unsigned int block_size_;
	unsigned char block_count_;	
	ACE_TString peer_;

	/* MMAP memory objects */
	ACE_Lite_MMAP_Memory_Pool *mem_pool;
	size_t requested_size;
	size_t given_size;
	ACE_TString flow_mmap_filename;
	

	/* Network objects */
#if (defined (ACE_WIN32))
	ACE_Acceptor<DUP_Local_Consumer, ACE_SOCK_Acceptor> local_consumer; /* always open */
	ACE_Acceptor<DUP_Local_Provider, ACE_SOCK_Acceptor> local_provider; /* open if is_provider */
	ACE_UINT64 port_provider;
	bool get_provider_and_consumer_ports(int &port_consumers);
#else
	ACE_Acceptor<DUP_Local_Consumer, ACE_LSOCK_Acceptor> local_consumer; /* always open */
	ACE_Acceptor<DUP_Local_Provider, ACE_LSOCK_Acceptor> local_provider; /* open if is_provider */
	ACE_TString flow_socket_filename;
	ACE_TString flow_provider_socket_filename;
#endif

	ACE_Acceptor  <DUP_Remote_Consumer, ACE_SOCK_Acceptor>  remote_consumer;   /* open if is_provider */
	ACE_Connector <DUP_Remote_Provider, ACE_SOCK_Connector> remote_provider;  /* open if ! is_provider */

	ACE_Connector<DUP_Server, ACE_SOCK_Connector> server;

	/* Peer management */
	CONSUMER_LIST consumers_connected;
	DUP_Provider *provider;

	/**/
	ACE_SOCK_Stream *socket_server;

	/* Block management */
	struct block_struct {
	  unsigned int size;
	};

	int consumer_block_release(long consumer_block);
	struct block_struct blocks[2];
	unsigned char current_provider_block;
	unsigned char current_consumer_block;

	//numbers of consumers expected to connect
	//For a duplicator consumer, isSynchronized should always be 0
	//If the duplicator has a provider part, 0 means the flow is not synchronized
	//A isSynchronized = 1, means the data transfer will only begin when we have the
	//required number of consumers have effectively connected, i.e we wait for the
	//server to tell us to start the transfer
	int isSynchronized_;

	//This integer is only use if isSynchronized == 1
	//if serverRequestDataTransfer == 1, that means the server send the request to start data transfer
	//if serverRequestDataTransfer == 0 (and  isSynchronized == 1), that means we are still
	//waiting for the request from the server.
	int serverRequestDataTransfer;

	//the number of remote consumers (i.e the number of duplicators ) connected to the duplicator 
	int nbRemoteConsumers;

	//true if the local provider is connected to the duplicator
	//if serverRequestDataTransfer == 1, that means the local provider is connected, 0 otherwise
	int isLocalProviderConnected_;

	/* Statistics */
	//unsigned long long block_seen;
	struct {
	  unsigned long long blocks_read;
	  unsigned long long blocks_written;
	} stats;

	Log_Manager *logManager;

 public:
	int connect(DUP_Consumer *c);
	int disconnect(DUP_Consumer *c);

	int block_release(DUP_Consumer *c);
	
	DUP_Provider *get_provider(void) { return provider; }

	Log_Manager *getLog_Manager() { return logManager; }
	
	int connect(DUP_Provider *p);
	int disconnect(DUP_Provider *p);
	int block_release(DUP_Provider *p, unsigned int size = 0, int block_number = 0);
	void *provider_block(void);

	int connect(DUP_Server *s);
	int disconnect(DUP_Server *s);

	//By Imad
	int displayAllConsumersConnected();
	int displayAllConsumersSwitchedOn();
	int pause(ACE_TString clientID);
	int pause_dupProvider_dupConsumer(ACE_TString host);
	int wake_up_host(ACE_TString host_name);
	int restart(ACE_TString clientID);

	//Antoine
	ACE_TString getFlowID() { return flowID; }
	int sendXMLMessageToServer(ACE_TString xml);
	int sendConnectionMessageToServer();
	void setServerSocket(ACE_SOCK_Stream* socket_);
	ACE_SOCK_Stream* getServerSocket();
	int setSynchronized(int val);
	int isSynchronized() { return isSynchronized_; }
	int setLocalProviderConnected(int val);
	int isLocalProviderConnected() { return isLocalProviderConnected_; }
	int setRequestDataTransfer(int val);
	int getRequestDataTransfer() { return serverRequestDataTransfer; }
	void increaseNumberOfRemoteConsumerConnected() { nbRemoteConsumers++; }
	void decreaseNumberOfRemoteConsumerConnected() { nbRemoteConsumers--; }


	int sendToServerTheNumberOfRemoteDuplicatorsConnected();
	int sendLogToServer(const std::string &logMessage);
};

typedef ACE_Singleton<Duplicator, ACE_Null_Mutex> TheDuplicator;

#endif

