#include "Log_Manager.h"
#include "Peer_Event_Handler.h"
#include "ace/Get_Opt.h"

#include "ace/Time_Value.h"
#include <iostream>
#include <set>

#define DEFAULT_HIGH_WATERMARK 131072

Log_Callback::Log_Callback(Log_Manager *manager) : ACE_Log_Msg_Callback() {
	//ACE_TRACE(ACE_TEXT("Log_Callback::Log_Callback"));
	manager_ = manager;
}

Log_Callback::~Log_Callback() {
	ACE_TRACE(ACE_TEXT("Log_Callback::~Log_Callback"));
	manager_ = NULL;
}

void Log_Callback::log(ACE_Log_Record &log_record) {
	//first we disable the callback to avoid an infinite loop...
	if ( manager_->forwardLog() )
		ACE_LOG_MSG->clr_flags (ACE_Log_Msg::MSG_CALLBACK);	
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Callback::log"));
	//we need to send the log to the server here

	putLogInQueue(log_record);
	sendLogToServer();
	
	//job is done, we can enable the callback again...
	if ( manager_->forwardLog() )
		ACE_LOG_MSG->set_flags (ACE_Log_Msg::MSG_CALLBACK);
}

void Log_Callback::putLogInQueue(ACE_Log_Record &log_record) {
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Callback::putLogInQueue"));
	
	//we get the size of the message
	size_t size = log_record.length();

	//first, we create a message block to store the log
	ACE_Message_Block *mb = new ACE_Message_Block(size);
	if ( mb )
		//we copy the log in the message block
		ACE_OS::memcpy( mb->wr_ptr(), log_record.msg_data(), size);
	else 
		return;
	
	//we tell the message block how much data we put in it
	mb->wr_ptr(size);


	//ACE_Time_Value timeout (ACE_OS::time (0) + 1, 0);
	ACE_Time_Value timeout(ACE_OS::gettimeofday());
	//It is pretty easy to fill the queue up... so we don't want to slow down the server....
	ACE_Message_Queue<ACE_SYNCH> *tmp_queue = manager_->getQueue();
	if ( tmp_queue ) {
		//we enqueue the message, if the queue is full or blocked,
		//this call will return after one seconde.
		if ( manager_->getQueue()->enqueue_tail(mb, &timeout) == -1 ) {
			//it failed, we forget about this message and delete the message block		
			delete mb;
		}
	}
	//ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ SF2D Log_Callback::putLogInQueue ] queue size=%d\n"), tmp_queue->message_count()));
	
	return;
}

int Log_Callback::sendLogToServer() {
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Callback::sendLogToServer"));

	//here we need to format the message properly and send it to the server
	ACE_Message_Queue<ACE_SYNCH> *tmp_queue = manager_->getQueue();
	if ( !tmp_queue )
		return -1;
	else {
		//we don't want to notify the reactor all the time, only every 5 message or 3 secondes
		if ( ( tmp_queue->message_count() >= 10 ) && !manager_->isNotificationSent() ) {
			if ( manager_->getReactor() != NULL ) {
				manager_->getReactor()->notify(manager_->getLogHandler(), ACE_Event_Handler::READ_MASK);
				manager_->setNotificationSent(1);
			}
		}
	}
	return 0;
}

/*******************************************************************/

Log_Manager::Log_Manager(Log_Event_Handler *logHandler) {
	ACE_TRACE(ACE_TEXT("Log_Manager::Log_Manager"));
	log_handler = logHandler;
	Log_Manager();
}

Log_Manager::Log_Manager() {
	ACE_TRACE(ACE_TEXT("Log_Manager::Log_Manager"));
	cb = NULL;
	file = NULL;
	react = NULL;
	notificationSent = 0;
	logForwardEnabled = false;
}

Log_Manager::~Log_Manager() {
	ACE_TRACE(ACE_TEXT("Log_Manager::~Log_Manager"));
	
	if ( cb ) {
		ACE_LOG_MSG->clr_flags (ACE_Log_Msg::MSG_CALLBACK);
		delete cb;
	}
	cb = NULL;
	if ( file ) {
		ACE_LOG_MSG->clr_flags (ACE_Log_Msg::OSTREAM);
		delete file;
	}
	file = NULL;
	log_handler = NULL;
	logQueue.deactivate();
	logQueue.flush();
}

int Log_Manager::init(const char* name) {
	ACE_TRACE(ACE_TEXT("Log_Manager::init"));

	//init the message queue
	//init the proper size here, with max message size
	logQueue.high_water_mark(DEFAULT_HIGH_WATERMARK);
	logQueue.low_water_mark(0);
	logQueue.flush();

	//we need to set up our debug level here
	//need to get the info from the PreferenceManager
	ACE_TString tmp_level = "loglevel";
	//we need to get the parameter from the command line to get 
	//the debug level (or log severity)
	//ACE_TString log_level = ACE_TEXT_CHAR_TO_TCHAR("info");
	ACE_TString log_level = LOG_PREF::instance()->getPref(tmp_level);
	if ( log_level == "debug" ) {
		//the log level is highest, we get the errors, warnings and debug info;
		ACE_LOG_MSG->priority_mask(  LM_TRACE | LM_DEBUG | LM_WARNING | LM_ERROR, ACE_Log_Msg::PROCESS);
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Severity set = debug\n")));
	}
	else if ( log_level == "warning" ) {
		//the log leve is medium, we get ther errors and warnings
		ACE_LOG_MSG->priority_mask( LM_TRACE | LM_WARNING | LM_ERROR, ACE_Log_Msg::PROCESS);
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Severity set = warning\n")));
	}
	else {
		//the log level is not specified or is the default one, we get the errors only
		ACE_LOG_MSG->priority_mask( LM_TRACE | LM_ERROR, ACE_Log_Msg::PROCESS);
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Severity set = error\n")));
	}

	//used to invoke a callback for logging operations
	cb = new Log_Callback(this);

	ACE_TString tmp_syslog = "syslog";	
	ACE_TString hasSyslog = LOG_PREF::instance()->getPref(tmp_syslog);
	if ( hasSyslog == "1" ) {
		//We are logging to the syslog of the system and stderr
		ACE_LOG_MSG->open(name, ACE_Log_Msg::SYSLOG, name);
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Logging to Syslog On\n")));
	}
	else {
		ACE_LOG_MSG->open(name);
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Logging to Syslog Off\n")));
	}

	//the log is going to stderr
	ACE_LOG_MSG->set_flags (ACE_Log_Msg::STDERR);
	
	ACE_TString tmp_callback = "logforward";
	ACE_TString isCallbackOn = LOG_PREF::instance()->getPref(tmp_callback);
	if ( isCallbackOn == "0" ) {
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Log Propagation Off\n")));
	}
	else {
		//enable the logging from the callback
		ACE_LOG_MSG->set_flags (ACE_Log_Msg::MSG_CALLBACK);
		//setting the callback object
		ACE_LOG_MSG->msg_callback(cb);
		logForwardEnabled = true;
		log_handler->enableTimer();
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Log Propagation On\n")));
	}

	//we need to enable the log info to a file
	//need to get the info from the PreferenceManager
	ACE_TString tmp_file = "logfile";
	//we need to get the parameter from the command line to see if we need to send
	//the log in a file
	//ACE_TString log_file = ACE_TEXT_CHAR_TO_TCHAR("c:\\logserver.txt");
	ACE_TString log_file = LOG_PREF::instance()->getPref(tmp_file);
	if ( log_file.length() != 0 ) {
		file = new std::ofstream(log_file.c_str());
		if ( !file ) 
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Log_Manager_ACE::init ] Log file couldn't be opened")));
		else {
			//the file has successfuly been opened
			//setting the output strean

			//need to check again what 0 means
			ACE_LOG_MSG->msg_ostream(file);
			//enable the logging to the output stream
			ACE_LOG_MSG->set_flags (ACE_Log_Msg::OSTREAM);
			ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Logging to file: %s\n"), log_file.c_str()));
		}
	}
	return 0;
}

/*******************************************************************/

Log_Preferences::Log_Preferences() {
	prefs.clear();
}

Log_Preferences::~Log_Preferences() {}
		
int Log_Preferences::initPreferences(int &argc, char** &argv) {
  ACE_TRACE(ACE_TEXT("Log_Preferences::initPreferences"));
  ACE_TString fullstring, keystring, valuestring;
  if (( argc == 0) || (argv == NULL))
    return 0;
  ACE_TCHAR **argv_;
  
  argv_ = (ACE_TCHAR **) malloc(sizeof(ACE_TCHAR*) * argc);
  
  //we make a local copy of the command line parameters
  for (int i = 0; i < argc; i++) {
    argv_[i] = ACE_TEXT_CHAR_TO_TCHAR(argv[i]);
  }

  std::set<int> index;
  index.clear();

  ACE_Get_Opt pref(argc, argv_, ACE_TEXT(""), 1, 0, ACE_Get_Opt::RETURN_IN_ORDER, 0);
  pref.long_option(ACE_TEXT("sf"),ACE_Get_Opt::ARG_REQUIRED);
  for ( int i; ( i = pref() ) != -1; )
    if ( i == 0 ) { /* A long option was found */
      fullstring = pref.opt_arg();
      index.insert(pref.opt_ind() - 2);
      index.insert(pref.opt_ind() - 1);
      unsigned int pos = fullstring.find(ACE_TEXT("="),0);
      if ( pos >=   1 ) {
		keystring = fullstring.substr(0, pos);
		if (  pos != ( fullstring.length() - 1 ) )
			valuestring = fullstring.substr(pos+1, fullstring.length()-1);
		else 
			valuestring = ACE_TEXT("");
		if ( ( (keystring.length() != 0) && (valuestring.length() != 0) ) ) {
			addPreference(keystring, valuestring);
			keystring = ACE_TEXT("");
			valuestring = ACE_TEXT("");
		}
	}
  }

  char **new_argv;
  int new_argc = 0;
  std::string listargs = "";

  //we make a list of the remaining arguments
  new_argv = (char **) malloc( sizeof(char*) * (argc - index.size()));
  for ( int i = 0; i < argc; i++ ) {
	if ( index.count(i) == 0 ) {
		new_argv[new_argc] = argv[i];
		//we don't need the first argument, which is the name of the client node.
		if ( i != 0 ) {
			listargs.append(argv[i]);
			listargs.append(" ");
		}
		new_argc++;
	}
  }
  
  //we don't want to keep sf options, we put the list of arguments, which don't concern the SF in argv.
  argv = new_argv;
  argc = new_argc;
  
  return 0;
}

void Log_Preferences::addPreference(ACE_TString key, ACE_TString value){
	ACE_TRACE(ACE_TEXT("Log_Preferences::addPreference"));
	prefs[key] = value;
}

void Log_Preferences::remPreference(ACE_TString key){
	ACE_TRACE(ACE_TEXT("Log_Preferences::remPreference"));
	if ( exists(key) ) 
		prefs.erase(key);
}

ACE_TString Log_Preferences::getPref(ACE_TString key){
	ACE_TRACE(ACE_TEXT("Log_Preferences::getPref"));
	if ( exists(key) )
		return prefs[key];
	else
		return "";

}

bool Log_Preferences::exists(ACE_TString &key) {
	ACE_TRACE(ACE_TEXT("Log_Preferences::exists"));
	std::map<ACE_TString, ACE_TString>::iterator iter;
	iter = prefs.find(key);
	if (iter != prefs.end()) 
		return true;
	else
		return false;
}

void Log_Preferences::displayPrefs() {
	ACE_TRACE(ACE_TEXT("Log_Preferences::displayPrefs"));
	std::map<ACE_TString, ACE_TString>::iterator iter;
	cout << "Preference list :" << endl;
	for(iter = prefs.begin(); iter != prefs.end(); ++iter)
		cout << "Key: " << iter->first << "\t" << "Value: " << iter->second << endl;
}

