#include "ace/Thread_Manager.h"

#include "Peer_Event_Handler.h"
#include "Smartflow.h"
#include "XML_Parser.h"
#include "sf2d_common_ACE.h"

#include "list/ClientList.h"
#include "list/Flow.h"
#include "list/FlowList.h"
#include "list/GlobalList.h"
#include "list/DuplicatorList.h"
#include "list/SocketList.h"
#include "list/ServerList.h"
#include "list/LastModifications.h"

#include "Thread_Server.h"
#include "Log_Manager.h"


ACE_HANDLE Peer_Event_Handler::get_handle(void) const
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::get_handle"));
	return remote_peer_.get_handle();
}

int Peer_Event_Handler::open()
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::open"));
	remote_peer_.get_remote_addr(remoteaddr); 
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::open ] Open With : %s \n"),remoteaddr.get_host_name()));
	return (reactor()->register_handler(this, ACE_Event_Handler::READ_MASK));
}

int Peer_Event_Handler::handle_close(ACE_HANDLE, ACE_Reactor_Mask)
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::handle_close"));
	remote_peer_.close();
	return 0;
}

int Peer_Event_Handler::handle_input(ACE_HANDLE)
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::handle_input"));
	ACE_TString doc = receive_xml();
	if (doc != ACE_TEXT(""))
	{
		if (actions_switch(doc) != 0)
		{
			return -1;
		}
	}
	else
	{
		return -1;
	}
	return 0;
}

int Peer_Event_Handler::handle_output(ACE_HANDLE) 
{
	//Here we dequeue the message partially sent
	//A message is stored using 2 or 3 blocks,
	//The first block tells us what failed to be sent: the message size and the message itself
	//or just the message (the size has been sent properly)
	//The second block contain either the size of the message or the message itself
	//The third block can only contain the message. This one is used only if the second one contains the size of the message
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::handle_output"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::handle_output ]\n")));
	ACE_Message_Block *mb1, *mb2, *mb3;
	ACE_Time_Value nowait (ACE_OS::gettimeofday ());
	bool keepGoing = true;		
		
	while ( keepGoing &&  !message_queue.is_empty() ) {
		//first we dequeue the first block containing the boolean
		char isSizeSent[2] = {0};
		if ( message_queue.dequeue_head(mb1, &nowait) <= 0 ) {
			//the message block has been properly dequeued
			//we copy the content of the block to the boolean
			ACE_OS::memcpy(isSizeSent, mb1->rd_ptr(), mb1->length());
			
			if ( ACE_OS::strcmp(isSizeSent,"0") == 0 ) {
				//the size has not been sent completely, there are therefore 2 more blocks to dequeue
				nowait = ACE_OS::gettimeofday();
				if ( message_queue.dequeue_head(mb2, &nowait) <= 0 ) {
					//So far so good, we got the size of the message in mb2 (can be partial data)
					nowait = ACE_OS::gettimeofday();
					if ( message_queue.dequeue_head(mb3, &nowait) <= 0 ) {
						//We dequeued the last piece of information to send
						ssize_t send_cnt = remote_peer_.send(mb2->rd_ptr(), mb2->length());
						if ( send_cnt == -1 ) {
							ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::handle_output ] Error sending message\n")));
							//we discard the message, the connection seems broken
							mb1->release();
							mb2->release();
							mb3->release();
							return -1;
						}
						else {
							//read pointer repositioned to reflect what has been sent
							mb2->rd_ptr(static_cast<size_t> (send_cnt));
						}
						//did we send everything?
						if (mb2->length() > 0) {
							//no we didn't, we need to requeue everything
							message_queue.enqueue_head(mb3);
							message_queue.enqueue_head(mb2);
							message_queue.enqueue_head(mb1);
							keepGoing = false;
						}
						else {
							//the size has been sent successfully, we know send the message
							send_cnt = remote_peer_.send(mb3->rd_ptr(), mb3->length());
							if ( send_cnt == -1 ) {
								ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::handle_output ] Error sending message\n")));
								//we discard the message, the connection seems broken
								mb1->release();
								mb2->release();
								mb3->release();
								return -1;
							}
							else {
								//read pointer repositioned to reflect what has been sent
								mb3->rd_ptr(static_cast<size_t> (send_cnt));
							}
							//did we send everything?
							if (mb3->length() > 0) {
								//no we didn't, we need to requeue the first and last block
								//the value of the boolean must change to reflect that the size has been sent
								char newValueForBlock[] = "1";
								ACE_OS::memcpy(mb1->rd_ptr(), newValueForBlock, sizeof(newValueForBlock)); 
								message_queue.enqueue_head(mb3);
								message_queue.enqueue_head(mb1);
								//we need to release mb2, no use for it anymore
								mb2->release();
								keepGoing = false;
							}
							else {
								//success! we release all the blocks
								mb1->release();
								mb2->release();
								mb3->release();
							}
						}
					} 
					else {
						//we failed to dequeue the last block, we therefore must requeue the other messages in reversed order. 
						message_queue.enqueue_head(mb2);
						message_queue.enqueue_head(mb1);
						keepGoing = false;
					}
				} 
				else {
					//we failed to dequeue the second block, we therefore must requeue the other messages in reversed order. 
					message_queue.enqueue_head(mb1);
					keepGoing = false;
				}
			} 
			else {
				//the size has been sent, there is therefore one more block to dequeue
				nowait = ACE_OS::gettimeofday();
				if ( message_queue.dequeue_head(mb2, &nowait) <= 0 ) {
					//we now have the partial message in mb2
					ssize_t send_cnt = remote_peer_.send(mb2->rd_ptr(), mb2->length());
					if ( send_cnt == -1 ) {
						ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::handle_output ] Error sending message\n")));
						//we discard the message, the connection seems broken
						mb1->release();
						mb2->release();
						return -1;
					}
					else {
						//read pointer repositioned to reflect what has been sent
						mb2->rd_ptr(static_cast<size_t> (send_cnt));
					}
					//did we send everything?
					if (mb2->length() > 0) {
						//no we didn't, we need to requeue everything
						message_queue.enqueue_head(mb2);
						message_queue.enqueue_head(mb1);
						keepGoing = false;
					}
					else {
						//success! we release all the blocks
						mb1->release();
						mb2->release();
					}
				}
				else {
					//we were not able to dequeue mb2, we must requeue mb1
					message_queue.enqueue_head(mb1);
					keepGoing = false;
				}
			} 
			
		} 
		else {
			//we failed to dequeue the first block, we must requeue it?
			keepGoing = false;
		}
		//was is the last block?
	}//while
	return (message_queue.is_empty ()) ? -1 : 0;
}

int Peer_Event_Handler::actions_switch(ACE_TString xml)
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::actions_switch"));
	//ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ] \n %s \n"),xml.c_str()));
	//It is very risky to display the xml message here if the logforward option is enabled
	//As a result, the XML message containing logs entry (bounded with CDATA tags) could contain messages bounded with CDATA tags
	//So the whole xml message becomes invalid and cannot be parsed by the Control Center.
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ] \n")));
	
	int type = 0;
	int target = 0;
	
	XML_Parser parser(xml);
	
	if (parser.getInfos(ACE_TEXT("type")) == ACE_TEXT(""))
    {
		return -1; //Socket shutdown
    }
	else
    {
		type = ACE_OS::atoi(parser.getInfos(ACE_TEXT("type")).c_str());
    }
	
	if (!(parser.getInfos(ACE_TEXT("target")) == ACE_TEXT("")))
    {
		target = ACE_OS::atoi(parser.getInfos(ACE_TEXT("target")).c_str());
    }
	
	switch (type)
    {
		case REQUEST:
			switch (target)
		{
			case FLOWLIST:
				break;
			case GETFLOW:
				getflow_request(xml);
				break;
			case REMOVEFLOW:
				remflow_request(xml); 
				break;
			case LAUNCH:
				launch_request(xml);
				break;
			case STOP:
				return stop_request(xml);
				break;
			case PAUSEFLOW:
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ][PAUSEFLOW REQUEST]\n")));
				notifyPauseFlowToAllServers(xml);
				break;
			case WAKE_UP_FOR_SERVER:
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ][WAKE_UP_FOR_SERVER ]\n")));
				notify_dupli_to_wakup_host(xml);
				break;
			case RESTARTFLOW:
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ][RESTARTFLOW REQUEST]\n")));
				notifyRestartFlowToAllServers(xml);
				break;
			default:
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ] Server Invalid Command\n")));
				break;
		}
			break;
		case ANNOUNCE:
			switch (target)
		{
			case REMOTEAPPLICATION:		  
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ] An existing server tries to send us its application description\n")));
				manage_remote_application_description(xml);
				break;
			case PAUSEFLOW:
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ][PAUSEFLOW ANNOUNCE]\n")));
				pauseFlow(xml);
				break;
			case RESTARTFLOW:
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ][RESTARTFLOW ANNOUNCE]\n")));
				restartFlow(xml);
				break;
			case LAUNCH_ERROR:
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ][LAUNCH ERROR ANNOUNCE]\n")));
				// We just add the received message to the modifications list so the local CC will be noticed
			{
				LastModifications *lm = LAST_MODIFICATIONS::instance(); 
				lm->addModifications(xml);
			}
				break;
			case CONSUMER_CONNECTED:		  
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ] CONSUMER_CONNECTED ANNOUNCE\n")));
				announceNewConsumerNodeConnectedToDuplicator(xml);
				break;
			case LOG_FROM_CLIENT :
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ][LOG_FROM_CLIENT] \n")));
				logAnnounce(xml);
				break;
			case LOG_FROM_DUPLI :
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ][LOG_FROM_DUPLI] \n")));
				logAnnounce(xml);
				break;
			case LOG_FROM_SERVER :
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ][LOG_FROM_SERVER] \n")));
				logAnnounce(xml);
				break;
			default:
				ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ] Server Invalid Command\n")));
				break;
		}
			break;
		case CONNECT:
			connect_request(xml);
			break;
		case DISCONNECT:
			disconnect_request(xml);
			break;
		default:
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::actions_switch ] Server Invalid Command\n")));
			break;
    }
	return 0;
}

int Peer_Event_Handler::send_xml(ACE_TString xml)
{	
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::send_xml"));
	/* As always in the Smartflow2 system, we first then the size then the data */
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::send_xml ]\n")));
	char size[512] = {0};
	ACE_OS::snprintf(size, sizeof(size), "%d", xml.length());
	remote_peer_.send_n(size, sizeof(size));
	
	const char *xml_message = ACE_TEXT_ALWAYS_CHAR(xml.c_str());
	
	//ACE_DEBUG((LM_DEBUG, ACE_TEXT("[ Peer_Event_Handler::send_xml ]Sent message:%s\n"), xml_message));
	
	remote_peer_.send_n(xml_message, ACE_OS::strlen(xml_message));
	
	return 0;
}


ACE_TString Peer_Event_Handler::receive_xml()
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::receive_xml"));
	//ACE_DEBUG((LM_DEBUG, ACE_TEXT("[ Peer_Event_Handler::receive_xml ]\n")));
	ACE_TString doc = ACE_TEXT("");
	char size_[512] = {0};  
	if (remote_peer_.recv_n(size_,sizeof(size_)) != sizeof(size_))
	{
		//A server has just quit. We will delete his entry in the server list and then delete the entries of his flow/clients in the GlobalList 
		//And finally destroy unnecessary object		
		ACE_TString serverIP = ACE_TEXT_CHAR_TO_TCHAR (remoteaddr.get_host_addr());
		
		ServerList *serverList = SERVER_LIST::instance();
		Server *server = serverList->get(serverIP, ServerList::IP);
		
		serverList->remove(serverIP, ServerList::IP);
		
		GlobalList *gl = GLOBAL_LIST::instance();
		
		gl->removeAllClientsOnServer( server->getName() );	
		//Update modifications
		LastModifications *lm = LAST_MODIFICATIONS::instance(); 
		ACE_TString modif = remove_serverxml(server->getName());
		lm->addModifications(modif.c_str());
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("Server '%s' has just quit.\n"), server->getName().c_str()));
		
		//deleting the server object
		delete server;
		
		return doc;
	}
	ACE_TString size = ACE_TEXT_CHAR_TO_TCHAR(size_);
	int length = ACE_OS::atoi(size.c_str());
	//ACE_DEBUG((LM_DEBUG, "[ Peer_Event_Handler::receive_xml ] Size : %d\n", length));
	if (length == 0) return doc;	
	
	//We are allocating here the memory to receive the data
	char *xml = new char[length+1];
	ACE_OS::memset(xml,0,length + 1);
	
	if ( remote_peer_.recv_n(xml, length) != length)
	{
		//A server has just quit. We will delete his entry in the server list and then delete the entries of his flow/clients in the GlobalList 
		//And finally destroy unnecessary object
		ACE_TString serverIP = ACE_TEXT_CHAR_TO_TCHAR (remoteaddr.get_host_addr());
		
		ServerList *serverList = SERVER_LIST::instance();
		Server *server = serverList->get(serverIP, ServerList::IP);
				
		serverList->remove(serverIP, ServerList::IP);
		
		GlobalList *gl = GLOBAL_LIST::instance();
		
		gl->removeAllClientsOnServer( server->getName() );	
		//Update modifications
		LastModifications *lm = LAST_MODIFICATIONS::instance(); 
		ACE_TString modif = remove_serverxml(server->getName());
		lm->addModifications(modif.c_str());
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("Server '%s' has just quit.\n"), server->getName().c_str()));
		
		//deleting the server object
		delete server;
		
		return doc;
	}
	doc = ACE_TEXT_CHAR_TO_TCHAR(xml); // copy of the data into the ACE_TString object
	//ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::receive_xml ] Data : \n \n %s \n \n"), doc.c_str()));
	//ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::receive_xml ] END \n")));
	delete [] xml;
	return doc;
}

int Peer_Event_Handler::disconnect_request(ACE_TString xml)
{ 
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::disconnect_request"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ]\n")));
	int result;
	GlobalList *gl = GLOBAL_LIST::instance();        
	XML_Parser parser(xml);
	ACE_TString clientname = (parser.getInfos(ACE_TEXT("clientname"))).c_str(); 
	ACE_TString clientgroup= (parser.getInfos(ACE_TEXT("clientgroup"))).c_str();
	ACE_TString clientid = Client::getIDFromGroupAndName(clientgroup, clientname);
	
	if (gl->findClient(clientid)== -1)
    {
		// Should not happen
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] Error Synchronization, Server Base corrupted\n")));
		return -1;     
    }
	
	//result = dm.notify_duplicators_to_die_peer( clientid );
	
	ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] Disconnection : %s\n"), clientid.c_str()));  
	
	// By Imad **********************************************************************************************************************************************
	// if duplicator provider server : test1
	// we verify all the flows consumed by this client : test2 
	// if it was the last client switched on and others clients are paused => pause between local and remote duplicators  : test3
	// smart detection of the duplicator provider
	Flow *f;
	ACE_TCHAR namelocal[256];
	//ACE_SOCK_Stream *dupli = NULL;
	Duplicator_Event_Handler *dupli = NULL ;
	//int buffer_size = dm.get_buffer_size();
	DuplicatorList *dl = DUPLICATOR_LIST::instance(); 
	ServerList *serverList = SERVER_LIST::instance();
	//ACE_TCHAR host_name_message[MAX_MESSAGE_SIZE]={0};
	
	if (ACE_OS::hostname(namelocal,256) == 1)
    {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::restartFlow ] ACE_OS::hostname \n")));
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("Peer_Handler restartFlow : ACE_OS::hostname"));
		send_xml(createxmlfailure.c_str());
		return 0;
    }
	ACE_TString localaddr = namelocal ;
	ACE_TString serverclientrequest = gl->getClient(clientid)->getServer();
	
	
	for (FLOW_ITERATOR fi = gl->getFirstFlow(); fi != gl->getLastFlow(); fi++ ) {
		f = GlobalList::getFlowFromIterator(fi);
		if (f == NULL) break; //Should never happens
		
		ACE_TString serverrequest = f->getServerProducerID();
		if (serverrequest == localaddr) { //test1
			ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] Server of remote duplicator detected  \n")));
			if (f->containsConsumer(clientid) != -1){ // test 2 
				ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] Client %s ==> flow %s \n"), clientid.c_str(), f->getID().c_str() ));
				//test3
				if ( (f->countSwitchedOffConsumersOnServer(serverclientrequest) > 0) && (f->countSwitchedOnConsumersOnServer(serverclientrequest) - 1 == 0) ) {
					ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] send xml sleep message to the duplicator   \n")));
					//flow_status = pending 
					//producer_status = waiting 
					if (f->setStatus(FLOW_STATUS_PENDING) == -1)
						ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] error set flow status \n")));
					if (f->setProducerStatus(CLIENT_STATUS_WAITING) == -1 )
						ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] error set producer status\n")));
					
					dupli = dl->getDupEventHandler(f->getID());
					if ( dupli == NULL ) {
						ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] ERROR: dupli==NULL\n")));
						return -1;
						//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] ERROR: dupli==NULL\n")), -1);
					}
					
					size_t size_needed = sizeof(PAUSE_DUPLI_MESSAGE) + (serverList->get(serverclientrequest.c_str(), ServerList::NAME)->getIP()).length() + 1; 
					ACE_TCHAR *host_name_message = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
					ACE_OS::memset(host_name_message, 0, size_needed);
					ACE_OS::sprintf(host_name_message,PAUSE_DUPLI_MESSAGE,(serverList->get(serverclientrequest.c_str(), ServerList::NAME)->getIP()).c_str());
					
					//result = dm.send_message(dupli, host_name_message);
					result = dupli->send_xml(host_name_message);
					
					ACE_OS::free(host_name_message);
					
					//if (result < buffer_size) {
					if (result == -1) {
						ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] ERROR: send not successful\n"));
						send_xml(createxmlfailure.c_str());
						ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] ERROR: send not successful\n")));
						return -1;
						//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] ERROR: send not successful\n")), -1);
					}
					
					
				}
			}
		}
		
	}
	// ********************************************************************************************************************************************************
	
	
	// Disconnection
	
	Client *tmp_client = gl->getClient(clientid);
	result = gl->removeClient(clientid);
	if ( tmp_client ) {
		delete tmp_client;
	}
		
	if (result != 0)
    {
		ACE_ERROR((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::disconnect_request ] ERROR: removeClient not successfull \n")));
    }
	
	LastModifications *lm = LAST_MODIFICATIONS::instance();
	ACE_TString modif ;
	modif = xml.c_str() ;
	lm->addModifications(modif) ;	
	
	return 0;
}


int Peer_Event_Handler::connect_request(ACE_TString xml)
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::connect_request"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::connect_request ]\n")));
	GlobalList *gl = GLOBAL_LIST::instance();          
	XML_Parser parser(xml);
	ACE_TString clientname = (parser.getInfos(ACE_TEXT("clientname"))).c_str(); 
	ACE_TString clientgroup = (parser.getInfos(ACE_TEXT("clientgroup"))).c_str();
	
	ACE_TString clientid = Client::getIDFromGroupAndName(clientgroup, clientname);
	if (gl->findClient(clientid) == 0)
	{
		//Cannot happen
	    ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::connect_request ] Error Synchronization, Server Base corrupted\n")));
	    return -1;     
	}
	Client *myclient; 
	myclient = new Client(clientgroup, clientname);
	
	myclient->setServer((parser.getInfos(ACE_TEXT("clientserver"))).c_str());
	myclient->setParams((parser.getInfos(ACE_TEXT("clientparam"))).c_str());
	myclient->setPath((parser.getInfos(ACE_TEXT("clientpath"))).c_str());
	myclient->setFlowInfo((parser.getInfos(ACE_TEXT("flowinfo"))).c_str());
	gl->addClient(myclient);
	
	LastModifications *lm = LAST_MODIFICATIONS::instance(); 
	ACE_TString modif ;
	modif = xml.c_str() ; 
	lm->addModifications(modif) ;	
	
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::connect_request ] END \n") ));
	return 0;
}

int Peer_Event_Handler::getflow_request(ACE_TString xml)
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::getflow_request"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::getflow_request ]\n")));
	GlobalList *gl = GLOBAL_LIST::instance();
	XML_Parser parser(xml);
	ACE_TString flowgroup = (parser.getInfos(ACE_TEXT("group"))).c_str();
	ACE_TString flowtype = (parser.getInfos(ACE_TEXT("typeflow"))).c_str();
	ACE_TString flowname = (parser.getInfos(ACE_TEXT("name"))).c_str();
	ACE_TString flowid = Flow::getIDFromGroupAndType(flowgroup, flowtype);
	ACE_TString port = (parser.getInfos(ACE_TEXT("port"))).c_str();
	
	//Provider
	if (ACE_OS::atoi( (parser.getInfos(ACE_TEXT("provider"))).c_str()  ) == smartflow::FLOW_OUTPUT ) 
	{
		ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::getflow_request ] It is a provider\n")));
		
		//Does the flow exist?
		if (gl->findFlow(flowid) == 0) 
		{
			//Yes, the flow exists.
			ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::getflow_request ] We found the flow\n")));
			
			gl->getFlow(flowid)->setName(flowname);
			gl->getFlow(flowid)->setProducer((parser.getInfos(ACE_TEXT("clientid"))).c_str());
			gl->getFlow(flowid)->setServerProducerID( (parser.getInfos(ACE_TEXT("serverid"))).c_str() );
			gl->getFlow(flowid)->setPortNumber(ACE_OS::atoi((parser.getInfos(ACE_TEXT("port"))).c_str()));
			
			DuplicatorList *dl = DUPLICATOR_LIST::instance();
			// We check if we have a consumer duplicator on our machine
			if (dl->find(flowid) == 0) {
				//We do have a consumer duplicator on our machine
				//ACE_SOCK_Stream *dupli = NULL;
				Duplicator_Event_Handler *dupli = NULL;
				dupli = dl->getDupEventHandler(flowid);
				if ( dupli == NULL ) {
					ACE_DEBUG((SF2D_WARNING_PREFIX ACE_TEXT("[ Peer_Event_Handler::getflow_request ] Duplicator for flow not connected to server yet! It will be notified to connect to the distant duplicator when it connects to the server.\n")));
					dl->setMustConnectToDistantDupli(flowid, true);
				}
				else {
					// ask the consumer dupli to connect to the provider dupli
					ACE_TString serverName = parser.getInfos(ACE_TEXT("serverid"));
					ServerList *serverList = SERVER_LIST::instance();
					ACE_TString serverIP = serverList->get(serverName, ServerList::NAME)->getIP();
					
					size_t size_needed = sizeof(CONNECT_TO_DUPLI_REQUEST_MESSAGE) + serverIP.length() + port.length() + 1; 
					ACE_TCHAR *connectToDupliRequestMessage = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
					ACE_OS::memset(connectToDupliRequestMessage, 0, size_needed);
					
					int thePort = ACE_OS::atoi(port.c_str());
					//ACE_TCHAR connectToDupliRequestMessage[MAX_MESSAGE_SIZE]={0};
					ACE_OS::sprintf(connectToDupliRequestMessage, CONNECT_TO_DUPLI_REQUEST_MESSAGE, serverIP.c_str(), thePort);
					
					//dm.send_message(dupli,connectToDupliRequestMessage);
					dupli->send_xml(connectToDupliRequestMessage);
					
					ACE_OS::free(connectToDupliRequestMessage);
				}
			}
			LastModifications *lm = LAST_MODIFICATIONS::instance(); 
			ACE_TString modif ;
			modif = xml.c_str() ; 
			lm->addModifications(modif) ;
			return 0;
			
		}
		//If we are here, that means the flow doesn't exist
		Flow *flow1 = new Flow( flowgroup , flowtype );
		flow1->setName(flowname);
		flow1->setBlockSize(ACE_OS::atoi((parser.getInfos(ACE_TEXT("blocksize"))).c_str()));
		flow1->setBlockCount(ACE_OS::atoi((parser.getInfos(ACE_TEXT("blockcount"))).c_str()));
		flow1->setProducer((parser.getInfos(ACE_TEXT("clientid"))).c_str());
		flow1->setServerProducerID( (parser.getInfos(ACE_TEXT("serverid"))).c_str() );
		flow1->setPortNumber(ACE_OS::atoi((parser.getInfos(ACE_TEXT("port"))).c_str()));
		gl->addFlow(flow1);  
	}// end if (ACE_OS::atoi( (parser.getInfos(ACE_TEXT("provider"))).c_str()  ) == smartflow::FLOW_OUTPUT ) 
	
	//Consumer
	else
	{
		Flow *f = NULL;
		if (gl->findFlow(flowid) == 0) 
		{
			f = gl->getFlow(Flow::getIDFromGroupAndType(flowgroup, flowtype));
			if (f == NULL)
			{	
				ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("Error Synchronization, Server Base corrupted\n") ));
				return -1;
			}
		}
		else
		{
			f = new Flow( flowgroup , flowtype );
			f->setName(flowname);
			f->setBlockSize( ACE_OS::atoi( parser.getInfos(ACE_TEXT("blocksize")).c_str() ));
			f->setBlockCount( ACE_OS::atoi( parser.getInfos(ACE_TEXT("blockcount")).c_str()));
			gl->addFlow(f);
		}
		
		Client *c = gl->getClient((parser.getInfos(ACE_TEXT("clientid"))).c_str());
		
		ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::getflow_request ] Add Consumer !  \n")));
		if (f->addConsumer(c->getID(), flowname.c_str() ))
		{
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::getflow_request ] Error : f->addConsumer \n") ));
		}
	} //end consumer part
	
	LastModifications *lm = LAST_MODIFICATIONS::instance();
	ACE_TString modif;
	modif = xml.c_str();
	lm->addModifications(modif);	
	
  	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::getflow_request ] END \n")));
	
	return 0;
}

int Peer_Event_Handler::remflow_request(ACE_TString xml) {
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::remflow_request"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::remflow_request ]\n")));
	
	//first we remove the flow from the table if necessary
	XML_Parser parser(xml);
	ACE_TString theFlowGroup = (parser.getInfos(ACE_TEXT("flowgroup"))).c_str();
	ACE_TString theFlowType = (parser.getInfos(ACE_TEXT("flowtype"))).c_str();
	ACE_TString theClientGroup = (parser.getInfos(ACE_TEXT("clientgroup"))).c_str();
	ACE_TString theClientName = (parser.getInfos(ACE_TEXT("clientname"))).c_str();
	
	//Is the flowgroup field present in the XML message?
	if (theFlowGroup.length() == 0)
	{
		//No, the flow has no group. It shouldn't happen.
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::remflow_request : Error XML Description ]"));
		send_xml(createxmlfailure.c_str());
		return -1;
	}
	
	//Is the flowtype field present in the XML message?
	if (  theFlowType.length() == 0 )
	{
		//No, the flow has no type. It shouldn't happen.
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::remflow_request : Error XML Description ]"));
		send_xml(createxmlfailure.c_str());
		return -1;
	}
	
	//Is the clientgroup field present in the XML message?
	if (theClientGroup.length() == 0)
	{
		//No, the flow has no group. It shouldn't happen.
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::remflow_request : Error XML Description ]"));
		send_xml(createxmlfailure.c_str());
		return -1;
	}
	
	//Is the clientname field present in the XML message?
	if (  theClientName.length() == 0 )
	{
		//No, the flow has no type. It shouldn't happen.
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::remflow_request : Error XML Description ]"));
		send_xml(createxmlfailure.c_str());
		return -1;
	}
	
	//We generate an ID for the flow from its name and group.
	ACE_TString theFlowID = Flow::getIDFromGroupAndType( theFlowGroup, theFlowType);
	ACE_TString theClientID = Client::getIDFromGroupAndName( theClientGroup, theClientName);
	GlobalList *gl = GLOBAL_LIST::instance();
	
	if (gl->findFlow(theFlowID) == 0) {
		// the flow already exists
		Flow *f = gl->getFlow(theFlowID);
		
		//we removed the client from the flow's client list. 
		if ( f->getProducer() == theClientID ) {
			f->unsetProducer();		
		}
		else {
			if ( f->containsConsumer( theClientID ) == 0)
				f->removeConsumer( theClientID);
		}
		
		if ( ( f->countConsumers() == 0 ) && ( f->getProducer() == "" ) ) {
			//the flow doesn't have any producer, nor consumers..
			//we remove it from the table.
			gl->removeFlow(theFlowID);
			delete f;
		}
		
		//If we are here, that means the flow has been properly removed from the tables.
		//We add the changes in the last modification list.
		LastModifications *lm = LAST_MODIFICATIONS::instance();
		ACE_TString modif ;
		//need to declare the method in the class
		modif = xml.c_str();
		lm->addModifications(modif) ;
	}
	
	return 0;
}

int Peer_Event_Handler::launch_request(ACE_TString xml)
{    
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::launch_request"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::launch_request ]\n")));
	XML_Parser parser(xml);
	ACE_TString tospawn = parser.getInfos(ACE_TEXT("clientpath")).c_str() ;
	tospawn += ACE_TEXT(" ");
	ACE_TString arguments = parser.getInfos(ACE_TEXT("args")).c_str() ;
	tospawn += arguments;
	tospawn += ACE_TEXT(" ");
	tospawn += parser.getInfos(ACE_TEXT("clientparam")).c_str() ;    
	
	int indice1 = arguments.find(ACE_TEXT("--sf clientname="));
	indice1 += 16; // length of "--sf clientname="
	int indice2 = arguments.find(ACE_TEXT(" "), indice1);
	ACE_TString clientname = arguments.substr(indice1, indice2-indice1);
	
	indice1 = arguments.find(ACE_TEXT("--sf clientgroup="));
	indice1 += 17; // length of "--sf clientgroup="
	indice2 = arguments.find(ACE_TEXT(" "), indice1);
	ACE_TString clientgroup = arguments.substr(indice1, indice2-indice1);
	
	Thread_Args* args = new Thread_Args( tospawn, clientname, clientgroup );	  
	ACE_Thread_Manager::instance()->spawn( (ACE_THR_FUNC) Thread_Server::run_svc, ACE_static_cast(void*, args), THR_DETACHED | THR_SCOPE_SYSTEM );
	//check to do ? Valid Path ?
	return 0; 
}

// Another server ask us to stop a client running on our machine
int Peer_Event_Handler::stop_request(ACE_TString xml)
{    
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::stop_request"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::stop_request ]\n")));
	// The xml message is parsed to obtain the different informations
	// regarding the client to stop
	XML_Parser parser(xml);
	ACE_TString clientname = (parser.getInfos(ACE_TEXT("clientname"))).c_str();
	ACE_TString clientgroup = (parser.getInfos(ACE_TEXT("clientgroup"))).c_str();
  	ACE_TString clientID = Client::getIDFromGroupAndName( clientgroup , clientname ); 
	
	// We search for the client in the global list
	GlobalList *gl = GLOBAL_LIST::instance();
	
	if ( gl->findClient(clientID) == 0 ) {
		// Find the socket of the client and send him a message
		SocketList *sl = SOCKET_LIST::instance();  
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::stop_request ] Client %s with Socket %x will be shutdown ! \n"), clientID.c_str(), sl->getClient(clientID) ));	 
		ACE_TCHAR size[512] = {0};
		ACE_OS::snprintf(size, sizeof(size), ACE_TEXT("%d"), STOP );
		if (sl->getClient(clientID) == 0 ) { //can't happen ... in theory ;)
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::stop_request ] Error socket Client not found  ! Internal Server Problem\n")));
			return 0;
		}		
		sl->getClient(clientID)->send_n(size, sizeof(size));
		ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::stop_request ] Client %s with Socket %x has been notified of the shutdown request ! \n"), clientID.c_str(), sl->getClient(clientID) ));		
	}
	else {  // Should never happen
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::stop_request ] Client not found  !\n")));
		return 0;
	}
	return 0;
}

int Peer_Event_Handler::logAnnounce(ACE_TString xml) {
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::logAnnounce"));
	//we just get a log message from an other server.
	//it can have been generated either by a server, a client node or a duplicator
	
	//We add the message in the last modification list.
	LastModifications *lm = LAST_MODIFICATIONS::instance();
	ACE_TString modif ;
	modif = xml.c_str() ;
	lm->addModifications(modif);
	
	return 0;
}

int Peer_Event_Handler::construct_and_send_local_application_description() {
	
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::construct_and_send_local_application_description"));
	// We save the name of the current machine (the machine on which the Smartflow Server is running)
	ACE_TCHAR hostname[99+1];
	if (ACE_OS::hostname (hostname,99) == -1) 
		ACE_ERROR ((LM_ERROR,ACE_TEXT("%p\n"),ACE_TEXT("hostname")));
	
#if (defined (ACE_WIN32))
	char os[] = "WinXP";
#elif (defined (ACE_CONFIG_MACOSX_TIGER_H))
	char os[] = "OSX Tiger";
#elif (defined (ACE_CONFIG_MACOX_LEOPARD_H))
	char os[] = "OSX Leopard";
#else
	char os[] = "Linux";
#endif
	
	GlobalList *gl = GLOBAL_LIST::instance();
	ACE_TString xml = ACE_TEXT("");
	xml += ACE_TEXT("<?xml version='1.0'?>\n");
	xml += ACE_TEXT("<smartspace>\n");
	xml += ACE_TEXT("<type>5</type>\n");
	xml += ACE_TEXT("<target>18</target>\n");
	xml += ACE_TEXT("<servername>");
	xml += hostname;
	xml += ACE_TEXT("</servername>\n");
	xml += ACE_TEXT("<serveros>");
	xml += os;
	xml += ACE_TEXT("</serveros>\n");
	xml += ACE_TEXT("<LocalList>\n");
	
	ACE_TString localname = hostname;
	
	// First we save the local clients
	xml += ACE_TEXT("    <ClientList>\n");
	for (CLIENT_ITERATOR ci = gl->getFirstClient(); ci != gl->getLastClient(); ci++) {
		Client *client = gl->getClientFromIterator(ci);
		ACE_TString servername = client->getServer();
		
		// We test if the client is running on the current server or on another
		if (servername == localname) {
			
			// The client is running on the current server, we add it to the LocalList
			xml += (*ci).int_id_->toXmlWithIndent(2);
		}
	}
	xml += ACE_TEXT("    </ClientList>\n");
	
	// Then we save the local provided flows
	xml += ACE_TEXT("    <FlowList>\n");
	xml += ACE_TEXT("        <ProducedFlows>\n");
	for (FLOW_ITERATOR fi = gl->getFirstFlow(); fi != gl->getLastFlow(); fi++) {
		Flow *flow = gl->getFlowFromIterator(fi);
		if (flow->getPortNumber() != 0) {
			Client *producer = gl->getClient(flow->getProducer());
			if (producer != NULL) {
				ACE_TString servername = producer->getServer();
				
				// We test if the flow is provided by a client running on the current server or on another
				if (servername == localname) {
					
					// The flow is provided by a client running on the current server, we add it to the LocalList
					xml += (*fi).int_id_->toXmlWithoutConsumers(2);
				}
			}
		}
	}
	xml += ACE_TEXT("        </ProducedFlows>\n");
	xml += ACE_TEXT("        <ConsumedFlows>\n");
	// Finally we save the local consumed flows
	for (FLOW_ITERATOR fi = gl->getFirstFlow(); fi != gl->getLastFlow(); fi++) {
		Flow *flow = gl->getFlowFromIterator(fi);
		bool has_local_consumer = false;
		for (CONSUMER_ITERATOR ci = flow->firstConsumer(); ci != flow->lastConsumer(); ci++) {
			Client *consumer = gl->getClient(Flow::getConsumerIDFromIterator(ci));
			ACE_TString servername = consumer->getServer();
			// We test if the client is running on the current server or on another
			if (servername == localname) {
				has_local_consumer = true;
				break;
			}
		}
		
		if (has_local_consumer == true) {
			// The flow is consumed by one or more clients running on the current server, we add it to the LocalList
			xml += (*fi).int_id_->toXmlWithoutProducer(2);
		}
	}
	xml += ACE_TEXT("        </ConsumedFlows>\n");
	xml += ACE_TEXT("    </FlowList>\n");
	xml += ACE_TEXT("</LocalList>\n");
	xml += ACE_TEXT("</smartspace>\n");
	//ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::construct_and_send_local_application_description ]\n %s\n"), xml.c_str()));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::construct_and_send_local_application_description ] We send our local application description\n")));
	send_xml(xml);
	return 0;
}

int Peer_Event_Handler::manage_remote_application_description(ACE_TString xml) {
	
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::manage_remote_application_description"));
	// We parse the received xml file to obtain the name of the distant server
	// We add it in the socket list and in the server list
	//SocketList *sl = SOCKET_LIST::instance();
	
	int startPosition = xml.strstr(ACE_TEXT("<servername>"));
	startPosition += 12;
	int endPosition = xml.strstr(ACE_TEXT("</servername>"));
	ACE_TString peerName = xml.substr(startPosition, endPosition - startPosition);
	
	//sl->addServer( peerName ,&this->peer()); 
	
	startPosition = xml.strstr(ACE_TEXT("<serveros>"));
	startPosition += 10;
	endPosition = xml.strstr(ACE_TEXT("</serveros>"));
	ACE_TString peerOS = xml.substr(startPosition, endPosition - startPosition);
	
	ServerList *serverList = SERVER_LIST::instance();
	Server *peer = new Server(peerName, ACE_TEXT_CHAR_TO_TCHAR (remoteaddr.get_host_addr()), peerOS);
	peer->setHandler(this);
	serverList->add(peer);
	
	LastModifications *lm = LAST_MODIFICATIONS::instance();
	ACE_TString modif ;
	modif = add_server_message(peerName);
	lm->addModifications(modif) ;
	
	// ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::manage_remote_application_description ]\n %s\n"), xml.c_str()));
	GlobalList *gl = GLOBAL_LIST::instance();	
	// First we add the clients
	startPosition = xml.strstr(ACE_TEXT("<ClientList>"));
	startPosition += 12;
	endPosition = xml.strstr(ACE_TEXT("</ClientList>"));
	ACE_TString clientsXML = xml.substr(startPosition, endPosition - startPosition);
	while ((startPosition = clientsXML.strstr(ACE_TEXT("<Client"))) != -1) {
		endPosition = clientsXML.strstr(ACE_TEXT("</Client>"));
		ACE_TString clientXML = clientsXML.substr(startPosition, endPosition - startPosition + 9);
		//ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::manage_remote_application_description ]clientsXML:\n %s\n"), clientXML.c_str()));
		Client *client = new Client (clientXML);
		gl->addClient(client);
		clientsXML = clientsXML.substr(endPosition + 9);
	}
	// Then we add the produced flows
	startPosition = xml.strstr(ACE_TEXT("<ProducedFlows>"));
	startPosition += 15;
	endPosition = xml.strstr(ACE_TEXT("</ProducedFlows>"));
	ACE_TString producedFlowsXML = xml.substr(startPosition, endPosition - startPosition);
	while ((startPosition = producedFlowsXML.strstr(ACE_TEXT("<Flow"))) != -1) {
		endPosition = producedFlowsXML.strstr(ACE_TEXT("</Flow>"));
		ACE_TString flowXML = producedFlowsXML.substr(startPosition, endPosition - startPosition + 7);
		// ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::manage_remote_application_description ]flowsXML:\n %s\n"), flowXML.c_str()));
		if (gl->containsFlow(flowXML) == 0) { // The flow already exists, we update it
			//ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::manage_remote_application_description ]The flow already exists in the list, we update it\n")));
			gl->updateProducerOfFlow(flowXML);
		}
		else {
			Flow *flow = new Flow (flowXML);
			//ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::manage_remote_application_description ]The flow doesn't exist in the list, we add it\n")));
			gl->addFlow(flow);
		}
		producedFlowsXML = producedFlowsXML.substr(endPosition + 7);
	}
	
	// Then we add the consumed flows
	startPosition = xml.strstr(ACE_TEXT("<ConsumedFlows>"));
	startPosition += 15;
	endPosition = xml.strstr(ACE_TEXT("</ConsumedFlows>"));
	ACE_TString consumedFlowsXML = xml.substr(startPosition, endPosition - startPosition);
	while ((startPosition = consumedFlowsXML.strstr(ACE_TEXT("<Flow"))) != -1) {
		endPosition = consumedFlowsXML.strstr(ACE_TEXT("</Flow>"));
		ACE_TString flowXML = consumedFlowsXML.substr(startPosition, endPosition - startPosition + 7);
		// ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::manage_remote_application_description ]flowsXML:\n %s\n"), flowXML.c_str()));
		if (gl->containsFlow(flowXML) == 0) { // The flow already exists, we update it
			// ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::manage_remote_application_description ]The flow already exists in the list, we update it\n")));
			gl->updateConsumersOfFlow(flowXML);
		}
		else {
			Flow *flow = new Flow (flowXML);
			//ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::manage_remote_application_description ]The flow doesn't exist in the list, we add it\n")));
			gl->addFlow(flow);
		}
		consumedFlowsXML = consumedFlowsXML.substr(endPosition + 7);
	}
	return 0;
}

ACE_TString Peer_Event_Handler::add_server_message(ACE_TString name_)
{ 
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::add_server_message"));
	ACE_TString doc ;
	doc += ACE_TEXT("<?xml version='1.0'?>\n<smartspace>\n<type>3</type>\n<target/>\n<servername>");
	doc +=   name_ ;
	doc += ACE_TEXT("</servername>\n</smartspace>");
	return doc;
	
}

ACE_TString Peer_Event_Handler::create_xmlsuccess()
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::create_xmlsuccess"));
	ACE_TString doc; 
	doc = ACE_TEXT("<?xml version='1.0'?>\n<smartspace>\n<type>2</type>\n<target/>\n<code>Successful</code>\n</smartspace>");
	return doc;
}

ACE_TString Peer_Event_Handler::create_xmlfailure(ACE_TString des_)
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::create_xmlfailure"));
	ACE_TString doc;
	doc = ACE_TEXT("<?xml version='1.0'?>\n<smartspace>\n<type>2</type>\n<target/>\n<code>Failure</code>\n<description>");
	doc += des_;
	doc += ACE_TEXT("</description>\n</smartspace>");
	return doc;
}

ACE_TString Peer_Event_Handler::remove_serverxml(ACE_TString name_)
{
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::remove_serverxml"));
	ACE_TString doc;
	doc = ACE_TEXT("<?xml version='1.0'?>\n<smartspace>\n<type>4</type>\n<target/>\n<servername>");
	doc += name_ ;
	doc += ACE_TEXT("</servername>\n</smartspace>");
	return doc;
}

/*int Peer_Event_Handler::send_xml_to_servers(ACE_SOCK_Stream *peer_server, ACE_TString xml) {
	
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::send_xml_to_servers"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::send_xml_to_servers ]\n")));	
	char size[512] = {0};
	ACE_OS::snprintf(size, sizeof(size), "%d", xml.length());
	peer_server->send_n(size, sizeof(size));
	
	const char *xml_message = ACE_TEXT_ALWAYS_CHAR(xml.c_str());
	peer_server->send_n(xml_message, ACE_OS::strlen(xml_message));
	
	return 0;
}*/

//original
/*int Peer_Event_Handler::send_xml_to_servers(Server *peer_server, ACE_TString xml) {
	
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::send_xml_to_servers"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::send_xml_to_servers ]\n")));
	if ( (peer_server != NULL) && (peer_server->getHandler() != NULL) ) {
		ACE_SOCK_Stream *sock = &(peer_server->getHandler()->peer());
		if ( sock ) {
			//here we are trying to send, if it is not successful,
			//we test for EWOULDBLOCK, we queue the message to send it later.
			char size[512] = {0};
			ACE_OS::snprintf(size, sizeof(size), "%d", xml.length());
			ssize_t send_cnt = sock->send(size, sizeof(size));
			const char *xml_message = ACE_TEXT_ALWAYS_CHAR(xml.c_str());
			if ( send_cnt == sizeof(size) ) {
				//we successfully sent the size, we know send the message itself
				send_cnt = sock->send(xml_message, ACE_OS::strlen(xml_message));
				if ( send_cnt == ACE_OS::strlen(xml_message) )
					//success, the message has been sent properly
					return 0;
				else if ( send_cnt == -1 && ACE_OS::last_error() == EWOULDBLOCK ) {
					//we queue the message
					return peer_server->getHandler()->putMessageInQueue(xml_message);
				}
				else {
					//there is definetely a problem
					return -1;
				}
			}
			else if ( send_cnt == -1 && ACE_OS::last_error() == EWOULDBLOCK ) {
				//we queue the message
				return peer_server->getHandler()->putMessageInQueue(xml_message);
			}
			else {
				return -1;
			}
		}
	}
	return 0;
}*/

//new
int Peer_Event_Handler::send_xml_to_servers(Server *peer_server, ACE_TString xml) {
	
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::send_xml_to_servers"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::send_xml_to_servers ]\n")));
	if ( (peer_server != NULL) && (peer_server->getHandler() != NULL) ) {
		ACE_SOCK_Stream *sock = &(peer_server->getHandler()->peer());
		if ( sock ) {
			//here we are trying to send, if it is not successful,
			//we test for EWOULDBLOCK, we queue the message to send it later.
			char size[512] = {0};
			ACE_OS::snprintf(size, sizeof(size), "%d", xml.length());
			const char *xml_message = ACE_TEXT_ALWAYS_CHAR(xml.c_str());
			ssize_t send_cnt = sock->send(size, sizeof(size));
			if ( send_cnt == sizeof(size) ) {
				//we successfully sent the size, we know send the message itself
				send_cnt = sock->send(xml_message, ACE_OS::strlen(xml_message));
				ssize_t messageLength = ACE_static_cast(size_t, ACE_OS::strlen(xml_message));
				if ( send_cnt == messageLength )
					//success, the message has been sent properly
					return 0;
				else if ( send_cnt == -1 && ACE_OS::last_error() != EWOULDBLOCK )
					//something really wrong is happening...
					return -1;
				
				//the socket is not closed but the message couldn't be send properly
				if ( send_cnt == -1 )
					//nothing has been sent so far... we adjust send_cnt to reflect the proper number of byte sent so far
					send_cnt = 0;
				
				//we know need to queue the message to send it later. 
				size_t remaining = ACE_static_cast(size_t, (ACE_OS::strlen(xml_message) - send_cnt));
				return peer_server->getHandler()->putMessageInQueue(xml_message, remaining, true);
			}
			else if ( send_cnt == -1 && ACE_OS::last_error() != EWOULDBLOCK )
				//something really wrong is happening...
				return -1;
			//the socket is not closed but the message couldn't be send properly
			if ( send_cnt == -1 )
				//nothing has been sent so far... we adjust send_cnt to reflect the proper number of byte sent so far
				send_cnt = 0;
			//we know need to queue the message to send it later. 
			size_t remaining = ACE_static_cast(size_t, (sizeof(size) - send_cnt));
			return peer_server->getHandler()->putMessageInQueue(xml_message, remaining, false);	
		}
	}
	return -1;
}


int Peer_Event_Handler::putMessageInQueue(const char* message, ssize_t send_cnt, bool isSizeSent) {
	//this method enqueues message blocks...
	//it enqueues 2 or 3 block
	//block 1: a boolean telling if we couldn't send the size of message or the message itself
	//block 2: the size of the message or the message itself
	//block 3: the message itself (unused if the size has been sent properly)
	
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::putMessageInQueue"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::putMessageInQueue ]\n")));
	//this method enqueue the message and post an event on the handler.
	int output_off = message_queue.is_empty();
	ACE_Message_Block *mb1, *mb2, *mb3;
	//we create the first block containing the boolean coded as a character
	const char *isSizeSentChar = NULL;
	if ( isSizeSent ) 
		isSizeSentChar = "1";
	else
		isSizeSentChar = "0";
	ACE_NEW_RETURN(mb1, ACE_Message_Block(isSizeSentChar, sizeof(isSizeSentChar)), -1);
	if ( !isSizeSent ) {
		//we need our three blocks
		//we put in the second block the remaining of the message size not sent.
		char size[512] = {0};
		ACE_OS::snprintf(size, sizeof(size), "%d", ACE_OS::strlen(message));
		//we calculate what is the remaining size of the data to send
		size_t remaining = ACE_static_cast(size_t, (sizeof(size) - send_cnt));
		ACE_NEW_RETURN(mb2, ACE_Message_Block(&size[send_cnt], remaining), -1);
		//we put in the third block the message itself
		ACE_NEW_RETURN(mb3, ACE_Message_Block(message, ACE_OS::strlen(message)), -1);
		//finally we put the messages in the queue
		ACE_Time_Value nowait(ACE_OS::gettimeofday());
		if ( message_queue.enqueue_tail(mb1, &nowait) == -1 ) {
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::putMessageInQueue ] Failed to enqueue message\n")));
			//we cannot enqueue... we release the blocks... This is not suppose to happen...
			mb1->release();
			mb2->release();
			mb3->release();
			return -1;
		}
		nowait = ACE_OS::gettimeofday();
		if ( message_queue.enqueue_tail(mb2, &nowait) == -1 ) {
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::putMessageInQueue ] Failed to enqueue message\n")));
			//we cannot enqueue... we release the block... This is not suppose to happen...
			//we need to dequeue 1 and release all blocks
			
			mb1->release();
			mb2->release();
			mb3->release();
			return -1;
		}
		nowait = ACE_OS::gettimeofday();
		if ( message_queue.enqueue_tail(mb3, &nowait) == -1 ) {
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::putMessageInQueue ] Failed to enqueue message\n")));
			//we cannot enqueue... we release the block... This is not suppose to happen...
			//we need to dequeue 1 and 2 and release all blocks
			
			mb1->release();
			mb2->release();
			mb3->release();
			return -1;
		}
		
	} else {
		//we only need 2 blocks
		//we put in the rest of the message not sent
		size_t remaining = ACE_static_cast(size_t, (ACE_OS::strlen(message) - send_cnt));
		ACE_NEW_RETURN(mb2, ACE_Message_Block(&message[send_cnt], remaining), -1);
		//finally we put the messages in the queue
		ACE_Time_Value nowait(ACE_OS::gettimeofday());
		if ( message_queue.enqueue_tail(mb1, &nowait) == -1 ) {
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::putMessageInQueue ] Failed to enqueue message\n")));
			//we cannot enqueue... we release the blocks... This is not suppose to happen...
			mb1->release();
			mb2->release();
			return -1;
		}
		nowait = ACE_OS::gettimeofday();
		if ( message_queue.enqueue_tail(mb2, &nowait) == -1 ) {
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::putMessageInQueue ] Failed to enqueue message\n")));
			//we cannot enqueue... we release the block... This is not suppose to happen...
			//we need to dequeue 1 and release all blocks
			mb1->release();
			mb2->release();
			return -1;
		}
	}
	//we properly enqueued the blocks
	//if the queue was empty, that means we need to notify the reactor to call the handle_output method
	//if the queue was not empty, this is not necessary, the reactor already knows about it.
	if ( output_off )
		return reactor()->register_handler(this, ACE_Event_Handler::WRITE_MASK);
	return 0;
}

// Request
int Peer_Event_Handler::notifyPauseFlowToAllServers(ACE_TString xml) {
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::notifyPauseFlowToAllServers"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ]\n")));
	XML_Parser parser(xml);
	
	ACE_TString clientName = (parser.getInfos(ACE_TEXT("clientname"))).c_str();
	ACE_TString clientGroup = (parser.getInfos(ACE_TEXT("clientgroup"))).c_str();
	ACE_TString clientID = clientGroup + ACE_TEXT(":") + clientName;
	
	ACE_TString flowType = (parser.getInfos(ACE_TEXT("flowtype"))).c_str();
	ACE_TString flowGroup = (parser.getInfos(ACE_TEXT("flowgroup"))).c_str();
	ACE_TString flowID = flowGroup + ACE_TEXT(":") + flowType;
	
	ACE_TString hostName = (parser.getInfos(ACE_TEXT("hostname"))).c_str();
	
	GlobalList *gl = GLOBAL_LIST::instance();
	//SocketList *sl = SOCKET_LIST::instance();
	ServerList *sl = SERVER_LIST::instance(); 
	DuplicatorList *dl = DUPLICATOR_LIST::instance();
	
	
	if ( gl->findClient(clientID) != 0 )
    {
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("Client not found !"));
		send_xml(createxmlfailure.c_str());
		return -1;
    }
	
	
	// send 'pause' to the duplicator of this flow 
	//ACE_SOCK_Stream *dupli = NULL;
	Duplicator_Event_Handler *dupli;
	int result;
	//int buffer_size = dm.get_buffer_size();
	Flow *f;
	f = gl->getFlow(flowID);
	if (f == NULL) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] flow == NULL \n") ));
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] flow == NULL \n"));
		send_xml(createxmlfailure.c_str());
		return -1;
	}
	
	dupli = dl->getDupEventHandler(f->getID());
	if ( dupli == NULL ) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] ERROR: dupli==NULL\n")));
		return -1;
		//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] ERROR: dupli==NULL\n")), -1);
	}
	// inform the dupli
	//result = dm.send_message(dupli,xml.c_str());
	result = dupli->send_xml(xml.c_str());
	//if (result < buffer_size) {
	if (result == -1) {
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] ERROR: send not successful\n"));
		send_xml(createxmlfailure.c_str());
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] ERROR: send not successful\n")));
		return -1;
		//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] ERROR: send not successful\n")), -1);
	}
	else
		ACE_DEBUG((SF2D_ANNOUNCE_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] Sent 'pause' xml message to the duplicator of flow '%s'\n"), f->getID().c_str()));
	
	// Inform all others servers
	//ACE_TCHAR serverPauseAnnounceMessage[MAX_MESSAGE_SIZE]={0};
	size_t size_needed = sizeof(PAUSE_ANNOUNCE_MESSAGE) + clientName.length() + clientGroup.length() + flowType.length() + flowGroup.length() + hostName.length() + 1; 
	ACE_TCHAR *serverPauseAnnounceMessage = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
	ACE_OS::memset(serverPauseAnnounceMessage, 0, size_needed);
	ACE_OS::sprintf(serverPauseAnnounceMessage, PAUSE_ANNOUNCE_MESSAGE, clientName.c_str(), clientGroup.c_str(), flowType.c_str(), flowGroup.c_str(), hostName.c_str());
	
	//for (SOCKET_ITERATOR si = sl->firstServer(); si != sl->lastServer(); si++) 
	//	send_xml_to_servers(sl->getServer(SocketList::getIDFromIterator(si)) , serverPauseAnnounceMessage); 
	
	for (SERVER_ITERATOR si = sl->first(); si != sl->last(); si++)
		send_xml_to_servers((*si).int_id_, serverPauseAnnounceMessage);
	// update 
	
	// 1. get the Flow
	// 2. switched_OFF client for this flow
	// 3. if no consumers for this flow then flow_status = pending and producer_status = waiting 
	
	if (f->setConsumerStatus(clientID, CLIENT_STATUS_SWITCHED_OFF) == -1 ) 
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] error set consumer status \n")));
	
	// no more consumer
	if (f->countSwitchedOnConsumers() == 0 ) {
		ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers] no more consumer, provider going to sleep ... \n")));
		if (f->setStatus(FLOW_STATUS_PENDING) == -1)
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] error set flow status \n")));
		if (f->setProducerStatus(CLIENT_STATUS_WAITING) == -1 )
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyPauseFlowToAllServers ] error set producer status\n")));
	}  
	
	// Update modifications
	LastModifications *lm = LAST_MODIFICATIONS::instance(); 
	lm->addModifications(serverPauseAnnounceMessage);
	
	ACE_OS::free(serverPauseAnnounceMessage);
	return 0;
}

// Request
int Peer_Event_Handler::notifyRestartFlowToAllServers(ACE_TString xml) {
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::notifyRestartFlowToAllServers"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyRestartFlowToAllServers ]\n")));
	XML_Parser parser(xml);
	
	ACE_TString clientName = (parser.getInfos(ACE_TEXT("clientname"))).c_str();
	ACE_TString clientGroup = (parser.getInfos(ACE_TEXT("clientgroup"))).c_str();
	ACE_TString clientID = clientGroup + ACE_TEXT(":") + clientName;
	
	ACE_TString flowType = (parser.getInfos(ACE_TEXT("flowtype"))).c_str();
	ACE_TString flowGroup = (parser.getInfos(ACE_TEXT("flowgroup"))).c_str();
	ACE_TString flowID = flowGroup + ACE_TEXT(":") + flowType;
	
	ACE_TString hostName = (parser.getInfos(ACE_TEXT("hostname"))).c_str();
	
	GlobalList *gl = GLOBAL_LIST::instance();
	//SocketList *sl = SOCKET_LIST::instance();
	ServerList *sl = SERVER_LIST::instance(); 
	DuplicatorList *dl = DUPLICATOR_LIST::instance();
	
	if ( gl->findClient(clientID) != 0 ) {
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("Client not found !"));
		send_xml(createxmlfailure.c_str());
		return 0;
	}
	
	//ACE_SOCK_Stream *dupli = NULL;
	Duplicator_Event_Handler *dupli = NULL;
	int result;
	//int buffer_size = dm.get_buffer_size();
	Flow *f;
	f = gl->getFlow(flowID);
	if (f == NULL) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyRestartFlowToAllServers ] flow == NULL \n") ));
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::notifyRestartFlowToAllServers ] flow == NULL \n"));
		send_xml(createxmlfailure.c_str());
		return -1;
	}
	
	// the request to the duplicator
	dupli = dl->getDupEventHandler(f->getID());
	if ( dupli == NULL ) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyRestartFlowToAllServers ] ERROR: dupli==NULL\n")));
		return -1;
		//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::notifyRestartFlowToAllServers ] ERROR: dupli==NULL\n")), -1);
	}
	//result = dm.send_message(dupli,xml.c_str());
	result = dupli->send_xml(xml.c_str());
	//if (result < buffer_size) {
	if (result == -1) {
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::restartFlow ] ERROR: send not successful\n"));
		send_xml(createxmlfailure.c_str());
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyRestartFlowToAllServers ] ERROR: send not successful\n")));
		return -1;
		//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::restartFlow ] ERROR: send not successful\n")), -1);
	}
	
	// infom all servers
	//ACE_TCHAR serverRestartAnnounceMessage[MAX_MESSAGE_SIZE]={0};
	size_t size_needed = sizeof(RESTART_ANNOUNCE_MESSAGE) + clientName.length() + clientGroup.length() + flowType.length() + flowGroup.length() + hostName.length() + 1; 
	ACE_TCHAR *serverRestartAnnounceMessage = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
	ACE_OS::memset(serverRestartAnnounceMessage, 0, size_needed);
	ACE_OS::sprintf(serverRestartAnnounceMessage, RESTART_ANNOUNCE_MESSAGE, clientName.c_str(), clientGroup.c_str(), flowType.c_str(), flowGroup.c_str(), hostName.c_str());
	
	
	/*for (SOCKET_ITERATOR si = sl->firstServer(); si != sl->lastServer(); si++) 
		send_xml_to_servers(sl->getServer(SocketList::getIDFromIterator(si)) , serverRestartAnnounceMessage);*/
		
	for (SERVER_ITERATOR si = sl->first(); si != sl->last(); si++)
		send_xml_to_servers((*si).int_id_, serverRestartAnnounceMessage);
		
	// update flow status
	if (f->setConsumerStatus(clientID, CLIENT_STATUS_SWITCHED_ON) == -1 ) 
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyRestartFlowToAllServers ] error set consumer status \n")));
	if (f->setStatus(FLOW_STATUS_RUNNING) == -1)
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyRestartFlowToAllServers ] error set flow status \n")));
	if (f->setProducerStatus(CLIENT_STATUS_SWITCHED_ON) == -1 )
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notifyRestartFlowToAllServers ] error set producer status\n")));
	
	// Update modifications
	LastModifications *lm = LAST_MODIFICATIONS::instance(); 
	lm->addModifications(serverRestartAnnounceMessage);
	
	ACE_OS::free(serverRestartAnnounceMessage);
	return 0; 
}

// update all servers (Announce)
int Peer_Event_Handler::pauseFlow(ACE_TString xml) {
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::pauseFlow"));
	
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::pauseFlow]\n")));
	XML_Parser parser(xml);
	
	ACE_TString clientName = (parser.getInfos(ACE_TEXT("clientname"))).c_str();
	ACE_TString clientGroup = (parser.getInfos(ACE_TEXT("clientgroup"))).c_str();
	ACE_TString clientID = clientGroup + ACE_TEXT(":") + clientName;
	
	ACE_TString flowType = (parser.getInfos(ACE_TEXT("flowtype"))).c_str();
	ACE_TString flowGroup = (parser.getInfos(ACE_TEXT("flowgroup"))).c_str();
	ACE_TString flowID = flowGroup + ACE_TEXT(":") + flowType;
	
	ACE_TString host_name = (parser.getInfos(ACE_TEXT("hostname"))).c_str();
	//ACE_TCHAR host_name_message[MAX_MESSAGE_SIZE]={0};
	
	GlobalList *gl = GLOBAL_LIST::instance();
	DuplicatorList *dl = DUPLICATOR_LIST::instance();  
	//SocketList *sl = SOCKET_LIST::instance();
	ServerList *sl = SERVER_LIST::instance();
	
	Duplicator_Event_Handler *dupli = NULL;
	//ACE_SOCK_Stream *dupli = NULL;
	int result;
	//int buffer_size = dm.get_buffer_size();
	
	Flow *f;
	f = gl->getFlow(flowID);
	if (f == NULL) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::pauseFlow ] flow == NULL \n") ));
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::pauseFlow ] flow == NULL \n"));
		send_xml(createxmlfailure.c_str());
		return -1;
	}
	
	if ( gl->findClient(clientID) != 0 )
    {
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("Client not found !"));
		send_xml(createxmlfailure.c_str());
		return 0;
    }
	
	// update 
	
	// 1. get the Flow
	// 2. switched_OFF client for this flow
	// 3. if no consumers for this flow  then flow_status = pending and producer_status = wating 
	
	
	if (f->setConsumerStatus(clientID, CLIENT_STATUS_SWITCHED_OFF) == -1 ) 
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::pause_flow ] error set consumer status \n")));
	
	
	if (f->countSwitchedOnConsumers() == 0 ) {
		if (f->setStatus(FLOW_STATUS_PENDING) == -1)
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::pause_flow ] error set flow status \n")));
		if (f->setProducerStatus(CLIENT_STATUS_WAITING) == -1 )
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::pause_flow ] error set producer status\n")));
	}
	
	
	
	ACE_TCHAR namelocal[256];
	if (ACE_OS::hostname(namelocal,256) ==1)
    {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::pause_flow ] ACE_OS::hostname \n")));
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("Peer_Handler_Launch::pause_flow : ACE_OS::hostname"));
		send_xml(createxmlfailure.c_str());
		return 0;
    }	
	ACE_TString localaddr = namelocal ;
	ACE_TString serverrequest = f->getServerProducerID() ; 
	
	// smart detection of the duplicator provider
	if (serverrequest == localaddr) {
		ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::pause_flow ] Server of remote duplicator detected \n")));
		
		for (SERVER_ITERATOR si = sl->first(); si != sl->last(); si++) { 
		//for (SOCKET_ITERATOR si = sl->firstServer(); si != sl->lastServer(); si++) {
			//if ( (f->countSwitchedOnConsumersOnServer(sl->getIDFromIterator(si)) == 0) && (f->countSwitchedOffConsumersOnServer(sl->getIDFromIterator(si)) > 0)  ){
			if ( (f->countSwitchedOnConsumersOnServer((*si).int_id_->getName()) == 0) && (f->countSwitchedOffConsumersOnServer((*si).int_id_->getName()) > 0)  ){ 
				dupli = dl->getDupEventHandler(f->getID());
				if ( dupli == NULL ) {
					ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::pause_flow ] ERROR: dupli==NULL\n")));
					return -1;
					//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::pause_flow ] ERROR: dupli==NULL\n")), -1);
				}
				
				//size_t size_needed = sizeof(PAUSE_DUPLI_MESSAGE) + (serverList->get(sl->getIDFromIterator(si).c_str(), ServerList::NAME)->getIP()).length() + 1; 
				//size_t size_needed = sizeof(PAUSE_DUPLI_MESSAGE) + (sl->get((*si).int_id_->getName().c_str(), ServerList::NAME)->getIP()).length() + 1; 
				size_t size_needed = sizeof(PAUSE_DUPLI_MESSAGE) + ((*si).int_id_->getIP()).length() + 1; 
				ACE_TCHAR *host_name_message = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
				ACE_OS::memset(host_name_message, 0, size_needed);
				//ACE_OS::sprintf(host_name_message,PAUSE_DUPLI_MESSAGE,(serverList->get(sl->getIDFromIterator(si).c_str(), ServerList::NAME)->getIP()).c_str());
				//ACE_OS::sprintf(host_name_message,PAUSE_DUPLI_MESSAGE,(sl->get((*si).int_id_->getName().c_str(), ServerList::NAME)->getIP()).c_str());
				ACE_OS::sprintf(host_name_message,PAUSE_DUPLI_MESSAGE, ((*si).int_id_->getIP()).c_str());
				
				result = dupli->send_xml(host_name_message);
				ACE_OS::free(host_name_message);
				//if (result < buffer_size) {
				if (result == -1) {
					ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::pauseFlow ] ERROR: send not successful\n"));
					send_xml(createxmlfailure.c_str());
					ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::pauseFlow ] ERROR: send not successful\n")));
					return -1;
					//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::pauseFlow ] ERROR: send not successful\n")), -1);
				}
			}
		}
	}
	
	// Update modifications
	LastModifications *lm = LAST_MODIFICATIONS::instance(); 
	lm->addModifications(xml.c_str());
	
	return 0;
}

// update all servers (Announce)
int Peer_Event_Handler::restartFlow(ACE_TString xml) {
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::restartFlow"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::restartFlow]\n")));
	XML_Parser parser(xml);
	
	ACE_TString clientName = (parser.getInfos(ACE_TEXT("clientname"))).c_str();
	ACE_TString clientGroup = (parser.getInfos(ACE_TEXT("clientgroup"))).c_str();
	ACE_TString clientID = clientGroup + ACE_TEXT(":") + clientName;
	
	ACE_TString flowType = (parser.getInfos(ACE_TEXT("flowtype"))).c_str();
	ACE_TString flowGroup = (parser.getInfos(ACE_TEXT("flowgroup"))).c_str();
	ACE_TString flowID = flowGroup + ACE_TEXT(":") + flowType;
	
	//ACE_SOCK_Stream *dupli = NULL;
	Duplicator_Event_Handler *dupli = NULL;
	int result;
	//int buffer_size = dm.get_buffer_size();
	
	GlobalList *gl = GLOBAL_LIST::instance();
	DuplicatorList *dl = DUPLICATOR_LIST::instance();  
	ServerList *serverList = SERVER_LIST::instance();  
	
	
	
	if ( gl->findClient(clientID) != 0 )
    {
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("Client not found !"));
		send_xml(createxmlfailure.c_str());
		return 0;
    }
	
	
	
	Flow *f;
	f = gl->getFlow(flowID);
	if (f == NULL) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::restartFlow ] flow = NULL \n") ));
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::pauseFlow ] flow == NULL \n"));
		send_xml(createxmlfailure.c_str());
		return -1;
	}
	
	// smart detection of the duplicator provider
	
	ACE_TCHAR namelocal[256];
	if (ACE_OS::hostname(namelocal,256) ==1)
    {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::restartFlow ] ACE_OS::hostname \n")));
		ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("Peer_Handler restartFlow : ACE_OS::hostname"));
		send_xml(createxmlfailure.c_str());
		return 0;
    }
	ACE_TString localaddr = namelocal ;
	ACE_TString serverrequest = f->getServerProducerID() ;
	ACE_TString serverclientrequest = gl->getClient(clientID)->getServer();
	
	
	
	if (serverrequest == localaddr) {
		ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::restartFlow ] Server of remote duplicator detected  \n")));
		
		if ( (f->countSwitchedOffConsumersOnServer(serverclientrequest) > 0) && (f->countSwitchedOnConsumersOnServer(serverclientrequest) == 0) ) { 
			
			//ACE_TCHAR wakeUpDupliRequestMessage[MAX_MESSAGE_SIZE]={0};
			size_t size_needed = sizeof(WAKE_UP_DUPLI_MESSAGE) + (serverList->get(serverclientrequest.c_str(), ServerList::NAME)->getIP()).length() + 1; 
			ACE_TCHAR *wakeUpDupliRequestMessage = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
			ACE_OS::memset(wakeUpDupliRequestMessage, 0, size_needed);
			ACE_OS::sprintf(wakeUpDupliRequestMessage, WAKE_UP_DUPLI_MESSAGE, (serverList->get(serverclientrequest.c_str(), ServerList::NAME)->getIP()).c_str()); 
			
			dupli = dl->getDupEventHandler(f->getID());
			if ( dupli == NULL ) {
				ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::restartFlow ] ERROR: dupli==NULL\n")));
				return -1;
				//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::restartFlow ] ERROR: dupli==NULL\n")), -1);
			}
			//result = dm.send_message(dupli, wakeUpDupliRequestMessage);
			result = dupli->send_xml(wakeUpDupliRequestMessage);
			ACE_OS::free(wakeUpDupliRequestMessage);
			if (result == -1) {
				//if (result < buffer_size) {
				ACE_TString createxmlfailure = create_xmlfailure(ACE_TEXT("[ Peer_Event_Handler::restartFlow ] flow == NULL \n"));
				send_xml(createxmlfailure.c_str());
				ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::restartFlow ] ERROR: send not successful\n")));
				return -1;
				//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::restartFlow ] ERROR: send not successful\n")), -1);
			}      
		}
	}
	
	
	// update
	if (f->setConsumerStatus(clientID, CLIENT_STATUS_SWITCHED_ON) == -1 ) 
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::restart_flow ] error set consumer status \n")));
	if (f->setStatus(FLOW_STATUS_RUNNING) == -1)
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::restart_flow ] error set flow status \n")));
	if (f->setProducerStatus(CLIENT_STATUS_SWITCHED_ON) == -1 )
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::restart_flow ] error set producer status\n")));
	
	
	// Update modifications
	LastModifications *lm = LAST_MODIFICATIONS::instance(); 
	lm->addModifications(xml.c_str());
    
	return 0;
}

int Peer_Event_Handler::notify_dupli_to_wakup_host(ACE_TString xml) {
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::notify_dupli_to_wakup_host"));
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::notify_dupli_to_wakup_host ]\n")));
	XML_Parser parser(xml);
	
	DuplicatorList *dl = DUPLICATOR_LIST::instance(); 
	GlobalList *gl = GLOBAL_LIST::instance();
	ServerList *serverList = SERVER_LIST::instance();
	
	ACE_TString host_name = (parser.getInfos(ACE_TEXT("hostname"))).c_str();
	ACE_TString flowID = (parser.getInfos(ACE_TEXT("flowid"))).c_str();
	
	//ACE_TCHAR host_name_message[MAX_MESSAGE_SIZE]={0};
	size_t size_needed = sizeof(WAKE_UP_DUPLI_MESSAGE) + (serverList->get(host_name.c_str(), ServerList::NAME)->getIP()).length() + 1; 
	ACE_TCHAR *host_name_message = (ACE_TCHAR *) ACE_OS::malloc(size_needed);
	ACE_OS::memset(host_name_message, 0, size_needed);
	ACE_OS::sprintf(host_name_message,WAKE_UP_DUPLI_MESSAGE, serverList->get(host_name.c_str(), ServerList::NAME)->getIP().c_str());
	
	//ACE_SOCK_Stream *dupli = NULL;
	Duplicator_Event_Handler *dupli = NULL;
	int result;
	//int buffer_size = dm.get_buffer_size();
	
	Flow *f;
	f = gl->getFlow(flowID);
	if (f == NULL) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notify_dupli_to_wakup_host ] flow = NULL \n") ));
		return -1;
	}
	
	dupli = dl->getDupEventHandler(f->getID());
	if ( dupli == NULL ) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notify_dupli_to_wakup_host ] ERROR: dupli==NULL\n")));
		return -1;
		//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::notify_dupli_to_wakup_host ] ERROR: dupli==NULL\n")), -1);
	}
	result = dupli->send_xml(host_name_message);
	ACE_OS::free(host_name_message);
	//if (result < buffer_size)
	if (result == -1) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::notify_dupli_to_wakup_host ] ERROR: send not successful\n")));
		return -1;
		//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::notify_dupli_to_wakup_host ] ERROR: send not successful\n")), -1);
	}
	
	return 0;
}

int Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator(ACE_TString xml) {
	ACE_TRACE(ACE_TEXT("Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator"));
	//the server need to store this information somewhere and broadcast it if necessary
	
	ACE_DEBUG((SF2D_DEBUG_PREFIX ACE_TEXT("[ Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator ]\n")));
    
	//first we get the infos from the XML file
	XML_Parser parser(xml);
	
	ACE_TString clientID = (parser.getInfos(ACE_TEXT("clientID"))).c_str();
	ACE_TString flowID = (parser.getInfos(ACE_TEXT("flowID"))).c_str();
	
	//then we check the info are correct
	if (  flowID.length() == 0 )
	{
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator ] flowID.length==0\n")));	
		return -1;	
	}	
	if (  clientID.length() == 0 )
	{
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator ] clientID.length==0\n")));	
		return -1;	
	}	
	
	//then we update the tables
	//First we find the flow
	GlobalList *gl = GLOBAL_LIST::instance();
	if ( !gl ) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator ] GLOBAL_LIST::instance()==NULL\n")));	
		return -1;	
	}
	
	if ( gl->findFlow(flowID) == -1 ) {
		ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator ] Flow Not Found\n")));	
		return -1;
	}
	
	Flow *tmp_flow = gl->getFlow(flowID);
	if ( tmp_flow ) {
		if ( tmp_flow->containsConsumer(clientID) == -1 ) {
			ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator ] Consumer Not Found\n")));	
			return -1;
		}
		else {
			//if we are here, the flow and the consumer exists in the list
			tmp_flow->setConsumerConnectionToDuplicatorVerified(clientID, 1);
			//the table of our local server is now up-to-date
			
			//HERE WE NEED TO CHECK IF THE FLOW IS SYNCHRONIZED AND IF WE HAVE TO START THE TRANSFER
			if ( ( tmp_flow->getNumberOfExpectedConsumers() > 0 ) && ( tmp_flow->isSynchronizedFlowMustBeStarted() == 0 ) ) {	
				DuplicatorList *dl = DUPLICATOR_LIST::instance();
				Duplicator_Event_Handler *dupli = NULL;
				dupli = dl->getDupEventHandler(tmp_flow->getID());
				if ( dupli == NULL ) {
					ACE_DEBUG((SF2D_ERROR_PREFIX ACE_TEXT("[ Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator ] ERROR: dupli==NULL\n")));
					return -1;
					//ACE_ERROR_RETURN((LM_ERROR, ACE_TEXT("[ Peer_Event_Handler::announceNewConsumerNodeConnectedToDuplicator ] ERROR: dupli==NULL\n")), -1);	
				}
				dupli->notifyDuplicatorToStartDataTransfer();
			}
			//the job is done :)
			return 0;
		}
	}
	//If we are that means something wrong...
	return -1;
}
