#ifndef _CLIENT_H_
#define _CLIENT_H_

#include "ace/ACE.h"
#include "ace/SString.h"
#include "ace/Log_Msg.h"
#include "ace/Containers_T.h"

/// This class describes a client object.

class Client
{	
	protected:
		int init(ACE_TString group_, ACE_TString name_);
		int init(ACE_TString xml_);
		int fini();
		
		/// ID of the client
		ACE_TString id;

		/// Group the client belongs to, cannot be changed after the initialization
		ACE_TString group;
		
		/// Name of the client, cannot be changed after the initialization
		ACE_TString name;
		
		/// Description of the client
		ACE_TString description;
		
		/// DNS name of the server the client is connected to
		ACE_TString server;
		
		/// Path of the client 
		ACE_TString path;
		
		/// Parameters the client has been launched with
		ACE_TString params;
		
		/// Infos about the client 
		ACE_TString flowinfo;
		
		

	public:
		
		/// If the group of the client is "admin", this list will be filled with modifications on the system
		ACE_Unbounded_Queue<ACE_TString> modifications;
		
		/// Create a client giving a group and a name
		Client(ACE_TString group_, ACE_TString name_);
		
		/// Create a client giving an Xml string
		Client(ACE_TString xml_);
		
		/// Destroy the client
		virtual ~Client();

		/// Return the ID of a client given its group and name
		
		/**
		 * @param group_ group of the client
 		 * @param name_ name of the client
		 * @return the ID of the client
		 */
		static ACE_TString getIDFromGroupAndName(ACE_TString group_, ACE_TString name_);

		/// Set the description of the client
		
		/**
		 * @param description_ description of the client
		 * @return 0 if the description was set succesfully, and -1 if not
		 */
		int setDescription(ACE_TString description_);

		/// Set the server the client connects to
		
		/**
		 * @param server_ DNS name of the server the client connects to
		 * @return 0 if the server was set succesfully, and -1 if not
		 */
		int setServer(ACE_TString server_);


		/// Set the path of the client
		
		/**
		 * @param Path of the client 
		 * @return 0 if the path was set succesfully, and -1 if not
		 */
		int setPath(ACE_TString path_);


		/// Set the parameters for the client
		
		/**
		 * @param params_ parameters the client has been launched with
		 * @return 0 if the parameters were set succesfully, and -1 if not
		 */
		int setParams(ACE_TString params_);

	
		/// Set information about the flows the client will use
		
		/**
		 * @param flowinfo_ information about the flows the client will use
		 * @return 0 if the parameters were set succesfully, and -1 if not
		 */
		int setFlowInfo(ACE_TString flowinfo_);


		/// Get the ID of the client
		
		/**
		 * @return the ID of the client
		 */
		ACE_TString getID();

		/// Get the group of the client
		
		/**
		 * @return the group of the client
		 */
		ACE_TString getGroup();


		/// Get the path of the client
		
		/**
		 * @return the path of the client
		 */
		ACE_TString getPath();
		

		/// Get the name of the client
		
		/**
		 * @return the name of the client
		 */
		ACE_TString getName();

		/// Get the description of the client
		
		/**
		 * @return the description of the client
		 */
		ACE_TString getDescription();

		/// Get the DNS name of the server the client connects to
		
		/**
		 * @return the DNS name of the server the client connects to
		 */
		ACE_TString getServer();

		/// Get the parameters of the client
		
		/**
		 * @return the parameters of the client
		 */
		ACE_TString getParams();


		/// Get the flows the client will use
		
		/**
		 * @return the flows the client will use
		 */
		ACE_TString getFlowInfo();


		/// Return the client represented as an Xml string
		
		/**
		 * @return the client represented as an Xml string
		 */	
		ACE_TString toXml();

		/// Return the client represented as an Xml string shifted at the specified offset
		
		/**
		 * @param n_ the offset of the shift
		 * @return the client represented as an Xml string shifted at the specified offset
		 */
		ACE_TString toXmlWithIndent(int n_);
		
		
		/// Add a modification to the modifications list
		
		/**
		 * @return 0 if success 
		 */	
		int addModifications(ACE_TString modif_);

		/// Clear the modifications list
		/**
		 * @return 0 if success 
		 */		

		int clear();	
};

#endif //_CLIENT_H_
