#ifndef _DUPLICATORLIST_H_
#define _DUPLICATORLIST_H_

#include "ace/ACE.h"
#include "ace/SString.h"
#include "ace/Log_Msg.h"
#include "ace/Singleton.h"
#include "ace/Null_Mutex.h"
#include "ace/SOCK_Stream.h"
#include "ace/SOCK_Acceptor.h"
#include "ace/Hash_Map_Manager_T.h"

class Duplicator_Event_Handler;

struct DuplicatorInfo
{
	DuplicatorInfo();
	// socket the server has to use to talk with the duplicator
	//ACE_SOCK_Stream *serverSocket;
	Duplicator_Event_Handler *dupli;
	// boolean set to true if the duplicator has been spawned with the "noprovider" 
	// argument and that the provider flow information has reached the local server
	// before the newly spawned duplicator has connected to the local server.
	bool mustConnectToDistantDupli;
	// boolean set to true if the duplicator has been spawned with the "noprovider" 
	// argument and that a local provider has connected to the local server before 
	// the newly spawned duplicator has connected to the local server.
	bool mustOpenProviderPart;
};

typedef ACE_Hash_Map_Manager_Ex<ACE_TString, DuplicatorInfo *, ACE_Hash<ACE_TString>, ACE_Equal_To<ACE_TString>, ACE_Null_Mutex> DUPLICATOR_MAP;
typedef ACE_Hash_Map_Iterator_Ex<ACE_TString, DuplicatorInfo *, ACE_Hash<ACE_TString>, ACE_Equal_To<ACE_TString>, ACE_Null_Mutex> DUPLICATOR_ITERATOR;

/// This class implements a list which associates a flow ID with the port number
/// the clients have to use to subscribe to this flow, and the socket the server
/// uses to communicate with the duplicator

/**
 * This class implements the design pattern Singleton.
 * In order to make sure that only the singleton instance is created,
 * and that users cannot create their own instances,
 * the constructor is protected and the singleton is a friend of this class
 */

class DuplicatorList
{
	friend class ACE_Singleton<DuplicatorList, ACE_Null_Mutex>;
	
	protected:
		int init();//u_short minPortNumber_);
		int fini();
		
		unsigned int minPortNumber;
		unsigned int currentPortNumber;

		/// Add a flow to the list

		/**
		 * This methods adds a flow to the list, and is used by addProducer
		 * and addConsumer. addProducer will automatically find an available
		 * port number, whereas addConsumer will set the port number to 0.
		 * @param flowID_ ID of the flow to add
		 * @return 0 if the flow was added successfully, -1 if not
		 */
		int add(ACE_TString flowID_);
		
		/// Return a XML description of the duplicator
		
		/**
		 * This methods returns a XML description of the duplicator
		 * @param flowID_ ID of the flow to add
		 * @return 0 if the flow was added successfully, -1 if not
		 */
		ACE_TString dupInfoToXmlWithIndent(int n_, ACE_TString dup_id_, ACE_TString server_name_);

		/// Internal map used to store the duplicators
		DUPLICATOR_MAP duplicators;
		
	public:
		/// Create an empty duplicator list
		DuplicatorList();//u_short minPort_);
		
		/// Destroy the duplicator list
		virtual ~DuplicatorList();

		/// Add a new producer flow to the list

		/**
		 * The flow will be added to the list and associated with a new
		 * available port.
		 * This method does not take care of initializing a socket listening on
		 * the available port associated with the flow.
		 * @param flowID_ ID of the flow to add
		 * @return 0 if the flow was added succesfully, 1 if the flow already
		 * exists in the list, and -1 if it was not added
		 */
		int addProducer(ACE_TString flowID_);

		/// Add a new consumer flow to the list

		/**
		 * The flow will be added to the list and associated 0 as port in order
		 * to know this is a consumer.
		 * @param flowID_ ID of the flow to add
		 * @return 0 if the flow was added succesfully, 1 if the flow already
		 * exists in the list, and -1 if it was not added
		 */
		int addConsumer(ACE_TString flowID_);

		/// Remove a flow from the list

		/**
		 * The flow and its associated port number will be removed from the
		 * list.
		 * This method does not take care of closing the socket listening on the
		 * port associated with the flow.
		 * @param flowID_ ID of the flow to remove
		 * @return 0 if the flow was removed succesfully, and -1 if it was not
		 * removed
		 */
		int remove(ACE_TString flowID_);
		
		/// Check if a flow is in the list giving its ID
		
		/**
		 * @param flowID_ ID of the flow to check
		 * @return 0 if the flow is in the list, and -1 if not
		 */
		int find(ACE_TString flowID_);

		/// Set the Duplicator_Event_Handler associated to the duplicator producing the flow
		
		/**
		 * @param flowID_ ID of the flow
		 * @param handler Duplicator_Event_Handler which the server will use to talk with the duplicator
		 * @return 0 if success, -1 if not
		 */
		int setDupEventHandler(ACE_TString flowID_, Duplicator_Event_Handler *handler);

		/// Get the Duplicator_Event_Handler associated to the duplicator producing the flow
	
		/**
		 * @param flowID_ ID of the flow
		 * @return a pointer to Duplicator_Event_Handler if success, NULL if not
		 */
		Duplicator_Event_Handler* getDupEventHandler(ACE_TString flowID_);

		/**
		 * @param flowID_ ID of the flow
		 * @param mustConnectToDistantDupli boolean
		 * @return 0 if success, -1 if not
		 */
		int setMustConnectToDistantDupli(ACE_TString flowID_, bool mustConnectToDistantDupli);

		/**
		 * @param flowID_ ID of the flow
		 * @return a the boolean associated with the flowID_ if success, false if not
		 */

		bool mustConnectToDistantDupli(ACE_TString flowID_);

		/**
		 * @param flowID_ ID of the flow
		 * @param mustOpenProviderPart boolean
		 * @return 0 if success, -1 if not
		 */
		int setMustOpenProviderPart(ACE_TString flowID_, bool mustOpenProviderPart);

		/**
		 * @param flowID_ ID of the flow
		 * @return a the boolean associated with the flowID_ if success, false if not
		 */

		bool mustOpenProviderPart(ACE_TString flowID_);

		/// Return the first available port number

		/**
		 * This methods returns the next port that is supposed to be available 
		 * @return the next suppositely available port number
		 */
		unsigned int nextPortSupposedToBeAvailable();

		/**
		 * This methods defines the next suppositely available port
		 * @param portUsed the next suppositely available port number
		 */
		void setNextAvailablePort(unsigned int portUsed);

		/// Count the number of flows in the list

		/**
		 * @return the number of flows in the list
		 */
		int count();

		/// Return an iterator on the first flow of the list
		
		/**
		 * @return iterator on the first flow of the list
		 */
		DUPLICATOR_ITERATOR first();

		/// Return an iterator on the last flow of the list
		
		/**
		 * @return iterator on the last flow of the list
		 */
		DUPLICATOR_ITERATOR last();

		/// Return the duplicator list represented as an Xml string
		
		/**
		 * @return the duplicator list represented as an Xml string
		 */
		ACE_TString toXml();

		/// Return the duplicator list represented as an Xml string shifted at the specified offset
		
		/**
		 * @param n_ the offset of the shift
		 * @return the duplicator list represented as an Xml string shifted at the specified offset
		 */
		ACE_TString toXmlWithIndent(int n_);
		
		/// Return the Duplicator_Event_Handler of the entry pointed by the iterator
		
		/**
		 * @param duplicatorIterator_ the duplicator iterator
		 * @return the a pointer to the Duplicator_Event_Handler
		 */	
		static Duplicator_Event_Handler* getDupEventHandlerFromIterator(DUPLICATOR_ITERATOR duplicatorIterator_);

		/// Return the ID of the flow of the entry pointed by the iterator
		
		/**
		 * @param duplicatorIterator_ the duplicator iterator
		 * @return the ID of the flow
		 */
		static ACE_TString getIDFromIterator(DUPLICATOR_ITERATOR duplicatorIterator_);
};

typedef ACE_Singleton<DuplicatorList, ACE_Null_Mutex> DUPLICATOR_LIST;

#endif //_DUPLICATORLIST_H_
