#ifndef _FLOW_H_
#define _FLOW_H_

#include "ace/ACE.h"
#include "ace/SString.h"
#include "ace/Log_Msg.h"
#include "ace/Null_Mutex.h"
#include "ace/OS_NS_stdlib.h"
#include "ace/Hash_Map_Manager_T.h"

#include "sf2d_common_ACE.h"

struct ClientInfo
{
  ClientStatus status;
  ACE_TString givenFlowName;
  int connectionToDuplicatorValidated;
};

typedef ACE_Hash_Map_Manager_Ex<ACE_TString, ClientInfo *, ACE_Hash<ACE_TString>, ACE_Equal_To<ACE_TString>, ACE_Null_Mutex> CONSUMER_LIST;
typedef ACE_Hash_Map_Iterator_Ex<ACE_TString, ClientInfo *, ACE_Hash<ACE_TString>, ACE_Equal_To<ACE_TString>, ACE_Null_Mutex> CONSUMER_ITERATOR;


/// This class describes a flow object.

class Flow
{
	protected:
		int init(ACE_TString group_, ACE_TString type_);
		int init(ACE_TString xml_);
		int fini();

		/// ID of the flow
		ACE_TString id;
	
		/// Group of the flow, cannot be changed after the initialization
		ACE_TString group;

		/// Type of the flow, cannot be changed after the initialization
		ACE_TString type;

		/// Name of the flow
		ACE_TString name;
	
		/// Description of the flow
		ACE_TString description;

		/// Port number which is used by the producer to provide the flow
		u_short portNumber;
		
		/// Size of the blocks
		u_short blockSize;
		
		/// Number of blocks
		u_short blockCount;
		
		/// Status of the flow
		FlowStatus status;

		/// ID of the server of the producer of the flow
		ACE_TString serverProducerID;

		/// ID of the client producing the flow
		ACE_TString producerID;

		/// Status of the client producing the flow
		ClientStatus producerStatus;
		
		/// List of the IDs and statuses of the clients consuming the flow
		CONSUMER_LIST consumerList;

		///The number of consumer expected, 0 means that the flow is not synchronized
		///if nbConsumersExpected > 0, that means the flow is synchronized and the data
		///transfer will begin when nbConsumersExpected consumers will connect to the flow
		unsigned int nbConsumersExpected;

		//this information is only useful when we handle a synchronized flow. This field will
		//never be broadcasted to other servers
		unsigned int nbRemoteConsumersConnectedToLocalDuplicator;

	public:
		/// Create a flow giving a group and a type
		Flow(ACE_TString group_, ACE_TString type_);
		
		/// Create a flow giving an Xml string
		Flow(ACE_TString xml_);

		/// Destroy the flow
		virtual ~Flow();

		/// Return the ID of a flow given its group and type
		
		/**
		 * @param group_ group of the flow
 		 * @param type_ type of the flow
		 * @return the ID of the flow
		 */
		static ACE_TString getIDFromGroupAndType(ACE_TString group_, ACE_TString type_);
		
		/// Set the name of the flow
		
		/**
		 * @param name_ name of the flow
		 * @return 0 if the name was set succesfully, and -1 if not
		 */
		int setName(ACE_TString name_);
		
		/// Set the description of the flow
		
		/**
		 * @param description_ description of the flow
		 * @return 0 if the description was set succesfully, and -1 if not
		 */
		int setDescription(ACE_TString description_);
		
		/// Set the port number the producer is using to provide the flow
		
		/**
		 * @param portNumber_ port number of the flow
		 * @return 0 if the port number was set succesfully, and -1 if not
		 */
		int setPortNumber(u_short portNumber_);
		
		/// Set the size of blocks of the flow
		
		/**
		 * @param blockSize_ size of blocks of the flow
		 * @return 0 if the size of blocks was set succesfully, and -1 if not
		 */
		int setBlockSize(u_short blockSize_);
		
		/// Set the number of blocks of the flow
		
		/**
		 * @param blockCount_ number of blocks of the flow
		 * @return 0 if the number of blocks was set succesfully, and -1 if not
		 */
		int setBlockCount(u_short blockCount_);

		/// Set the status of the flow
		
		/**
		 * @param status_ status of the flow
		 * @return 0 if the status was set succesfully, and -1 if not
		 */
		int setStatus(FlowStatus status_);
		
		///Set the number of expected consumers.

		/**
		 * @param nbConsumersExpected_ the minimum number of consumers, which should connect to the flow for the data transfer to begin
		 * @return 0 if the number of consumers was set successfully, and -1 otherwise.
		 */
		int setNumberOfExpectedConsumers(unsigned int nbConsumersExpected_);
		
		///Set the number of remote duplicator connected to a local duplicator handling a synchronized flow.
		/**
		 * @param nbConsumersExpected_ the minimum number of consumers, which should connect to the flow for the data transfer to begin
		 * @return 0 if the number of consumers was set successfully, and -1 otherwise.
		 */
		int setNumberOfRemoteConsumersConnectedToLocalDuplicator(unsigned int nbRemoteConsumers);

		///Return the number of remote duplicator connected to a local duplicator handling a synchronized flow.
		/**
		 * @return the number of remote duplicator connected to a local duplicator handling a synchronized flow.
		 */
		unsigned int getNumberOfRemoteConsumersConnectedToLocalDuplicator();

		/// Check if the data transfer of the synchronized flow must be started.
		/**
		* @return 0 for Success, -1 otherwise
		*/  
		int isSynchronizedFlowMustBeStarted();

		/// Get the ID of the flow
		
		/**
		 * @return the ID of the flow
		 */
		ACE_TString getID();

		/// Get the group of the flow
		
		/**
		 * @return the group of the flow
		 */
		ACE_TString getGroup();

		/// Get the type of the flow
		
		/**
		 * @return the type of the flow
		 */
		ACE_TString getType();

		/// Get the name of the flow
		
		/**
		 * @return the name of the flow
		 */
		ACE_TString getName();

		/// Get the description of the flow
		
		/**
		 * @return the description of the flow
		 */
		ACE_TString getDescription();

		/// Get the port of the flow
		
		/**
		 * @return the port number of the flow
		 */
		u_short getPortNumber();

		/// Get the size of the blocks of the flow
		
		/**
		 * @return the size of the blocks of the flow
		 */
		u_short getBlockSize();

		/// Get the number of blocks of the flow
		
		/**
		 * @return the number of blocks of the flow
		 */
		u_short getBlockCount();

		/// Get the status of the flow
		
		/**
		 * @return the status of the flow
		 */
		FlowStatus getStatus();

		///Get the number of expected consumers.

		/**
		 * @return the number of consumers, which need to connect to the flow before the data transfer starts.
		 */
	    unsigned int getNumberOfExpectedConsumers();

		/// Get the ID of the producer of the flow
		
		/**
		 * @return the ID of the producer of the flow
		 */
		ACE_TString getProducer();

		/// Get the ID of the server of producer of the flow
		
		/**
		 * @return the ID of the server of the producer of the flow
		 */

		ACE_TString getServerProducerID();


		/// Set the server of the producer of the flow
		
		/**
		 * @param serverID_ ID of the server of the producer of the flow
		 * @return 0 if the ID of the producer was set succesfully, and -1 if not
		 */
		int setServerProducerID(ACE_TString producerID_);

		/// Set the producer of the flow
		
		/**
		 * @param producerID_ ID of the producer of the flow
		 * @return 0 if the ID of the producer was set succesfully, and -1 if not
		 */
		int setProducer(ACE_TString producerID_);

		/// Unset the producer of the flow
		
		/**
		 * @return 0 if the producer was unset succesfully, and -1 if not
		 */
		int unsetProducer();


		/// Set the status of the producer concerning the flow
		
		/**
		 * @param producerStatus_ status of the producer
		 * @return 0 if success, -1 if failure
		 */
		int setProducerStatus(ClientStatus producerStatus_);

		/// Get the status of the producer concerning the flow
		
		/**
		 * @return status of the producer concerning the flow
		 */
		ClientStatus getProducerStatus();

		/// Add a consumer to the flow
		
		/**
		 * @param consumerID_ ID of the consumer to add
		 * @param givenFlowName_ name of the flow given by the client
		 * @return 0 if success, -1 if failure
		 */
		int addConsumer(ACE_TString consumerID_, ACE_TString givenFlowName_);

		/// Remove a consumer of the flow
		
		/**
		 * @param consumerID_ ID of the consumer to remove
		 * @return 0 if success, -1 if failure
		 */
		int removeConsumer(ACE_TString consumerID_);

		/// Check if a client is a consumer of the flow
		
		/**
		 * @param consumerID_ ID of the consumer to check
		 * @return 0 if success, -1 if failure
		 */
		int containsConsumer(ACE_TString consumerID_);

		/// Count the number of consumers of the flow
		
		/**
		 * @return the number of consumers of the flow
		 */
		int countConsumers();
		
		// By Imad
		int countSwitchedOffConsumers();
		int countSwitchedOnConsumers();
		int countSwitchedOffConsumersOnServer(ACE_TString serverID_);
		int countSwitchedOnConsumersOnServer( ACE_TString serverID_ );
		


		/// Count the number of consumers of the flow on a given server
		
		/**
		 * @return the number of consumers of the flow
		 */
		int countConsumersOnServer( ACE_TString serverID_ );

		/// Count the number of server handling the flow (servers, where there is at least a client node, which subscribes to the flow)
		/**
		 * @return the number of server handling the flow
		 */
		unsigned int countServersHandlingFlow();

		/// Return an iterator on the first consumer of the flow
		
		/**
		 * @return iterator on the first consumer of the flow
		 */
		CONSUMER_ITERATOR firstConsumer();

		/// Return an iterator on the last consumer of the flow
		
		/**
		 * @return iterator on the last consumer of the flow
		 */
		CONSUMER_ITERATOR lastConsumer();

		/// Return the ID of the consumer of the entry pointed by the iterator
		
		/**
		 * @param consumerIterator_ the consumer iterator
		 * @return the ID of the consumer
		 */	
		static ACE_TString getConsumerIDFromIterator(CONSUMER_ITERATOR consumerIterator_);

		/// Return the status of the consumer of the entry pointed by the iterator
		
		/**
		 * @param consumerIterator_ the consumer iterator
		 * @return the status of the consumer
		 */	
		static ClientStatus getConsumerStatusFromIterator(CONSUMER_ITERATOR consumerIterator_);

		/// Return the name of the flow given by the consumer of the entry pointed by the iterator
		
		/**
		 * @param consumerIterator_ the consumer iterator
		 * @return the name of the flow
		 */	

		static ACE_TString getConsumerGivenFlowNameFromIterator(CONSUMER_ITERATOR consumerIterator_);

		//Check if the the connection between the consumer given by the entry pointed by the iterator and the duplicator has been verified
		/**
		 * @param consumerID_ ID of the consumer
		 * @return 1 if the connection between the duplicator and the consumer has been confirmed, 0 otherwise
		 */
		static int isConsumerConnectionToDuplicatorVerifiedFromIterator(CONSUMER_ITERATOR consumerIterator_);

		/// Set the status of a consumer concerning the flow
		
		/**
		 * @param consumerID_ ID of the consumer
		 * @param consumerStatus_ status of the consumer
		 * @return 0 if success, -1 if failure
		 */
		int setConsumerStatus(ACE_TString consumerID_, ClientStatus consumerStatus_);

		/// Set the name of the flow given by the consumer
		
		/**
		 * @param consumerID_ ID of the consumer
		 * @param givenFlowName_ name of the flow given by the consumer
		 * @return 0 if success, -1 if failure
		 */
		int setConsumerGivenFlowName(ACE_TString consumerID_, ACE_TString givenFlowName_);

		///Set the validation of the connection between the duplicator and the consumer client 
		/**
		 * @param consumerID_ ID of the consumer
		 * @param givenFlowName_ name of the flow given by the consumer
		 * @return 0 if success, -1 if failure
		 */
		int setConsumerConnectionToDuplicatorVerified(ACE_TString consumerID_, int connectionOK);
		
		/// Get the status of a consumer concerning the flow
		
		/**
		 * @param consumerID_ ID of the consumer
		 * @return status of the consumer concerning the flow
		 */
		ClientStatus getConsumerStatus(ACE_TString consumerID_);

		/// Get the name of the flow given by the consumer
		
		/**
		 * @param consumerID_ ID of the consumer
		 * @return the name of the flow
		 */	
		ACE_TString getConsumerGivenFlowName(ACE_TString consumerID_);
		
		//Check if the the connection between the consumer and the duplicator has been verified
		/**
		 * @param consumerID_ ID of the consumer
		 * @return true if the connection between the duplicator and the consumer has been confirmed, false otherwise
		 */

		int isConsumerConnectionToDuplicatorVerified(ACE_TString consumerID_);

		/// Update the flow (producer information) giving its xml representation
		/// Used for server synchronization
		
		/**
		 * @param flowXML_ XML representation of the flow
		 * @return 0 if the flow was updated succesfully, -1 if it is wrong flow
		 */
		int updateProducer(ACE_TString flowXML_);
		
		/// Update the flow (consumers information) giving its xml representation
		/// Used for server synchronization

		/**
		 * @param flowXML_ XML representation of the flow
		 * @return 0 if the flow was updated succesfully, -1 if it is wrong flow
		 */
		int updateConsumers(ACE_TString flowXML_);

		/// Return the flow represented as an Xml string
		
		/**
		 * @return the flow represented as an Xml string
		 */
		ACE_TString toXml();

		/// Return the flow represented as an Xml string shifted at the specified offset
		
		/**
		 * @param n_ the offset of the shift
		 * @return the flow represented as an Xml string shifted at the specified offset
		 */
		ACE_TString toXmlWithIndent(int n_);

		/// Return the flow represented as an Xml string without the consumerlist 
		/// Used for the synchronisation

		/**
		 * @param n_ the offset of the shift
		 * @return the flow represented as an Xml string without the consumerList
		 */
		ACE_TString toXmlWithoutConsumers(int n_);

		/// Return the flow represented as an Xml string without the producer
		/// Used for the synchronisation

		/**
		 * @param n_ the offset of the shift
		 * @return the flow represented as an Xml string without the producer
		 */
		ACE_TString toXmlWithoutProducer(int n_);
};

#endif //_FLOW_H_
