#ifndef _FLOWLIST_H_
#define _FLOWLIST_H_

#include "ace/ACE.h"
#include "ace/SString.h"
#include "ace/Log_Msg.h"
#include "ace/Null_Mutex.h"
#include "ace/Hash_Map_Manager_T.h"

#include "Flow.h"

typedef ACE_Hash_Map_Manager_Ex<ACE_TString, Flow *, ACE_Hash<ACE_TString>, ACE_Equal_To<ACE_TString>, ACE_Null_Mutex> FLOW_MAP;
typedef ACE_Hash_Map_Iterator_Ex<ACE_TString, Flow *, ACE_Hash<ACE_TString>, ACE_Equal_To<ACE_TString>, ACE_Null_Mutex> FLOW_ITERATOR;

/// This class describes a flow list object.

class FlowList
{
	protected:
		int init();
		//int init(ACE_TString xml_);
		int fini();

		/// Internal map used to store data
		FLOW_MAP flows;
		
	public:
		/// Create an empty flow list
		FlowList();
		
		/// Create a flow list giving an Xml string
		//FlowList(ACE_TString xml_);

		/// Destroy the flow list
		virtual ~FlowList();
		
		/// Add a flow to the flow list
		
		/**
		 * @param flow_ pointer to the flow to add
		 * @return 0 if the flow was added succesfully, 1 if the flow already
		 * exists in the list, and -1 if it was not added
		 */
		int add(Flow *flow_);
		
		/// Remove a flow from the flow list
		
		/**
		 * @param flowID_ ID of the flow to remove
		 * @return 0 if the flow was removed succesfully, -1 if not
		 */
		int remove(ACE_TString flowID_);
		
		/// Check if a flow is in the flow list giving its ID
		
		/**
		 * @param flowID_ ID of the flow to check
		 * @return 0 if the flow is in the list, and -1 if not
		 */
		int find(ACE_TString flowID_);

		/// Check if a flow is in the flow list giving its group and type
		
		/**
		 * @param group_ group of the flow to check
		 * @param type_ type of the flow to check
		 * @return 0 if the flow is in the list, and -1 if not
		 */
		int find(ACE_TString group_, ACE_TString type_);

		/// Check if a flow is in the list giving its xml representation
		/// Used for server synchronization

		/**
		 * @param flowXML_ XML representation of the flow to check
		 * @return 0 if the flow is in the list, and -1 if not
		 */
		int contains(ACE_TString flowXML_);

		/// Update a flow (information regarding producer) in the list giving its xml representation
		/// Used for server synchronization
		
		/**
		 * @param flowXML_ XML representation of the flow to update
		 * @return 0 if the flow was updated succesfully, -1 if the flow wasn't found in the list
		 */
		int updateProducerOfFlow(ACE_TString flowXML_);		

		/// Update a flow (information regarding consumers) in the list giving its xml representation
		/// Used for server synchronization
		
		/**
		 * @param flowXML_ XML representation of the flow to update
		 * @return 0 if the flow was updated succesfully, -1 if the flow wasn't found in the list
		 */
		int updateConsumersOfFlow(ACE_TString flowXML_);	

		/**
		 * @param flowID_ ID of the flow to get
		 * @return a pointer to the flow, null pointer if flow is not in the flow list
		 */

		/// Return a pointer to a flow giving its ID

		Flow *get(ACE_TString flowID_);

		/// Count the number of flows in the flow list
		
		/**
		 * @return the number of flows in the flow list
		 */
		int count();

		/// Return an iterator on the first flow of the flow list
		
		/**
		 * @return iterator on the first flow of the flow list
		 */
		FLOW_ITERATOR first();

		/// Return an iterator on the last flow of the flow list
		
		/**
		 * @return iterator on the last flow of the flow list
		 */
		FLOW_ITERATOR last();

		/// Return a pointer to the flow of the entry pointed by the iterator
		
		/**
		 * @param flowIterator_ the flow iterator
		 * @return a pointer to the flow
		 */	
		static Flow *fromIterator(FLOW_ITERATOR flowIterator_);

		/// Return the flow list represented as an Xml string
		
		/**
		 * @return the flow list represented as an Xml string
		 */
		ACE_TString toXml();

		/// Return the flow list represented as an Xml string shifted at the specified offset
		
		/**
		 * @param n_ the offset of the shift
		 * @return the flow list represented as an Xml string shifted at the specified offset
		 */
		ACE_TString toXmlWithIndent(int n_);
};

#endif //_FLOWLIST_H_
