#ifndef _GLOBALLIST_H_
#define _GLOBALLIST_H_

#include "ace/ACE.h"
#include "ace/Log_Msg.h"
#include "ace/Singleton.h"
#include "ace/Vector_T.h"

#include "ClientList.h"
#include "FlowList.h"

typedef ACE_Vector<ACE_TString> StringVector;
typedef ACE_Vector<ACE_TString>::Iterator StringIterator;

/// This class describes a global list object.

/** This class implements the design pattern Singleton.
 * In order to make sure that only the singleton instance is created,
 * and that users cannot create their own instances,
 * the constructor is protected and the singleton is a friend of this class
 */

class GlobalList
{
	friend class ACE_Singleton<GlobalList, ACE_Null_Mutex>;
	
	protected:
		int init();
		//int init(ACE_TString xml_);
		int fini();

		/// List used to store the clients
		ClientList *clientList;
		
		/// List used to store the flows
		FlowList *flowList;
	
	public:
		/// Create an empty global list
		GlobalList();
		
		/// Create a global list giving an Xml string
		//GlobalList(ACE_TString xml_);
		
		/// Destroy the global list
		virtual ~GlobalList();

		/// Return a pointer to a flow giving its ID
		
		/**
		 * @param flowID_ ID of the flow to get
		 * @return a pointer to the flow, null pointer if flow is not in the list
		 */
		Flow *getFlow(ACE_TString flowID_);
		
		/// Check if a flow is in the list giving its ID
		
		/**
		 * @param flowID_ ID of the flow to check
		 * @return 0 if the flow is in the list, and -1 if not
		 */
		int findFlow(ACE_TString flowID_);

		/// Check if a flow is in the list giving its xml representation
		/// Used for server synchronization
		
		/**
		 * @param flowXML_ XML representation of the flow to check
		 * @return 0 if the flow is in the list, and -1 if not
		 */
		int containsFlow(ACE_TString flowXML_);

		/// Update a flow (only information regarding consumers) in the list giving its xml representation
		/// Used for server synchronization
		
		/**
		 * @param flowXML_ XML representation of the flow to update
		 * @return 0 if the flow was updated succesfully, -1 if the flow wasn't found in the list
		 */
		int updateConsumersOfFlow(ACE_TString flowXML_);

		/// Update a flow (only information regarding producer) in the list giving its xml representation
		/// Used for server synchronization
		
		/**
		 * @param flowXML_ XML representation of the flow to update
		 * @return 0 if the flow was updated succesfully, -1 if the flow wasn't found in the list
		 */
		int updateProducerOfFlow(ACE_TString flowXML_);

		/// Add a flow to the list
		
		/**
		 * @param flow_ pointer to the flow to add
		 * @return 0 if the flow was added succesfully, 1 if the flow already
		 * exists in the list, and -1 if it was not added
		 */
		int addFlow(Flow *flow_);

		/// Remove a flow from the list
		
		/**
		 * @param flowID_ ID of the flow to remove
		 * @return 0 if the flow was removed succesfully, and -1 if not
		 */
		int removeFlow(ACE_TString flowID_);

		/// Return an iterator on the first flow of list
		
		/**
		 * @return iterator on the first flow of the list
		 */
		FLOW_ITERATOR getFirstFlow();

		/// Return an iterator on the last flow of the list
		
		/**
		 * @return iterator on the last flow of the list
		 */
		FLOW_ITERATOR getLastFlow();

		/// Return a pointer to the flow of the entry pointed by the iterator
		
		/**
		 * @param flowIterator_ the flow iterator
		 * @return a pointer to the flow
		 */	
		static Flow *getFlowFromIterator(FLOW_ITERATOR flowIterator_);

		/// Return a pointer to a client giving its ID
		
		/**
		 * @param clientID_ ID of the client to get
		 * @return a pointer to the client, null pointer if client is not in the list
		 */
		Client *getClient(ACE_TString clientID_);

		/// Check if a client is in the client giving its ID
		
		/**
		 * @param clientID_ ID of the client to check
		 * @return 0 if the client is in the list, and -1 if not
		 */
		int findClient(ACE_TString clientID_);

		/// Add a client to the list
		
		/**
		 * @param client_ pointer to the client to add
		 * @return 0 if the client was added succesfully, 1 if the client already
		 * exists in the list, and -1 if it was not added
		 */
		int addClient(Client *client_);

		/// Remove a client from the list
		
		/**
		 * @param clientID_ ID of the client to remove
		 * @return 0 if the client was removed succesfully, and -1 if not
		 */
		int removeClient(ACE_TString clientID_);

		/// Remove all the clients connected to a server
		
		/**
		 * @param serverID_ ID of the server
		 * @return 0 if the clients were removed succesfully,
		 * and -n if 'n' clients were not removed succesfully
		 */
		int removeAllClientsOnServer(ACE_TString serverID_);

		/// Return an iterator on the first client of the list
		
		/**
		 * @return iterator on the first client of the list
		 */
		CLIENT_ITERATOR getFirstClient();

		/// Return an iterator on the last client of the list
		
		/**
		 * @return iterator on the last client of the list
		 */
		CLIENT_ITERATOR getLastClient();

		/// Return a pointer to the client of the entry pointed by the iterator
		
		/**
		 * @param clientIterator_ the client iterator
		 * @return a pointer to the client
		 */	
		static Client *getClientFromIterator(CLIENT_ITERATOR clientIterator_);

		/// Return a pointer to the client list
		/**
		 * @return a pointer to the client list
		 */	
		ClientList *getClientList();
		
		/// Return a pointer to the flow list
		/**
		 * @return a pointer to the flow list
		 */		
		FlowList *getFlowList();

		/// Return the global list represented as an Xml string
		
		/**
		 * @return the global list represented as an Xml string
		 */	
		ACE_TString toXml();
};

typedef ACE_Singleton<GlobalList, ACE_Null_Mutex> GLOBAL_LIST;

#endif //_GLOBALLIST_H_
