#ifndef _SOCKETLIST_H_
#define _SOCKETLIST_H_

#include "ace/ACE.h"
#include "ace/SString.h"
#include "ace/Log_Msg.h"
#include "ace/Singleton.h"
#include "ace/Null_Mutex.h"
#include "ace/SOCK_Stream.h"
#include "ace/Hash_Map_Manager_T.h"

typedef ACE_Hash_Map_Manager_Ex<ACE_TString, ACE_SOCK_Stream *, ACE_Hash<ACE_TString>, ACE_Equal_To<ACE_TString>, ACE_Null_Mutex> SOCKET_MAP;
typedef ACE_Hash_Map_Iterator_Ex<ACE_TString, ACE_SOCK_Stream *, ACE_Hash<ACE_TString>, ACE_Equal_To<ACE_TString>, ACE_Null_Mutex> SOCKET_ITERATOR;

/// This class implements a list which associates client IDs with sockets.

/**
 * This class implements the design pattern Singleton.
 * In order to make sure that only the singleton instance is created,
 * and that users cannot create their own instances,
 * the constructor is protected and the singleton is a friend of this class
 */

class SocketList
{
	friend class ACE_Singleton<SocketList, ACE_Null_Mutex>;
	
	protected:
		int init();
		int fini();

		/// Internal map used to store the client sockets
		SOCKET_MAP socketsClient;
		
	public:
		/// Create an empty socket list
		SocketList();
		
		/// Destroy the socket list
		virtual ~SocketList();

		/// Add a client to the list

		/**
		 * @param clientID_ ID of the client to add
		 * @param socket_ pointer to the socket to associate with the client
		 * @return 0 if the client was added successfully, -1 if not
		 */									
		int addClient(ACE_TString clientID_, ACE_SOCK_Stream *socket_);

		/// Remove a client from the list

		/**
		 * The client and its associated socket will be removed from the
		 * list.
		 * This method does not take care of closing the socket.
		 * @param clientID_ ID of the client to remove
		 * @return 0 if the client was removed succesfully, and -1 if not
		 */
		int removeClient(ACE_TString clientID_);

		/// Check if a client is in the list giving its ID
		
		/**
		 * @param clientID_ ID of the client to check
		 * @return 0 if the client is in the list, and -1 if not
		 */
		int findClient(ACE_TString clientID_);

		/// Get the socket associated to a client
		
		/**
		 * @param clientID_ ID of the client
		 * @return a pointer to the socket associated to the client
		 */
		ACE_SOCK_Stream *getClient(ACE_TString clientID_);

		/// Count the number of clients in the list

		/**
		 * @return the number of clients in the list
		 */
		int countClient();

		/// Return an iterator on the first client of the list
		
		/**
		 * @return iterator on the first client of the list
		 */
		SOCKET_ITERATOR firstClient();

		/// Return an iterator on the last client of the list
		
		/**
		 * @return iterator on the last client of the list
		 */
		SOCKET_ITERATOR lastClient();
		
		/// Return the socket of the entry pointed by the iterator
		
		/**
		 * @param socketIterator_ the socket iterator
		 * @return a pointer to the socket
		 */
		static ACE_SOCK_Stream *getSocketFromIterator(SOCKET_ITERATOR socketIterator_);

		/// Return the ID (either client of server) of the entry pointed by the iterator
		
		/**
		 * @param socketIterator_ the socket iterator
		 * @return the ID (either client of server)
		 */
		static ACE_TString getIDFromIterator(SOCKET_ITERATOR socketIterator_);
};

typedef ACE_Singleton<SocketList, ACE_Null_Mutex> SOCKET_LIST;

#endif //_SOCKETLIST_H_
