/* This client node is a consumer. It consumes a Flow_Block_Test flow. A long and a string are retrieved from the Buffer and displayed on the standard output. We don't know in which order are stored the string and the long in the Buffer, custom metadatas tell us this information. In this example, we retrieved the data out of the Buffer using a BufferIterator */

#include <iostream>
//#include <string.h>
#include "ace/OS.h"

//Required to handle the kill signal and exit properly from the client node.
#include <signal.h>

//For the timestamping of the data
#include "ace/Time_Value.h"

//here are the required includes to use the NSDFS II API in your client node.
//Smartflow.h is required
#include "Smartflow.h"

//Buffer.h is required.
#include "Buffer.h"

//Each type of flow has an include file. If you use a type of flow in your client node, your need to include the proper file of the corresponding flow.
//Every type of flow should be put in folder like that: $SMARTFLOW/src/flows/Flow_YourOwnFlow/Flow_YourOwnFlow.h
#include "Flow_BlockTest/Flow_BlockTest.h"

using namespace std;

//We have a smartflow namespace, if you don't declare you will use it, you have to prefix your NSDFS II method with  Smartflow:: 
using namespace smartflow;

//Boolean used to quit the main loop when required
bool running = 1;

//A flow of Flow_Block_Test type is used to transport the long. An other option is to create a specific flow to transport a long.
//this variable is in the global scope of the program, so it is visible from the exit_function()
Flow_BlockTest *flow = NULL;

static void exit_function() { 
	cout << "EXIT FUNCTION" << endl; 
	if ( flow )
		//we stop the flow. That call disconnects the flow, it also makes the flow->getBuffer() returns
		//even if no data is available.
		flow->stop();
	//we set running to 0 to cleanly exit from the while loop
	running = 0;
}

//here we are declaring the two types of custom metadata used in this example
//The value assigned MUST be an integer >= 10
//If you assigned a value < 10, it will create conflicts with the NSDFS II API
#define LONGMETADATA 10
#define STRINGMETADATA 11

int main(int argc, char **argv)
{
	//A Smartflow object is required in each client. This object is used to communicate with the server throught the NSDFS II API. It is used to manage flows as well.
	Smartflow *sf = NULL;
	
	//We instantiate the Smartflow object.
	sf = new Smartflow();
	
	//We checked the Smartflow object is correctly instantiated.
	if ( ! sf ) return 1;
	
	//We init the Smartflow object internally. This initialisation is required even if you don't plan to use the arguments of the client node.
	//We strongly advise to pass the parameter from the command line in every client. Otherwise, you will lose a lot of dynamicity of the NSDFS II system.
	//The parameter contains the information for connecting clients nodes together. This call is also required if you want to control the clients from the NSDFS Control Center.
	sf->init(argc,argv);
	
	//We check everything went fine in the initialisation. This call is required.
	if ( !sf->readyToRun() )   {
		//If any method the returns an error (i.e.: a value different from 0). You can get a string describing this error by calling the method Smartflow::strError().
		cout << sf->strError() << endl;
		delete sf; return 1;
	}
	
	//This call is as well required in order for the client node to connect to the server.
	//You can give an argument to this method. The argument is the name of the client. If you don't give any, the name of the executable on the disk
	//will be used by default. You cannot have two client with the same name running at the same time, the Smartflow Server doesn't allow it and the 
	//connection is refused. So if you want to launch two times the same client node, you need to overide the name 
	//of it for at least one client node (so you will get two different names).
	//To overide dynamically the name at the execution, just pass this argument on the command line when you launch a client node : "--sf clientname=newName"
	//where "newName" is the name you want to give to your client node. The name has to be unique.
	if ( !sf->connectToApplicationServer() ) {
		cout << sf->strError() << endl;
		delete sf; return 1;
	}
	
	// We set up the callback, which needs to be called to exit properly. This function will be called if SIGINT or SIGKILL is sent to the client node.
	sf->setUserExitFunction(exit_function);
	
	//At this point the client node is connected to the server and visible on the NSDFS network.
	//We can know take care of the flow...
	
	//This client node is a consumer. We create an input flow of Flow_Block_Test flow.
	//You must NEVER use the Flow constructor method. ALWAYS create your flows using methods Smartflow::makeInputFlow() or Smartflow::makeOutputFlow().
	//You need to cast your flow to the proper type because Smartflow::makeInputFlow() and Smartflow::makeOutputFlow() returns a flow of type 'Flow*'.
	//You can use a dynamic_cast<> to be sure your flow object is properly created and has the right type. However the dynamic_cast<> is not portable on
	//every C++ compiler. A classic cast is good enought, but you will need to make an extra test to ensure your flow has the right type.
	//The first parameter is the Type of the flow you want to create. This type must exist. If you need a new type, it is possible to create it.
	//The second parameter is the name of the flow. This parameter is used to identify flows for the developper of the client. This name will appear in
	//the control center. 
	//The third parameter is used to pass argument to the flow. This is an optionnal parameter. A really useful argument is "blocksize", it allows to change dynamically the size
	//of the buffer you want to use (blocksize allows you to specify the maximal size of the buffer). Each flow has a default size, if you want more or less space,
	//you can specify it here AT THE CREATION OF THE FLOW. The provider used a buffer of 512 bytes So we use the same size on the consumer side. The maximum size of the Buffer can
	//be bigger than the size declared by the provider, but NEVER smaller. Otherwise, you might miss some data. 
	//The fourth argument is a string used as a connection ID: if an input flow and an output flow in a NSDFS II application have the same type and the same connection ID,
	//they will be connected. If you don't have several input flows (or output flows) of the same kind, this argument is optionnal, otherwise you need to use it in order 
	//for the server to know which flows it should connect together.
	//This fourth argument can be overriden when you launch a client node from the command line: "--sf flow-*=* (i.e.: "--sf flow-theNameYouGiveToTheFlow=theID")
	//If you want to change the connection ID of this flow, for this specific client, you need to launch the client node with this paramater:  "--sf flow-myFlowConsumer=myNewID". 
	//In this case, "ID_custom_metadata" will be overidden by "myNewID".
	flow = (Flow_BlockTest*)(sf->makeInputFlow("Flow_BlockTest", "myFlowConsumer", "blocksize=512", "ID_custom_metadata"));
	
	//We check the flow is correcty created
	if ( ! flow) {
		cout << sf->strError() << endl;
		delete sf; return 1;
	}
	
	//We advised you to check the real type of your flow if you use a regular cast to create your flow. If you use a dynamic_cast<>, this is not required. 
	//With a dynamic_cast, the casted object is null if it is not of the proper type. 
	if ( flow->getFlowType() != "Flow_BlockTest" ) {
		cout << "Flow Cast Failed" << endl;
		delete sf; return 1;
	} 
    
	//You need to start your flow. This method is required, it will launch the duplicator if needed.
	//After this call, you can begin to exchange data.
	if ( !flow->start() ) {
		cout << sf->strError() << endl;
		delete sf; return 1;
	}
	
	//You take the data out of the flow using a Buffer object
	Buffer *myBuffer = NULL;
	
	//In this example, we use a BufferIterator object in order to read the Buffer.
	//We create a BufferIterator object to iterate and retrieve data from the Buffer.
	BufferIterator *iterator = new BufferIterator();
	
	//We use this integer internally in order to know if a long or a string is the next thing in the Buffer. 1 for a long, -1 for a string
	int isLongOrString = 0;
	
	//This is the main loop
	while ( running ) {
		
		//You get the Buffer, which contains the data, from the flow.
		//Same as the flow, you NEVER instantiate a Buffer object using its constructor method. 
		//This method returns when there is a buffer available to be consumed
		myBuffer = flow->getBuffer();
		
		//We checked we have the buffer.  After this point, you are ready to read or take data out of the Buffer.
		if ( !myBuffer ) {
			cout << "My buffer is null" << endl;
			cout << sf->strError() << endl;
		}
		else {
			//Once we have the Buffer, we pass it to the BufferIterator object using the method BufferIterator::setBuffer() method. 
			//Each time we get a new Buffer from the flow, we need to pass it to the BufferIterator this way.
			iterator->setBuffer(myBuffer);
			
			//Then we init the iterator. This call is required before beginning the iteration. You can't call twice this method about the same Buffer. If you do it, the second call doesn't do anything
			iterator->init();
			
			//We are know ready to iterate on the Buffer.
			
			ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("--------------BUFFER BEGIN----------------\n")));
			
			
			//While we haven't reached the end of the buffer, we keep iterating.
			//The method BufferIterator::end() returns true when we reached the end of the Buffer.
			while ( !iterator->end() ) {
				//This BufferIterator::getDataType() method gives us the dataType of the current element in the buffer. The data type can be DATA, DATAWITHTIMESTAMP
				//or your customs metadata type (i.e: LONGMETADATA and STRINGMETADATA).
				switch(iterator->getDataType()) {
					case DATA:
						//we displayed a message to know what we are about to read in the buffer.
						ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("We received a Data\n")));
						break;
					case DATAWITHTIMESTAMP:
						ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("We received a timestamped data\n")));
						break;
					case LONGMETADATA:
						ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("We received a LONGMETADATA\n")));
						//we set isLongOrString to 1, so we know we will have to deal with a long in the next block of the Buffer.
						isLongOrString = 1;
						break;
					case STRINGMETADATA:
						ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("We received a STRINGMETADATA\n")));
						//we set isLongOrString to -1, so we know we will have to deal with a string in the next block of the Buffer. 
						isLongOrString = -1;
						break;
					default:
						break;
				}
				
				//The method BufferIterator::getTimestamp() returns the timestamp associated to the data. If you didn't timestamp your data, you get a timestamp anyway with tv_sec = tv_nsec = 0
				if ( iterator->hasTimestamp() ) {
					ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("Value of the timestamp:\n")));
					//Once we get the timestamp, we can display its second and nanosecond values.
					ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("   Second=%d\n"), iterator->getTimestamp().sec()));
					ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("   Microsecond=%d\n"), iterator->getTimestamp().usec()));
				}
				else {
					ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("There is a NO Timestamp\n")));
				}
				
				//Each data or metadata has a size. In order to get it, we use the BufferIterator::getDataSize() method.
				//This methods gives you the size of the data or the metadata you are about to read from the Buffer.
				ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("The size of the data/metadata is = %d\n"), iterator->getDataSize()));
				
				//Here we check if the current block contains DATA OR METADATA
				if ( (iterator->getDataType() == DATA) || (iterator->getDataType() == DATAWITHTIMESTAMP) ) {
					//The block contains DATA
					if ( isLongOrString == -1) {
						//The block contains a string
						
						//We declare an object big enougt to contain the data. In order to know the size of the data we use the BufferIterator::getDataSize() method.
						//This call doesn't  work on VS2005
						//char result[iterator->getDataSize()];
						//dynamic memory allocation for VS2005
						char *result= (char *)malloc(iterator->getDataSize()*sizeof(char));
						
						//Using the BufferIterator::getData(), we get the data from the Buffer and store it in 'result'.
						iterator->getData(result);
						
						//Then we display it.
						ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("Data = '%s'\n\n"), result));
						
						//We don't need isLongOrString anymore, so we set it to 0.
						isLongOrString = 0;
						
						//we don't need result anymore
						free(result);
					} //end if( isLongOrString == -1) 
					else if ( isLongOrString == 1){
						//The block contains a long
						long result;
						
						//Using the BufferIterator::getData(), we get the data from the Buffer and store it in 'result'.
						iterator->getData(&result);
						
						//We get a network representation of the long. We want to be able to read it. So we get a host representation of this long.
						long longToDisplay = ntohl(result);
						
						//Then we display it.
						ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("Data = %d\n\n"), longToDisplay));
						
						//We don't need isLongOrString anymore, so we set it to 0.
						isLongOrString = 0;
					} //end else if ( isLongOrString == 1)
				} //end if ( (iterator->getDataType() == DATA) || (iterator->getDataType() == DATAWITHTIMESTAMP) )
				else 
					//The block contains a METADATA, we don't do anything about it. We already knows the exact type of the metadata
					//You can notice we never take the value out of the metadata because we don't need it.
					cout << endl;
				
				//we finished to work on the current block of the Buffer, we can know take care of the next one by calling the BufferIterator::next() method.
				iterator->next();
			}
			
			ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("--------------BUFFER END------------------\n\n")));
			
			//We took the data out of the buffer, we don't need this buffer anymore. So we released it.
			flow->releaseBuffer(myBuffer); 
		}
		//We set the Buffeer object to NULL, no need to destroy it. The NSDFS II API takes care of that
		myBuffer = NULL;	
	}//end while
	
	//We are out of the loop, we can destroy the objects before quitting the client node.
	//We destroy the flow. ALWAYS destroy your flows using the Smartflow::destroyFlow() method. Using this method will destroy all objects associated to the flow (like its buffer).
	sf->destroyFlow(flow); 
	flow = NULL;
	
	//Finally we delete the Smartflow Object invoking the destructor method of the object.
	delete sf; sf = NULL;
	
	return 0;
}
