/* This client node is a consumer. It consumes a Flow_Block_Test flow. A long is retrieved from the Buffer of the flow and displayed on the standard output  */

#include <iostream>
#include <ace/OS.h>

//Required to handle the kill signal and exit properly from the client node.
#include <signal.h>

//here are the required includes to use the NSDFS II API in your client node.
//Smartflow.h is required
#include <Smartflow.h>

//Buffer.h is required.
#include <Buffer.h>

//Each type of flow has an include file. If you use a type of flow in your client node, your need to include the proper file of the corresponding flow.
//Every type of flow should be put in folder like that: $SMARTFLOW/src/flows/Flow_YourOwnFlow/Flow_YourOwnFlow.h
#include <Flow_BlockTest/Flow_BlockTest.h>

using namespace std;

//We have a smartflow namespace, if you don't declare you will use it, you have to prefix your NSDFS II method with  Smartflow:: 
using namespace smartflow;

//A flow of Flow_Block_Test type is used to transport the long. An other option is to create a specific flow to transport a long.
//this variable is in the global scope of the program, so it is visible from the exit_function()
Flow_BlockTest *flow = NULL;

//Boolean used to quit the main loop when required
bool running = 1;

static void exit_function() { 
  cout << "EXIT FUNCTION" << endl; 
  if ( flow )
	//we stop the flow. That call disconnects the flow, it also makes the flow->getBuffer() returns
	//even if no data is available.
	flow->stop();
  //we set running to 0 to cleanly exit from the while loop
  running = 0;
}

int main(int argc, char **argv)
{
  //A Smartflow object is required in each client. This object is used to communicate with the server throught the NSDFS II API. It is used to manage flows as well.
  Smartflow *sf = NULL;
  
  //We instantiate the Smartflow object.
  sf = new Smartflow();

  //We checked the Smartflow object is correctly instantiated.
  if ( ! sf ) return 1;

  //We init the Smartflow object internally. This initialisation is required even if you don't plan to use the arguments of the client node.
  //We strongly advise to pass the parameter from the command line in every client. Otherwise, you will lose a lot of dynamicity of the NSDFS II system.
  //The parameter contains the information for connecting clients nodes together. This call is also required if you want to control the clients from the NSDFS Control Center.
  sf->init(argc,argv);
  
  //We check everything went fine in the initialisation. This call is required.
  if ( !sf->readyToRun() )   {
    //If any method the returns an error (i.e.: a value different from 0). You can get a string describing this error by calling the method Smartflow::strError().
    cout << sf->strError() << endl;
    delete sf; return 1;
  }
  
  //This call is as well required in order for the client node to connect to the server.
  //You can give an argument to this method. The argument is the name of the client. If you don't give any, the name of the executable on the disk
  //will be used by default. You cannot have two client with the same name running at the same time, the Smartflow Server doesn't allow it and the 
  //connection is refused. So if you want to launch two times the same client node, you need to overide the name 
  //of it for at least one client node (so you will get two different names).
  //To overide dynamically the name at the execution, just pass this argument on the command line when you launch a client node : "--sf clientname=newName"
  //where "newName" is the name you want to give to your client node. The name has to be unique.
  if ( !sf->connectToApplicationServer() ) {
    cout << sf->strError() << endl;
    delete sf; return 1;
  }

  // We set up the callback, which needs to be called to exit properly. This function will be called if SIGINT or SIGKILL is sent to the client node.
  sf->setUserExitFunction(exit_function);

  //At this point the client node is connected to the server and visible on the NSDFS network.
  //We can know take care of the flow...

  //This client node is a consumer. We create an input flow of Flow_Block_Test flow.
  //You must NEVER use the Flow constructor method. ALWAYS create your flows using methods Smartflow::makeInputFlow() or Smartflow::makeOutputFlow().
  //You need to cast your flow to the proper type because Smartflow::makeInputFlow() and Smartflow::makeOutputFlow() returns a flow of type 'Flow*'.
  //You can use a dynamic_cast<> to be sure your flow object is properly created and has the right type. However the dynamic_cast<> is not portable on
  //every C++ compiler. A classic cast is good enought, but you will need to make an extra test to ensure your flow has the right type.
  //The first parameter is the Type of the flow you want to create. This type must exist. If you need a new type, it is possible to create it.
  //The second parameter is the name of the flow. This parameter is used to identify flows for the developper of the client. This name will appear in
  //the control center. 
  //The third parameter is used to pass argument to the flow. This is an optionnal parameter. A really useful argument is "blocksize", it allows to change dynamically the size
  //of the buffer you want to use (blocksize allows you to specify the maximal size of the buffer). Each flow has a default buffer size, if you want more or less space,
  //you can specify it here AT THE CREATION OF THE FLOW. The provider use a Buffer of 16 bytes. So we use the same size on the consumer side. The maximum size of the Buffer can
  //be bigger than the size declared by the provider, but NEVER smaller. Otherwise, you might miss some data.
  //The fourth argument is a string used as a connection ID: if an input flow and an output flow in a NSDFS II application have the same type and the same connection ID,
  //they will be connected (i.e: exchange data). If you don't have several input flows (or output flows) of the same kind, this argument is optionnal, otherwise you need to use it in order 
  //for the server to know which flows it should connect together. 
  //This fourth argument can be overriden when you launch a client node from the command line: "--sf flow-*=* (i.e.: "--sf flow-theNameYouGiveToTheFlow=theID")
  //If you want to change the connection ID of this flow, for this specific client, you need to launch the client node with this paramater:  "--sf flow-myFlowConsumer=myNewID". 
  //In this case, "ID_long_memcpy" will be overidden by "myNewID".
  flow = (Flow_BlockTest*)(sf->makeInputFlow("Flow_BlockTest", "myFlowConsumer", "blocksize=16", "ID_long_memcpy"));
	
  //We check the flow is correcty created
  if ( ! flow) {
    cout << sf->strError() << endl;
    delete sf; return 1;
  }

  //We advised you to check the real type of your flow if you use a regular cast to create your flow. If you use a dynamic_cast<>, this is not required. 
  //With a dynamic_cast, the casted object is null if it is not of the proper type. 
  if ( flow->getFlowType() != "Flow_BlockTest" ) {
    cout << "Flow Cast Failed" << endl;
    delete sf; return 1;
  } 
    
  //You need to start your flow. This method is required, it will launch the duplicator if needed.
  //After this call, you can begin to exchange data.
  if ( !flow->start() ) {
    cout << sf->strError() << endl;
    delete sf; return 1;
  }

  //You take the data out of the flow using a Buffer object
  Buffer *myBuffer = NULL;

  //This is the main loop
  while ( running ) {
    
    //You get the Buffer, which contains the data, from the flow.
    //Same as the flow, you NEVER instantiate a Buffer object using its constructor method.
    //This method returns when there is a buffer available to be consumed
    myBuffer = flow->getBuffer();

    //We checked we have the buffer. After this point, you are ready to read or take data out of the Buffer.
    if ( !myBuffer ) {
	  cout << "My buffer is null" << endl;
      cout << sf->strError() << endl;
      //delete sf; return 1;
    }
    else {
		//At this point we have a Buffer object ready to be read.
		//We need now to get the data out the buffer, there is several philosophy. In this client node, we use the memcpy() method.

		//Each buffer contains one long in this example.
		long myLong = 0;
    
		//Here we get the size of the Buffer. This is NOT the maximum size of the Buffer, but the size of the data the provider put in that Buffer. This size may be different for the next Buffer.
		//This size change if the provider put more or less data in the buffer.
		int bufferSize = myBuffer->getSize();

		//We use the memcpy() method from ACE for portability issue, but you can use an other memcpy method.
		//Here, we write in myLong the data contained in the Buffer. We know bufferSize represents the size of a long because we know what the provider put in the Buffer.
		ACE_OS::memcpy(&myLong, myBuffer->data() , bufferSize);

		//We get a network representation of the long. We want to be able to read it. So we get a host representation of this long.
		long longToDisplay = ntohl(myLong);

		//Once we get the long value, we display it to check everything work fine
		ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("Long Value received: %d\n"), longToDisplay));
		//ACE_DEBUG((USER_PREFIX ACE_TEXT("Long Value received: %d\n"), longToDisplay));
		//We have done the work on this buffer (i.e.: we took the data out of it), we don't it anymore. So we released it.
		flow->releaseBuffer(myBuffer); 
    }
    //We set the Buffer object to NULL, no need to destroy it. The NSDFS II API took care of that
	 myBuffer = NULL;	
  }
  
  //We are out of the loop, we can destroy the objects before quitting the client node.
  //We destroy the flow. ALWAYS destroy your flows using the Smartflow::destroyFlow() method. Using this method will destroy all objects associated to the flow (like its buffer).
  sf->destroyFlow(flow); 
  flow = NULL;
  
  //Finally we delete the Smartflow Object invoking the destructor method of the object.
  delete sf; sf = NULL;
  
  cout << "Exiting client" << endl;
  return 0;
}
