/* This client node is a provider. It provides a Flow_Block_Test flow. Using this flow we send a long and a string. Sometimes the long is put in the data first, sometimes it is the string. In order for the consumer to know what kind of data it is about to consume, we put before each long and each string a metadata in the buffer telling us what follows (a long or a string). */

#include <iostream>
#include <string.h>
#include "ace/OS.h"

//Required to handle the kill signal and exit properly from the client node.
#include <signal.h>		

//here are the required includes to use the NSDFS II API in your client node.
//Smartflow.h is required
#include <Smartflow.h>

//Buffer.h is required.
#include <Buffer.h>

//Each type of flow has an include file. If you use a type of flow in your client node, your need to include the proper file of the corresponding flow.
//Every type of flow should be put in folder like that: $SMARTFLOW/src/flows/Flow_YourOwnFlow/Flow_YourOwnFlow.h
#include <Flow_BlockTest/Flow_BlockTest.h>

using namespace std;

//We have a smartflow namespace, if you don't declare you will use it, you have to prefix your NSDFS II method with  Smartflow:: 
using namespace smartflow;

//Boolean used to quit the main loop when required
bool running = 1;

//We will use a Flow_Block_Test flow to provide the long, the string, and the associated metadatas.
Flow_BlockTest *flow = NULL;

static void exit_function() { 
	cout << "EXIT FUNCTION" << endl; 
	if ( flow )
		flow->stop();
	running = 0;
}

int main(int argc, char **argv)
{
	
	//A Smartflow object is required in each client. This object is used to communicate with the server throught the NSDFS II API. It is used to manage flows as well.
	Smartflow *sf = NULL;
	
	//We instantiate the Smartflow object.
	sf = new Smartflow();
	
	//We checked the Smartflow object is correctly instantiated.
	if ( ! sf ) return 1;
	
	//We init the Smartflow object internally. This initialisation is required even if you don't plan to use the arguments of the client node.
	//We strongly advise to pass the parameter from the command line in every client. Otherwise, you will lose a lot of dynamicity of the NSDFS II system.
	//The parameter contains the information for connecting clients nodes together. This call is also required if you want to control clients from the NSDFS Control Center.
	sf->init(argc,argv);
	
	//We check everything went fine in the initialisation. This call is required.
	if ( ! sf->readyToRun() )   {
		//If any method the returns an error (i.e.: a value different from 0). You can get a string describing this error by calling the method Smartflow::strError().
		cout << sf->strError() << endl;
		delete sf; return 1;
	}
	
	//This call is as well required in order for the client node to connect to the server.
	//You can give an argument to this method. The argument is the name of the client. If you don't give any, the name of the executable on the disk
	//will be used by default. You cannot have two client with the same name running at the same time, the Smartflow Server doesn't allow it and the 
	//connection is refused. So if you want to launch two times the same client node, you need to overide the name 
	//of it for at least one client node (so you will get two different names).
	//To overide dynamically the name at the execution, just pass this argument on the command line when you launch a client node : "--sf clientname=newName"
	//where "newName" is the name you want to give to your client node. The name has to be unique.
	if ( !sf->connectToApplicationServer() ) {
		cout << sf->strError() << endl;
		delete sf; return 1;
	}
	
	// We set up the callback, which needs to be called to exit properly. This function will be called if SIGINT or SIGKILL is sent to the client node.
	sf->setUserExitFunction(exit_function);
	
	//At this point the client node is connected to the server and visible on the NSDFS network.
	//We can know take care of the flow...
	
	//This client node is a provider. We create an output flow of Flow_Block_Test flow to transport the string, the long and metadatas.
	//You must NEVER use the Flow constructor method. ALWAYS create your flows using methods Smartflow::makeInputFlow() or Smartflow::makeOutputFlow().
	//You need to cast your flow to the proper type because Smartflow::makeInputFlow() and Smartflow::makeOutputFlow() returns a flow of type 'Flow*'.
	//You can use a dynamic_cast<> to be sure your flow object is properly created and has the right type. However the dynamic_cast<> is not portable on
	//every C++ compiler. A classic cast is good enought, but you will need to make an extra test to ensure your flow has the right type.
	//The first parameter is the Type of the flow you want to create. This type must exist. If you need a new type, it is possible to create it.
	//The second parameter is the name of the flow. This parameter is used to identify flows for the developper of the client. This name will appear in
	//the control center. 
	//The third parameter is used to pass argument to the flow. This is an optionnal parameter. A really useful argument is "blocksize", it allows to change dynamically the size
	//of the buffer you want to use (blocksize allows you to specify the maximal size of the buffer). Each flow has a default size, if you want more or less space,
	//you can specify it here AT THE CREATION OF THE FLOW. Here 512 bytes are enough to contain what we want to put inside. Only the part of the buffer you filled out with data is actually send,
	//so if you declare a flow with a buffer giving you to much space regarding your needs, it is not really a big deal.
	//The fourth argument is a string used as a connection ID: if an input flow and an output flow in a NSDFS II application have the same type and the same connection ID (and opposite directions),
	//they will be connected. If you don't have several input flows (or output flows) of the same kind, this argument is optionnal, otherwise you need to use it in order 
	//for the server to know which flows it should connect together.
	//This fourth argument can be overriden when you launch a client node from the command line: "--sf flow-*=* (i.e.: "--sf flow-theNameYouGiveToTheFlow=theID")
	//If you want to change the connection ID of this flow, for this specific client, you need to launch the client node with this paramater:  "--sf flow-myFlowProvider=myNewID". 
	//In this case, "ID_multipush" will be overidden by "myNewID".
	flow = (Flow_BlockTest*)(sf->makeOutputFlow("Flow_BlockTest", "myFlowProvider", "blocksize=512", "ID_multipush"));
	
	//We check the flow is correcty created
	if ( ! flow) {
		cout << sf->strError() << endl;
		delete sf; return 1;
	}
	
	//We advised you to check the real type of your flow if you use a regular cast to create your flow. If you use a dynamic_cast<>, this is not required. 
	//With a dynamic_cast, the casted object is null if it is not of the proper type. 
	if ( flow->getFlowType() != "Flow_BlockTest" ) {
		cout << "Flow Cast Failed" << endl;
		delete sf; return 1;
	}   
	
	//You need to start your flow. This method is required, it will launch the duplicator if needed.
	//After this call, you can begin to exchange data.
	if ( !flow->start() ) {
		cout << sf->strError() << endl;
		delete sf; return 1;
	}
	
	//You put the data in the flow using a Buffer object
	Buffer *myBuffer = NULL;
	
	//This flow transport a long. We increase its value from 0.1 everytime we iterate in the loop
	ACE_UINT32 myInt = 0;
	
	//This flow transport a string as well. This is the string we want to send.
	char myText[] = "This is an example of a string";
	
	//This is the value of the metadata. When the consumer will find a metadata, and this metadata is "Following is a string", it will let it know that after this metadata, there is a string in the Buffer.
	char myMetadata4theString[] = "Following is a string";
	
	//This is the value of the metadata. When the consumer will find a metadata, and this metadata is "Following is a long", it will let it know that after this metadata, there is a string in the Buffer.
	char myMetadata4theInt[] = "Following is an integer";
	
	//This is the main loop
	while ( running ) {
		
		//You get the Buffer, which contains the data, from the flow.
		//Same as the flow, you NEVER instantiate a Buffer object using its constructor method. 
		//This method returns when there is a buffer available to be consumed
		myBuffer = flow->getBuffer();
		
		//We checked we have the buffer.  After this point, you are ready to read or take data out of the Buffer.
		if ( !myBuffer ) {
			cout << sf->strError() << endl;
			//delete sf; return 1;
		}
		else {
			//If you use the method Buffer::push_data() to write in your buffer, you need to call Buffer::init_buf() first.
			//After that, you can call push_data() as many times you want to (be carefull to not put to much data in your Buffer!!!)
			myBuffer->initBuf();
			
			//At this point we have a Buffer object ready to written.
			//We need now to put the data in the buffer, there is several philosophy. In this client node, we use the Buffer::push_data() method.
			//We will use several times the Smartflow::push_data() methods. You can use it as many time you want before releasing the buffer,
			//as long your buffer is not full.
			//Here we put sometimes a long at first in the buffer, sometimes it is a string. Basically the consumer has no way to know what will
			//be first in the buffer. In order to tell that information to the consumer, we use metadatas.
			
			//This code is just to have a random number between 0 and 1. If the result res is bigger than 0.5, we send the long first. Otherwise
			//we send the string first
			double resRand = rand();
			double res = resRand/RAND_MAX;
			
			//In this branch, we put the long before the string
			if ( res > 0.5 ) {
				//We first put a metadata in the buffer in order to specify we have a long following
				//To do that, we use the Smartflow::push_metadata() method. The metadata is a string, which contains "Following is a long".
				//We can put whatever we want. In this case, the consumer will look for this string whenever it gets metadata. If it finds it,
				//it knows that the next data poped out of the buffer is a long.
				//Adding metadata in the buffer is pretty easy, you just need to use the Buffer::push_metadata() method.
				//The first parameter is the address of the data you want to put in the buffer.
				//The second one is the size of the metadata itself
				if (myBuffer->pushMetadata(&myMetadata4theInt, sizeof(myMetadata4theInt)) == false ) {
					cout << "Error while pushing the data into the buffer" << endl;	
				}
				
				//We could put the data in the buffer directly, but we want to make sure the data will be readable for everyone. 
				//So we sent the network representation of the long.
				ACE_UINT32 intToSend = htonl(myInt);
				
				//After adding the metadata about the long, we can put the long itself in the buffer
				if ( myBuffer->pushData(&intToSend, sizeof(ACE_UINT32)) == false) {
					cout << "Error while pushing the data into the buffer" << endl;	
				}
				
				//Once we add the metadata about the long and then the long itself in the buffer, we can add the metadata about the string.
				//Here the metadata about the string is a string, which contains "Following is a long". So if the consumer gets a metadata and 
				//the information of this metadata is the string "Following is a long", the consumer knows that the next data poped is a string.
				if (myBuffer->pushMetadata(&myMetadata4theString, sizeof(myMetadata4theString)) == false ) {
					cout << "Error while pushing the data into the buffer" << endl;	
				}
				
				//After adding the metadata about the string in the buffer, we add the string itself. This time, the data we have is timestamped.
				//Using Buffer::push_data_ts() add a timestamp to the metadata. If you don't provide any timestamp, the timestamp used comes from
				//the local machine. If you want to specify your own timestamp, you can pass it as a third parameter to the method. 
				if ( myBuffer->pushDataTS(&myText, sizeof(myText)) == false) {
					cout << "Error while pushing the data into the buffer" << endl;	
				}
			}
			//In this branch, we put the string before the long in the buffer
			else {
				//First, we put the metadata associated to the string in the Buffer.
				if (myBuffer->pushMetadata(&myMetadata4theString, sizeof(myMetadata4theString)) == false ) {
					cout << "Error while pushing the data into the buffer" << endl;	
				}
				
				//Then the string itself
				if ( myBuffer->pushData(&myText, sizeof(myText)) == false) {
					cout << "Error while pushing the data into the buffer" << endl;	
				}
				
				//Once we took care of the string, we add the long in the Buffer. But first we add the metadata associated to the long to let the consumer know what it will read from the Buffer.
				if (myBuffer->pushMetadata(&myMetadata4theInt, sizeof(myMetadata4theInt)) == false ) {
					cout << "Error while pushing the data into the buffer" << endl;	
				}
				
				//We could put the data in the buffer directly, but we want to make sure the data will be readable for everyone. 
				//So we sent the network representation of the long.
				ACE_UINT32 intToSend = htonl(myInt);
				
				//Finally we add the long in the Buffer. In this branch, we decided to timestamp the long and not the string.
				if ( myBuffer->pushDataTS(&intToSend, sizeof(ACE_UINT32)) == false) {
					cout << "Error while pushing the data into the buffer" << endl;	
				}
			}
			
			//we make a pause otherwise it is really to fast to follow!!!
			ACE_OS::sleep(1);
			
			//We put the data in the buffer, we have finished to write in the Buffer, so we don't need this it anymore. So we released it.
			//For an output flow, this call will send the buffer using the NSDFS II network.
			//If the release is successfull, the job is done. Otherwise we cancel the buffer to avoid memory leaks.
			if ( !flow->releaseBuffer(myBuffer)  )
				flow->cancelBuffer(myBuffer);
			else
				ACE_DEBUG((USER_PREFIX_EMPTY ACE_TEXT("Buffer sent\n")));
			
			//We set the Buffer object to NULL, no need to destroy it. The NSDFS II API takes care of that using the releaseBuffer() call.
			myBuffer = NULL;	
			
			//we increment the long.
			myInt += 1;
		}
	}//end for
	
	//We are out of the loop, we can destroy the objects before quitting the client node.
	//We destroy the flow. ALWAYS destroy your flows using the Smartflow::destroyFlow() method. Using this method will destroy all objects associated to the flow (like its buffer).
	sf->destroyFlow(flow); 
	flow = NULL;
	
	//Finally we delete the Smartflow Object invoking the destructor method of the object.
	delete sf; sf = NULL;
	
	return 0;
}
