/* This client node is a provider. It provides a Flow_Block_Test flow. A double is sent using the NSDFS II network. Smartflow methods are used to fill out the Buffer.*/

//here are the required includes to use the NSDFS II API in your client node.
//Smartflow.h is required
#include <Smartflow.h>

//Each type of flow has an include file. If you use a type of flow in your client node, your need to include the proper file of the corresponding flow.
//Every type of flow should be put in folder like that: $SMARTFLOW/src/flows/Flow_YourOwnFlow/Flow_YourOwnFlow.h
#include <Flow_BlockTest/Flow_BlockTest.h>

#include <iostream>
#include <string>

using namespace std;

//We have a smartflow namespace, if you don't declare you will use it, you have to prefix your NSDFS II method with  Smartflow:: 
using namespace smartflow;

// User defined exit function
// This function will be called (back) by the Smartflow API
// when the Smartflow server will asked the client to quit
// (typically when the user decides to stop the client using
// the control center)
void my_exit_function()
{
    cout << "\n\n\n\n\nMy exit function is called by the Smartflow API\n\n\n\n";
	exit(0);
}

int main(int argc, char **argv)
{
  
  //A Smartflow object is required in each client. This object is used to communicate with the server throught the NSDFS II API.  It is used to manage flows as well.
  Smartflow *sf = NULL;

  //We will use a Flow_Block_Test flow to provide the double.
  Flow_BlockTest *flow = NULL;
  
  //We instantiate the Smartflow object.
  sf = new Smartflow();

  //We checked the Smartflow object is correctly instantiated.
  if ( ! sf ) 
	  return 1;
  
  //We init the Smartflow object internally. This initialisation is required even if you don't plan to use the arguments of the client node.
  //We strongly advise to pass the parameter from the command line in every client. Otherwise, you will lose a lot of dynamicity of the NSDFS II system.
  //The parameter contains the information for connecting clients nodes together. This call is also required if you want to control clients from the NSDFS Control Center.
  sf->init(argc,argv);
  
  //We check everything went fine in the initialisation. This call is required.
  if ( !sf->ready_to_run() )   {
    //If any method the returns an error (i.e.: a value different from 0). You can get a string describing this error by calling the method Smartflow::strError().
    cout << sf->strError() << endl;
    delete sf; return 1;
  }
  
  //This call is as well required in order for the client node to connect to the server.
  //You can give an argument to this method. The argument is the name of the client. If you don't give any, the name of the executable on the disk
  //will be used by default. You cannot have two client with the same name running at the same time, the Smartflow Server doesn't allow it and the 
  //connection is refused. So if you want to launch two times the same client node, you need to overide the name 
  //of it for at least one client node (so you will get two different names).
  //To overide dynamically the name at the execution, just pass this argument on the command line when you launch a client node : "--sf clientname=newName"
  //where "newName" is the name you want to give to your client node. The name has to be unique.
  if ( !sf->connectToApplicationServer() ) {
    cout << sf->strError() << endl;
    delete sf; return 1;
  }

  // We set up the callback
  sf->setUserExitFunction(my_exit_function);

  //At this point the client node is connected to the server and visible on the NSDFS network.
  //We can know take care of the flow...

  //This client node is a provider. We create an output flow of Flow_Block_Test flow. This flow will contain the double we want to send.
  //You must NEVER use the Flow constructor method. ALWAYS create your flows using methods Smartflow::makeInputFlow() or Smartflow::makeOutputFlow().
  //You need to cast your flow to the proper type because Smartflow::makeInputFlow() and Smartflow::makeOutputFlow() returns a flow of type 'Flow*'.
  //You can use a dynamic_cast<> to be sure your flow object is properly created and has the right type. However the dynamic_cast<> is not portable on
  //every C++ compiler. A classic cast is good enought, but you will need to make an extra test to ensure your flow has the right type.
  //The first parameter is the Type of the flow you want to create. This type must exist. If you need a new type, it is possible to create it.
  //The second parameter is the name of the flow. This parameter is used to identify flows for the developper of the client. This name will appear in
  //the control center. 
  //The third parameter is used to pass argument to the flow. This is an optionnal parameter. A really useful argument is "blocksize", it allows to change dynamically the size
  //of the buffer you want to use (blocksize allows you to specify the maximal size of the buffer). Each flow has a default size, if you want more or less space,
  //you can specify it here AT THE CREATION OF THE FLOW. Here 16 byte is plenty of space for a double.
  //The fourth argument is a string used as a connection ID: if an input flow and an output flow in a NSDFS II application have the same type and the same connection ID,
  //they will be connected. If you don't have several input flows (or output flows) of the same kind, this argument is optionnal, otherwise you need to use it in order 
  //for the server to know which flows it should connect together.
  //This fourth argument can be overriden when you launch a client node from the command line: "--sf flow-*=* (i.e.: "--sf flow-theNameYouGiveToTheFlow=theID")
  //If you want to change the connection ID of this flow, for this specific client, you need to launch the client node with this paramater:  "--sf flow-myFlowProvider=myNewID". 
  //In this case, "ID_double_pushpop" will be overidden by "myNewID".

  flow = (Flow_BlockTest*)(sf->makeOutputFlow("Flow_BlockTest", "myFlowProvider", "blocksize=16", "ID_double_pushpop"));
	
  //We check the flow is correcty created
  if ( !flow) {
    cout << sf->strError() << endl;
    delete sf; return 1;
  }
      
  //We advised you to check the real type of your flow if you use a regular cast to create your flow. If you use a dynamic_cast<>, this is not required. 
  //With a dynamic_cast, the casted object is null if it is not of the proper type. 
  if ( flow->getFlowType() != "Flow_BlockTest" ) {
    cout << "Flow Cast Failed" << endl;
    delete sf; return 1;
  }  
 
  //You need to start your flow. This method is required, it will launch the duplicator if needed.
  //After this call, you can begin to exchange data.
  if ( !flow->start() ) {
    cout << sf->strError() << endl;
    delete sf; return 1;
  }

  //You put the data in the flow using a Buffer object
  Buffer *myBuffer = NULL;
  
  //This flow will transport double. We increase its value from 0.1 everytime we iterate in the loop
  double myDouble = 0;

  //This is the main loop
  while ( 1 ) {
    
    //You get the Buffer, which contains the data, from the flow.
    //Same as the flow, you NEVER instantiate a Buffer object using its constructor method. 
    //This method returns when there is a buffer available to be consumed
    myBuffer = flow->getBuffer();

    //We checked we have the buffer.  After this point, you are ready to read or take data out of the Buffer.
    if ( !myBuffer ) {
      cout << sf->strError() << endl;
      delete sf; return 1;
    }
    
    //If you use the method Buffer::push_data() to write in your buffer, you HAVE TO call Buffer::init_buf() first.
    //After that, you can call push_data() as many times you want to (be carefull to not put to much data in your Buffer!!!)
    myBuffer->initBuf();

    //At this point we have a Buffer object ready to written.
    //We need now to put the data in the buffer, there is several philosophy. In this client node, we use the Buffer::push_data() method.

    //We use the push_data() method to write in the buffer. The first parameter is the adress of what you want to write in the Buffer.
    //The second parameter is the size of data you want to write
    //Here we put a double in the Buffer.
    if ( !myBuffer->pushData(&myDouble, sizeof(myDouble)) ) {
      cout << "Error while pushing the data into the buffer" << endl;	
    }

    //we make a pause otherwise it is really to fast to follow!!!
	ACE_OS::sleep(1);

    //We don't need to call the method Buffer::set_size() because we used methods from the API to fill out the Buffer. So the Buffer
    //knows how much data it needs to send.

    //We put the data in the buffer, we have finished to write in the Buffer, so we don't need this it anymore. So we released it.
    //For an output flow, this call will send the buffer using the NSDFS II network.
    flow->releaseBuffer(myBuffer); 

    //We display the double we are sending to check everything work fine.
    cout << "Double value sent: " << myDouble << endl;
    
    //We set the Buffer object to NULL, no need to destroy it. The NSDFS II API takes care of that using the releaseBuffer() call.
    myBuffer = NULL;	

    //we increment the double.
    myDouble += 0.1;
  }//end while
  
  return 0;

}

