#ifndef _Buffer_H
#define _Buffer_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "ace/Time_Value.h"

#include "Smartflow.h"
#include "MemoryManager.h"

#define ENDOFBUFFER	0
#define DATA 		1
#define DATAWITHTIMESTAMP 	2
#define METADATA 	3
#define SF_COMMAND 	4

namespace smartflow {
	
	/** \brief Class Buffer .
	 * Interface for the Buffer object of the Smartflow system.
	 * The basic piece of information of a flow, different ways are provided 
	 * for getting and setting the data and the size of the piece of data.
	 * It is NOT to be created directly by the user, but rather throught the flow 
	 * object for a correct management within the library.
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 */
	
	typedef ACE_UINT32 data_type_t;
	
	class ACE_Svc_Export BufferIterator;
	
	class ACE_Svc_Export Buffer {
	public:
		/** \brief Class constructor. 
		 * \warning This method does not initialise the object, the init() method
		 * must be used for this purpose.
		 */
		Buffer();
		
		/** \brief Class destructor.
		 */
		virtual ~Buffer();
		
		/** \brief Class initialiser. Must be called just after the object
		 * construction. 
		 * \param manager The manager in which this object will be registered.
		 * \return true if the object is correctly initialised, otherwise false. If possible,
		 * an error will be set via Smartflow::seterror() to explain the error.
		 */
		virtual bool init(MemoryManager *manager,Flow * flow);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * getSmartflow(void);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \warning This method is just a wrapper for the getSmartflow() method following the C naming guidelines.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * get_smartflow(void);
		
		/** \brief Return the instance of the MemoryManager associated whith this Buffer.
		 * \return A pointer to the MemoryManager manager object, otherwise NULL.
		 */ 
		MemoryManager * getManager(void);
		
		/** \brief Return the instance of the MemoryManager associated whith this Buffer.
		 * \warning This method is just a wrapper for the getManager() method following the C naming guidelines.
		 * \return A pointer to the MemoryManager manager object, otherwise NULL.
		 */   
		MemoryManager * get_manager(void);
		
		/** \brief Return the instance of the Flow that created this Buffer.
		 * \return A pointer to a Flow object, otherwise NULL.
		 */        
		Flow * getFlow(void);
		
		/** \brief Return the instance of the Flow that created this Buffer.
		 * \warning This method is just a wrapper for the getFlow() method following the C naming guidelines.
		 * \return A pointer to a Flow object, otherwise NULL.
		 */  
		Flow * get_flow(void);
		
		/** \brief Return an address pointing to the data area of this Buffer.
		 * The maximum size of this memory space is given by getMaxsize().
		 * The current filled size is given by getSize().
		 */
		virtual void * data(void) = 0; 
		
		/** \brief Return the maximum size of the Buffer.
		 * \return A buffer_size_t (an integer)
		 */
		virtual buffer_size_t getMaxSize(void) = 0;
		
		/** \brief Return the maximum size of the Buffer.
		 * \warning This method is just a wrapper for the getMaxSize() method following the C naming guidelines.
		 * \return A buffer_size_t (an integer)
		 */ 
		buffer_size_t get_max_size(void);
		
		/** \brief Return the current size of the Buffer.
		 * This size is the size presently used in the Buffer.
		 * \return A buffer_size_t (an integer)
		 */
		virtual buffer_size_t getSize(void) = 0;
		
		/** \brief Return the current size of the Buffer.
		 * \warning This method is just a wrapper for the getSize() method following the C naming guidelines.
		 * This size is the size presently used in the Buffer.
		 * \return A buffer_size_t (an integer)
		 */ 
		buffer_size_t get_size(void);
		
		/** \brief Set the size of the buffer
		 * This method allows you to specify how much data you put in the Buffer if needed.
		 * You need this method if you fill out your buffer using memcpy.
		 * \param new_size An interger representing the new size of the Buffer
		 * \return true if successful, false otherwise.
		 */
		virtual bool setSize(buffer_size_t new_size) = 0;
		
		/** \brief Set the size of the buffer
		 * \warning This method is just a wrapper for the setSize() method following the C naming guidelines.
		 * This method allows you to specify how much data you put in the Buffer if needed.
		 * You need this method if you fill out your buffer using memcpy.
		 * \param new_size An interger representing the new size of the Buffer
		 * \return true if successful, false otherwise.
		 */
		bool set_size(buffer_size_t new_size);
		
		/** \brief Clear the data area in the Buffer.
		 * \return true if successful, false otherwise.
		 */
		virtual bool clear(void) = 0;
		
		/** \brief Initialize the Buffer for writing before using pushData() method.
		 * This method is required if you will write in your buffer using the push_*() methods.
		 * \return true if sucessful, false otherwise.
		 */
		virtual bool initBuf(void) = 0;
		
		/** \brief Initialize the Buffer for writing before using pushData() method.
		 * \warning This method is just a wrapper for the initBuf() method following the C naming guidelines.
		 * This method is required if you will write in your buffer using the push_*() methods.
		 * \return true if sucessful, false otherwise.
		 */
		bool init_buf(void);
		
		struct timespec *ts ;
		
		/** \brief It gives you the datatype of the current block in the Buffer.
		 * You get a data type only if you use the push_*() methods to fill your buffer
		 * \return the datatype: DATA, METADATA, your custom metadata types...
		 */
		virtual data_type_t getNextDataType() = 0;
		
		/** \brief It gives you the datatype of the current block in the Buffer.
		 * \warning This method is just a wrapper for the getNextDataType() method following the C naming guidelines.
		 * You get a data type only if you use the push_*() methods to fill your buffer
		 * \return the datatype: DATA, METADATA, your custom metadata types...
		 */
		data_type_t get_next_data_type();
		
		/** \brief It gives you the size of the current block in the Buffer
		 * You get the data size only if you use the push*() methods to fill your buffer.
		 * Each call of push*() creates a new block in the Buffer.
		 * \return the size of the current block in the buffer.
		 */
		virtual buffer_size_t getNextDataSize() = 0;
		
		/** \brief It gives you the size of the current block in the Buffer
		 * \warning This method is just a wrapper for the getNextDataSize() method following the C naming guidelines.
		 * You get the data size only if you use the push*() methods to fill your buffer.
		 * Each call of push*() creates a new block in the Buffer.
		 * \return the size of the current block in the buffer.
		 */ 
		buffer_size_t get_next_data_size();
		
		/** \brief It gives you the timestamp associated with the current block.
		 * You get the timestamp only if you use the push*() methods, which add an associated timestamp.
		 * \return a timestamp initialized. If there is no timestamp for the current block, it gives you a timestamp with a 0 value.
		 */
		virtual ACE_Time_Value getNextTimeStamp() = 0;
		
		/** \brief It gives you the timestamp associated with the current block.
		 * \warning This method is just a wrapper for the getNextTimeStamp() method following the C naming guidelines.
		 * You get the timestamp only if you use the push*() methods, which add an associated timestamp.
		 * \return a timestamp initialized. If there is no timestamp for the current block, it gives you a timestamp with a 0 value.
		 */
		ACE_Time_Value get_next_time_stamp();
		
		/** \brief This method allows you to jump to the next block in the Buffer.
		 * \return true if we jump to the next block, false otherwise (i.e.: we already reached the end of the buffer)
		 */
		virtual bool gotoNextBlock() = 0; 
		
		/** \brief This method allows you to jump to the next block in the Buffer.
		 * \warning This method is just a wrapper for the gotoNextBlock() method following the C naming guidelines.
		 * \return true if we jump to the next block, false otherwise (i.e.: we already reached the end of the buffer)
		 */
		bool go_to_next_block(); 
		
		/** \brief This method allows you to put a data in the buffer.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your data block.
		 * \param size The size of your data block.
		 * \return True if succesful, otherwise false.
		 */
		virtual bool pushData(void *src, buffer_size_t size) = 0;
		
		/** \brief This method allows you to put a data in the buffer.
		 * \warning This method is just a wrapper for the pushData() method following the C naming guidelines.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your data block.
		 * \param size The size of your data block.
		 * \return True if succesful, otherwise false.
		 */
		bool push_data(void *src, buffer_size_t size);
		
		/** \brief This method allows you to put a data block in the Buffer.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your data block.
		 * \param size The size of your data block.
		 * \return True if succesful, otherwise false.
		 */
		virtual bool pushData(const void *src, buffer_size_t size) = 0;
		
		/** \brief This method allows you to put a data block in the Buffer.
		 * \warning This method is just a wrapper for the pushData() method following the C naming guidelines.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your data block.
		 * \param size The size of your data block.
		 * \return True if succesful, otherwise false.
		 */
		bool push_data(const void *src, buffer_size_t size);
		
		/** \brief This method allows you to put a timestamped data block in the Buffer. 
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your data block.
		 * \param size The size of your data block.
		 * \param ts The address of your timestamp. If you don't provide any timestamp, a timestamp initilized with the computer clock is provided.
		 * \return a boolean True if succesful, otherwise false.
		 */
		virtual bool pushDataTS(void *src, buffer_size_t size, ACE_Time_Value *ts = 0) = 0;
		
		/** \brief This method allows you to put a timestamped data block in the Buffer. 
		 * \warning This method is just a wrapper for the pushDataTS() method following the C naming guidelines.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your data block.
		 * \param size The size of your data block.
		 * \param ts The address of your timestamp. If you don't provide any timestamp, a timestamp initilized with the computer clock is provided.
		 * \return a boolean True if succesful, otherwise false.
		 */
		bool push_data_ts(void *src, buffer_size_t size, ACE_Time_Value *ts = 0);
		
		/** \brief This method allows you to put a timestamped data block in the Buffer. 
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your data block.
		 * \param size The size of your data block.
		 * \param ts The address of your timestamp. If you don't provide any timestamp, a timestamp initilized with the computer clock is provided.
		 * \return a boolean True if succesful, otherwise false.
		 */
		virtual bool pushDataTS(const void *src, buffer_size_t size, ACE_Time_Value *ts = 0) = 0;
		
		/** \brief This method allows you to put a timestamped data block in the Buffer. 
		 * \warning This method is just a wrapper for the pushDataTS() method following the C naming guidelines.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your data block.
		 * \param size The size of your data block.
		 * \param ts The address of your timestamp. If you don't provide any timestamp, a timestamp initilized with the computer clock is provided.
		 * \return a boolean True if succesful, otherwise false.
		 */
		bool push_data_ts(const void *src, buffer_size_t size, ACE_Time_Value *ts = 0);
		
		/** \brief  This method allows you to put a metadata in the Buffer.
		 * If no type is provided, the METADATA type is used. Your custom type of metadata is represented using an integer greater than 9.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your metadata.
		 * \param size The size of your metadata.
		 * \param type The type of your metadata.
		 * \return a boolean True if succesful, otherwise false.
		 */
		virtual bool pushMetadata(void *src, buffer_size_t size, data_type_t type = 0) = 0;
		
		/** \brief  This method allows you to put a metadata in the Buffer.
		 * \warning This method is just a wrapper for the pushMetadata() method following the C naming guidelines.
		 * If no type is provided, the METADATA type is used. Your custom type of metadata is represented using an integer greater than 9.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your metadata.
		 * \param size The size of your metadata.
		 * \param type The type of your metadata.
		 * \return a boolean True if succesful, otherwise false.
		 */
		bool push_metadata(void *src, buffer_size_t size, data_type_t type = 0);
		
		/** \brief  This method allows you to put a metadata in the Buffer.
		 * If no type is provided, the METADATA type is used. Your custom type of metadata is represented using an integer greater than 9.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your metadata.
		 * \param size The size of your metadata.
		 * \param type The type of your metadata.
		 * \return a boolean True if succesful, otherwise false.
		 */
		virtual bool pushMetadata(const void *src, buffer_size_t size, data_type_t type = 0) = 0;
		
		/** \brief  This method allows you to put a metadata in the Buffer.
		 * \warning This method is just a wrapper for the pushMetadata() method following the C naming guidelines.
		 * If no type is provided, the METADATA type is used. Your custom type of metadata is represented using an integer greater than 9.
		 * You must call the initBuf() method once to initialize the writing area of the Buffer before using this method.
		 * \warning Do not use this method if you decide to fill out you Buffer using the memcpy method.
		 * \param src The address of your metadata.
		 * \param size The size of your metadata.
		 * \param type The type of your metadata.
		 * \return a boolean True if succesful, otherwise false.
		 */
		bool push_metadata(const void *src, buffer_size_t size, data_type_t type = 0);
		
		/** \brief This method allows you to take a data block out of the Buffer.
		 * \warning Do not use this method if you filled out you Buffer using the memcpy method.
		 * \param src The address of your data block. src need to be allocated before calling the method.
		 * \param size The size of the data block you want to pop out of the Buffer.
		 * \return a boolean True if succesful, otherwise false.
		 */
		virtual bool popData(void *src,buffer_size_t size) = 0;
		
		/** \brief This method allows you to take a data block out of the Buffer.
		 * \warning This method is just a wrapper for the popData() method following the C naming guidelines.
		 * \warning Do not use this method if you filled out you Buffer using the memcpy method.
		 * \param src The address of your data block. src need to be allocated before calling the method.
		 * \param size The size of the data block you want to pop out of the Buffer.
		 * \return a boolean True if succesful, otherwise false.
		 */
		bool pop_data(void *src,buffer_size_t size);
		
		/** \brief This method allows you to take metadata out of the Buffer.
		 * \warning Do not use this method if you filled out you Buffer using the memcpy method.
		 * \param src The address of your metadata. src need to be allocated before calling the method.
		 * \param size The size of the metadata you want to pop out of the buffer.
		 * \return a boolean True if succesful, otherwise false.
		 */
		virtual bool popMetadata(void *src,buffer_size_t size ) = 0;   
		
		/** \brief This method allows you to take metadata out of the Buffer.
		 * \warning This method is just a wrapper for the popMetadata() method following the C naming guidelines.
		 * \warning Do not use this method if you filled out you Buffer using the memcpy method.
		 * \param src The address of your metadata. src need to be allocated before calling the method.
		 * \param size The size of the metadata you want to pop out of the buffer.
		 * \return a boolean True if succesful, otherwise false.
		 */
		bool pop_metadata(void *src,buffer_size_t size );    
		
		/** \brief  This method allows you to take a timestamped data block of the Buffer.
		 * \warning Do not use this method if you filled out you Buffer using the memcpy method.
		 * \param src A pointer on your data. src need to be allocated before calling the method.
		 * \param size The size of the data you want to pop out of the buffer.
		 * \param ts The address of your timestamp. The timestamp must be allocated before calling the method. When the method returns, the timestamps is properly initialized.
		 * \return a boolean True if succesful, otherwise false.
		 */
		virtual bool popDataTS(void *src,buffer_size_t size, ACE_Time_Value *ts )  = 0 ; 
		
		/** \brief  This method allows you to take a timestamped data block of the Buffer.
		 * \warning This method is just a wrapper for the popDataTS() method following the C naming guidelines.
		 * \warning Do not use this method if you filled out you Buffer using the memcpy method.
		 * \param src A pointer on your data. src need to be allocated before calling the method.
		 * \param size The size of the data you want to pop out of the buffer.
		 * \param ts The address of your timestamp. The timestamp must be allocated before calling the method. When the method returns, the timestamps is properly initialized.
		 * \return a boolean True if succesful, otherwise false.
		 */
		bool pop_data_ts(void *src,buffer_size_t size, ACE_Time_Value *ts ); 
		
	private:
		MemoryManager * manager_;
		Flow          * flow_;
	}; // class Buffer
	
	
	/*! \class BufferIterator 
	 * \brief Class BufferIterator .
	 * Interface for the BufferIterator object of the Smartflow system.
	 * Provide an easy way to read the Buffer. Some methods are given to get the data, 
	 * its timestamp if any, the size of the data, and/or metadata.
	 * This object needs to be created directly by the user and associated to a Buffer object.
	 * \author Nist Smartspace Group
	 * \date 01-26-2005
	 */
	
	class ACE_Svc_Export BufferIterator {
	public:
		
		/* \brief The default constructor.
		 * Create a BufferIterator object used to take data block out of the Buffer.
		 * \return a BufferIterator object.
		 */
		BufferIterator();
		
		/* \brief The default destructor.
		 * Delete the BufferIterator.
		 */
		~BufferIterator();
		
		/* \brief Return whether the current block of the Buffer has a timestamp.
		 * \return a boolean representing whether the current block of the Buffer has a timestamp.
		 */
		bool hasTimestamp();
		
		/* \brief Return whether the current block of the Buffer has a timestamp.
		 * \warning This method is just a wrapper for the hasTimestamp() method following the C naming guidelines.
		 * \return a boolean representing whether the current block of the Buffer has a timestamp.
		 */
		bool has_timestamp();
		
		/* \brief Return the timestamp associated with the current block of the Buffer.
		 * \return struct timespec reprensenting the timestamp associated with the current block in the Buffer. If the current block of the Buffer doesn't have any timestamp, the fields of the timestamp are set to 0.
		 */
		ACE_Time_Value getTimestamp();
		
		/* \brief Return the timestamp associated with the current block of the Buffer.
		 * \warning This method is just a wrapper for the getTimestamp() method following the C naming guidelines.
		 * \return struct timespec reprensenting the timestamp associated with the current block in the Buffer. If the current block of the Buffer doesn't have any timestamp, the fields of the timestamp are set to 0.
		 */
		ACE_Time_Value get_timestamp();
		
		/* \brief Return the data type associated with the current block of the Buffer. The type can be be DATA, METADATA, TIMESTAMP (means timestamped data), SF_COMMAND, ENDOFBUFFER, or a custom metadata value created by the user.
		 * Customs metadata created by the user MUST HAVE a data_type_t value greater than 9. Otherwise, it will generate an error.
		 * \return The data_type_t associated with the current block in the Buffer. 
		 */
		data_type_t getDataType();
		
		/* \brief Return the data type associated with the current block of the Buffer. The type can be be DATA, METADATA, TIMESTAMP (means timestamped data), SF_COMMAND, ENDOFBUFFER, or a custom metadata value created by the user.
		 * \warning This method is just a wrapper for the getDataType() method following the C naming guidelines.
		 * Customs metadata created by the user MUST HAVE a data_type_t value greater than 9. Otherwise, it will generate an error.
		 * \return The data_type_t associated with the current block in the Buffer. 
		 */
		data_type_t get_data_type();
		
		/* \brief Return the size of the data stored in current block. This size is usefull to know. It allows the user to know how much space he should allocate in order to get its data block.
		 * \return an unsigned long representing the the data size of the current block from the Buffer.
		 */
		ACE_UINT32 getDataSize();
		
		/* \brief Return the size of the data stored in current block. This size is usefull to know. It allows the user to know how much space he should allocate in order to get its data block.
		 * \warning This method is just a wrapper for the getDataSize() method following the C naming guidelines.
		 * \return an unsigned long representing the the data size of the current block from the Buffer.
		 */
		ACE_UINT32 get_data_size();
		
		/* \brief Copy the data from the current block of the Buffer to the memory areas pointed by the data parameter and return whether everything worked fine.
		 * \param the address of where the data block should be copied. data has to be allocated with enough space otherwise an error occured. In order to know how much space is required, you can use the getDataSize() method.
		 * \return a true if the data block has been copied without problem, false otherwise.
		 */
		bool getData(void *data);
		
		/* \brief Copy the data from the current block of the Buffer to the memory areas pointed by the data parameter and return whether everything worked fine.
		 * \warning This method is just a wrapper for the getData() method following the C naming guidelines.
		 * \param the address of where the data block should be copied. data has to be allocated with enough space otherwise an error occured. In order to know how much space is required, you can use the getDataSize() method.
		 * \return a true if the data block has been copied without problem, false otherwise.
		 */
		bool get_data(void *data);
		
		/* \brief This method associates the iterator with a Buffer. This is the first method to call after the instanciation of the BufferIterator object. 
		 * After invoking this method, the BufferIterator object can iterate on _buffer.
		 * \param _buffer a pointer to the Buffer you want to iterate on.
		 */
		void setBuffer(Buffer *_buffer);
		
		/* \brief This method associates the iterator with a Buffer. This is the first method to call after the instanciation of the BufferIterator object. 
		 * \warning This method is just a wrapper for the setBuffer() method following the C naming guidelines.
		 * After invoking this method, the BufferIterator object can iterate on _buffer.
		 * \param _buffer a pointer to the Buffer you want to iterate on.
		 */
		void set_buffer(Buffer *_buffer);
		
		/* \brief This method allows you to jump to the next data block of the Buffer associated with the boolean.
		 */
		void next();
		
		/* \brief Return whether you reached the end of the Buffer. True means you are at the end of the buffer. False means there are still data block(s) to read in the Buffer.
		 * \return a boolean telling you if you reached the end of the Buffer.
		 */
		bool end();
		
		/* \brief Initialize the BufferIterator on the Buffer. This call is required before iterating on a Buffer. This method is only call once on each Buffer.
		 * This method needs to be invoked after calling setBuffer().
		 */
		void init();
		
	private:
		bool _hasTimestamp;
		ACE_Time_Value ts;
		data_type_t dataType;
		ACE_UINT32 size;
		Buffer *buffer;
		bool endOfBuffer;
		bool isDataPoped;
		bool newBuffer;
		
		void initBuffer(Buffer *_buffer);
		
	};// class BufferIterator
	
}  // namespace smartflow
#endif
