#ifndef _ConnectionManager_H
#define _ConnectionManager_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "Smartflow.h"

namespace smartflow {
	
	/** \brief Class ConnectionManager .
	 * This class is used to for client/server communication. 
	 * Most of these methods shouldn't be called directly but rather invoked using the Smartflow object.
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 */
	class ACE_Svc_Export ConnectionManager {
		
	public:
		
		/** \brief Class constructor. 
		 * \warning This method does not initialise the object, the init()
		 * method must be used for this purpose.
		 */
		ConnectionManager();
		
		/** \brief Class destructor.
		 * Take care of any cleanup that is necessary to free the object.
		 */
		virtual ~ConnectionManager();
		
		/** \brief Class initialiser. Must be called just after the object
		 * construction. 
		 * \param sf The Smartflow object in which this object will be be registered.
		 * \return true if the object is correctly initialised, otherwise false. If possible,
		 * an error will be set via Smartflow::setError() to explain the error.
		 */
		virtual bool init(Smartflow *sf);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * getSmartflow(void);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \warning This method is just a wrapper for getSmartflow() method following the C naming guidelines.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * get_smartflow(void);
		
		/** \brief Attempt a connection with the NDFS-II Server.
		 * The init() method must have been called on this instance.
		 * \param clientname the name the client use to register to the NDFS-II network.
		 * \param clientgroup the group the client use to register to the NDFS-II network.
		 * \return true if the connection is successful, false otherwise (and set an error 
		 * with the Smartflow's setError() method if available
		 */ 
		virtual bool connectToApplicationServer(const std::string &clientname, const std::string &clientgroup) = 0;
		
		/** \brief Attempt a connection with the NDFS-II Server.
		 * \warning This method is just a wrapper for the connectToApplicationServer() method following the C naming guidelines.
		 * The init() method must have been called on this instance.
		 * \param clientname the name the client use to register to the NDFS-II network.
		 * \param clientgroup the group the client use to register to the NDFS-II network.
		 * \return true if the connection is successful, false otherwise (and set an error 
		 * with the Smartflow's setError() method if available
		 */
		bool connect_to_application_server(const std::string &clientname, const std::string &clientgroup);
		
		
		/** \brief Disconnect from the NDFS-II server.
		 * \return true if the connection was established and successfully closed,
		 * false if an error occured.
		 */
		virtual bool disconnectFromApplicationServer(void) = 0;
		
		/** \brief Disconnect from the NDFS-II server.
		 * \warning This method is just a wrapper for the disconnectFromApplicationServer() method following the C naming guidelines.
		 * \return true if the connection was established and successfully closed,
		 * false if an error occured.
		 */
		bool disconnect_from_application_server(void);
		
		/** \brief Try to register a new flow to the server. 
		 * \warnning This method should never be called directly. The user should use the flow creation methods provided by the Smartflow object.
		 * The server will actively try to register this particular flow, spawn the 
		 * duplicator, and propagate the information to the rest of the NDFS-II network.
		 * \param type The "Type" of the flow : Flow_Audio, Flow_Video, .... This is the 
		 * Flow class instance which will be instanciated
		 * \param name The "Name" of the flow, used to distinguish flows if there is 
		 * several flows of the same type in the smartflow.
		 * \param group An optional group for the flow. (Type+Group form an unique flow within the full application)
		 * \param args Flow-specific arguments.
		 * \param direction The direction of the flow (input or output).
		 * \param blockcount The number of data blocks created in shared memory to handle the data transfer.
		 * \param blocksize The maximum size of a block.
		 * \param check A boolean use to check the status of a flow without trying to register it.
		 * \param numberOfConsumersExpected indicates the number of consumer that needs to register to this flow before the data transfer starts (if producer). 0 means the transfer will start asap.
		 * \return true The flow was successfully registered with the current NDFS-II server.
		 * \return false An error occured (An error code has been set), the situation
		 * can happen if the server answer that another client is already providing this flow.
		 */
		virtual bool getFlow(const std::string &type, const std::string &name, const std::string &group , const std::string &arg, flow_direction_t direction, int blockcount, int blocksize, int check, unsigned int numberOfConsumersExpected)= 0 ;
		
		/** \brief Try to register a new flow to the server. 
		 * \warning This method is just a wrapper for the getFlow() method following the C naming guidelines.
		 * \warnning This method should never be called directly. The user should use the flow creation methods provided by the Smartflow object.
		 * The server will actively try to register this particular flow, spawn the 
		 * duplicator, and propagate the information to the rest of the NDFS-II network.
		 * \param type The "Type" of the flow : Flow_Audio, Flow_Video, .... This is the 
		 * Flow class instance which will be instanciated
		 * \param name The "Name" of the flow, used to distinguish flows if there is 
		 * several flows of the same type in the smartflow.
		 * \param group An optional group for the flow. (Type+Group form an unique flow within the full application)
		 * \param args Flow-specific arguments.
		 * \param direction The direction of the flow (input or output).
		 * \param blockcount The number of data blocks created in shared memory to handle the data transfer.
		 * \param blocksize The maximum size of a block.
		 * \param check A boolean use to check the status of a flow without trying to register it.
		 * \param numberOfConsumersExpected indicates the number of consumer that needs to register to this flow before the data transfer starts (if producer). 0 means the transfer will start asap.
		 * \return true The flow was successfully registered with the current NDFS-II server.
		 * \return false An error occured (An error code has been set), the situation
		 * can happen if the server answer that another client is already providing this flow.
		 */
		bool get_flow(const std::string &type, const std::string &name, const std::string &group , const std::string &arg, flow_direction_t direction, int blockcount, int blocksize, int check, unsigned int numberOfConsumersExpected);
		
		/** \brief Notify the server to remove a flow. 
		 * \arg type The data type of the flow.
		 * \arg name The name of the flow.
		 * \arg group The group (or connectionID of the flow).
		 * The server will try to remove this flow from its tables and propagate the information to its peer servers.
		 * \return true The flow was successfully removed from the application.
		 * \return false An error occured (An error code has been set), the flow has not been found.
		 */
		virtual bool removeFlow(const std::string &type, const std::string &name, const std::string &group)=0;
		
		/** \brief Notify the server to remove a flow. 
		 * \warning This method is just a wrapper for the removeFlow() method following the C naming guidelines.
		 * \arg type The data type of the flow.
		 * \arg name The name of the flow.
		 * \arg group The group (or connectionID of the flow).
		 * The server will try to remove this flow from its tables and propagate the information to its peer servers.
		 * \return true The flow was successfully removed from the application.
		 * \return false An error occured (An error code has been set), the flow has not been found.
		 */
		bool remove_flow(const std::string &type, const std::string &name, const std::string &group);
		
		/** \brief Check if a flow exist.
		 * \arg type The data type of the flow,
		 * \arg name The name of the flow,
		 * \arg group The group of the flow (one particular instance of a flow)
		 * \return true if the flow exists, false otherwise.
		 */
		virtual bool checkFlow(const std::string &type, const std::string &name, const std::string &group)=0;
		
		/** \brief Check if a flow exist.
		 * \warning This method is just a wrapper for the checkFlow() method following the C naming guidelines.
		 * \arg type The data type of the flow,
		 * \arg name The name of the flow,
		 * \arg group The group of the flow (one particular instance of a flow)
		 * \return true if the flow exists, false otherwise.
		 */
		bool check_flow(const std::string &type, const std::string &name, const std::string &group);
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center 
		 * to launch a client node on the local or a remote server. The method provided by the Smartflow object should be used instead.
		 * \arg clientpath a string representing the path of the client executable on the machine it is supposed to run.
		 * \arg server a string representing the server on which the client is supposed to run.
		 * \arg arguments a string representing the command line arguments to be given to the client.
		 * \arg parameters a string representing parameters for the client (internally used by the data flow)
		 * \return true if successful, otherwise false.
		 */
		virtual bool startClient(const std::string &clientpath, const std::string &server, const std::string &arguments, const std::string &parameters)=0;
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center 
		 * to launch a client node on the local or a remote server. The method provided by the Smartflow object should be used instead.
		 * \warning This method is just a wrapper for the startClient() method following the C naming guidelines.
		 * \arg clientpath a string representing the path of the client executable on the machine it is supposed to run.
		 * \arg server a string representing the server on which the client is supposed to run.
		 * \arg arguments a string representing the command line arguments to be given to the client.
		 * \arg parameters a string representing parameters for the client (internally used by the data flow)
		 * \return true if successful, otherwise false.
		 */
		bool start_client(const std::string &clientpath, const std::string &server, const std::string &arguments, const std::string &parameters);
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center 
		 *	stop a client on the local or a remote server.  The method provided by the Smartflow object should be used instead.
		 * \arg clientname a string representing the client's name.
		 * \arg clientgroup a string representing the clientgroup. Can be used to make a difference between instances of client nodes.
		 * \return true if successful, otherwise false.
		 */
		virtual bool stopClient(const std::string &clientname, const std::string &clientgroup)=0;
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center 
		 *	stop a client on the local or a remote server.  The method provided by the Smartflow object should be used instead.
		 * \warning This method is just a wrapper for the stopClient() method following the C naming guidelines.
		 * \arg clientname a string representing the client's name.
		 * \arg clientgroup a string representing the clientgroup. Can be used to make a difference between instances of client nodes.
		 * \return true if successful, otherwise false.
		 */
		bool stop_client(const std::string &clientname, const std::string &clientgroup);
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center
		 * to do the discovery process. The method provided by the Smartflow object should be used instead.
		 * \arg flowlist a string containing (when method returns) the XML representation of the current NDFS-II application running on the network. 
		 * \return true if successful, otherwise false.
		 */
		virtual bool getApplicationDescription(std::string &flowlist)=0;
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center
		 * to do the discovery process. The method provided by the Smartflow object should be used instead.
		 * \warning This method is just a wrapper for the getApplicationDescription() method following the C naming guidelines.
		 * \arg flowlist a string containing (when method returns) the XML representation of the current NDFS-II application running on the network. 
		 * \return true if successful, otherwise false.
		 */
		bool get_application_description(std::string &flowlist);
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center
		 * to get the list of the running servers on the Smartflow Network. The method provided by the Smartflow object should be used instead.
		 * \arg serverlist a string containing (when method returns) the XML representation of the different hosts participating in the NDFS-II network. 
		 * \return true if successful, otherwise false.
		 */
		virtual bool getServerList(std::string &serverlist)=0;
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center
		 * to get the list of the running servers on the Smartflow Network. The method provided by the Smartflow object should be used instead.
		 * \warning This method is just a wrapper for the getServerList() method following the C naming guidelines.
		 * \arg serverlist a string containing (when method returns) the XML representation of the different hosts participating in the NDFS-II network. 
		 * \return true if successful, otherwise false.
		 */
		bool get_server_list(std::string &serverlist);
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center
		 * to get all the informations about what recently happened on the Smartflow Network. The method provided by the Smartflow object should be used instead.
		 * \arg modifications a string containing  (when method returns) the XML representation of the last changes that happened in the NDFS-II network since its last invokation.
		 * \return true if successful, otherwise false.
		 */
		virtual bool getLastModifications(std::string &modifications)=0;
		
		/** \brief This method is used internally. It has been developped to enable the NSDFS Control Center and the Web Control Center
		 * to get all the informations about what recently happened on the Smartflow Network. The method provided by the Smartflow object should be used instead.
		 * \warning This method is just a wrapper for the getLastModifications() method following the C naming guidelines.
		 * \arg modifications a string containing  (when method returns) the XML representation of the last changes that happened in the NDFS-II network since its last invokation.
		 * \return true if successful, otherwise false.
		 */
		bool get_last_modifications(std::string &modifications);
		
		/** \brief This method is used internally. It is used to send log message to the server.
		 * This method is intended to be used internally.
		 * \arg a string representing the log message to send to the server. 
		 * \return true if successful, otherwise false.
		 */
		virtual bool sendLog(const std::string &logMessage)=0;
		
		/** \brief This method is used internally. It is used to send log message to the server.
		 * This method is intended to be used internally.
		 * \warning This method is just a wrapper for the sendLog() method following the C naming guidelines.
		 * \arg a string representing the log message to send to the server. 
		 * \return true if successful, otherwise false.
		 */
		bool send_log(const std::string &logMessage);
		
		/** \brief This method is used internally. It is used to pause a flow at the client level.
		 * This method provided by the Smartflow object or Flow object should be used instead.
		 * \arg clientname a string representing the name of the client that subscribes to the flow we want to pause. 
		 * \arg clientgroup a string representing the group of the client that subscribes to the flow we want to pause.
		 * \arg flowtype a string representing the type of the flow we want to pause.
		 * \arg flowgroup a string representing the group of the flow we want to pause.
		 * \return true if successful, otherwise false.
		 */
		virtual bool pauseFlow(const std::string &clientname, const std::string &clientgroup, const std::string &flowtype, const std::string &flowgroup)=0;
		
		/** \brief This method is used internally. It is used to pause a flow at the client level.
		 * The method provided by the Smartflow object or Flow object should be used instead.
		 * \warning This method is just a wrapper for the pauseFlow() method following the C naming guidelines.
		 * \arg clientname a string representing the name of the client that subscribes to the flow we want to pause. 
		 * \arg clientgroup a string representing the group of the client that subscribes to the flow we want to pause.
		 * \arg flowtype a string representing the type of the flow we want to pause.
		 * \arg flowgroup a string representing the group of the flow we want to pause.
		 * \return true if successful, otherwise false.
		 */
		bool pause_flow(const std::string &clientname, const std::string &clientgroup, const std::string &flowtype, const std::string &flowgroup);
		
		/** \brief This method is used internally. It is used to restart the paused input flow flowID of the
		 * client node clientID. The method provided by the Smartflow object or Flow object should be used instead.
		 * \arg clientname a string representing the name of the client that subscribes to the flow we want to restart. 
		 * \arg clientgroup a string representing the group of the client that subscribes to the flow we want to restart.
		 * \arg flowtype a string representing the type of the flow we want to pause.
		 * \arg flowgroup a string representing the group of the flow we want to pause.
		 * \return true if successful, otherwise false.
		 */
		virtual bool restartFlow(const std::string &clientname, const std::string &clientgroup, const std::string &flowtype, const std::string &flowgroup)=0;
		
		/** \brief This method is used internally. It is used to restart the paused input flow flowID of the
		 * client node clientID. The method provided by the Smartflow object or Flow object should be used instead.
		 * \warning This method is just a wrapper for the restartFlow() method following the C naming guidelines.
		 * \arg clientname a string representing the name of the client that subscribes to the flow we want to restart. 
		 * \arg clientgroup a string representing the group of the client that subscribes to the flow we want to restart.
		 * \arg flowtype a string representing the type of the flow we want to pause.
		 * \arg flowgroup a string representing the group of the flow we want to pause.
		 * \return true if successful, otherwise false.
		 */
		bool restart_flow(const std::string &clientname, const std::string &clientgroup, const std::string &flowtype, const std::string &flowgroup);
		
	protected:
		Smartflow * sf_;
	}; // class ConnectionManager
}  // namespace SMARTFLOW
#endif 
