#ifndef _ErrorManager_H
#define _ErrorManager_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "ace/SOCK_Stream.h"

#include "Smartflow.h"

#include <string>

namespace smartflow {
	
	enum duplicator_error_enum {
		NO_ERRORS = 0,
		INVALID_APPLICATION_NAME,
		APPLICATION_ALREADY_EXIST,
		SMARTFLOW_NOT_READY,
		INVALID_FLOW,
		INVALID_PREFERENCE_FILE_XML,
		INVALID_PREFERENCE_FILE_XML_NOWRITE,
		INVALID_PREFERENCE_PAIR,
		SERVER_CANNOT_CONNECT,
		SERVER_IS_NOT_CONNECTED,
		SERVER_INVALID_PARAMETER,
		SERVER_INVALID_PROTOCOL,
		SERVER_ERROR_RESPONSE,
		FLOW_NOT_SET_ON_BUFFER,
		FLOW_MANAGER_NOT_EMPTY_ON_STARTUP,
		FLOW_ALREADY_EXIST,
		FLOW_NOT_EXIST,
		FLOW_CANNOT_FIND_DLL,
		FLOW_INVALID_DLL,
		FLOW_DLL_INSTANCE_NOT_VALID,
		FLOW_INVALID_FLOW_STATE,
		FLOW_CANNOT_CREATE_IMPL,
		FLOWIMPL_CANNOT_SETUP_HISTORY,
		FLOWIMPL_CANNOT_RUN_THREAD,
		FLOWIMPL_INVALID_FLOW_STATE,
		FLOWIMPL_INVALID_QUEUE_STATE,
		FLOWIMPL_CANNOT_OPEN_DUPLICATOR,
		FLOWIMPL_SERVICE_UNEXPECTED_QUIT,
		BUFFER_INVALID_SIZE,
		FLOW_PROVIDER_PAUSE_ERROR,
		FLOW_QUEUE_POLICY_INVALID,
		FLOW_SYNCHRONIZED_INVALID,
		BUFFER_ALLOCATION_FAILED,
		NOT_YET_IMPLEMENTED,
		UNKNOWN_ERROR = -1
	};
	
	struct Smartflow_Error {
		duplicator_error_enum code;
		const std::string message;
	};
	
	/** \brief Class ErrorManager .
	 * This class hold all the error code and message of the Smartflow system.
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 */
	class ACE_Svc_Export ErrorManager {
		
	public:
		
		/** \brief Class constructor. 
		 * \warning This method does not initialise the object, the init() method
		 * must be used for this purpose.
		 */
		ErrorManager();
		
		/** \brief Class destructor.
		 * Take care of any cleanup that is necessary to free the object.
		 */
		virtual ~ErrorManager();
		
		/** \brief Class initialiser. Must be called just after the object
		 * construction. 
		 * \param sf The Smartflow object in which this object will be be registered.
		 * \return true if the object is correctly initialised, otherwise false. If possible,
		 * an error will be set via Smartflow::seterror() to explain the error.
		 */
		virtual bool init(Smartflow *sf);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * getSmartflow(void);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \warning This method is just a wrapper for the getSmartflow() method following the C naming guidelines.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * get_smartflow(void);
		
		/** \brief Set an error code
		 * \param e the error code to set.
		 */
		void setError(long e) { error_ = e; }
		
		/** \brief Set an error code
		 * \warning This method is just a wrapper for the setError() method following the C naming guidelines.
		 * \param e the error code to set.
		 */
		void set_error(long e) { setError(e); }
		
		/** \brief Get the error number currently set.
		 *  \return The error number 
		 */
		long error(void)           { return error_; }
		
		/** \brief Return a string explaining the error currently set.
		 *  There is no end of line char at the end for the string 
		 * (neither CR or LF).
		 * \return a string object containing the error message associated with the error code presently set.
		 */
		const std::string & strError(void);
		
		/** \brief Return a string explaining the error currently set.
		 * \warning This method is just a wrapper for the strError() method following the C naming guidelines.
		 *  There is no end of line char at the end for the string 
		 * (neither CR or LF).
		 * \return a string object containing the error message associated with the error code presently set.
		 */
		const std::string & str_error(void);
		
		
	private:
		Smartflow * sf_;
		long error_;
		static const Smartflow_Error duplicator_error[];
		
	}; // class ErrorManager
	
}  // namespace SMARTFLOW
#endif
