#ifndef _Flow_H
#define _Flow_H

#include "ace/svc_export.h" // This is used for dynamic loading and  for exported symbols (windows)

#include "Smartflow.h"
#include "FlowManager.h"
#include "FlowImpl.h"
#include "Buffer.h"

#include <string>

/* This is used for dynamic loading */
#define Smartflow_Export ACE_Svc_Export
#define SF_FLOW_CREATION_SYMBOL ACE_TEXT("create_flow")
typedef smartflow::Flow*  (*Flow_Creator) (void);

namespace smartflow {
	
	/** \brief Class Flow .
	 * Represent a data flow, either input or output.
	 * It provide the basic methods common to each flow but is not supposed to be used nor created 
	 * directly. Instead, the Smartflow object's makeFlow methods should be used
	 * to create and initialise it.
	 * Also, this class is a template for dynamic plugins that are type-specific.
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 */
	class ACE_Svc_Export Flow {
	public:
		/** \brief Class constructor.newflo 
		 * \warning This method does not initialise the object, the init() method
		 * must be used for this purpose.
		 * \warning This method should not be called directly by users, use
		 * Smartflow::makeInputFlow() or makeOutputFlow() instead. 
		 */
		Flow();
		
		/** \brief Class destructor.
		 */
		virtual ~Flow();
		
		/** \brief Class initialiser. Must be called just after the object
		 * construction. 
		 * This method is not supposed to be called by users, who should 
		 * create a flow by using the makeInputFlow() or makeOutputFlow()
		 * methods of the Smartflow class.
		 * \param manager The manager in which this object will be registered.
		 * \param type The type of this instance of the flow (e.g. Flow_Audio).
		 * \param name The name of this instance of the flow.
		 * \param group The group (or connection ID) of this instance of the flow.
		 * \param args The arguments of this instance of the flow.
		 * \param mode FLOW_INPUT or FLOW_OUTPUT.
		 * \param policy The policy mode applied to the queue: DO_NOT_DROP_BUFFER (the flow is blocking),
		 * DROP_MOST_RECENT_BUFFER (the most recent buffer is dropped when a new one is received if the history of the flow is running full),
		 * DROP_OLDEST_BUFFER (the oldest buffer is dropped when a new one is received if the history of the flow is running full).
		 * \param isSynchroized specify if the flow is synchronized (i.e every single data buffer sent will be received, regardless of when the consumers are joining the network).
		 * \return true if the object is correctly initialised, otherwise false. If possible,
		 * an error will be set via Smartflow::seterror() to explain the error.
		 */
		virtual bool init(FlowManager *manager, const std::string &type, const std::string &name, const std::string &rgroup, const std::string &args, flow_direction_t mode, flow_policy_t policy, bool isSynchronized);
		
		/** \brief Start the data transfer if available. 
		 * \return true if success, otherwise false.
		 */
		virtual bool start(void);
		
		/** \brief Stop & cleanup method. Must be called when exiting, even if init
		 * was not successful.
		 * Invoking this method will make the getBuffer() method returns even if there is no buffer available.
		 * \param waitForEmptyHistory False by default. If true, wait for the history to be empty before stopping the flow.
		 * \return true if successfully close, otherwise false.
		 */
		virtual bool stop(bool waitForEmptyHistory = false);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */ 
		Smartflow * getSmartflow(void);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \warning This method is just a wrapper for the getSmartflow() method following the C naming guidelines.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */ 
		Smartflow * get_smartflow(void);
		
		/** \brief Return the instance of the FlowManager class where is object is registered.
		 * \return A pointer to the FlowManager manager object, otherwise NULL.
		 */ 
		FlowManager * getManager(void);
		
		/** \brief Return the instance of the FlowManager class where is object is registered.
		 * \warning This method is just a wrapper for the getManager() method following the C naming guidelines.
		 * \return A pointer to the FlowManager manager object, otherwise NULL.
		 */ 
		FlowManager * get_manager(void);
		
		/** \brief Get the type of the flow as specified in the first argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \return An std::string representing the name of the flow.
		 */
		virtual std::string & getFlowType(void)=0;
		
		/** \brief Get the type of the flow as specified in the first argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \warning This method is just a wrapper for the getFlowType() method following the C naming guidelines.
		 * \return An std::string representing the name of the flow.
		 */
		std::string & get_flow_type(void);
		
		/** \brief Get the Description of the flow, an string whose
		 * purpose is to give a textual description of the flow.
		 * The size of the text is unspecified, and it can include 
		 * spaces. (It is not used beside the description itself)
		 * Must be reimplemented by Type-specific flows classes.
		 * \return a string providing a textual description of the flow.
		 */
		virtual std::string & getFlowDesc(void)=0;
		
		/** \brief Get the Description of the flow, an string whose
		 * purpose is to give a textual description of the flow.
		 * The size of the text is unspecified, and it can include 
		 * spaces. (It is not used beside the description itself).
		 * \warning This method is just a wrapper for the getFlowDesc() method following the C naming guidelines.
		 * \return a string providing a textual description of the flow.
		 */
		std::string & get_flow_desc(void);
		
		/** \brief Get the the direction of the flow (input or output)
		 * \return a flow_direction_t giving the direction of the flow.
		 */
		flow_direction_t getMode(void);
		
		/** \brief Get the the direction of the flow (input or output)
		 * \warning This method is just a wrapper for the getMode() method following the C naming guidelines.
		 * \return a flow_direction_t giving the direction of the flow.
		 */
		flow_direction_t get_mode(void);
		
		/** \brief Get the flow policy applied to the flow
		 * \return flow_policy_t the flow policy applied to the flow: it can be
		 * not dropping any buffer, dropping most recent buffer if
		 * queue full, dropping oldest buffer if queue full.
		 */
		flow_policy_t getFlowQueuePolicy(void);
		
		/** \brief Get the flow policy applied to the flow
		 * \warning This method is just a wrapper for the getFlowQueuePolicy() method following the C naming guidelines.
		 * \return flow_policy_t the flow policy applied to the flow: it can be
		 * not dropping any buffer, dropping most recent buffer if
		 * queue full, dropping oldest buffer if queue full.
		 */
		flow_policy_t get_flow_queue_policy(void);
		
		/** \brief Return the unique ID of the flow. This ID is made
		 * using the flow name and flow group.
		 * \return An std::string representing the ID of the flow.
		 */
		std::string & getId(void);
		
		/** \brief Return the unique ID of the flow. This ID is made
		 * using the flow name and flow group.
		 * \warning This method is just a wrapper for the getId() method following the C naming guidelines.
		 * \return An std::string representing the ID of the flow.
		 */
		std::string & get_id(void);
		
		/** \brief Get the name of the flow as specified in the second argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \return a std::string representing the name of the flow.
		 */
		std::string & getName(void);
		
		/** \brief Get the name of the flow as specified in the second argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \warning This method is just a wrapper for the getName() method following the C naming guidelines.
		 * \return a std::string representing the name of the flow.
		 */
		std::string & get_name(void);
		
		/** \brief Get the type of the flow as specified in the first argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \return An std::string representing the name of the flow.
		 */
		std::string & getType(void);
		
		/** \brief Get the type of the flow as specified in the first argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \warning This method is just a wrapper for the getType() method following the C naming guidelines.
		 * \return An std::string representing the name of the flow.
		 */
		std::string & get_type(void);
		
		/** \brief Get the args of the flow as specified in the third argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \return An std::string representing the type of the flow.
		 */
		std::string & getArgs(void);
		
		/** \brief Get the args of the flow as specified in the third argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \warning This method is just a wrapper for the getArgs() method following the C naming guidelines.
		 * \return An std::string representing the type of the flow.
		 */
		std::string & get_args(void);
		
		/** \brief Get the group of the flow as specified in the fourth argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \return An std::string representing the group of the flow.
		 */
		std::string & getGroup(void);
		
		/** \brief Get the group of the flow as specified in the fourth argument of 
		 * the Smartflow::makeInputFlow or Smartflow::makeOutputFlow methods.
		 * \warning This method is just a wrapper for the getGroup() method following the C naming guidelines.
		 * \return An std::string representing the group of the flow.
		 */
		std::string & get_group(void);
		
		/* \brief True means the flow is Synchronized, false means the flow is not synchronized.
		 * A synchronized output flow waits for its consumer before starting the data transfer
		 * in order to ensure that no data block is lost.
		 * An input flow cannot be synchronized. This method will always return false for an input flow.
		 * \return True if synchronized, false otherwise
		 */
		bool isSynchronized() { return isSynchronized_; };
		
		/* \brief True means the flow is Synchronized, false means the flow is not synchronized.
		 * \warning This method is just a wrapper for the isSynchronized() method following the C naming guidelines.
		 * A synchronized output flow waits for its consumer before starting the data transfer
		 * in order to ensure that no data block is lost.
		 * An input flow cannot be synchronized. This method will always return false for an input flow.
		 * \return True if synchronized, false otherwise
		 */
		bool is_synchronized();
		
		/** \brief Get the history size of this flow.
		 * This is the number of Buffer in the queue (or history) of this flow.
		 * \return a long giving the history size in Buffers.
		 */
		long getHistorySize(void);
		
		/** \brief Get the history size of this flow.
		 * \warning This method is just a wrapper for the getHistorySize() method following the C naming guidelines.
		 * This is the number of Buffer in the queue (or history) of this flow.
		 * \return a long giving the history size in Buffers.
		 */
		long get_history_size(void);
		
		/** \brief Set the history size of this flow.
		 * This is the maximum number of Buffers in the queue (or history) of this flow
		 * at any time.
		 * \param size giving the desired size of the queue (or history).
		 * \return true is properly set, false otherwise.
		 */
		bool setHistorySize(long size);
		
		/** \brief Set the history size of this flow.
		 * \warning This method is just a wrapper for the setHistorySize() method following the C naming guidelines.
		 * This is the maximum number of Buffers in the queue (or history) of this flow
		 * at any time.
		 * \param size giving the desired size of the queue (or history).
		 * \return true is properly set, false otherwise.
		 */
		bool set_history_size(long size);
		
		/** \brief Set the size of the data area of a buffer.
		 * This is the size of the usable payload for the client user.
		 * It cannot go over the size returned by getMaxBufferSize().
		 * \param size representing the maximum Buffers size.
		 * \return true is properly set, false otherwise.
		 */
		bool setMaxBufferSize(buffer_size_t size);
		
		/** \brief Set the size of the data area of a buffer.
		 * \warning This method is just a wrapper for the setMaxBufferSize() method following the C naming guidelines.
		 * This is the size of the usable payload for the client user.
		 * It cannot go over the size returned by getMaxBufferSize().
		 * \param size representing the maximum Buffers size.
		 * \return true is properly set, false otherwise.
		 */
		bool set_max_buffer_size(buffer_size_t size);
		
		/** \brief Get the maximum size of the Buffer.
		 * \return a buffer_size_t representing the maximum size of the Buffer. 
		 */
		buffer_size_t getMaxBufferSize(void);
		
		/** \brief Get the maximum size of the Buffer.
		 * \warning This method is just a wrapper for the getMaxBufferSize() method following the C naming guidelines.
		 * \return a buffer_size_t representing the maximum size of the Buffer. 
		 */
		buffer_size_t get_max_buffer_size(void);
		
		/** \brief Set the number of blocks the duplicator will use.
		 * \param size The number of shared memory blocks used by the duplicator. 
		 * \return true is properly set, false otherwise. 
		 */
		bool setBlockCount(int size);
		
		/** \brief Set the number of blocks the duplicator will use.
		 * \warning This method is just a wrapper for the setBlockCount() method following the C naming guidelines.
		 * \param size The number of shared memory blocks used by the duplicator. 
		 * \return true is properly set, false otherwise. 
		 */
		bool set_block_count(int size);
		
		/** \brief Get the number of blocks the duplicator will use.
		 * \return an integer, the number of blocks used by the duplicator. 
		 */
		int getBlockCount(void);
		
		/** \brief Get the number of blocks the duplicator will use.
		 * \warning This method is just a wrapper for the getBlockCount() method following the C naming guidelines.
		 * \return an integer, the number of blocks used by the duplicator. 
		 */
		int get_block_count(void);
		
		/** \brief Return the watermark of the flow queue (a double between 0 and 1)
		 * Each instance of a flow has a queue (or history). The watermark is used if the queue get full.
		 * If the queue is full, no more Buffer can be enqueued until the level of the queue
		 * drop under the watermark. By default, the watermark is set to 0.8 (80%). That means
		 * that if the queue get full, data will be enqueue only when the level of the queue 
		 * decrease of 20%.
		 * This method is for advanced users. This method allows to tune the queue of the flow.
		 * This queue is never handled directly by the user.
		 * \Return a double between 0 and 1 representing the watermark.
		 */
		double getWaterMark(void);
		
		/** \brief Return the watermark of the flow queue (a double between 0 and 1)
		 * \warning This method is just a wrapper for the getWaterMark() method following the C naming guidelines.
		 * Each instance of a flow has a queue (or history). The watermark is used if the queue get full.
		 * If the queue is full, no more Buffer can be enqueued until the level of the queue
		 * drop under the watermark. By default, the watermark is set to 0.8 (80%). That means
		 * that if the queue get full, data will be enqueue only when the level of the queue 
		 * decrease of 20%.
		 * This method is for advanced users. This method allows to tune the queue of the flow.
		 * This queue is never handled directly by the user.
		 * \Return a double between 0 and 1 representing the watermark.
		 */
		double get_water_mark(void);
		
		/** \brief Set the watermark of the flow queue (a double between 0 and 1)
		 * Each instance of a flow has a queue (or history). The watermark is used if the queue get full.
		 * If the queue is full, no more Buffer can be enqueued until the level of the queue
		 * drop under the watermark. By default, the watermark is set to 0.8 (80%). That means
		 * that if the queue get full, data will be enqueue only when the level of the queue 
		 * decrease of 20%.
		 * This method is for advanced users. This method allows to tune the queue of the flow.
		 * This queue is never handled directly by the user
		 * \param water_mark, a double between 0 and 1 included.
		 * \return true if properly set, false otherwise.
		 */
		bool setWaterMark(double water_mark);
		
		/** \brief Set the watermark of the flow queue (a double between 0 and 1)
		 * \warning This method is just a wrapper for the setWaterMark() method following the C naming guidelines.
		 * Each instance of a flow has a queue (or history). The watermark is used if the queue get full.
		 * If the queue is full, no more Buffer can be enqueued until the level of the queue
		 * drop under the watermark. By default, the watermark is set to 0.8 (80%). That means
		 * that if the queue get full, data will be enqueue only when the level of the queue 
		 * decrease of 20%.
		 * This method is for advanced users. This method allows to tune the queue of the flow.
		 * This queue is never handled directly by the user
		 * \param water_mark, a double between 0 and 1 included.
		 * \return true if properly set, false otherwise.
		 */
		bool set_water_mark(double water_mark);
		
		/** \brief Return a double between 0 and 1 representing the filling level of the queue (or history).
		 * 0 means the queue is empty, 1 means the queue is full.
		 * This method is used for debugging purpose.
		 * \return A double beetween 0 and 1 representing the filling level of the flow queue.
		 */
		double getFillingLevelOfQueue();
		
		/** \brief Return a double between 0 and 1 representing the filling level of the queue (or history).
		 * \warning This method is just a wrapper for the getFillingLevelOfQueue() method following the C naming guidelines.
		 * 0 means the queue is empty, 1 means the queue is full.
		 * This method is used for debugging purpose.
		 * \return A double beetween 0 and 1 representing the filling level of the flow queue.
		 */
		double get_filling_level_of_queue();
		
		/** \brief Return true is there is a Buffer available.
		 * The method Flow::getBuffer() doesn't return until there
		 * is a buffer available. This method allows to test if there is a buffer ready.
		 * \return A boolean true if there is a buffer available, false otherwise
		 */
		bool isBufferAvailable();
		
		/** \brief Return true is there is a Buffer available.
		 * \warning This method is just a wrapper for the isBufferAvailable() method following the C naming guidelines.
		 * The method Flow::getBuffer() doesn't return until there
		 * is a buffer available. This method allows to test if there is a buffer ready.
		 * \return A boolean true if there is a buffer available, false otherwise
		 */
		bool is_buffer_available();
		
		/** \brief Get an available Buffer from the NDFS-II library.
		 * For an output flow, the Buffer is empty.
		 * For an input flow, the Buffer is filled with data.
		 * \warning In both cases, the function will block if no buffer can 
		 * be returned.
		 * \return The returned buffer, of NULL (and and error is set with setError()).
		 */
		virtual Buffer	* getBuffer(void);
		
		/** \brief Get an available Buffer from the NDFS-II library.
		 * \warning This method is just a wrapper for the getBuffer() method following the C naming guidelines.
		 * For an output flow, the Buffer is empty.
		 * For an input flow, the Buffer is filled with data.
		 * \warning In both cases, the function will block if no buffer can 
		 * be returned.
		 * \return The returned buffer, of NULL (and and error is set with setError()).
		 */
		Buffer	* get_buffer(void);
		
		/** \brief Return a Buffer to the NDFS-II library.
		 * For an output flow, the buffer will be put in line for delivery.
		 * For an input flow, the buffer will be reused for another data bunch.
		 * \param buf The buffer to return.
		 * \return true if no errors, otherwise false (and an error is set), in particular
		 * if the user return a pointer on a buffer not given by getBuffer().
		 */
		virtual bool releaseBuffer(Buffer * buf);
		
		/** \brief Return a Buffer to the NDFS-II library.
		 * \warning This method is just a wrapper for the releaseBuffer() method following the C naming guidelines.
		 * For an output flow, the buffer will be put in line for delivery.
		 * For an input flow, the buffer will be reused for another data bunch.
		 * \param buf The buffer to return.
		 * \return true if no errors, otherwise false (and an error is set), in particular
		 * if the user return a pointer on a buffer not given by getBuffer().
		 */
		bool release_buffer(Buffer * buf);
		
		/** \brief Return a buffer to the Smartflow library.
		 * The buffer will be put back into the free list.
		 * There is no difference between this method and releaseBuffer() for an
		 * input flow. However, for an output flow, it mean the buffer will *not*
		 * be delivered.
		 * \param buf The buffer to return.
		 * \return true if no errors, otherwise false (and an error is set), in particular
		 * if the user return a pointer on a buffer not given by getBuffer().         
		 */
		virtual bool cancelBuffer(Buffer * buf);
		
		/** \brief Return a buffer to the Smartflow library.
		 * \warning This method is just a wrapper for the cancelBuffer() method following the C naming guidelines.
		 * The buffer will be put back into the free list.
		 * There is no difference between this method and releaseBuffer() for an
		 * input flow. However, for an output flow, it mean the buffer will *not*
		 * be delivered.
		 * \param buf The buffer to return.
		 * \return true if no errors, otherwise false (and an error is set), in particular
		 * if the user return a pointer on a buffer not given by getBuffer().         
		 */
		bool cancel_buffer(Buffer * buf);
		
		/* \brief Put the current flow in pause.
		 * A paused flow doesn't get anymore data untill the restart method is called.
		 * This method can only be used with input flows.
		 * All data not received are lost.
		 * \return true if succesful, otherwis false.
		 */
		bool pause();
		
		/* \brief Restart the current flow if it is paused.
		 * After a restart call, the client node can receive data again from the flow.
		 * You can only restart input flows.
		 * \return true if succesful, otherwis false.
		 */
		bool restart();
		
	private:
		FlowManager * manager_;
		flow_direction_t      mode_;
		flow_policy_t         policy_;
		//a synchronized flow only makes sense for output flows, it means the transfer of data starts when every expected consumers are effectively connected
		bool					  isSynchronized_;
		std::string           args_;
		std::string           type_;
		std::string           id_;
		std::string           name_;
		std::string           rgroup_;
		
	protected:
		
		/** \brief Extract a value from a string. 
		 * The arg string must be in the form of : "key=value key2=value2 [...]"
		 * this method return a value from a key. If the key does not exists, then
		 * the empty string is returned.
		 * \param keyword The key to be located
		 * \param arg The string that will be parsed
		 * \return the value string, or an empty string
		 */
		static std::string parseArgs(std::string keyword, std::string arg);
		
		/** \brief This is a pointer on the Implementation class, following the 
		 * "Delegation design pattern". It mean that all low-level operations
		 * on flows will be deferred to this implementation object. 
		 * \internal
		 */
		FlowImpl           *flowimpl ;
	}; // class Flow
	
}  // namespace SMARTFLOW
#endif
