#ifndef _FlowManager_H
#define _FlowManager_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "Smartflow.h"

namespace smartflow {
	
	/** \brief Class FlowManager .
	 * This class is an interface for the FlowManager, the class that take care
	 * of the registration, management, loading, etc... of flows instances.
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 */
	class ACE_Svc_Export FlowManager {
		
	public:
		
		/** \brief Class constructor. 
		 * \warning This method does not initialise the object, the init() method
		 * must be used for this purpose.
		 */
		FlowManager();
		
		/** \brief Class destructor.
		 * Take care of any cleanup that is necessary to free the object.
		 */
		virtual ~FlowManager();
		
		/** \brief Class initialiser. Must be called just after the object
		 * construction. 
		 * \param sf The Smartflow object in which this object will be be registered.
		 * \return true if the object is correctly initialised, otherwise false. If possible,
		 * an error will be set via Smartflow::setError() to explain the error.
		 */
		virtual bool init(Smartflow *sf);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * getSmartflow(void);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \warning This method is just a wrapper for the getSmartflow() method following the C naming guidelines.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * get_smartflow(void);
		
		/** \brief Create a flow object.
		 * The Flow object is dynamic : The system check the type, then look on 
		 * the default path on the disk for a DLL file of the corresponding name
		 * (eg. FLOWTYPE.DLL, libflowtype.so ). It load it if not already in memory,
		 * and run it's init() function after dynamic link edition.
		 * This function in turn create a Flow object of the real type, calling its 
		 * constructor and returning a pointer to it.
		 * \param type The type of the flow
		 * \param name The name of the flow (used to refer to the flow)
		 * \param group The group of the flow (to distinguish 2 flow of the same type)
		 * \param arg  optional arguements to pass to the flow (Flow-specific)
		 * \param direction Wether the flow is input or output.
		 * \param policy A flag to specify the flow policy applied to the flow queue if the queue is full.
		 * \param numberOfConsumersExpected The number of consumer, which need 
		 *  to be connected to this flow before the data transfer starts. If numberOfConsumersExpected  is
		 *  equal to 0, the flow is not synchronized. If numberOfConsumersExpected is greater than 0, this
		 *  means that the flow is synchronized, the data transfer will only begin when the number of expected
		 *  consumers are actually connected.
		 * \return a flow pointer of the requiered type (dynamic_cast<> is necessary)
		 */
		virtual Flow *makeFlow(const std::string &type, const std::string &name, const std::string &group, const std::string &arg, flow_direction_t direction, flow_policy_t policy, unsigned int numberOfConsumersExpected)=0;
		
		/** \brief Create a flow object.
		 * \warning This method is just a wrapper for the makeFlow() method following the C naming guidelines.
		 * The Flow object is dynamic : The system check the type, then look on 
		 * the default path on the disk for a DLL file of the corresponding name
		 * (eg. FLOWTYPE.DLL, libflowtype.so ). It load it if not already in memory,
		 * and run it's init() function after dynamic link edition.
		 * This function in turn create a Flow object of the real type, calling its 
		 * constructor and returning a pointer to it.
		 * \param type The type of the flow
		 * \param name The name of the flow (used to refer to the flow)
		 * \param group The group of the flow (to distinguish 2 flow of the same type)
		 * \param arg  optional arguements to pass to the flow (Flow-specific)
		 * \param direction Wether the flow is input or output.
		 * \param policy A flag to specify the flow policy applied to the flow queue if the queue is full.
		 * \param numberOfConsumersExpected The number of consumer, which need 
		 *  to be connected to this flow before the data transfer starts. If numberOfConsumersExpected  is
		 *  equal to 0, the flow is not synchronized. If numberOfConsumersExpected is greater than 0, this
		 *  means that the flow is synchronized, the data transfer will only begin when the number of expected
		 *  consumers are actually connected.
		 * \return a flow pointer of the requiered type (dynamic_cast<> is necessary)
		 */
		Flow *make_flow(const std::string &type, const std::string &name, const std::string &group, const std::string &arg, flow_direction_t direction, flow_policy_t policy, unsigned int numberOfConsumersExpected);
		
		/** \brief Locate an existing flow by name within a client.
		 * \param fname the full name of the flow (the full name is built with using the flow name and flow group.
		 * Return the Flow if found or NULL otherwise.
		 */
		virtual Flow * findFlow(const std::string &fname)=0;
		
		/** \brief Locate an existing flow by name within a client.
		 * \warning This method is just a wrapper for the findFlow() method following the C naming guidelines.
		 * \param fname the full name of the flow (the full name is built with using the flow name and flow group.
		 * Return the Flow if found or NULL otherwise.
		 */
		Flow * find_flow(const std::string &fname);
		
		/** \brief Remove a flow from the manager. 
		 * After the call, the object is not a runnable flow anymore, and cannot 
		 * be reused (the last this to do is delete on it).
		 * Calling this method also stops the flow.
		 * \param f the Flow instance to destroy.
		 * \return true for success, false otherwise.
		 */
		virtual bool destroyFlow(Flow *f)=0;
		
		/** \brief Remove a flow from the manager. 
		 * \warning This method is just a wrapper for the destroyFlow() method following the C naming guidelines.
		 * After the call, the object is not a runnable flow anymore, and cannot 
		 * be reused (the last this to do is delete on it).
		 * Calling this method also stops the flow.
		 * \param f the Flow instance to destroy.
		 * \return true for success, false otherwise.
		 */
		bool destroy_flow(Flow *f);
		
		/** \brief Stop all the flows of the client. 
		 * \return true for success, false otherwise.
		 */
		virtual bool stopAllFlows()=0;
		
		/** \brief Stop all the flows of the client. 
		 * \warning This method is just a wrapper for the stopAllFlows() method following the C naming guidelines.
		 * \return true for success, false otherwise.
		 */
		bool stop_all_flows();
		
		
	private:
		Smartflow * sf_;
	}; // class FlowManager
}  // namespace SMARTFLOW
#endif
