#ifndef _MemoryManager_H
#define _MemoryManager_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "Smartflow.h"

namespace smartflow {
	
	/** \brief Class MemoryManager .
	 * This is an abstaction for a memory manager who provide buffers.
	 * The implementation can then uses various tricks to recycle buffer,
	 * do batch allocation or more elaborate functions.
	 * This is for local buffers only - not for the shared memory buffers that are 
	 * taken care of by the duplicator only.
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 */
	class ACE_Svc_Export MemoryManager {
		
	public:
		
		/** \brief Class constructor. 
		 * \warning This method does not initialise the object, the init() method
		 * must be used for this purpose.
		 */
		MemoryManager();
		
		/** \brief Class destructor.
		 * Take care of any cleanup that is necessary to free the object.
		 */
		virtual ~MemoryManager();
		
		/** \brief Class initialiser. Must be called just after the object
		 * construction. 
		 * \param sf The Smartflow object  which to be be registered with.
		 * \return true if the object is correctly initialised, otherwise false. If possible,
		 * an error will be set via Smartflow::setError() to explain the error.
		 */
		virtual bool init(Smartflow *sf);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * get_smartflow(void);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \warning This method is just a wrapper for the getSmartflow() method following the C naming guidelines.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * getSmartflow(void);
		
		/** \brief Return a allocated buffer for the flow.
		 * \param A pointer to the flow.
		 * \param sizeOfBuffer the size of the buffer.
		 * \return A pointer to an allocated buffer object.
		 */
		virtual Buffer * mallocBuffer(Flow * flow, buffer_size_t sizeOfBuffer = 0)=0;
		
		/** \brief Return a allocated buffer for the flow.
		 * \warning This method is just a wrapper for the mallocBuffer() method following the C naming guidelines.
		 * \param A pointer to the flow.
		 * \param sizeOfBuffer the size of the buffer.
		 * \return A pointer to an allocated buffer object.
		 */
		Buffer * malloc_buffer(Flow * flow, buffer_size_t sizeOfBuffer = 0);
		
		/** \brief Free the buffer
		 * \param buf The Buffer to free.
		 */
		virtual void freeBuffer(Buffer *buf)=0;
		
		/** \brief Free the buffer
		 * \warning This method is just a wrapper for the freeBuffer() method following the C naming guidelines.
		 * \param buf The Buffer to free.
		 */
		void free_buffer(Buffer *buf);
		
	private:
		Smartflow * sf_;
	}; // class MemoryManager
}  // namespace SMARTFLOW
#endif
