#ifndef _Preference_H
#define _Preference_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "Smartflow.h"
#include "PreferenceManager.h"

#include <string>

using namespace std;

namespace smartflow {
	
	/** \brief Class Preference .
	 * Represent one way of accessing preferences, like an XML file, or a command-line
	 * argument.
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 */
	class ACE_Svc_Export Preference {
	public:
		
		/** \brief Class constructor. 
		 * \warning This method does not initialise the object, the init() method
		 * must be used for this purpose.
		 */
		Preference();
		
		/** \brief Class destructor.
		 * Take care of any cleanup that is necessary to free the object.
		 */
		virtual ~Preference();
		
		/** \brief Class initialiser. Must be called just after the object
		 * construction. 
		 * \param manager The manager with which this object will be registered.
		 * \return true if the object is correctly initialised, otherwise false. If possible,
		 * an error will be set via Smartflow::setError() to explain the error.
		 */
		virtual bool init(PreferenceManager *manager);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * getSmartflow(void);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \warning This method is just a wrapper for the getSmartflow() method following the C naming guidelines.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * get_smartflow(void);
		
		/** \brief Return the instance of the PreferenceManager class where is object is registered.
		 * \return A pointer to the PreferenceManager manager object, otherwise NULL.
		 */ 
		PreferenceManager * getManager(void);
		
		/** \brief Return the instance of the PreferenceManager class where is object is registered.
		 * \warning This method is just a wrapper for the getManager() method following the C naming guidelines.
		 * \return A pointer to the PreferenceManager manager object, otherwise NULL.
		 */
		PreferenceManager * get_manager(void);
		
		/** \brief This function allows to add a preference without a comment. 
		 * A preference is represented with 2 strings. The first one is the key, 
		 * and the second one is the value. If the preference already exists, 
		 * this functions does nothing. If the key already exists and the new value
		 * is different of the old one, the new value will replace the old one.
		 * The arguments (the preference) are stored in the PreferenceManager.
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 * \param value This is the value (a string) associated with the key.
		 */
		void addPreference(string key, string value);
		
		/** \brief This function allows to add a preference without a comment. 
		 * \warning This method is just a wrapper for the addPreference() method following the C naming guidelines.
		 * A preference is represented with 2 strings. The first one is the key, 
		 * and the second one is the value. If the preference already exists, 
		 * this functions does nothing. If the key already exists and the new value
		 * is different of the old one, the new value will replace the old one.
		 * The arguments (the preference) are stored in the PreferenceManager.
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 * \param value This is the value (a string) associated with the key.
		 */
		void add_preference(string key, string value);
		
		/** \brief This function allows to add a preference with a comment.
		 * A preference is represented with 3 strings. The first one is the key, 
		 * and the second one is the value. The third one is a comment of this preference.
		 * If the preference already exists, this functions does nothing. If the key already 
		 * exists and the new value is different of the old one, the new value will replace 
		 * the old one. If the comment doesn't exist, it is added. If it already exists, 
		 * it is modified.
		 * The arguments (the preference) are stored in the PreferenceManager.
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 * \param value This is the value (a string) associated with the key.
		 * \param comment This is a string, which comments the preference.
		 */
		void addPreference(string key, string value, string comment);
		
		/** \brief This function allows to add a preference with a comment.
		 * \warning This method is just a wrapper for the addPreference() method following the C naming guidelines.
		 * A preference is represented with 3 strings. The first one is the key, 
		 * and the second one is the value. The third one is a comment of this preference.
		 * If the preference already exists, this functions does nothing. If the key already 
		 * exists and the new value is different of the old one, the new value will replace 
		 * the old one. If the comment doesn't exist, it is added. If it already exists, 
		 * it is modified.
		 * The arguments (the preference) are stored in the PreferenceManager.
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 * \param value This is the value (a string) associated with the key.
		 * \param comment This is a string, which comments the preference.
		 */
		void add_preference(string key, string value, string comment);
		
	private:
		PreferenceManager * manager_;
	}; // class Preference
}  // namespace SMARTFLOW
#endif
