#ifndef _PreferenceManager_H
#define _PreferenceManager_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "Smartflow.h"

#include <string>

using namespace std;
namespace smartflow {
	
	/** \brief Class PreferenceManager .
	 * An abstraction for a list of preference, with methods to access them.
	 * A preference is defined as a key=value tuple.
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 */
	class ACE_Svc_Export PreferenceManager {
		
	public:
		
		/** \brief Class constructor. 
		 * \warning This method does not initialise the object, the init() method
		 * must be used for this purpose.
		 * 
		 */
		PreferenceManager();
		
		/** \brief Class destructor.
		 * Take care of any cleanup that is necessary to free the object.
		 */
		virtual ~PreferenceManager();
		
		/** \brief Class initialiser. Must be called just after the object
		 * construction. 
		 * \param sf The Smartflow object in which this object will be be registered.
		 * \return true if the object is correctly initialised, otherwise false. If possible,
		 * an error will be set via Smartflow::setError() to explain the error.
		 */
		virtual bool init(Smartflow *sf);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * getSmartflow(void);
		
		/** \brief Return the Smartflow object associated with an instance of the class.
		 * \warning This method is just a wrapper for the getSmartflow() method following the C naming guidelines.
		 * \return A pointer to the smartflow object, or NULL if an error occured.
		 */
		Smartflow * get_smartflow(void);
		
		/** This function allows to add a preference with an optional comment.
		 * A preference is represented with 3 strings. The first one is the key, 
		 * and the second one is the value. The third one is a comment of this preference.
		 * If the preference already exists, this functions does nothing. If the key already 
		 * exists and the new value is different of the old one, the new value will replace 
		 * the old one. If the comment doesn't exist, it is added. If it already exists, 
		 * it is modified. 
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 * \param value This is the value (a string) associated with the key.
		 * \param comment This is a string, which comments the preference.
		 */
		virtual void addPreference(string key, string value, string comment="");
		
		/** This function allows to add a preference with an optional comment.
		 * \warning This method is just a wrapper for the addPreference() method following the C naming guidelines.
		 * A preference is represented with 3 strings. The first one is the key, 
		 * and the second one is the value. The third one is a comment of this preference.
		 * If the preference already exists, this functions does nothing. If the key already 
		 * exists and the new value is different of the old one, the new value will replace 
		 * the old one. If the comment doesn't exist, it is added. If it already exists, 
		 * it is modified. 
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 * \param value This is the value (a string) associated with the key.
		 * \param comment This is a string, which comments the preference.
		 */
		void add_preference(string key, string value, string comment="");
		
		/** This function allows to remove a preference.
		 * If the preference doesn't exist, nothing append.
		 * \param key This is the key of the preference, which will be remove.
		 */   
		virtual void remPreference(string key);
		
		/** This function allows to remove a preference.
		 * \warning This method is just a wrapper for the remPreference() method following the C naming guidelines.
		 * If the preference doesn't exist, nothing append.
		 * \param key This is the key of the preference, which will be remove.
		 */
		void rem_preference(string key);
		
		/** This functions allows to get the value of a preference.
		 * The preference is identified by its key. If the key doesn't exist, 
		 * the function returns an empty string.
		 * \return string The value of the preference associated with the key. 
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 */
		virtual string getPreferenceValue(string key);
		
		/** This functions allows to get the value of a preference.
		 * \warning This method is just a wrapper for the getPreferenceValue() method following the C naming guidelines.
		 * The preference is identified by its key. If the key doesn't exist, 
		 * the function returns an empty string.
		 * \return string The value of the preference associated with the key. 
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 */
		string get_preference_value(string key);
		
		/** This functions allows to get the integer value of a preference.
		 * If the preference is an alphanumeric string, the function returns 0 
		 * and a warning is raised. If the preference is a float, the function 
		 * returns the integer value of the float truncated.
		 * \return int The integer value of a preference. 
		 * \param key This is a string, which allows to identify a preference.
		 */ 
		int getPreferenceIntValue(string key);
		
		/** This functions allows to get the integer value of a preference.
		 * \warning This method is just a wrapper for the getPreferenceIntValue() method following the C naming guidelines.
		 * If the preference is an alphanumeric string, the function returns 0 
		 * and a warning is raised. If the preference is a float, the function 
		 * returns the integer value of the float truncated.
		 * \return int The integer value of a preference. 
		 * \param key This is a string, which allows to identify a preference.
		 */ 
		int get_preference_int_value(string key);
		
		/** This functions allows to get the double value of a preference. 
		 * If the preference is an alphanumeric string, the function returns
		 * 0 and a warning is raised.
		 * \return double The double value of a preference.
		 * \param key This is a string, which allows to identify a preference.
		 */
		double getPreferenceDoubleValue(string key);
		
		/** This functions allows to get the double value of a preference. 
		 * \warning This method is just a wrapper for the getPreferenceDoubleValue() method following the C naming guidelines.
		 * If the preference is an alphanumeric string, the function returns
		 * 0 and a warning is raised.
		 * \return double The double value of a preference.
		 * \param key This is a string, which allows to identify a preference.
		 */
		double get_preference_double_value(string key);
		
		/** This functions allows to get the optionnal comment of a preference.
		 * The comment (a string) is identified by the key. If there is no comment 
		 * associted with the key or the key doesn't exist, the function returns an empty string.
		 * \return string The comment associated with the key. 
		 * \param key This is a string, which allows to identify a preference.
		 */ 
		string getPreferenceComment(string key);
		
		/** This functions allows to get the optionnal comment of a preference.
		 * \warning This method is just a wrapper for the getPreferenceComment() method following the C naming guidelines.
		 * The comment (a string) is identified by the key. If there is no comment 
		 * associted with the key or the key doesn't exist, the function returns an empty string.
		 * \return string The comment associated with the key. 
		 * \param key This is a string, which allows to identify a preference.
		 */
		string get_preference_comment(string key);
		
		/** This function allows to see the list of the preferences on stdout.
		 */
		virtual void viewPrefs()=0;
		
		/** This function allows to see the list of the preferences on stdout.
		 * \warning This method is just a wrapper for the viewPrefs() method following the C naming guidelines.
		 */
		void view_prefs();
		
		/** This function allows to get the key of a preference using his number.
		 * If the number is invalid (upper or lower), the function returns
		 * an empty key. The first key has the number 0.
		 * \return string The key of the preference identified by his number.
		 * \param num The number of the key.
		 */
		string getPreferenceKey(int num);
		
		/** This function allows to get the key of a preference using his number.
		 * \warning This method is just a wrapper for the getPreferenceKey() method following the C naming guidelines.
		 * If the number is invalid (upper or lower), the function returns
		 * an empty key. The first key has the number 0.
		 * \return string The key of the preference identified by his number.
		 * \param num The number of the key.
		 */
		string get_preference_key(int num);
		
		/** This function returns the count of the preferences.
		 * \return int The total number of the preferences. 
		 */
		virtual int count()=0;
		
		/** This function uses the command-line arguments to setup the 
		 * couples (key,values) that the user has given.
		 * The argc/argv arguments are modified to remove the used arguments.
		 * \return -1 if there has been an error in setting the 
		 * preferences. The function does not return false on a parse error,
		 * since the user might want to use argc,argv for his own usage.
		 */
		virtual bool initArgs(int& argc, char**& argv, bool keepSFOptions)=0;
		
		/** This function uses the command-line arguments to setup the 
		 * couples (key,values) that the user has given.
		 * \warning This method is just a wrapper for the initArgs() method following the C naming guidelines.
		 * The argc/argv arguments are modified to remove the used arguments.
		 * \return -1 if there has been an error in setting the 
		 * preferences. The function does not return false on a parse error,
		 * since the user might want to use argc,argv for his own usage.
		 */
		bool init_args(int& argc, char**& argv, bool keepSFOptions);
		
	private:
		Smartflow * sf_;
		
		/** This is a private function, which allows to add a preference without a comment.
		 * The developper never uses this functions. It is called by the method 
		 * addPreference of the Preferencemanager class.
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 * \param value This is the value (a string) associated with the key.
		 */
		virtual void addPref(string &key, string &value)=0;
		
		/** This is a private function, which allows to add a preference with a comment.
		 * The developper never uses this functions. It is called by the method 
		 * addPreference of the PreferenceManager class.
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 * \param value This is the value (a string) associated with the key.
		 * \param comment This is a string, which comments the preference.
		 */
		virtual void addPref(string &key, string &value, string &comment)=0;
		
		/** This is a private function, which allows to remove a preference.
		 * The developper never uses this functions. It is called by the method 
		 * remPreference of the Preferencemanager class.
		 * \param key This is the key of the preference, which will be remove.
		 */
		virtual void remPref(string &key)=0;
		
		/** This is a private function, which allows to get a preference (a string).
		 * A preference is identified by its key.
		 * The developper never uses this functions. It is called by the method 
		 * getPreferenceValue of the Preferencemanager class.
		 * \return string The value of the preference associated with the key. 
		 * \param key This is the ID of the preference. A key, which is a string, is unique.
		 */
		virtual string getPrefValue(string &key)=0;
		
		/** This is a private function, which allows to get the integer value of
		 * a preference. A preference is identified by its key.
		 * The developper never uses this functions. It is called by the method 
		 * getPreferenceIntValue of the Preferencemanager class.
		 * \return int The integer value of a preference.
		 * \param key This is a string, which allows to identify a preference.
		 */
		virtual int getPrefIntValue(string &key)=0;
		
		/** This is a private function, which allows to get the float value of a preference. 
		 * A preference is identified by its key.
		 * The developper never uses this functions. It is called by the method 
		 * getPreferenceDoubleValue of the Preferencemanager class.
		 * \return double The double value of a preference.
		 * \param key This is a string, which allows to identify a preference.
		 */
		virtual double getPrefDoubleValue(string &key)=0;
		
		/** This is a private function, which allows to get the key of a preference.
		 * A preference is identified by its number.    
		 * This number must be > 0 and less than the total of the preferences
		 * The developper never uses this functions. It is called by the method 
		 * getPreferenceKey of the Preferencemanager class.
		 * \return string The key of the preference identified by his number.
		 * \param num The number of the key.
		 */ 
		virtual string getPrefKey(int num)=0;
		
		/** This is a private function, which allows to get the comment of a preference.
		 * A preference is identified by its key.
		 * The developper never uses this functions. It is called by the method 
		 * getPreferenceComment of the Preferencemanager class.
		 * \return string The comment associated with the key. 
		 * \param key This is a string, which allows to identify a preference.
		 */
		virtual string getPrefComment(string &key)=0;
		
		/** This private function is used to see if a comment exist for the 
		 * preference identified by the key.
		 * \param key This is a string, which allows to identify a preference.
		 */
		virtual bool existsComment(string &key)=0;
		
		/** This private function is used in the class PreferenceManager_XXX
		 * to check is the preference identified by the key exist in the 
		 * list of preferences.
		 * \param key This is a string, which allows to identify a preference.
		 */
		virtual bool exists(string &key)=0;
		
	}; // class PreferenceManager
}  // namespace SMARTFLOW
#endif
