#ifndef _SMARTFLOW_H
#define _SMARTFLOW_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "ace/Synch.h"

#include <string>
#include <map>

static ACE_Thread_Mutex m; 

/** \namespace smartflow All the smartflow classes belongs to this namespace.
 This helps to reduce name collision when several libraries are used.
 Users can add the instruction "using namespace SMARTFLOW" at the begining
 of their files to use all the classes for the Smartflow 2 API
 */
namespace smartflow {
	
	/* Smartflow API classes pre-declarations */
	class ACE_Svc_Export Smartflow;
	/* Managers */
	class ACE_Svc_Export ErrorManager;
	class ACE_Svc_Export MemoryManager;
	class ACE_Svc_Export FlowManager;
	class ACE_Svc_Export PreferenceManager;
	class ACE_Svc_Export ConnectionManager;
	class ACE_Svc_Export MessageManager;
	class ACE_Svc_Export Log_Manager;
	
	
	/* Objects */
	class ACE_Svc_Export Buffer;
	class ACE_Svc_Export Flow;
	class ACE_Svc_Export FlowImpl;
	class ACE_Svc_Export Preference;
	class ACE_Svc_Export Connection;
	class ACE_Svc_Export Message;
	
	typedef long long address_type_t;
	typedef ACE_UINT32 buffer_size_t;
	
	/** Describe a flow direction, from a client point of view.
	 * An "INPUT FLOW" is a flow that a client consume.
	 * An "OUTPUT FLOW" is a a flow that a client provide.
	 * For a client, a given flow can be either INPUT or OUTPUT, but not both.
	 */
	enum flow_direction_t {
		FLOW_INVALID = 0,
		FLOW_INPUT,
		FLOW_OUTPUT
	};	 
	
	/**Describe what is the policy applied to the flow queue.
	 * Each flow has the queue, which can contains some Buffers. 
	 * The user never manipulate the queue. However some queue 
	 * options can be specified at the flow creation in order to
	 * set the queue policy.
	 * "DO_NOT_DROP_BUFFER" is the default policy mode, 
	 * if the queue is full and a new buffer is received, it will 
	 * only be enqueued when there is room in the queue.
	 * "DROP_MOST_RECENT_BUFFER": if the queue is full and a new
	 * buffer comes, the buffer is dropped.
	 * "DROP_MOST_OLDEST_BUFFER": if the queue is full and a new
	 * buffer comes, the oldest buffer(s) in the queue is dropped 
	 * in order to make room for this new one.
	 */
	enum flow_policy_t {
		INVALID_POLICY = 0,
		DO_NOT_DROP_BUFFER,
		DROP_MOST_RECENT_BUFFER,
		DROP_OLDEST_BUFFER
	};
	
	
#define DEBUG_PREFIX		LM_DEBUG,		ACE_TEXT ("%I[CLI  DEBUG   %T] ") 
#define WARNING_PREFIX		LM_WARNING,		ACE_TEXT ("%I[CLI  WARNING %T] ")
#define ERROR_PREFIX		LM_ERROR,		ACE_TEXT ("%I[CLI  ERROR   %T] ")
#define USER_PREFIX		LM_ERROR,		ACE_TEXT ("%I[CLI  USER    %T] ")
#define USER_PREFIX_EMPTY	LM_ERROR,		ACE_TEXT ("")
#define ANNOUNCE_PREFIX		LM_ERROR,		ACE_TEXT ("%I[CLI  ANNOUNCE%T] ")
	
	
#define DEFAULT_FLOW_GROUP "default"
#define DEFAULT_CLIENT_GROUP "default"
#define DEFAULT_APPLICATION_NAME "default"
#define DEFAULT_EMPTY ""
	
} // namespace smartflow

namespace smartflow {
	
	/** NDFS-II library main class.
	 * This class is used as an "entry point" for all the other classes in the library, and
	 * all the other classes are created directly or indirectly by methods of a Smartflow
	 * object, and never directly form the user.
	 * For the configurations options, here the order followed :\n
	 * 1) Hard-coded defaults\n
	 * 2) Default configuration file\n
	 * 3) Client-given parameters\n
	 * 4) Command-line given parameter file\n
	 * 5) Command-line parameters\n
	 * ( See the PreferenceManager documentation for details )
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 */
	class ACE_Svc_Export Smartflow {
	public:
		/** \brief Default class constructor. 
		 * \warning This constructor does not initialise the object, the init() method
		 * must be used for this purpose.
		 */
		Smartflow();
		
		/** \brief Class destructor.
		 * Take care of any cleanup that is necessary to free the object correctly
		 */
		~Smartflow();
		
		/** \brief Class initialiser. Must be called just after the object
		 * construction.
		 * This method initialise the various "Managers" requiered by the Smartflow.
		 * It use the argc,argv, 
		 * \param argc argc from the main() function
		 * \param argv argv from the main() function
		 * \param keepSFOptions A boolean, if true, the SF parameters from the command line are kept from argv. 
		 * By default, they are remove, only leaving  the user specific arguments. 
		 * \param application the application name. There is a default if it is not
		 * provided, but it is strongly recomended to set one. The commandline can also
		 * be used to specify it.
		 * \return true if the object is correctly initialised, otherwise false. If possible,
		 * an error will be set via Smartflow::setError() to explain the error.
		 */        
		bool init(int &argc, char** &argv, bool keepSFOptions = false, const std::string application=DEFAULT_APPLICATION_NAME);
		
		/** \brief Create an input flow.
		 * \param type The type of the flow : Audio, Video, .... This is the 
		 * Flow class instance which will be instanciated
		 * \param name The name of the flow is used to distinguish flows if there is 
		 * several flows of the same type in the NDFS-II network.
		 * \param args Flow-specific arguments, such as blocksize or blockcount.
		 * \param group An optional group for the flow. (Type+Group form an unique flow within the full application)
		 * \param policy An optional flag to specify the flow policy applied to the flow queue if the queue is full.
		 * \return A pointer on a Flow instance, of the requiered type (cast is necessary)
		 * or NULL if an error occured, then an error code is available via the error() method.
		 */
		Flow * makeInputFlow( const std::string type, const std::string name, const std::string args=DEFAULT_EMPTY, const std::string group=DEFAULT_FLOW_GROUP, const flow_policy_t policy=DO_NOT_DROP_BUFFER);
		
		/** \brief Create an input flow.
		 * \warning This method is just a wrapper for the makeInputFlow() method following the C naming guidelines.
		 * \param type The type of the flow : Audio, Video, .... This is the 
		 * Flow class instance which will be instanciated
		 * \param name The name of the flow is used to distinguish flows if there is 
		 * several flows of the same type in the NDFS-II network.
		 * \param args Flow-specific arguments, such as blocksize or blockcount.
		 * \param group An optional group for the flow. (Type+Group form an unique flow within the full application)
		 * \param policy An optional flag to specify the flow policy applied to the flow queue if the queue is full.
		 * \return A pointer on a Flow instance, of the requiered type (cast is necessary)
		 * or NULL if an error occured, then an error code is available via the error() method.
		 */
		Flow * make_input_flow( const std::string type, const std::string name, const std::string args=DEFAULT_EMPTY, const std::string group=DEFAULT_FLOW_GROUP, const flow_policy_t policy=DO_NOT_DROP_BUFFER);
		
		/** \brief Create an input flow.
		 * \param type The type of the flow : Audio, Video, .... This is the 
		 * Flow class instance which will be instanciated
		 * \param name The name of the flow is used to distinguish flows if there is 
		 * several flows of the same type in the NDFS-II network.
		 * \param args Flow-specific arguments, such as blocksize or blockcount.
		 * \param group An optional group for the flow. (Type+Group form an unique flow within the full application)
		 * \param policy An optional flag to specify the flow policy applied to the flow queue if the queue is full.
		 * \return A pointer on a Flow instance, of the requiered type (cast is necessary)
		 * or NULL if an error occured, then an error code is available via the error() method.
		 */	
		Flow * makeOutputFlow(const std::string type, const std::string name, const std::string args=DEFAULT_EMPTY, const std::string group=DEFAULT_FLOW_GROUP, const flow_policy_t policy=DO_NOT_DROP_BUFFER);
		
		/** \brief Create an input flow.
		 * \warning This method is just a wrapper for the makeOutputFlow() method following the C naming guidelines.
		 * \param type The type of the flow : Audio, Video, .... This is the 
		 * Flow class instance which will be instanciated
		 * \param name The name of the flow is used to distinguish flows if there is 
		 * several flows of the same type in the NDFS-II network.
		 * \param args Flow-specific arguments, such as blocksize or blockcount.
		 * \param group An optional group for the flow. (Type+Group form an unique flow within the full application)
		 * \param policy An optional flag to specify the flow policy applied to the flow queue if the queue is full.
		 * \return A pointer on a Flow instance, of the requiered type (cast is necessary)
		 * or NULL if an error occured, then an error code is available via the error() method.
		 */
		Flow * make_output_flow(const std::string type, const std::string name, const std::string args=DEFAULT_EMPTY, const std::string group=DEFAULT_FLOW_GROUP, const flow_policy_t policy=DO_NOT_DROP_BUFFER);
		
		/** \brief Create a synchronized Output flow. Synchronized means data will never be lost if there is no consumer.
		 * The data transfer will only begin when all the expected consumer are connected to the input flow.
		 * You need to use this method if you use the data flow to transfer files.
		 * \param type The type of the flow : Audio, Video, .... This is the 
		 * Flow class instance which will be instanciated
		 * \param name The name of the flow is used to distinguish flows if there is 
		 * several flows of the same type in the NDFS-II network.
		 * \param numberOfConsumersExpected The number of consumer, which need to be connected to this flow before the data
		 * transfer starts. When Flow::start() is invoked, it may seem that some Buffers are transfered before every expected consumers 
		 * connect to the flow, but they are not and stored in the queue until they the number of expected consumers is reached.
		 * \param args Flow-specific arguments, such as blocksize or blockcount.
		 * \param group An optional group for the flow. (Type+Group form an unique flow within the full application)
		 * \return A pointer on a Flow instance, of the requiered type (cast is necessary)
		 * or NULL if an error occured, then an error code is available
		 * via the error() method.
		 */
		Flow * makeSynchronizedOutputFlow(const std::string type, const std::string name, unsigned int numberOfConsumersExpected, const std::string args=DEFAULT_EMPTY,const std::string group=DEFAULT_FLOW_GROUP);
		
		/** \brief Create a synchronized Output flow. Synchronized means data will never be lost if there is no consumer.
		 * \warning This method is just a wrapper for the makeSynchronizedOutputFlow() method following the C naming guidelines.
		 * The data transfer will only begin when all the expected consumer are connected to the input flow.
		 * You need to use this method if you use the data flow to transfer files.
		 * \param type The type of the flow : Audio, Video, .... This is the 
		 * Flow class instance which will be instanciated
		 * \param name The name of the flow is used to distinguish flows if there is 
		 * several flows of the same type in the NDFS-II network.
		 * \param numberOfConsumersExpected The number of consumer, which need to be connected to this flow before the data
		 * transfer starts. When Flow::start() is invoked, it may seem that some Buffers are transfered before every expected consumers 
		 * connect to the flow, but they are not and stored in the queue until they the number of expected consumers is reached.
		 * \param args Flow-specific arguments, such as blocksize or blockcount.
		 * \param group An optional group for the flow. (Type+Group form an unique flow within the full application)
		 * \return A pointer on a Flow instance, of the requiered type (cast is necessary)
		 * or NULL if an error occured, then an error code is available
		 * via the error() method.
		 */
		Flow * make_synchronized_output_flow(const std::string type, const std::string name, unsigned int numberOfConsumersExpected, const std::string args=DEFAULT_EMPTY,const std::string group=DEFAULT_FLOW_GROUP);
		
		/** \brief Delete a flow.
		 * Take care of disconnecting, waiting for threads to terminate, and deregistering the flow.
		 * Invoking this method automatically stopped the flow. 
		 * \param flow The flow to destroy.
		 * \return true if success, false otherwise.
		 */
		bool destroyFlow( Flow * flow );
		
		/** \brief Delete a flow.
		 * \warning This method is just a wrapper for the destroyFlow() method following the C naming guidelines.
		 * Take care of disconnecting, waiting for threads to terminate, and deregistering the flow.
		 * Invoking this method automatically stopped the flow. 
		 * \param flow The flow to destroy.
		 * \return true if success, false otherwise.
		 */
		bool destroy_flow( Flow * flow );
		
		/** \brief Return the status of the Smartflow.
		 * \return true if the Smartflow instance is ready to connect to the 
		 * application server, otherwise false. 
		 */
		bool readyToRun();
		
		/** \brief Return the status of the Smartflow.
		 * \warning This method is just a wrapper for the readyToRun() method following the C naming guidelines.
		 * \return true if the Smartflow instance is ready to connect to the 
		 * application server, otherwise false. 
		 */
		bool ready_to_run();
		
		/** \brief Stop this Smartflow instance.
		 * That mean remove the connected flows, disconnect from peers, an 
		 * free all the running objects.
		 * The object can be reused by calling init() with new parameters.
		 * \return true if the stopping is successful, otherwise false, in this case
		 * an error code is set.
		 * The destructor call this method if necessary.
		 */    
		bool stop();
		
		/** \brief Return the current application name.
		 * \return the application name, or an empty string is it is not set yet.
		 * this method never fails.
		 */	
		std::string & applicationName(void);
		
		/** \brief Return the current application name.
		 * \warning This method is just a wrapper for the applicationName() method following the C naming guidelines.
		 * \return the application name, or an empty string is it is not set yet.
		 * this method never fails.
		 */	
		std::string & application_name(void);
		
		/** \brief Connect to the application server.
		 * The name of the application to be used is given by application_name(), so
		 * it must be set prior to call this method. Also, ready_to_run() must return
		 * true.
		 * The method uses the connection manager to establish to connection, and to 
		 * get the status back. 
		 * \return true if the connection is successful, otherwise false, and in this case
		 * an error is set.
		 */    
		bool connectToApplicationServer(const std::string clientname=DEFAULT_EMPTY,
										const std::string clientgroup=DEFAULT_CLIENT_GROUP);
		
		/** \brief Connect to the application server.
		 * \warning This method is just a wrapper for the connectToApplicationServer() method following the C naming guidelines.
		 * The name of the application to be used is given by application_name(), so
		 * it must be set prior to call this method. Also, ready_to_run() must return
		 * true.
		 * The method uses the connection manager to establish to connection, and to 
		 * get the status back. 
		 * \return true if the connection is successful, otherwise false, and in this case
		 * an error is set.
		 */
		bool connect_to_application_server(const std::string clientname=DEFAULT_EMPTY,
										   const std::string clientgroup=DEFAULT_CLIENT_GROUP);
		
		/** \brief Disconnect from the currently connected application server. 
		 * \return true if the disconnection is successful, otherwise false.
		 */
		bool disconnectFromApplicationServer(void);
		
		/** \brief Disconnect from the currently connected application server. 
		 * \warning This method is just a wrapper for the disconnectFromApplicationServer() method following the C naming guidelines.
		 * \return true if the disconnection is successful, otherwise false.
		 */
		bool disconnect_from_application_server(void);
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center in order to 
		 * launch a client node on the local or a remote server.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \arg clientpath a string representing the path of the client executable on the machine it is supposed to run.
		 * \arg server a string representing the server on which the client is supposed to run.
		 * \arg arguments a string representing the command line arguments to be given to the client.
		 * \arg parameters a string representing parameters for the client (internally used by the data flow).
		 * \return true if successful, otherwise false.
		 */
		bool startClient(const std::string clientpath, 
						 const std::string server,
						 const std::string arguments=DEFAULT_EMPTY,
						 const std::string parameters=DEFAULT_EMPTY);
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center in order to 
		 * launch a client node on the local or a remote server.
		 * \warning This method is just a wrapper for the startClient() method following the C naming guidelines.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \arg clientpath a string representing the path of the client executable on the machine it is supposed to run.
		 * \arg server a string representing the server on which the client is supposed to run.
		 * \arg arguments a string representing the command line arguments to be given to the client.
		 * \arg parameters a string representing parameters for the client (internally used by the data flow).
		 * \return true if successful, otherwise false.
		 */
		bool start_client(const std::string clientpath, 
						  const std::string server,
						  const std::string arguments=DEFAULT_EMPTY,
						  const std::string parameters=DEFAULT_EMPTY);
		
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center in order to 
		 * stop a client on the local or a remote host.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \arg clientname a string representing the client's name.
		 * \arg clientgroup a string representing the clientgroup. Can be used to make a difference between instances of client nodes.
		 * \return true if successful, otherwise false.
		 */
		bool stopClient(const std::string clientname,const std::string clientgroup);
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center in order to 
		 * stop a client on the local or a remote host.
		 * \warning This method is just a wrapper for the stopClient() method following the C naming guidelines.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \arg clientname a string representing the client's name.
		 * \arg clientgroup a string representing the clientgroup. Can be used to make a difference between instances of client nodes.
		 * \return true if successful, otherwise false.
		 */
		bool stop_client(const std::string clientname,const std::string clientgroup);
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center to 
		 * do the discovery process. It returns a XML description of the current application running
		 * on the NDFS-II network.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \param flowlist a string containing (when method returns) the XML representation of the current NDFS-II application running on the network.
		 * \return true if success, false otherwise. 
		 */
		bool getApplicationDescription(std::string &flowlist);
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center to 
		 * do the discovery process. It returns a XML description of the current application running
		 * on the NDFS-II network.
		 * \warning This method is just a wrapper for the getApplicationDescription() method following the C naming guidelines.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \param flowlist a string containing (when method returns) the XML representation of the current NDFS-II application running on the network.
		 * \return true if success, false otherwise. 
		 */
		bool get_application_description(std::string &flowlist);
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center to 
		 * get the list of the running servers on the NDFS-II network.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \arg serverlist a string containing (when method returns) the XML representation of the different hosts participating in the NDFS-II network.
		 * \return true if success, false otherwise. 
		 */
		bool getServerList(std::string &serverlist);
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center to 
		 * get the list of the running servers on the NDFS-II network.
		 * \warning This method is just a wrapper for the getServerList() method following the C naming guidelines.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \arg serverlist a string containing (when method returns) the XML representation of the different hosts participating in the NDFS-II network.
		 * \return true if success, false otherwise. 
		 */
		bool get_server_list(std::string &serverlist);
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center in order to 
		 * get all the informations about what recently happened on the NDFS-II network.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \arg modifications a string containing  (when method returns) the XML representation of the last changes that happened in the NDFS-II network since its last invokation.
		 * \return true if success, false otherwise. 
		 */ 
		bool getLastModifications(std::string &lastmodifications);
		
		/** \brief This method is used by the NSDFS Control Center and the Web Control Center in order to 
		 * get all the informations about what recently happened on the NDFS-II network.
		 * \warning This method is just a wrapper for the getLastModifications() method following the C naming guidelines.
		 * \warning This method is intented to be used by client belonging to the 'admin' group.
		 * \arg modifications a string containing  (when method returns) the XML representation of the last changes that happened in the NDFS-II network since its last invokation.
		 * \return true if success, false otherwise. 
		 */ 
		bool get_last_modifications(std::string &lastmodifications);
		
		/** \brief Suspend the transport of data of the flow flowID to the consumer clientID.  
		 * This method only works with input flows.
		 * If you pause of flow, you miss data the time the flow is paused
		 * \arg clientname a string representing the name of the client that subscribes to the flow we want to pause. 
		 * \arg clientgroup a string representing the group of the client that subscribes to the flow we want to pause.
		 * \arg flowtype a string representing the type of the flow we want to pause.
		 * \arg flowgroup a string representing the group of the flow we want to pause.
		 * \return true if successful, otherwise false.
		 */
		bool pauseFlow(const std::string &clientname,const std::string &clientgroup,const std::string &flowtype,const std::string &flowgroup);
		
		/** \brief Suspend the transport of data of the flow flowID to the consumer clientID.  
		 * \warning This method is just a wrapper for the pauseFlow() method following the C naming guidelines.
		 * This method only works with input flows.
		 * If you pause of flow, you miss data the time the flow is paused
		 * \arg clientname a string representing the name of the client that subscribes to the flow we want to pause. 
		 * \arg clientgroup a string representing the group of the client that subscribes to the flow we want to pause.
		 * \arg flowtype a string representing the type of the flow we want to pause.
		 * \arg flowgroup a string representing the group of the flow we want to pause.
		 * \return true if successful, otherwise false.
		 */
		bool pause_flow(const std::string &clientname,const std::string &clientgroup,const std::string &flowtype,const std::string &flowgroup);
		
		/** \brief Restart the transport of data of the flow flowID to the consumer clientID.
		 * This method only work with input flows.
		 * Restart the flow, which have been previously paused. 
		 * When a consumer client node pauses a flow, all the data send by the provider to this client node
		 * via this flow are lost for the consumer client node.
		 * \arg clientname a string representing the name of the client that subscribes to the flow we want to restart. 
		 * \arg clientgroup a string representing the group of the client that subscribes to the flow we want to restart.
		 * \arg flowtype a string representing the type of the flow we want to pause.
		 * \arg flowgroup a string representing the group of the flow we want to pause.
		 * \return true if successful, otherwise false.
		 */
		bool restartFlow(const std::string &clientname,const std::string &clientgroup,const std::string &flowtype,const std::string &flowgroup);
		
		/** \brief Restart the transport of data of the flow flowID to the consumer clientID.
		 * \warning This method is just a wrapper for the restartFlow() method following the C naming guidelines.
		 * This method only work with input flows.
		 * Restart the flow, which have been previously paused. 
		 * When a consumer client node pauses a flow, all the data send by the provider to this client node
		 * via this flow are lost for the consumer client node.
		 * \arg clientname a string representing the name of the client that subscribes to the flow we want to restart. 
		 * \arg clientgroup a string representing the group of the client that subscribes to the flow we want to restart.
		 * \arg flowtype a string representing the type of the flow we want to pause.
		 * \arg flowgroup a string representing the group of the flow we want to pause.
		 * \return true if successful, otherwise false.
		 */
		bool restart_flow(const std::string &clientname,const std::string &clientgroup,const std::string &flowtype,const std::string &flowgroup);
		
		/** \brief Pointer to a callback function defined by the user (in the user code) 
		 * in order to exit properly the client when the order to quit is launched 
		 * by the server (typically asked by the control center). This function
		 * can also be called when the program quit because of a Control-C or 
		 * because the user closes the console window. 
		 */
		void (*user_exit_function) ();
		
		/** \brief Pointer to a callback function (with one parameter) defined by the user 
		 * (in the user code) in order to exit properly the client when the order 
		 * to quit is launched by the server (typically asked by the control center). 
		 * This function can also be called when the program quit because of a Control-C 
		 * or because the user closes the console window. 
		 * \param userarg argument given by the client
		 */
		void (*user_exit_function_with_parameters) (void *userarg);
		
		/** \brief Pointer to the parameter of the callback function defined by the user
		 * in the user code
		 */
		void *user_exit_func_argument;
		
		/** \brief Function used by the user in order to define the callback function 
		 * that must be called to exit properly the client when the order to quit 
		 * is launched by the server (typically asked by the control center) and
		 * eventually when  the program quit because of a Control-C 
		 * or because the user closes the console window.
		 * \param user_exit_function_ pointer to the callback function defined by the client
		 * \param controlC boolean which is true if the callback function must be called
		 * when the user exits the program using control-C or closes the console window.
		 * \return true if successful, otherwise false.
		 */
		bool setUserExitFunction(void (*user_exit_function_) (), bool controlC = true);
		
		/** \brief Function used by the user in order to define the callback function 
		 * that must be called to exit properly the client when the order to quit 
		 * is launched by the server (typically asked by the control center) and
		 * eventually when  the program quit because of a Control-C 
		 * or because the user closes the console window.
		 * \warning This method is just a wrapper for the setUserExitFunction() method following the C naming guidelines.
		 * \param user_exit_function_ pointer to the callback function defined by the client
		 * \param controlC boolean which is true if the callback function must be called
		 * when the user exits the program using control-C or closes the console window.
		 * \return true if successful, otherwise false.
		 */
		bool set_user_exit_function(void (*user_exit_function_) (), bool controlC = true);
		
		/** \brief Function used by the user in order to define the callback function 
		 * that must be called to exit properly the client when the order to quit 
		 * is launched by the server (typically asked by the control center) and
		 * eventually when  the program quit because of a Control-C 
		 * or because the user closes the console window.
		 * \param user_exit_function_ pointer to the callback function defined by the client
		 * \param arg pointer to the argument which will be used as parameter of the callback 
		 * function defined by the client
		 * \param controlC boolean which is true if the callback function must be called
		 * when the user exits the program using control-C or closes the console window/
		 * \return true if successful, otherwise false.
		 */
		bool setUserExitFunction(void (*user_exit_function_) (void*userarg), void *arg, bool controlC = true);
		
		/** \brief Function used by the user in order to define the callback function 
		 * that must be called to exit properly the client when the order to quit 
		 * is launched by the server (typically asked by the control center) and
		 * eventually when  the program quit because of a Control-C 
		 * or because the user closes the console window.
		 * \warning This method is just a wrapper for the setUserExitFunction() method following the C naming guidelines.
		 * \param user_exit_function_ pointer to the callback function defined by the client
		 * \param arg pointer to the argument which will be used as parameter of the callback 
		 * function defined by the client
		 * \param controlC boolean which is true if the callback function must be called
		 * when the user exits the program using control-C or closes the console window/
		 * \return true if successful, otherwise false.
		 */
		bool set_user_exit_function(void (*user_exit_function_) (void*userarg), void *arg, bool controlC = true);
		
		/** \brief Set an error code
		 * \param error the error code to set.
		 */
		void setError(long int error);
		
		/** \brief Set an error code
		 * \warning This method is just a wrapper for the setError() method following the C naming guidelines.
		 * \param error the error code to set.
		 */
		void set_error(long int error);
		
		/** \brief Return the error code currently set.
		 * \return The error number code. 0 if no error.
		 */
		long int error(void);
		
		/** \brief Return a message associated with the current error number.
		 * \return The error message.
		 */
		std::string strError(void);
		
		/** \brief Return a message associated with the current error number.
		 * \warning This method is just a wrapper for the strError() method following the C naming guidelines.
		 * \return The error message.
		 */
		std::string str_error(void);
		
		/** \brief This static method return a smartflow instance from the application
		 * name. 
		 * The Smartflow maintain a list of all the Smartflow objects created.
		 * they can be retreived with this method.
		 * There can be only one Smartflow object per application.
		 * \param application select the required application name. If no name is
		 * given, the default name is used. If a null pointer is passed, the first
		 * smartflow in the list is returned.
		 * \return A smartflow object pointer, or null if an error ocured. 
		 * An error can be that there is no smartflow object registered with
		 * the given name, or no smartflow object registered at all.
		 */
		static Smartflow * instance(const std::string application=DEFAULT_APPLICATION_NAME);
		
		/** \brief Return the Preference manager. */
		PreferenceManager	* getPreferenceManager();
		PreferenceManager	* get_preference_manager();
		/** \brief Return the FlowManager manager. */
		FlowManager		    * getFlowManager();
		FlowManager		    * get_flow_manager();
		/** \brief Return the Memory manager. */
		MemoryManager		* getMemoryManager();
		MemoryManager		* get_memory_manager();
		/** \brief Return the Connection manager. */
		ConnectionManager	* getConnectionManager();
		ConnectionManager	* get_connection_manager();
		/** \brief Return the Error manager. */
		ErrorManager		* getErrorManager();
		ErrorManager		* get_error_manager();
		/** \brief Return the Log manager. */
		Log_Manager		* getLogManager();
		Log_Manager		* get_log_manager();
		
		std::string & getClientName();
		std::string & get_client_name();
		
		std::string & getGroupName();
		std::string & get_group_name();
		
	private:
		std::string                 application_name_;
		std::string                 client_name_argv0;
		static std::map<std::string,Smartflow*> smartflow_map_;
		
		// By Imad
		std::string clientName;
		std::string clientGroup;
		
		/** True if all the managers has been successfully created and  initialised
		 */
		bool	manager_spawn_good_;
		
		/** True if the Smartflow object itself is in the ready-to-connect state
		 */
		bool    manager_init_good_;
		
		
		PreferenceManager	* preferencemanager_;
		FlowManager		* flowmanager_;
		MemoryManager		* memorymanager_;
		ConnectionManager	* connectionmanager_;
		ErrorManager		* errormanager_;
		Log_Manager       * logmanager_;
		
		/** Method to create the managers. Used by init, and separed to 
		 facilitate the maintenance 
		 \private
		 */
		bool    createManagers(void);
		
		
	}; //class Smartflow
}//namespace smartflow

#endif
