#ifndef _Buffer_ACE_H
#define _Buffer_ACE_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "ace/Message_Block.h"

#include "Smartflow.h"
#include "Buffer.h"

#define TIMESTAMP 	2

namespace smartflow {
	
	/** \brief Class Buffer_ACE .
	 * Implement the necessary method for a basic Buffer.
	 * This class inherit from the Smartflow Buffer class to implement the 
	 * Smartflow side of the library, and it also inherit from the ACE Message_Block
	 * class to allow it to be integrated in an ACE Task class, and thus be managed
	 * by a thread (for the locking problems). 
	 * \author Nist Smartspace Group
	 * \date 04-19-2004
	 * \internal
	 */
	class ACE_Svc_Export Buffer_ACE: public Buffer, public ACE_Message_Block  {
		/** Redefine Buffer (which is the ancestor of Buffer_ACE) as
		 \b super.
		 */
		typedef Buffer super;
		
	public:
		/** \brief Class constructor. 
         * \warning This method does not initialise the object, the init() method
         * must be used for this purpose.
		 */
		Buffer_ACE();
		
        /** \brief Class destructor.
		 */
        virtual ~Buffer_ACE();
        /** \brief Class initialiser. Must be called just after the object
         * construction. 
         * \param manager The manager in which this object will be registered
         * \return 0 if the object is correctly initialised, otherwise -1. If possible,
         * an error will be set via Smartflow::seterror() to explain the error.
		 */
        virtual bool init(MemoryManager *manager, Flow * flow, buffer_size_t max_size);   
        
        /** \brief return a pointer on the actual data, or on the space available
         * to put the data in. The size is given by get_size() and get_maxsize(),
         * respectively.
         */
        virtual void * data(void);
        /** \brief return the maximum size of the buffer.
         * It's by default the "default size" specified for the flow.
         */
        virtual buffer_size_t getMaxSize(void);
        /** \brief return the current, filled size of the buffer, or for a provider 
         * buffer, the maximum size to be filled.
         */ 
        virtual buffer_size_t getSize(void);
        /** \brief set the actual size of the data, for a provider flow.
         * It is to be used after putting the data into the buffer itself.
         * It does nothing for a consumer buffer.
         * If this is not called, the default is to take the maximum size for
         * sending.
         */
        virtual bool setSize(buffer_size_t new_size);
        /* set all the data for the buffer to 0. 
         */
        virtual bool clear(void);
        
        virtual bool initBuf(void);
		
    	virtual data_type_t getNextDataType();
        
		virtual buffer_size_t getNextDataSize();
		
		virtual ACE_Time_Value getNextTimeStamp();
		
		virtual bool gotoNextBlock();  
		
		virtual bool pushData(void *src, buffer_size_t size);
		
		virtual bool pushData(const void *src, buffer_size_t size);
		
		virtual bool pushDataTS(void *src, buffer_size_t size, ACE_Time_Value *ts = 0);
		
		virtual bool pushDataTS(const void *src, buffer_size_t size, ACE_Time_Value *ts = 0);
		
		virtual bool pushMetadata(void *src, buffer_size_t size, data_type_t type = 0);
		
		virtual bool pushMetadata(const void *src, buffer_size_t size, data_type_t type = 0);
		
		virtual bool addData(const void *src, buffer_size_t size, data_type_t type);
		
		virtual bool popData(void *src, buffer_size_t size);
		
		virtual bool popMetadata(void *src, buffer_size_t size );
		
		virtual bool popDataTS(void *src, buffer_size_t size, ACE_Time_Value *ts ) ;
		
		virtual bool removeData(void * dest, buffer_size_t dest_size);
        
	private:	
	}; // class Buffer_ACE
}  // namespace SMARTFLOW
#endif
