#ifndef _FlowImpl_H
#define _FlowImpl_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "Smartflow.h"
#include "Flow.h"


namespace smartflow {
  
 #define DEFAULT_WATER_MARK_LEVEL 0.8

  /** \brief Class FlowImpl .
   * This class is an interface for the FlowImpl delegation.
   * The class implementation must handle the communication with the duplicator.
   * \author Nist Smartspace Group
   * \date 04-19-2004
   * \internal
   */
  class ACE_Svc_Export FlowImpl {
  public:
	/** \brief Class constructor. 
         * \warning This method does not initialise the object, the init() method
         * must be used for this purpose.
        */
	FlowImpl();
        /** \brief Class destructor.
        */
    virtual ~FlowImpl();
        /** \brief Class initialiser. Must be called just after the object
         * construction. 
         * \param manager The flow owning this object this implementation. (1-1 relationship)       
         * \return 0 if the object is correctly initialised, otherwise -1. If possible,
         * an error will be set via Smartflow::seterror() to explain the error.
        */
        virtual bool init(Flow *manager);
        
        virtual bool start(void)=0;
        
        /** \brief Stop & cleanup method. Must be called when exiting, even if init
         * was not successful.
		 * \param waitForEmptyHistory If true, wait for the history to be empty before stopping the flow. 
         * \return 0 if successfully close, otherwise -1.
         */
        virtual bool stop(bool waitForEmptyHistory)=0;
        
        /** \brief Return the Smartflow object associated with an instance of the class.
         * \return A pointer to the smartflow object, or NULL if an error occured.
        */
        Smartflow * getSmartflow(void);
	    Smartflow * get_smartflow(void);

        /** \brief Return the instance of the Flow class where is object is registered.
         * \return A pointer to the Flow manager object, otherwise NULL.
        */ 
        Flow * getManager(void);
	    Flow * get_manager(void);

        
        /* buffer access */
     	virtual Buffer * getBuffer(void)=0;
     	
	    virtual bool releaseBuffer(Buffer *buf)=0;

     	virtual bool cancelBuffer(Buffer * buf)=0;

	
	/* Buffer parameters */
	virtual long getHistorySize(void)=0; 
	virtual bool setHistorySize(long history_size)=0;
	virtual bool setMaxBufferSize(buffer_size_t buffer_size)=0;
	virtual buffer_size_t getMaxBufferSize(void)=0;
        
	/** \brief Return the watermark of the flow queue (a double between 0 and 1)
	 *   Each instance of a flow has a queue. The watermark is used if the queue get full.
	 *   If the queue is full, no more buffer can be enqueued until the level of the queue
	 *   drop under the watermark. By default, the watermark is set to 0.8 (80%). That means
	 *   that if the queue get full, data will be enqueue only when the level of the queue 
	 *   decrease of 20%.
	 *   This method must NEVER be called by the user
	 *   This queue is never handled directly by the user
	 *   \Return a double between 0 and 1 representing the watermark
	 */
	virtual double getWaterMark(void)=0;
	
	/** \brief Set the watermark of the flow queue (a double between 0 and 1)
	 *   Each instance of a flow has a queue. The watermark is used if the queue get full.
	 *   If the queue is full, no more buffer can be enqueued until the level of the queue
	 *   drop under the watermark. By default, the watermark is set to 0.8 (80%). That means
	 *   that if the queue get full, data will be enqueue only when the level of the queue 
	 *   decrease of 20%.
	 *   This method must NEVER be called by the user. 
	 *   This queue is never handled directly by the user
	 *   \param water_mark, a double between 0 and 1 included.
	 *   \return a double between 0 and 1 representing the watermark
	 */
	virtual bool setWaterMark(double water_mark)=0;

	//Get the filling level of the flow queue.
	virtual double getFillingLevelOfQueue()=0;

	//Return true if there is a buffer available.
	virtual bool isBufferAvailable()=0;

	/* Number of blocks for the duplicator */
	virtual bool setBlockCount(int nbBlocks)=0;
	virtual int getBlockCount(void)=0;

        
protected:
	Flow * manager_;
}; // class FlowImpl
}  // namespace SMARTFLOW
#endif
