#include "ace/ACE.h"
#include "ace/SString.h"
#include "ace/OS.h"
#include "ace/Signal.h"
#include "ace/Time_Value.h"

#include "FlowImpl_ACE.h"
#include "ErrorManager.h"
#include "PreferenceManager.h"
#include "Log_Manager_ACE.h"

#include "duplicator_common_ACE.h"

#include <iostream>

using namespace std;

namespace smartflow {

FlowImpl_ACE::FlowImpl_ACE()
  :super(),history_size(0),buffer_size(0),direction(FLOW_INVALID),water_mark(DEFAULT_WATER_MARK_LEVEL),
	 policy(INVALID_POLICY),connector(),stream(),mem_pool(0),base_addr(0)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::FlowImpl_ACE"));
}

FlowImpl_ACE::~FlowImpl_ACE()
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::~FlowImpl_ACE"));
	//this->wait_empty_history();
	this->close();
}

bool FlowImpl_ACE::init(Flow *f)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::init"));
	/* Check that ancestor is correctly initized */
	if ( !super::init(f) ) {
         return false;
    }

    /* Setup local variables to values from the flow */
	//the flow is not initialized yet.
    history_size = f->getHistorySize();
    buffer_size  = f->getMaxBufferSize();
    block_count  = f->getBlockCount();
    direction = f->getMode();
    policy = f->getFlowQueuePolicy();

	isSynchronized_ = f->isSynchronized();

    /* Setup the message-queue */
    if ( msg_queue() == NULL ){
    	get_smartflow()->setError(FLOWIMPL_CANNOT_SETUP_HISTORY);
    	return false;
    }
    /* Set the History size */
    setHistorySize(history_size);

    /*Set the water mark */
    //We change the default watermark level if we use a policy which drops buffers.
    if ( policy != DO_NOT_DROP_BUFFER ) {
      setWaterMark(1.0); 
    }

    return true;
	
}

bool FlowImpl_ACE::start(void)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::start"));
    /*setup duplicator connection */
    if ( !open_duplicator() ) {
    	return false;
    }
    /* run the thread */
    if ( activate() != 0 ) {  
    	get_smartflow()->setError(FLOWIMPL_CANNOT_RUN_THREAD);
    	return false;
    }    
	return true;
}

bool FlowImpl_ACE::stop(bool waitForEmptyHistory)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::stop"));
	if ( thr_count() > 0 ) 
	{
		if ( waitForEmptyHistory ) // Wait for the queue to be empty before stopping the flow
			wait_empty_history();
	
		msg_queue()->deactivate(); /* Disable the queue, so that get() calls unblock
		                              with -1 on the tread */
		stream.close_reader();    /* This call shutdown(), which in effect cancel a 
		                           read() or recv in progress with -1 on the threads  */
		stream.close_writer();    /* This call is for Windows which doesn't behave the
								     same way as Linux. It has no effect under Linux */
		wait();                    /* Make sure the threads exited */
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ FlowImpl_ACE::stop ] Flow: %s stopped\n"), get_manager()->get_id().c_str()));
	}
	stream.close(); /* Repeated here to close the stream completely, just in case. */
	if ( mem_pool) {
		mem_pool->release(0);
		delete mem_pool; mem_pool = 0;
	}
	return true;
}	



long FlowImpl_ACE::getHistorySize(void)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::getHistorySize"));
	return history_size;
}
bool FlowImpl_ACE::setHistorySize(long history_size)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::setHistorySize"));
        this->history_size = history_size;
	/* Set the high watermark to the maximum size of the queue.
	 * If the queue size (i.e number of byte in the queue) > high watermark,
	 * it becomes impossible to put new Buffers in the queue.
	 */
	this->water_marks(ACE_IO_Cntl_Msg::SET_HWM,history_size*buffer_size);
	/* Set the low watermark to 80% or custom value of the queue size
	 * If the queue is full (or the size of the queue is greater than the high watermark),
	 * it becomes impossible to put new Buffers in the queue (putq) until the size of the queue 
	 * drops below the low watermark.
	 */
	this->water_marks(ACE_IO_Cntl_Msg::SET_LWM,(size_t)(history_size*buffer_size*this->water_mark));
	return true; 
}
	
bool FlowImpl_ACE::setMaxBufferSize(buffer_size_t buffer_size)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::setMaxBufferSize"));
	this->buffer_size = buffer_size;
	/* Set the high watermark to the maximum size of the queue.
	 * If the queue size (i.e number of byte in the queue) > high watermark,
	 * it becomes impossible to put new Buffers in the queue.
	 */
	this->water_marks(ACE_IO_Cntl_Msg::SET_HWM,history_size*buffer_size);
	/* Update the low watermark.
	 * If the queue is full (or the size of the queue is greater than the high watermark),
	 * it becomes impossible to put new Buffers in the queue (putq) until the size of the queue 
	 * drops below the low watermark.
	 */
	this->water_marks(ACE_IO_Cntl_Msg::SET_LWM,(size_t)(history_size*buffer_size*this->water_mark));
	return true;
}
buffer_size_t FlowImpl_ACE::getMaxBufferSize(void)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::getMaxBufferSize"));
	return buffer_size;
}
	
bool FlowImpl_ACE::setWaterMark(double water_mark) {
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::setWaterMark"));
  if ( (water_mark <= 1) && (water_mark >= 0) ) {
    this->water_mark = water_mark;
    this->water_marks(ACE_IO_Cntl_Msg::SET_LWM,(size_t)(history_size*buffer_size*this->water_mark));
    return true;
  }
  else 
    return false;
}	
	
double FlowImpl_ACE::getWaterMark(void) {
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::getWaterMark"));
        return water_mark;
}

double FlowImpl_ACE::getFillingLevelOfQueue() {
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::getFillingLevelOfQueue"));
  double bytes = (double)(this-> msg_queue()->message_bytes());
  double total = (double)(history_size*buffer_size);
  double res = bytes/total; 
  return res;
}

bool FlowImpl_ACE::isBufferAvailable()  {
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::isBufferAvailable"));
	switch(direction) {
	case FLOW_INPUT:
		return (!msg_queue()->is_empty());
	case FLOW_OUTPUT:
		return true;
	case FLOW_INVALID:
	  //should never be here.
	        return false;
	default:
	  //should never be here.
	       return false;
	}
}

bool FlowImpl_ACE::setBlockCount(int nbBlocks) 
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::setBlockCount"));
        this->block_count = nbBlocks;
        return true;
}

int FlowImpl_ACE::getBlockCount(void) {
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::getBlockCount"));
	return block_count;
}

Buffer * FlowImpl_ACE::getBuffer(void)
{	
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::getBuffer"));
	Buffer_ACE     * buffer_ace    = 0;
	switch (direction) {
		case FLOW_INPUT: 
			ACE_Message_Block * block;
			if ( getq(block) == -1 ) {
				//get_smartflow()->setError(FLOWIMPL_INVALID_QUEUE_STATE);
				return 0;
			}
			buffer_ace = dynamic_cast<Buffer_ACE *> (block);
			break;
		case FLOW_OUTPUT:
			buffer_ace = dynamic_cast<Buffer_ACE *> (get_smartflow()->getMemoryManager()->malloc_buffer(get_manager()));
			break;
	    default:
	    	get_smartflow()->setError(FLOWIMPL_INVALID_FLOW_STATE);
	    	break;
	}

    return buffer_ace;
}

bool FlowImpl_ACE::releaseBuffer(Buffer *buf)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::releaseBuffer"));
  if ( ! buf ) {
    ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ FlowImlp_ACE::releaseBuffer ] Buffer is NULL\n")));
    return false;
  }
  switch (direction) {
  case FLOW_INPUT:
    get_smartflow()->getMemoryManager()->free_buffer(buf);
    break;
  case FLOW_OUTPUT:		
	  if ( policy == DROP_MOST_RECENT_BUFFER ) {
		//if the queue is full, we don't try to put the buffer in the queue.
		  if ( enough_room_in_queue(buf->getSize()) == true ) {
			if ( putq( dynamic_cast<Buffer_ACE *> (buf) ) == -1 ) {
				//get_smartflow()->setError(FLOWIMPL_INVALID_QUEUE_STATE);
				ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ FlowImlp_ACE::releaseBuffer ] Error putting buffer in queue\n")));
				return false;
			}
		  }
		  else {
			  get_smartflow()->getMemoryManager()->free_buffer(buf);
			  ACE_TString flowID =  ACE_TEXT_CHAR_TO_TCHAR((manager_->get_id()).c_str());
			  ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ FlowImpl_ACE::releaseBuffer ] Queue full for the output flow: '%s', Most recent buffer dropped.\n" ), flowID.c_str()));
		  }
	  }
	  else if ( policy == DROP_OLDEST_BUFFER ) {
		  //the queue is full, we need to remove the oldest buffer(s) to make room for the new one.
		  while ( enough_room_in_queue(buf->getSize()) == false ) {
			 ACE_Message_Block  *block_to_drop=0;
			 getq(block_to_drop);
			 get_smartflow()->getMemoryManager()->free_buffer(dynamic_cast<Buffer_ACE *> (block_to_drop));
			 ACE_TString flowID =  ACE_TEXT_CHAR_TO_TCHAR((manager_->get_id()).c_str());
			 ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ FlowImpl_ACE::releaseBuffer ] Queue full for the output flow '%s', Oldest buffer dropped.\n" ), flowID.c_str())); 
		  }
		  //Now there is room in the queue to put the new buffer
		  if ( putq( dynamic_cast<Buffer_ACE *> (buf) ) == -1 ) {
		    //get_smartflow()->setError(FLOWIMPL_INVALID_QUEUE_STATE);
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ FlowImlp_ACE::releaseBuffer  ] Error putting buffer in queue\n")));
			return false;
		  }
	  }
	  else if ( policy == DO_NOT_DROP_BUFFER ) {
		if ( putq( dynamic_cast<Buffer_ACE *> (buf) ) == -1 ) {
		    //get_smartflow()->setError(FLOWIMPL_INVALID_QUEUE_STATE);
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ FlowImlp_ACE::releaseBuffer ] Error putting buffer in queue\n")));
			return false;
		}
	  }
    break;
  default:
	  ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ FlowImlp_ACE::releaseBuffer ] FLOWIMPL_INVALID_FLOW_STATE\n")));
    get_smartflow()->setError(FLOWIMPL_INVALID_FLOW_STATE);
    return false;
    break;
  }
  return true;
}

bool FlowImpl_ACE::cancelBuffer(Buffer * buf)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::cancelBuffer"));
	if ( buf ) {
		get_smartflow()->getMemoryManager()->free_buffer(buf);
	}
	return true;
}

void FlowImpl_ACE::wait_empty_history(void)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::wait_empty_history"));
	while ( ! msg_queue()->is_empty() ) 
	{
		ACE_OS::sleep(1);
	}
	
}

bool FlowImpl_ACE::enough_room_in_queue(buffer_size_t size)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::enough_room_in_queue"));
  double current_size = (double)(this-> msg_queue()->message_bytes());
  double total_size = (double)(history_size*buffer_size);
  if ( (current_size + size) < total_size )  
	  return true;
  else
	  return false;
}

/***********************************************************/
bool FlowImpl_ACE::open_duplicator(void)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::open_duplicator"));
    size_t mempool_size  = getMaxBufferSize() * block_count; 
	size_t mempool_real_size = 0;
	ACE_TCHAR tmp_filename[DUP_MAX_FILENAME_LEN+1] = {0};
	ACE_TString flow_mmap_filename;
	ACE_TString flow_filename_socket;
	int err = ACE::get_temp_dir(tmp_filename, DUP_MAX_FILENAME_LEN);
	if ( err == -1 )  {
		get_smartflow()->setError(FLOWIMPL_CANNOT_OPEN_DUPLICATOR);
		return false;
	}

	#if (defined (ACE_WIN32))
		// Some chracters are not allowed in a filename under windows
		// We replace them by '-'

		/* Find flow filename base */
		flow_mmap_filename = tmp_filename;
		flow_mmap_filename += DUP_FLOW_FILENAME_PREFIX;
		ACE_TString flowIDModified = ACE_TEXT_CHAR_TO_TCHAR(get_manager()->get_id().c_str());
			
		for (unsigned int i = 0; i < flowIDModified.length(); i++) {
			if ((flowIDModified[i] == ':') ||(flowIDModified[i] == '/') ||(flowIDModified[i] == '?') ||(flowIDModified[i] == '"') ||(flowIDModified[i] == '*') ||(flowIDModified[i] == '<') ||(flowIDModified[i] == '>') ||(flowIDModified[i] == '|'))
				flowIDModified[i] = '-';
		}
		
		flow_mmap_filename += flowIDModified;
		flow_mmap_filename += DUP_FLOW_FILENAME_MMAP;
	#else
		/* Find flow filename base */
		flow_mmap_filename = tmp_filename;
		flow_mmap_filename += DUP_FLOW_FILENAME_PREFIX;
		flow_mmap_filename += get_manager()->get_id().c_str();
		flow_mmap_filename += DUP_FLOW_FILENAME_MMAP;
	#endif
	
	#if (!(defined (ACE_WIN32)))
		flow_filename_socket = tmp_filename;
		flow_filename_socket += DUP_FLOW_FILENAME_PREFIX;
		flow_filename_socket += get_manager()->get_id().c_str();
		switch (direction) {
			case FLOW_INPUT: 
				flow_filename_socket += DUP_FLOW_FILENAME_SOCKET;
				break;
			case FLOW_OUTPUT:
				flow_filename_socket += DUP_FLOW_FILENAME_PSOCKET;
				break;
			default:
				get_smartflow()->setError(FLOWIMPL_INVALID_FLOW_STATE);
				return -1;
		}
		/* Open the communication socket with the Duplicator (Unix Socket) */
		ACE_UNIX_Addr local_addr( flow_filename_socket.c_str() );	
	#else
		ACE_TString flow_name (ACE_TEXT_CHAR_TO_TCHAR(get_manager()->get_id().c_str()));
		ACE_UINT64 port = 1;
		for (unsigned int i = 0; i < flow_name.length(); i++) {
			int valeur;
			switch (flow_name[i]) {
				case 'a' :
				case 'A' :
					valeur = 1;
					break;
				case 'b' :
				case 'B' :
					valeur = 2;
					break;
				case 'c' :
				case 'C' :
					valeur = 3;
					break;
				case 'd' :
				case 'D' :
					valeur = 4;
					break;
				case 'e' :
				case 'E' :
					valeur = 5;
					break;
				case 'f' :
				case 'F' :
					valeur = 6;
					break;
				case 'g' :
				case 'G' :
					valeur = 7;
					break;
				case 'h' :
				case 'H' :
					valeur = 8;
					break;
				case 'i' :
				case 'I' :
					valeur = 9;
					break;
				case 'j' :
				case 'J' :
					valeur = 10;
					break;
				case 'k' :
				case 'K' :
					valeur = 11;
					break;
				case 'l' :
				case 'L' :
					valeur = 12;
					break;
				case 'm' :
				case 'M' :
					valeur = 13;
					break;
				case 'n' :
				case 'N' :
					valeur = 14;
					break;
				case 'o' :
				case 'O' :
					valeur = 15;
					break;
				case 'p' :
				case 'P' :
					valeur = 16;
					break;
				case 'q' :
				case 'Q' :
					valeur = 17;
					break;
				case 'r' :
				case 'R' :
					valeur = 18;
					break;
				case 's' :
				case 'S' :
					valeur = 19;
					break;
				case 't' :
				case 'T' :
					valeur = 20;
					break;
				case 'u' :
				case 'U' :
					valeur = 21;
					break;
				case 'v' :
				case 'V' :
					valeur = 22;
					break;
				case 'w' :
				case 'W' :
					valeur = 23;
					break;
				case 'x' :
				case 'X' :
					valeur = 24;
					break;
				case 'y' :
				case 'Y' :
					valeur = 25;
					break;
				case 'z' :
				case 'Z' :
					valeur = 26;
					break;
				default :
					valeur = ACE_OS::atoi(&flow_name[i]);
					if (valeur <=0 )
						valeur = 1;
					break;
			}
			port *= valeur;
			port += i;
		}
		port %= 62357;
		port += 1217;
		switch (direction) {
			case FLOW_INPUT: 
				port++;
				break;
			case FLOW_OUTPUT:
				break;
			default:
				get_smartflow()->setError(FLOWIMPL_INVALID_FLOW_STATE);
				return false;
		}
		//AF_INET to force IPv4
		ACE_INET_Addr local_addr((u_short) port, "localhost", AF_INET);
	#endif

	err = connector.connect(stream, local_addr);
 
	while (err == -1)
	{
		get_smartflow()->setError(FLOWIMPL_CANNOT_OPEN_DUPLICATOR);
		ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ FlowImpl_ACE::open_duplicator ] Error Connect for flow '%s'.. trying again... \n"), get_manager()->get_id().c_str() ));
		ACE_Time_Value tmp(0, 500000);
		ACE_OS::sleep(tmp);		
		err = connector.connect(stream,local_addr);
	}
		
	if (err || (stream.get_handle() == ACE_INVALID_HANDLE))
	{
		get_smartflow()->setError(FLOWIMPL_CANNOT_OPEN_DUPLICATOR);
		return false;
	}
	
	//If this is an input flow, we send the ID of the client the duplicator handle 
	if ( direction == FLOW_INPUT ) {
	  //by Imad
	  ACE_TString client_name;
	  std::string tmp = "clientname";
	  client_name = ACE_TEXT_CHAR_TO_TCHAR((get_smartflow()->getPreferenceManager()->getPreferenceValue(tmp)).c_str());
	  if ( client_name.length() == 0 )
			client_name = ACE_TEXT_CHAR_TO_TCHAR(get_smartflow()->getClientName().c_str()); 

   	  ACE_TString client_group;
	  tmp = "clientgroup";
	  client_group = ACE_TEXT_CHAR_TO_TCHAR((get_smartflow()->getPreferenceManager()->getPreferenceValue(tmp)).c_str());
	  if ( client_group.length() == 0 )
		client_group = ACE_TEXT_CHAR_TO_TCHAR(get_smartflow()->getGroupName().c_str());
	  
	  ACE_TString client_ID;
	  client_ID = client_group + ACE_TEXT(":") + client_name;
	  
	  ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ FlowImpl_ACE::open_duplicator ] I try to send my ID '%s' to the duplicator\n"),client_ID.c_str()));
	  stream.send_n(client_ID.c_str(), ACE_OS::strlen(client_ID.c_str())+1, 0);
	
	  //we need to check that the message has been correctly send
	}
	//If this is an output flow, we send synchronization information to the server
	else if ( direction == FLOW_OUTPUT ) {
		//we just send an integer, 0 means the flow is not synchronized, 1 means the flow is synchronized
		ACE_TCHAR tmp[11];
		ACE_TString theSync = ACE_TEXT("");
		if ( isSynchronized_ ) {
			theSync += "1";
		}
		else {
			theSync += "0";
		}

		ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ FlowImpl_ACE::open_duplicator ] I try to send the synchronization info '%s' to the duplicator\n"), theSync.c_str()));
	    stream.send_n(theSync.c_str(), ACE_OS::strlen(theSync.c_str())+1, 0);

		//we need to check that the message has been correctly send
	}
	//Otherwise, we throw an error, this case should never happens
	else {
		get_smartflow()->setError(FLOWIMPL_INVALID_FLOW_STATE);
		return false;
	}
	
	/* open MMAPed file */
	mem_pool = new ACE_Lite_MMAP_Memory_Pool(flow_mmap_filename.c_str(),0);
	if ( ! mem_pool ) return false;	
	if ( (base_addr = (char *)mem_pool->init_acquire(mempool_size,mempool_real_size,err)) == NULL ) {
		get_smartflow()->setError(FLOWIMPL_CANNOT_OPEN_DUPLICATOR);
		return false;
	}
	/* Unix only : setup a SIGPIPE handler, to avoid having a killing signal when the 
	 * duplicator quit unexpectedly. By doing nothing in this handler, the send() or read() 
	 * will return -1 so the error will be managed correctly
	 */
	 ACE_Sig_Action    sigpipe_catch(SIG_IGN,SIGPIPE,0,SA_RESTART);
	return true;
}

/** Service Method : Entry point for a thread.
 * This method is run when activate() is called in the main thread.
 * It just select the method to call for a provider or a consumer.
 * \internal
 */
int FlowImpl_ACE::svc(void)
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::svc"));
	int return_code=-1;
	/* Sanity checks */
	if ( (mem_pool == NULL) ||  (stream.get_handle() == ACE_INVALID_HANDLE ) )
		return -1;
	switch (direction) {
		case FLOW_INPUT : return_code = svc_consumer(); break;
		case FLOW_OUTPUT : return_code = svc_provider(); break;
		default: get_smartflow()->setError(FLOWIMPL_INVALID_FLOW_STATE); break;
	}
	this->msg_queue()->deactivate(); /* this force an error on putq or getq on the other side */
	return return_code;
}

/** Service Method : Called in thread context, for a provider client
 * This method read from the duplicator an available buffer, then 
 * read from the internal queue a buffer, and copy it into the mmapped buffer
 * number that the provider indicated. Then it "release" the buffer to the 
 * duplicator to says it has finished.
 * If a read() call return 0 or -1 it mean the duplicator failed, so the thread
 * exit, same thing for the queue.
 * \internal
 */
int FlowImpl_ACE::svc_provider()
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::svc_provider"));
	int            err = -1;
	char         * mmap_base;
	ACE_Message_Block  * block=0;
	local_block_t  buf; memset(&buf,0,sizeof(buf));

	while (true)
	{
		/* 1/ Get the SHM Buffer number from the duplicator */
		err = stream.recv_n(&buf,sizeof(buf));
		if ( err <=0 ) {
			get_smartflow()->setError(FLOWIMPL_SERVICE_UNEXPECTED_QUIT);
			return -1;
		}
		
		/* 2/ Calculate the address of the buffer inside the SHM */
		mmap_base = base_addr + (buf.block_number * buffer_size) + sizeof(ACE_UINT32);
		
		/* 3/ Get a ACE_Message_Block ( = Smartflow Buffer ) from the ACE_Task. */
		/* This function can block if no buffer available at this time */
		err = getq(block);
		if ( err < 0 ) break ; /* normal quit request */
		if ( block == NULL ) {
			get_smartflow()->setError(FLOWIMPL_SERVICE_UNEXPECTED_QUIT);
			return -1;
		}
		
		/* 4/ Do the memcpy from the ACE_Message_Block data to the SHM */
		/* Size is set to what is reported from the ACE_Message_Block */
		block->rd_ptr(block->base());
		buf.block_size = block->length();
			
		ACE_OS::memcpy(mmap_base,block->rd_ptr(),buf.block_size);
		get_smartflow()->getMemoryManager()->free_buffer(dynamic_cast<Buffer_ACE *> (block));

		err = stream.send_n(&buf,sizeof(buf));
		if ( err <= 0 ) { 
			get_smartflow()->setError(FLOWIMPL_SERVICE_UNEXPECTED_QUIT);
			return -1;
		}
	};
	return 0;
}

/** Service Method : Called in thread context, for a consumer client
 * This method read from the duplicator an available buffer, then create a 
 * new Message_Block from the internal queue a buffer, and copy the buffer from
 * mmapped buffer number that the provider indicated into the new Message Block.
 * Then it "release" the buffer to the duplicator to says it has finished.
 * If a read() call return 0 or -1 it mean the duplicator failed, so the thread
 * exit, same thing for the queue.
 * \internal
 */
int FlowImpl_ACE::svc_consumer()
{
	ACE_TRACE(ACE_TEXT("FlowImpl_ACE::svc_consumer"));
  int            err = -1;
  char         * mmap_base;
  ACE_Message_Block  * block = 0;
  local_block_t  buf; memset(&buf,0,sizeof(buf));
  
  while (true)
    {
      err = stream.recv_n(&buf,sizeof(buf));
      if ( err <=0 ) {
	     get_smartflow()->setError(FLOWIMPL_SERVICE_UNEXPECTED_QUIT);
	     return -1;
      }
      
      mmap_base = base_addr + (buf.block_number * buffer_size) + sizeof(ACE_UINT32);
      block = dynamic_cast<ACE_Message_Block *> (get_smartflow()->getMemoryManager()->malloc_buffer(get_manager(), buf.block_size));
      if ( block == NULL )  return -1;
      
      block->wr_ptr(block->rd_ptr());
      ACE_OS::memcpy(block->wr_ptr(),mmap_base,buf.block_size);	
      block->wr_ptr(buf.block_size);

	  if ( policy == DO_NOT_DROP_BUFFER ) {
			err = putq(block);
			if ( err < 0 ) break; /* normal  quit request */
			if ( ! block ) {
				get_smartflow()->setError(FLOWIMPL_SERVICE_UNEXPECTED_QUIT);
				return -1;
			}
	  }
	  else if ( policy == DROP_MOST_RECENT_BUFFER) {
			if ( enough_room_in_queue(buf.block_size) == true ) {
				err = putq(block);
				if ( err < 0 ) break; /* normal  quit request */
				if ( ! block ) {
					get_smartflow()->setError(FLOWIMPL_SERVICE_UNEXPECTED_QUIT);
					return -1;
				}
			}
			else {
				get_smartflow()->getMemoryManager()->free_buffer(dynamic_cast<Buffer_ACE *> (block));
				ACE_TString flowID =  ACE_TEXT_CHAR_TO_TCHAR((manager_->get_id()).c_str());
				ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ FlowImpl_ACE::svc_consumer ] Queue full for the input flow '%s'. Most recent buffer dropped.\n" ), flowID.c_str()));
			}
	  }
	  else if ( policy == DROP_OLDEST_BUFFER) {
		  if ( enough_room_in_queue(buf.block_size) == true ) {
			   err = putq(block);
			   if ( err < 0 ) break; /* normal  quit request */
			   if ( ! block ) {
					get_smartflow()->setError(FLOWIMPL_SERVICE_UNEXPECTED_QUIT);
					return -1;
			   }
		  }
		  else {
			  //the queue is full, we need to remove the oldest buffer(s) to make room for the new one.
			  while ( enough_room_in_queue(buf.block_size) == false ) {	
				  ACE_Message_Block  *block_to_drop=0;
				  err = getq(block_to_drop);
				  if ( err < 0 ) break; /* normal quit request */
				  get_smartflow()->getMemoryManager()->free_buffer(dynamic_cast<Buffer_ACE *> (block_to_drop));
				  ACE_TString flowID =  ACE_TEXT_CHAR_TO_TCHAR((manager_->get_id()).c_str());
				  ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ FlowImpl_ACE::svc_consumer ] Queue full for the input flow '%s'. Oldest buffer dropped.\n" ), flowID.c_str())); 
			  }
			  //Now there is room in the queue to put the new buffer
		   	 err = putq(block);
			 if ( err < 0 ) break; /* normal  quit request */
			 if ( ! block ) {
				get_smartflow()->setError(FLOWIMPL_SERVICE_UNEXPECTED_QUIT);
				return -1;
			 }
		  }
	  }
	  else if ( policy == INVALID_POLICY) {
			get_smartflow()->setError(FLOW_QUEUE_POLICY_INVALID);
			return -1;
	  }
	  else {
			get_smartflow()->setError(FLOW_QUEUE_POLICY_INVALID);
			return -1;
	  }
      
      err = stream.send_n(&buf,sizeof(buf));
      if ( err <= 0 ) {
	     get_smartflow()->setError(FLOWIMPL_SERVICE_UNEXPECTED_QUIT);
	     return -1;
      }
      
  };
  return 0;
}


} // namespace SMARTFLOW
