#ifndef _FlowImpl_ACE_H
#define _FlowImpl_ACE_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "ace/Task_T.h"

#if (defined (ACE_WIN32))
	#include "ace/SOCK_Connector.h"
	#include "ace/SOCK_Stream.h"
#else
	#include "ace/LSOCK_Connector.h"
	#include "ace/LSOCK_Stream.h"
#endif /* ACE_WIN32 */

#include "ace/MMAP_Memory_Pool.h"

#include "Smartflow.h"
#include "FlowImpl.h"
#include "Buffer_ACE.h"

namespace smartflow {

/** \brief Class FlowImpl_ACE .
 * This class is a delegate of FlowManager, and is used to implement the 
 * communication with the duplicator, using ACE library for portability.
 * It is based on an ACE_Task using one thread to handle the communication
 * with the duplicator, and using the Task's queue to push the buffers
 * into the client's main thread (or the reverse for a provider flow).
 * \author Nist Smartspace Group
 * \date 04-19-2004
 * \internal
*/
class ACE_Svc_Export FlowImpl_ACE:public FlowImpl,ACE_Task<ACE_MT_SYNCH> {
/** Redefine FlowImpl (which is the ancestor of FlowImpl_ACE) as
    \b super . Facility taken from the Java language.
*/
typedef FlowImpl super;

public:
	/** \brief Class constructor. 
         * \warning This method does not initialise the object, the init() method
         * must be used for this purpose.
        */
	FlowImpl_ACE();
        /** \brief Class destructor.
        */
        virtual ~FlowImpl_ACE();
        /** \brief Class initialiser. Must be called just after the object
         * construction. 
         * \param f The The flow that this instance works for (1-1 relationship)
         * \return 0 if the object is correctly initialised, otherwise -1. If possible,
         * an error will be set via Smartflow::seterror() to explain the error.
        */
        virtual bool init(Flow *f);
        
        /** \brief Start the connection of the flow with his duplicator.
         * It open the shared-memory file and the local connections.
        */
        virtual bool start(void);
        
        /** Class de-initialiser. Force any threads to quit (and wait for them), 
         * and disconnect any attached stream.
		 * \param waitForEmptyHistory  If true, wait for the history to be empty before stopping the flow.
         * \return 0 si successfully close, or -1 if an error appear along the way.
         */
        virtual bool stop(bool waitForEmptyHistory);
        
        /* buffer access */
     	virtual Buffer * getBuffer(void);
     	virtual bool releaseBuffer(Buffer *buf);
     	virtual bool cancelBuffer(Buffer * buf);

	
	/* Buffer parameters */
	virtual long getHistorySize(void);
	virtual bool setHistorySize(long history_size);
	virtual bool setMaxBufferSize(buffer_size_t buffer_size);
	virtual buffer_size_t getMaxBufferSize(void);
	
	
	/** \brief Return the watermark of the flow queue (a double between 0 and 1)
	 *   Each instance of a flow has a queue. The watermark is used if the queue get full.
	 *   If the queue is full, no more buffer can be enqueued until the level of the queue
	 *   drop under the watermark. By default, the watermark is set to 0.8 (80%). That means
	 *   that if the queue get full, data will be enqueue only when the level of the queue 
	 *   decrease of 20%.
	 *   This method must NEVER be called by the user.
	 *   This queue is never handled directly by the user
	 *   \Return a double between 0 and 1 representing the watermark
	 */
	virtual double getWaterMark(void);
	
	/** \brief Set the watermark of the flow queue (a double between 0 and 1)
	 *   Each instance of a flow has a queue. The watermark is used if the queue get full.
	 *   If the queue is full, no more buffer can be enqueued until the level of the queue
	 *   drop under the watermark. By default, the watermark is set to 0.8 (80%). That means
	 *   that if the queue get full, data will be enqueue only when the level of the queue 
	 *   decrease of 20%.
	 *   This method must NEVER be called by the user.
	 *   This queue is never handled directly by the user
	 *   \param water_mark, a double between 0 and 1 included.
	 *   \return a double between 0 and 1 representing the watermark
	 */
	virtual bool setWaterMark(double water_mark);

	//Get the filling level of the flow queue.
	virtual double getFillingLevelOfQueue();

    //Return true if there is a buffer available.
	virtual bool isBufferAvailable();

	/* Number of blocks for the duplicator */
	virtual bool setBlockCount(int nbBlocks);
	virtual int getBlockCount(void);
	
  protected:
	/* These routine runs in a separate thread. It read buffers from the duplicator
	 * and put them into the queue (or the reverse for a provider)
	 */
	virtual int svc(void);
	int svc_provider(void);
	int svc_consumer(void);
	void wait_empty_history(void);
	
  private:
	long             history_size;
	buffer_size_t    buffer_size;
	buffer_size_t    block_count;
	flow_direction_t direction;
	double           water_mark;
	flow_policy_t    policy;
	bool				 isSynchronized_;
 
	
	bool open_duplicator(void);
	bool enough_room_in_queue(buffer_size_t size);
	
#if (defined (ACE_WIN32))
	ACE_SOCK_Connector connector;
	ACE_SOCK_Stream    stream;
#else
	ACE_LSOCK_Connector connector;
	ACE_LSOCK_Stream    stream;
#endif /* ACE_WIN32 */
	ACE_Lite_MMAP_Memory_Pool *mem_pool;
	char * base_addr;
	
}; // class FlowImpl_ACE
  

}  // namespace SMARTFLOW
#endif
