#include "ace/DLL.h"
#include "ace/Time_Value.h"
#include "ace/OS.h"

#include "FlowManager_STL.h"
#include "Flow.h"
#include "ConnectionManager.h"
#include "PreferenceManager.h"
#include "ErrorManager.h"
#include "duplicator_common_ACE.h"
#include "Log_Manager_ACE.h"

#include <iostream>

namespace smartflow {
FlowManager_STL::FlowManager_STL()
	:super()
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::FlowManager_STL"));
}
FlowManager_STL::~FlowManager_STL()
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::~FlowManager_STL"));
	/* destroy all flows */
	/* empty the STL Flow List */
	Flowlist_t::iterator it;
	for (it=flowlist.begin(); it != flowlist.end(); it++)
	{
		destroyFlow( it->second );
	}
}

bool FlowManager_STL::init(Smartflow *sf)
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::init"));
	if ( ! sf ) return false;
	if ( ! flowlist.empty() )
	{	sf->setError(FLOW_MANAGER_NOT_EMPTY_ON_STARTUP);
		return false;
	}
	return super::init(sf);
}


Flow * FlowManager_STL::makeFlow(const std::string & type,const std::string & name,const std::string &group,const std::string & arg,flow_direction_t direction, flow_policy_t policy, unsigned int numberOfConsumersExpected)
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::makeFlow"));
     Flow * newflow = NULL;
     bool flowresult = false;
     std::string flow_name;
     std::string full_name;
     std::string real_group;
     std::string tmp = "flow-";
     tmp.append(name);

	 bool isFlowSynchronized = false;
	
     real_group = get_smartflow()->getPreferenceManager()->getPreferenceValue(tmp);
	 if ( real_group.empty() ) 
	      real_group = group; 
     /* Ensure that the init() function has been passed */
     if ( ! get_smartflow() ) return NULL;
          
     full_name = real_group;
     full_name.append(":");
     full_name.append(type);
     
	 /* Check if a synchronized flow has been requested. If yes, it needs to be an output flow */
	 if ( (numberOfConsumersExpected > 0) && (direction != FLOW_OUTPUT ) ) {
		get_smartflow()->setError(FLOW_SYNCHRONIZED_INVALID); return NULL;
	 }

     /* Check if the flow already exist in THIS local client (either provider or consumer) */
     if ( findFlow(full_name) != NULL ){get_smartflow()->setError(FLOW_ALREADY_EXIST); return NULL; }
     
	 /* Initialise isFlowSynchronzed to 1 if we expect more than 1 consumer for the output flow */
	 if ( (numberOfConsumersExpected > 0) && (direction == FLOW_OUTPUT ) ) {
		isFlowSynchronized = true;
	 }

     /* Create a the flow instance. This will  instantiate the correct type of flow 
      * with the parameters passed. setError() is set by dynLoadFlow */
     newflow = dynLoadFlow(type); 
     if ( newflow == NULL ) { 
     	return NULL;
     }
	 
     /* Initialise the instance for the flow. If it does not work, give up here. */
     flowresult = newflow->init(this, type,name, real_group, arg, direction, policy, isFlowSynchronized);
     if ( flowresult == false ) {
	   ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ FlowManager_STL::makeFlow ] flow->init() failed\n")));
       delete newflow;
       return NULL; 
     }
	 
	 if (direction == FLOW_OUTPUT) {
	   while (get_smartflow()->getConnectionManager()->getFlow(type, name, real_group, arg, direction, newflow->getBlockCount() , newflow->getMaxBufferSize(), 0, numberOfConsumersExpected) == false)
	   {
	     ACE_Time_Value tmp(0,500000);
	     ACE_OS::sleep(tmp);	
	   }
	 }
	 else     
	 {
		/* Ask the connection manager if there is already a flow with this 
		type+name. If not, the flow name will be reserved. setError() can be set. */	  	  
		flowresult = get_smartflow()->getConnectionManager()->getFlow(type, name, real_group, arg, direction, newflow->getBlockCount() , newflow->getMaxBufferSize(), 1, numberOfConsumersExpected );
		if ( flowresult == false ) { 
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ FlowManager_STL::makeFlow ] ConnectionManager->getFlow() failed\n")));
			delete newflow; 
			return NULL;
		}
	 }	 
     /* Everything is good so far, register this flow in the list */
     insertFlow(full_name,newflow);
     return newflow;
}

bool FlowManager_STL::stopAllFlows() {
	ACE_TRACE(ACE_TEXT("FlowManager_STL::stopAllFlows"));
	Flowlist_t::iterator it;
	for (it = flowlist.begin(); it != flowlist.end(); it++)
	{
		it->second->stop();
	}
	
	return true;
}

bool FlowManager_STL::destroyFlow(Flow *f)
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::destroyFlow"));
     /* Locate the flow in the list */
     if ( f == 0 ) return false;
   
     /* Inform the connection manager that the flow is unavailable from now on */
	 bool res = get_smartflow()->getConnectionManager()->removeFlow(f->getFlowType() , f->getName(), f->getGroup());
     /* Politely ask the flow to shutdown his connections */
	 f->stop();
     /* Remove the flow from the flow list */
     removeFlow(f);
     delete f;
	 return res;
}


Flow * FlowManager_STL::findFlow(const std::string & fname)
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::findFlow"));
	Flowlist_t::iterator i;
	i = flowlist.find(fname);
	
	if ( i == flowlist.end() ) return NULL;
	return i->second;
}

void FlowManager_STL::insertFlow(const std::string & fname,Flow * f)
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::insertFlow"));
	flowlist[fname] = f;
}

void FlowManager_STL::removeFlow(Flow * f)
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::removeFlow"));
	flowlist.erase( f->get_id() );
}

void FlowManager_STL::removeFlow(const std::string fname)
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::removeFlow"));
	flowlist.erase( fname );
}

/* Do not close the shared lib when the dll object goes out of scope */
Flow * FlowManager_STL::dynLoadFlow(const std::string & type)
{
	ACE_TRACE(ACE_TEXT("FlowManager_STL::dynLoadFlow"));
	ACE_DLL dll;
	Flow * the_flow = 0;
	Flow_Creator fc=0;
	std::string  flow_plugin_path = get_smartflow()->getPreferenceManager()->getPreferenceValue("FlowPlugin_Path");
	
	if ( flow_plugin_path.empty())
	{
		flow_plugin_path = type;
	} else {
		flow_plugin_path.append("/lib");
		flow_plugin_path.append(type);
		flow_plugin_path.append(".so");
	}
	
	int err = dll.open(ACE_TEXT_CHAR_TO_TCHAR(flow_plugin_path.c_str()),ACE_DEFAULT_SHLIB_MODE,0); 
	if ( err ) {
		get_smartflow()->setError(FLOW_CANNOT_FIND_DLL);
		return NULL; /* Can't find the DLL */
	}
	
	fc = (Flow_Creator) dll.symbol(SF_FLOW_CREATION_SYMBOL);
	if ( fc == 0 ) { /* Can't find loading symbol in the DLL */
		dll.close();
		get_smartflow()->setError(FLOW_INVALID_DLL);
		return NULL;
	}
	the_flow =  fc();
	if (! the_flow) { /* Can't initialise the Flow instance */
	   dll.close();
	   get_smartflow()->setError(FLOW_INVALID_DLL);
	   return NULL;
	}
	
	return the_flow;
}

} // namespace smartflow
