#ifndef _FlowManager_STL_H
#define _FlowManager_STL_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "Smartflow.h"
#include "FlowManager.h"

#include <map>

namespace smartflow {

/** \brief Class FlowManager_STL .
 * This class implement the flow manager. 
 * It uses a STL map to store the flows for the current 
 * client. 
 * It also communicate with the server (using the communication manager) 
 * to register or remove flows from the network, and also 
 * uses the ACE library to dynamically instanciate flow classes.
 * 
 * \author Nist Smartspace Group
 * \date 04-19-2004
 * \internal
*/
class ACE_Svc_Export FlowManager_STL:public FlowManager {

typedef FlowManager super;

public:

	/** \brief Class constructor. 
         * \warning This method does not initialise the object, the init() method
         * must be used for this purpose.
        */
	FlowManager_STL();

    /** \brief Class destructor.
         * Take care of any cleanup that is necessary to free the object.
        */
    virtual ~FlowManager_STL();

        /** \brief Class initialiser. Must be called just after the object
         * construction. 
         * \param sf The Smartflow object in which this object will be registered
         * \return 0 if the object is correctly initialised, otherwise -1. If possible,
         * an error will be set via Smartflow::seterror() to explain the error.
        */
    virtual bool init(Smartflow *sf);
        
        /** \brief Create a flow object. 
         * Implementation from FlowManager::makeFlow().
        */
    virtual Flow *makeFlow(const std::string &type,const std::string &name,const std::string &group,const std::string &arg,flow_direction_t direction, flow_policy_t policy, unsigned int numberOfConsumersExpected);
        
        /** \brief Locate an existing flow within a client by full name
         * Implementation from FlowManager::find()
        */
   	virtual Flow * findFlow(const std::string &name);
	
	/** \brief Create a flow object. 
	 * Implementation from FlowManager::stopAllFlows().
	 * \return 0 if everything is fine, -1 otherwise.
	 */
	virtual bool stopAllFlows();
        
private:
	/* A map type ensure that each key is unique */
	typedef std::map<std::string,Flow *> Flowlist_t ;
	Flowlist_t flowlist;
	
	/* Insert the flow instance into the map list.
     * \warning No integrity checking of parameters or instance state !
     * \internal 
    */
	void insertFlow(const std::string & fname,Flow * f);

	/* remove a flow from the list (by instance pointer)
     * \warning No integrity checking of parameters or instance state !
     * \internal
    */
	void removeFlow(Flow * f);
	
    /* remove a flow from the list (by full-name)
     * \warning No integrity checking of parameters or instance state !
     * \internal
    */
    void removeFlow(const std::string fname);
    
    /* Un-register a flow from the system (from the server) */
    virtual bool destroyFlow(Flow *f);
    
   /* This method use the ACE library to dynamic load a Flow class of the type given.
    * The flow class is then initialised with the parameters provided.
    * \warning No integrity checking of parameters or instance state !
    * \internal
    */
    Flow * dynLoadFlow(const std::string & type);
    
        
}; // class FlowManager_STL
}  // namespace SMARTFLOW
#endif
