#include "Log_Manager_ACE.h"

#include "Smartflow.h"
#include "PreferenceManager.h"
#include "ConnectionManager.h"
#include "ace/Time_Value.h"
#include "ace/Reactor.h"
#include "ace/Event_Handler.h"
#include <iostream>

namespace smartflow {

#define DEFAULT_HIGH_WATERMARK 131072

Log_Callback::Log_Callback(Log_Manager_ACE *manager) : ACE_Log_Msg_Callback() {
	//ACE_TRACE(ACE_TEXT("Log_Callback::Log_Callback"));
	manager_ = manager;
}

Log_Callback::~Log_Callback() {
	//ACE_TRACE(ACE_TEXT("Log_Callback::~Log_Callback"));
	manager_ = NULL;
}

void Log_Callback::log(ACE_Log_Record &log_record) {
	//first we disable the callback to avoid an infinite loop...
	if ( manager_->forwardLog() )
		ACE_LOG_MSG->clr_flags (ACE_Log_Msg::MSG_CALLBACK);	
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Callback::log"));
	//we need to send the log to the server here

	putLogInQueue(log_record);
	sendLogToServer();
	
	//job is done, we can enable the callback again...
	if ( manager_->forwardLog() )
		ACE_LOG_MSG->set_flags (ACE_Log_Msg::MSG_CALLBACK);
}

void Log_Callback::putLogInQueue(ACE_Log_Record &log_record) {
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Callback::putLogInQueue"));
	
	//we get the size of the message
	size_t size = log_record.length();

	//first, we create a message block to store the log
	ACE_Message_Block *mb = new ACE_Message_Block(size);
	if ( mb )
		//we copy the log in the message block
		ACE_OS::memcpy( mb->wr_ptr(), log_record.msg_data(), size);
	else 
		return;
	
	//we tell the message block how much data we put in it
	mb->wr_ptr(size);

	ACE_Time_Value timeout(ACE_OS::gettimeofday());
	ACE_Message_Queue<ACE_SYNCH> *tmp_queue = manager_->getQueue();
	if ( tmp_queue ) {
		//we enqueue the message, if the queue is full or blocked,
		//this call will return after one seconde.
		if ( manager_->getQueue()->enqueue_tail(mb, &timeout) == -1 ) {
			//it failed, we forget about this message and delete the message block		
			delete mb;
		}
	}
	//std::cout << "*******Put log in queue, queue size=" << tmp_queue->message_count() << std::endl;
	
	return;
}

bool Log_Callback::sendLogToServer() {
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Callback::sendLogToServer"));
	if ( manager_->forwardLog() ) {
		//here we need to format the message properly and notify the reactor to invoke the callback that sends the log
		ACE_Message_Queue<ACE_SYNCH> *tmp_queue = manager_->getQueue();
		if ( !tmp_queue )
			return false;
		else {
			//we don't want to notify the reactor all the time, only every 5 message or 3 secondes
			if ( ( tmp_queue->message_count() >= 5 ) && ( !manager_->isNotificationSent() ) ) {
				manager_->getLogHandler()->reactor()->notify(manager_->getLogHandler(), ACE_Event_Handler::READ_MASK);
				manager_->setNotificationSent(true);
				//ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Callback::sendLogToServer ] Notification Sent\n")));
			}
		}
	}
	return true;
}

/************************************************************/

Log_Manager_ACE::Log_Manager_ACE() : super() {
	ACE_TRACE(ACE_TEXT("Log_Manager_ACE::Log_Manager_ACE"));
	cb = NULL;
	file = NULL;
	log_handler = NULL;
	notificationSent = false;
	logForwardEnabled = false;
}	
	
Log_Manager_ACE::Log_Manager_ACE(Log_Handler *logHandler) : super() {
	ACE_TRACE(ACE_TEXT("Log_Manager_ACE::Log_Manager_ACE"));
	Log_Manager_ACE();
	log_handler = logHandler;
}

Log_Manager_ACE::~Log_Manager_ACE() {
	ACE_TRACE(ACE_TEXT("Log_Manager_ACE::~Log_Manager_ACE"));
	if ( cb ) {
		ACE_LOG_MSG->clr_flags (ACE_Log_Msg::MSG_CALLBACK);
		delete cb;
	}
	cb = NULL;
	if ( file ) {
		ACE_LOG_MSG->clr_flags (ACE_Log_Msg::OSTREAM);
		delete file;
	}
	file = NULL;
	log_handler = NULL;
	logQueue.deactivate();
	logQueue.flush();
}

bool Log_Manager_ACE::init(Smartflow *sf, const char* name) {
	ACE_TRACE(ACE_TEXT("Log_Manager_ACE::init"));

	if ( !super::init(sf, name) )
		return false;

	//init the message queue
	logQueue.high_water_mark(DEFAULT_HIGH_WATERMARK);
	logQueue.low_water_mark(0);
	logQueue.flush();

	//we need to set up our debug level here
	//need to get the info from the PreferenceManager
	std::string tmp_level = "loglevel";
	ACE_TString log_level = ACE_TEXT_CHAR_TO_TCHAR((get_smartflow()->getPreferenceManager()->getPreferenceValue(tmp_level)).c_str());
	if ( log_level == "debug" ) {
		//the log level is highest, we get the errors, warnings and debug info;
		ACE_LOG_MSG->priority_mask( LM_TRACE | LM_DEBUG | LM_WARNING | LM_ERROR, ACE_Log_Msg::PROCESS);
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Severity set = debug\n")));
	}
	else if ( log_level == "warning" ) {
		//the log leve is medium, we get ther errors and warnings
		ACE_LOG_MSG->priority_mask( LM_TRACE | LM_WARNING | LM_ERROR, ACE_Log_Msg::PROCESS);
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Severity set = warning\n")));
	}
	else {
		//the log level is not specified or is the default one, we get the errors only
		ACE_LOG_MSG->priority_mask( LM_TRACE | LM_ERROR, ACE_Log_Msg::PROCESS);
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Severity set = error\n")));
	}

	//used to invoke a callback for logging operations
	cb = new Log_Callback(this);

	std::string tmp_syslog = "syslog";	
	ACE_TString hasSyslog = ACE_TEXT_CHAR_TO_TCHAR((get_smartflow()->getPreferenceManager()->getPreferenceValue(tmp_syslog)).c_str());
	if ( hasSyslog == "1" ) {
		//We are logging to the syslog of the system and stderr
		ACE_LOG_MSG->open(name, ACE_Log_Msg::SYSLOG, name);
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Logging to Syslog On\n")));

	}
	else {
		ACE_LOG_MSG->open(name);
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Logging to Syslog Off\n")));
	}

	//the log is going to stderr
	ACE_LOG_MSG->set_flags (ACE_Log_Msg::STDERR);
	
	std::string tmp_callback = "logforward";
	ACE_TString isCallbackOn = ACE_TEXT_CHAR_TO_TCHAR((get_smartflow()->getPreferenceManager()->getPreferenceValue(tmp_callback)).c_str());
	if ( isCallbackOn == "0" ) {
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Log Propagation Off\n")));
	}
	else {
		//enable the logging from the callback
		ACE_LOG_MSG->set_flags (ACE_Log_Msg::MSG_CALLBACK);
		//setting the callback object
		ACE_LOG_MSG->msg_callback(cb);
		//logForwardEnabled = true;
		ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Log Propagation On\n")));
	}
	
	//we need to enable the log info to a file
	//need to get the info from the PreferenceManager
	std::string tmp_file = "logfile";
	ACE_TString log_file = ACE_TEXT_CHAR_TO_TCHAR((get_smartflow()->getPreferenceManager()->getPreferenceValue(tmp_file)).c_str());
	if ( log_file.length() != 0 ) {
		file = new std::ofstream(log_file.c_str());
		if ( !file ) 
			ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Log_Manager_ACE::init ] Log file couldn't be opened")));
		else {
			//the file has successfuly been opened
			//setting the output strean

			//need to check again what 0 means
			ACE_LOG_MSG->msg_ostream(file);
			//enable the logging to the output stream
			ACE_LOG_MSG->set_flags (ACE_Log_Msg::OSTREAM);
			ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Log_Manager::init ] Logging to file: %s\n"), log_file.c_str()));
		}
	}
	return true;
}

	
void Log_Manager_ACE::setLogHandler(Log_Handler *lh) { 
	log_handler = lh; 
	if ( log_handler == NULL )
		logForwardEnabled = false;
	else
		logForwardEnabled = true;
}
	
/************************************************************/
	
Log_Handler::Log_Handler(ACE_Reactor *reactor, Smartflow *sf_) : ACE_Event_Handler(reactor) {
	sf = sf_;
	lm_ace =(Log_Manager_ACE *) (sf->getLogManager());
	lm_ace->setLogHandler(this);
	if ( reactor ) {
		this->reactor(reactor);
		this->reactor()->register_handler(this, ACE_Event_Handler::READ_MASK);
	}
	enableTimer();
}

Log_Handler::~Log_Handler() {}

int Log_Handler::handle_close(ACE_HANDLE, ACE_Reactor_Mask) {
	disableTimer();
	lm_ace->setLogHandler(NULL);
	delete this;
	return 0;
}

int Log_Handler::handle_input(ACE_HANDLE handle_) {
	//commented to avoid infinite loop
	//ACE_TRACE(ACE_TEXT("Log_Handler::handle_input"));
	ACE_UNUSED_ARG(handle_);
	ACE_TString messages;
	int result = 0;
	int stop = 0;
	unsigned int nbMessages = 0;
	ACE_Message_Block *mb;
	//we don't have access to LOG_MESSAGE from here and we need its size
	//this is really nasty to do that but the message is not supposed to change...
	size_t lengthMessage = 180;
	//we add the size of the parameter, meaning .....
	lengthMessage += sf->getClientName().length() + sf->getGroupName().length();
	
	//here we need to format the message properly and send it to the server
	ACE_Message_Queue<ACE_SYNCH> *messageQueue = NULL;
	if ( lm_ace )
		messageQueue = lm_ace->getQueue();
	if ( messageQueue ) {
		//here we need to dequeue, format the message and pass it to the method, 
		//which send the message to the server
		nbMessages = messageQueue->message_count();
		while ( (nbMessages != 0) && !stop ) {
			//we dequeue the message...
			ACE_Time_Value timeout(ACE_OS::gettimeofday());
			result = messageQueue->dequeue_head(mb, &timeout);
			if ( result == -1 )
				stop = 1;
			else {
				//get its size
				int lengthMB = mb->length();
				
				if ( (lengthMessage + lengthMB + 1)  > 131072 ) {
					stop = 1;
					//we put back the message in the queue
					ACE_Time_Value timeout(ACE_OS::gettimeofday());
					result = messageQueue->enqueue_head(mb, &timeout);
				}
				else {
					//set up its read pointer on the beginning of the message
					mb->rd_ptr(mb->base());
					//Allocate the buffer to copy the message
					ACE_TCHAR *dest = (ACE_TCHAR *) ACE_OS::malloc(lengthMB+1);
					//copy the message into dest
					ACE_OS::memcpy(dest, mb->rd_ptr(), lengthMB);
					
					ACE_TString messTString = dest;
					ACE_TString messInXML = create_XML_message(messTString);
					messages += messInXML;
					
					//we update the size
					lengthMessage += lengthMB;
					
					nbMessages--;
					
					//we don't need the message block anymore, we delete it
					delete mb;
					mb = NULL;
					//we free the dest variable
					ACE_OS::free(dest);
				}
			}
		}
		if ( messages.length() > 0 ) {
			//we have at least one log message int messages.	
			//now we send it
			sf->getConnectionManager()->sendLog(messages.c_str());
		}
		if ( nbMessages != 0 ) {
			//we didn't take care of everymessage, we will finish the job later
			//because we don't want to let the reactor blocked for too long..
			//so we notify the reactor to call again this method later
			reactor()->notify(this, ACE_Event_Handler::READ_MASK);
		}
		else {
			//here we notify the Log_Callback object that we received the notification
			//so it can notify the reactor to invoke this method later
			lm_ace->setNotificationSent(false);
		}
	}
	return 0;	
}


int Log_Handler::handle_timeout(const ACE_Time_Value &currentTime, const void *) {
	ACE_UNUSED_ARG(currentTime);
	ACE_Message_Queue<ACE_SYNCH> *messageQueue = NULL;
	if ( lm_ace )
		messageQueue = lm_ace->getQueue();
	if ( messageQueue ) {
		unsigned int nbMessages = messageQueue->message_count();
		if ( nbMessages != 0 && lm_ace && lm_ace->isNotificationSent() == false ) {
			reactor()->notify(this, ACE_Event_Handler::READ_MASK);
			lm_ace->setNotificationSent(true);
		}
	}
	return 0;
}

void Log_Handler::enableTimer() {
	ACE_Time_Value delay(1);
	ACE_Time_Value interval(1);
	if ( this->reactor() )
		this->reactor()->schedule_timer(this, 0, delay, interval);
}

void Log_Handler::disableTimer() {
	if ( this->reactor() )
		this->reactor()->cancel_timer(this);
}
	

ACE_TString Log_Handler::create_XML_message(ACE_TString message_){
	ACE_TRACE(ACE_TEXT("Log_Callback::create_XML_message"));
	ACE_TString doc ;
	doc += ACE_TEXT("<message>");
	doc += ACE_TEXT("<![CDATA[");
	doc += message_;
	doc += ACE_TEXT("]]>");
	doc += ACE_TEXT("</message>\n");
	
	return doc;
}


} // namespace SMARTFLOW

