#ifndef _LOG_MANAGER_ACE_H
#define _LOG_MANAGER_ACE_H



#include "ace/OS.h"
#include "ace/streams.h"
#include "ace/Log_Msg.h"
#include "ace/Log_Msg_Callback.h"
#include "ace/Log_Record.h"
#include "ace/SString.h"
#include "ace/Message_Queue.h"
#include "ace/Thread_Manager.h"

#include "Log_Manager.h"
#include "Smartflow.h"

namespace smartflow {

class Log_Handler;
class Log_Callback;
class Log_Manager_ACE;

/** \brief Class Log_Callback .
 * This class is used to implement a callback method
 * invoked when a logging operation occured.
 * It is used to propagate the log to the server
 * \author Nist Smartspace Group
 */
class Log_Callback : public ACE_Log_Msg_Callback {
	public:
		Log_Callback(Log_Manager_ACE *manager);
		~Log_Callback();
		void log(ACE_Log_Record &log_record);
		
    private:
		bool sendLogToServer();
		void putLogInQueue(ACE_Log_Record &log_record);
		Log_Manager_ACE *manager_;
};
	
class Log_Handler : public ACE_Event_Handler {
public:
	Log_Handler(ACE_Reactor *reactor, Smartflow *sf_);
	virtual ~Log_Handler();
	virtual int handle_input(ACE_HANDLE handle_ = ACE_INVALID_HANDLE);
	virtual int handle_timeout(const ACE_Time_Value &currentTime, const void * = 0 );
	virtual int handle_close(ACE_HANDLE = ACE_INVALID_HANDLE, ACE_Reactor_Mask = 0);
	void enableTimer();
	void disableTimer();
	void setLogManager( Log_Manager_ACE *lm_) { lm_ace = lm_; }
private:
	ACE_TString create_XML_message(ACE_TString message_);
	Smartflow *sf;
	Log_Manager_ACE* lm_ace;
};
	
/** \brief Class Log_Manager_ACE
 * This class is used to manage the log system.
 * It is used to define the desired severity.
 * It is also used to send the log info to the server, 
 * write them in a file if desired and print them on the stderr
 */
class ACE_Svc_Export Log_Manager_ACE : public Log_Manager {

   typedef Log_Manager super;

   public:
    
     /** \brief Class constructor. 
     * \warning This method does not initialise the object, the init() method
     * must be used for this purpose.
     * 
     */
	Log_Manager_ACE();
	
    Log_Manager_ACE(Log_Handler *logHandler);
    
	/** \brief Class destructor.
     * Take care of any cleanup that is necessary to free the object.
     */
	virtual ~Log_Manager_ACE();
    
	/** \brief Class initialiser. Must be called just after the object
     * construction. 
     * \param sf The Smartflow object in which this object will be be registered.
     * \return 0 if the object is correctly initialised, otherwise -1. If possible,
     * an error will be set via Smartflow::seterror() to explain the error.
     */
	virtual bool init(Smartflow *sf, const char* name);

	Log_Handler * getLogHandler() {return log_handler; }
	
	void setLogHandler(Log_Handler *lh);
	
	ACE_Message_Queue<ACE_SYNCH> *getQueue() { return &logQueue; }

    bool isNotificationSent() { return notificationSent;}
	
	void setNotificationSent(bool notif) { notificationSent = notif; }
	
	bool forwardLog() { return logForwardEnabled; }
	
private:
	Log_Callback *cb;
	Log_Handler *log_handler;
	ACE_OSTREAM_TYPE *file;
	ACE_Message_Queue<ACE_SYNCH> logQueue;
	bool notificationSent;
	bool logForwardEnabled;
	
};
	
}

#endif

