#ifndef _PreferenceManager_STL_H
#define _PreferenceManager_STL_H

#include "ace/svc_export.h" // for exported symbols (windows)

#include "Smartflow.h"
#include "PreferenceManager.h"

#include <string>
#include <map>

using namespace std;
namespace smartflow {

/** \brief Class PreferenceManager_STL .
 * Implementation of the PreferenceManager. This class hold and give access
 * to the actual list of preference.
 * It also provide facilities to parse files and command-line, and also to
 * write a preference file from the current key=value list.
 * \author Nist Smartspace Group
 * \date 04-19-2004
 * \internal
*/
class ACE_Svc_Export PreferenceManager_STL:public PreferenceManager {
/** Redefine PreferenceManager (which is the ancestor of PreferenceManager_STL) as
    \b super . Facility taken from the Java language.
*/
typedef PreferenceManager super;
public:
	/** \brief Class constructor. 
         * \warning This method does not initialise the object, the init() method
         * must be used for this purpose.
        */
		PreferenceManager_STL();
        /** \brief Class destructor.
        */
        virtual ~PreferenceManager_STL();
        /** \brief Class initialiser. Must be called just after the object
         * construction. 
         * \param sf The Smartflow object in which this object will be registered
         * \return 0 if the object is correctly initialised, otherwise -1. If possible,
         * an error will be set via Smartflow::seterror() to explain the error.
        */
        virtual bool init(Smartflow *sf);

/** This function allows to see the list of the preferences on stdout.
 */
virtual void viewPrefs();

/** This function returns the count of the preferences.
 *@return int The total number of the preferences. 
 */
virtual int count();

 /** This function uses the command-line arguments to setup the 
* couples (key,values) that the user has given.
* The argc/argv arguments are modified to remove the used arguments.
* @return -1 if there has been an error in setting the 
* preferences. The function does not return false on a parse error,
* since the user might want to use argc,argv for his own usage.
*/
 bool initArgs(int & argc, char** &argv, bool keepSFOptions);
  
private:

/** This is the map where are stored the preference.
 *  A preference is a couple of two strings, the first string is a key,
 *  which is unique. And the second string is the value associated with the
 *  key.
 */
map<string, string> thePrefs;

/** This map is used to store comments associated with preferences.
 *  A comment is associated with a key. 
 */
map<string, string> theComments;

/** This is a private function, which allows to add a preference without a comment.
 *  The developper never uses this functions. It is called by the method 
 *  addPreference of the Preferencemanager class.
 *@param key This is the ID of the preference. A key, which is a string, is unique.
 *@param value This is the value (a string) associated with the key.
 */
virtual void addPref(string &key, string &value);

/** This is a private function, which allows to add a preference with a comment.
 *  The developper never uses this functions. It is called by the method 
 *  addPreference of the Preferencemanager class.
 *@param key This is the ID of the preference. A key, which is a string, is unique.
 *@param value This is the value (a string) associated with the key.
 *@param comment This is a string, which comments the preference.
 */
virtual void addPref(string &key, string &value, string &comment);

/** This is a private function, which allows to remove a preference.
 *  The developper never uses this functions. It is called by the method 
 *  remPreference of the Preferencemanager class.
 *@param key This is the key of the preference, which will be remove.
 */
virtual void remPref(string &key);

/** This is a private function, which allows to get a preference (a string).
 *  A preference is identified by its key.
 *  The developper never uses this functions. It is called by the method 
 *  getPreferenceValue of the Preferencemanager class.
 *@return string The value of the preference associated with the key. 
 *@param key This is the ID of the preference. A key, which is a string, is unique.
 */
virtual string getPrefValue(string &key);

/** This is a private function, which allows to get the integer value of
 *  a preference. A preference is identified by its key.
 *  The developper never uses this functions. It is called by the method 
 *  getPreferenceIntValue of the Preferencemanager class.
 *@return int The integer value of a preference.
 *@param key This is a string, which allows to identify a preference.
 */
virtual int getPrefIntValue(string &key);

 /** This is a private function, which allows to get the double value of a preference. 
  *  A preference is identified by its key.
  *  The developper never uses this functions. It is called by the method 
  *  getPreferenceDoubleValue of the Preferencemanager class.
  *@return double The double value of a preference.
  *@param key This is a string, which allows to identify a preference.
  */
virtual double getPrefDoubleValue(string &key);

/** This is a private function, which allows to get the key of a preference.
 *  A preference is identified by its number.    
 *  This number must be > 0 and less than the total of the preferences
 *  The developper never uses this functions. It is called by the method 
 *  getPreferenceKey of the Preferencemanager class.
 *@return string The key of the preference identified by his number.
 *@param num The number of the key.
 */ 
virtual string getPrefKey(int num);

/** This is a private function, which allows to get the comment of a preference.
 *  A preference is identified by its key.
 *  The developper never uses this functions. It is called by the method 
 *  getPreferenceComment of the Preferencemanager class.
 *@return string The comment associated with the key. 
 *@param key This is a string, which allows to identify a preference.
 */
virtual string getPrefComment(string &key);

/** This private function is used to see if a comment exist for the 
 *  preference identified by the key.
 *@param key This is a string, which allows to identify a preference.
 */
virtual bool existsComment(string &key);

/** This private function is used in the class PreferenceManager_XXX
 *  to check is the preference identified by the key exist in the 
 *  list of preferences.
 *@param key This is a string, which allows to identify a preference.
 */
virtual bool exists(string &key);

}; // class PreferenceManager_STL
}  // namespace SMARTFLOW
#endif
