#include "Smartflow.h"
#include "ErrorManager.h"

#include "PreferenceManager_STL.h"
#include "ConnectionManager_ACE.h"
#include "MemoryManager_ACE.h"
#include "FlowManager_STL.h"
#include "Log_Manager_ACE.h"

#include <map>
#include <string>
#include <iostream>

using namespace std;

namespace smartflow {

/* Global list which hold a pointer on all instances of the smartflow
   for a given program. */
std::map<std::string,Smartflow*> Smartflow::smartflow_map_;

#if (!(defined (ACE_WIN32)))
Smartflow::Smartflow()
	:manager_spawn_good_(false),manager_init_good_(false),
	preferencemanager_(0), flowmanager_(0),
	memorymanager_(0),connectionmanager_(0),errormanager_(0) 
{
	ACE_TRACE(ACE_TEXT("Smartflow::Smartflow"));
	user_exit_function = NULL;
	user_exit_function_with_parameters = NULL;
	user_exit_func_argument = NULL;
}
#else
Smartflow::Smartflow() {
	ACE_TRACE(ACE_TEXT("Smartflow::Smartflow"));
	ACE::init();
	user_exit_function = NULL;
	user_exit_function_with_parameters = NULL;
	user_exit_func_argument = NULL;
	manager_spawn_good_ = false;
	manager_init_good_ = false;
	preferencemanager_ = 0;
	flowmanager_ = 0;
	memorymanager_ = 0;
	connectionmanager_ = 0;
	errormanager_ = 0;
	logmanager_ = 0;
}

#endif

Smartflow::~Smartflow()
{
	ACE_TRACE(ACE_TEXT("Smartflow::~Smartflow"));
	stop();
	
	smartflow_map_.erase(application_name_);
	application_name_.clear();
	
	delete flowmanager_; flowmanager_ = 0;
    delete memorymanager_; memorymanager_=0;
	delete logmanager_; logmanager_ = 0;
	delete connectionmanager_; connectionmanager_ = 0; 
	delete preferencemanager_; preferencemanager_ = 0;
    delete errormanager_; errormanager_=0;

    manager_spawn_good_ = false;
    manager_init_good_ = false;
	ACE::fini();
} //~Smartflow()


bool Smartflow::createManagers(void)
{
	ACE_TRACE(ACE_TEXT("Smartflow::createManagers"));
	errormanager_ = new ErrorManager();	
    preferencemanager_ = new PreferenceManager_STL();
    connectionmanager_ = new ConnectionManager_ACE();
	logmanager_ = new Log_Manager_ACE();
    memorymanager_ = new MemoryManager_ACE();
    flowmanager_ = new FlowManager_STL();
    if ( ! (errormanager_ && preferencemanager_
        && connectionmanager_ && memorymanager_ && flowmanager_ && logmanager_) ) {
        return false;
     }
    // Early initialisation of the Error Manager
	if ( !errormanager_->init(this) ) {  return false; }
    manager_spawn_good_ = true;
    return true; 
}	

bool Smartflow::init(int &argc,char** &argv, bool keepSFOptions, const std::string application)
{
	ACE_TRACE(ACE_TEXT("Smartflow::init"));
	createManagers();
	  
	if ( ! manager_spawn_good_ ) { 
		return false; 
	}

    // No application => abort 
	if ( application.empty() ) { 
		setError(INVALID_APPLICATION_NAME); 
		return false; 
	}
    // Look if the given application does not exist yet in the list 
	if ( smartflow_map_.find(application) !=  smartflow_map_.end() ) {
		setError(APPLICATION_ALREADY_EXIST); 
		return false; 
	}
	application_name_ = application;
	
	// Managers initialisations 
	if ( !preferencemanager_->init(this) ) { return false; }	
	// Early initialisation to allow other managers to use the preferences 
	if (argv != NULL) 
		client_name_argv0 = argv[0];
	if ( !preferencemanager_->initArgs(argc,argv, keepSFOptions) ) {	return false;}
	if ( !connectionmanager_->init(this) ) {  return false; }	
	if ( !memorymanager_->init(this) )     {  return false; }
	if ( !flowmanager_->init(this) )       {  return false; }
	if ( !logmanager_->init(this, client_name_argv0.c_str()) ) {  return false; }

	//Add this smartflow in the list of the instances for this application 
	smartflow_map_[application_name_] = this;
	// Good to go, clear the error and set good condition 
	manager_init_good_ = true;
	
	return true;
} //init()

bool Smartflow::pauseFlow(const std::string &clientname,const std::string &clientgroup,const std::string &flowtype,const std::string &flowgroup) {
    ACE_TRACE(ACE_TEXT("Smartflow::pauseFlow"));
	if ( readyToRun() && (connectionmanager_ != 0))
      return connectionmanager_->pauseFlow(clientname,clientgroup,flowtype,flowgroup);
    setError(SMARTFLOW_NOT_READY);
    return false;
  }

bool Smartflow::pause_flow(const std::string &clientname,const std::string &clientgroup,const std::string &flowtype,const std::string &flowgroup) {
	return this->pauseFlow(clientname, clientgroup, flowtype, flowgroup);
}
 
bool Smartflow::restartFlow(const std::string &clientname,const std::string &clientgroup,const std::string &flowtype,const std::string &flowgroup) {
    ACE_TRACE(ACE_TEXT("Smartflow::restartFlow"));
	if ( readyToRun() && (connectionmanager_ != 0))
		return connectionmanager_->restartFlow(clientname,clientgroup,flowtype,flowgroup);
    setError(SMARTFLOW_NOT_READY);
    return false;
  }
	
bool Smartflow::restart_flow(const std::string &clientname,const std::string &clientgroup,const std::string &flowtype,const std::string &flowgroup) {
	return this->restartFlow(clientname, clientgroup, flowtype, flowgroup);
}


  static void (*static_user_exit_function) ();
  static void (*static_user_exit_function_with_parameters) (void *userarg);
  static void *static_user_exit_func_argument;
  #if (defined (ACE_WIN32) && !defined(WINCE))
	
   // Windows Keyboard and console interrupt handler.
   static BOOL WINAPI ctrl_c_handler (DWORD type)
   {
	  // If I got a CTRL-C, a CTRL-BREAK or a CTRL-CLOSE-EVENT I want to do my
      // cleanup and exit gracefully now
      if (type == CTRL_C_EVENT || type == CTRL_BREAK_EVENT || type == CTRL_CLOSE_EVENT)
      {
		if (static_user_exit_function != NULL)
			static_user_exit_function();
		else if (static_user_exit_function_with_parameters != NULL && static_user_exit_func_argument != NULL)
			static_user_exit_function_with_parameters(static_user_exit_func_argument);
		else 
			// Notify to call other handlers.
			return FALSE;
      }

      // Notify to not call other handlers.
      return TRUE;
   }
  #else 
   static void ctrl_C(int data) { 
     if (static_user_exit_function != NULL)
       static_user_exit_function();
     else if (static_user_exit_function_with_parameters != NULL && static_user_exit_func_argument != NULL)
       static_user_exit_function_with_parameters(static_user_exit_func_argument);
   }
  #endif

  /** Functions used by the user in order to define the callback function 
   * that must be called to exit properly the client when the order to quit 
   * is launched by the server (typically asked by the control center) and
   * eventually when  the program quit because of a Control-C or because the 
   * user closes the console window.
   */

bool Smartflow::setUserExitFunction(void (*user_exit_function_) (), bool controlC) {
	ACE_TRACE(ACE_TEXT("Smartflow::setUserExitFunction"));
	user_exit_function = user_exit_function_;
	if (user_exit_function == NULL) {
		return false;
	}
	if (controlC) {
		static_user_exit_function = user_exit_function_;
		#if (defined (ACE_WIN32) && !defined(WINCE))
			if (!SetConsoleCtrlHandler (ctrl_c_handler, TRUE))
				return false;
		#else 
			#if (!defined(WINCE))
				//this code is required to handle the Ctrl+C signal under linux.
				struct sigaction sa;
				sa.sa_handler = ctrl_C;
				sigemptyset(&sa.sa_mask);
				sigaction(SIGINT, &sa, 0);
				sigaction(SIGTERM,&sa, 0);
			#endif
		#endif
	}
	return true;
}

bool Smartflow::set_user_exit_function(void (*user_exit_function_) (), bool controlC) {
	return setUserExitFunction(user_exit_function_, controlC);
}
	
bool Smartflow::setUserExitFunction(void (*user_exit_function_) (void*userarg), void *arg, bool controlC) {
	ACE_TRACE(ACE_TEXT("Smartflow::setUserExitFunction"));
	user_exit_function_with_parameters = user_exit_function_;
	user_exit_func_argument = arg;
	if (user_exit_function_with_parameters == NULL || user_exit_func_argument == NULL) {
		return false;
	}
	if (controlC) {
		static_user_exit_function_with_parameters = user_exit_function_;
		static_user_exit_func_argument = arg;
		#if (defined (ACE_WIN32) && !defined(WINCE))
			if (!SetConsoleCtrlHandler (ctrl_c_handler, TRUE))
				return false;
		#else
			#if (!defined(WINCE))
				//this code is required to handle the Ctrl+C signal under linux.
				struct sigaction sa;
				sa.sa_handler = ctrl_C;
				sigemptyset(&sa.sa_mask);
				sigaction(SIGINT, &sa, 0);
				sigaction(SIGTERM,&sa, 0);
			#endif
		#endif
	}
	return true;
}

bool Smartflow::set_user_exit_function(void (*user_exit_function_) (void*userarg), void *arg, bool controlC) {
	return setUserExitFunction(user_exit_function_, arg, controlC);
}
	
std::string & Smartflow::getClientName() {
	ACE_TRACE(ACE_TEXT("Smartflow::getClientName"));
	return clientName;
}

std::string & Smartflow::get_client_name() {
	return this->getClientName();
}


std::string & Smartflow::getGroupName() {
	ACE_TRACE(ACE_TEXT("Smartflow::getGroupName"));
	return clientGroup;
}
  
std::string & Smartflow::get_group_name() {
	return this->getGroupName();
}

bool Smartflow::readyToRun() 
{	
	ACE_TRACE(ACE_TEXT("Smartflow::readyToRun"));
	return manager_spawn_good_ && manager_init_good_; 
}
	
bool Smartflow::ready_to_run() {
	return this->readyToRun();
}

bool Smartflow::stop()
{	
	ACE_TRACE(ACE_TEXT("Smartflow::stop"));
	return disconnect_from_application_server();
}

std::string & Smartflow::applicationName(void)
{ 	
	ACE_TRACE(ACE_TEXT("Smartflow::applicationName"));
	return application_name_;
}

std::string & Smartflow::application_name(void) {
	return this->applicationName();
}

bool Smartflow::connectToApplicationServer(const std::string clientname, const std::string clientgroup)
{
	ACE_TRACE(ACE_TEXT("Smartflow::connectToApplicationServer"));
	if ( readyToRun() && (connectionmanager_ != 0)) { 
		if ( clientname.empty() ) {
			clientName = client_name_argv0;
			clientGroup = clientgroup;
			bool result = connectionmanager_->connectToApplicationServer(client_name_argv0,clientgroup);
			if ( result == true ) 
				ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Smartflow::connectToApplicationServer ] Connection Successful\n")));
			return result;
		}
		else
		{
			clientName = clientname;
			clientGroup = clientgroup; 
			bool result = connectionmanager_->connectToApplicationServer(clientname,clientgroup);
			if ( result == true ) 
				ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Smartflow::connectToApplicationServer ] Connection Successful\n")));
			return result;
		}
	}
	setError(SMARTFLOW_NOT_READY);
	return false;
}
	
bool Smartflow::connect_to_application_server(const std::string clientname, const std::string clientgroup) {
	return this->connectToApplicationServer(clientname, clientgroup);
}

bool Smartflow::disconnectFromApplicationServer()
{
	ACE_TRACE(ACE_TEXT("Smartflow::disconnectFromApplicationServer"));
	if ( readyToRun() && (connectionmanager_ != 0)) {
		bool result =  connectionmanager_->disconnectFromApplicationServer();
		if ( result == true ) {
			ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Smartflow::disconnectFromApplicationServer ] Disconnection Successful\n")));
		}
		else {
			ACE_DEBUG((ANNOUNCE_PREFIX ACE_TEXT("[ Smartflow::disconnectFromApplicationServer ] Disconnection Failed\n")));
		}
		return result;
	}
	setError(SMARTFLOW_NOT_READY);
	return false;
}
	
bool Smartflow::disconnect_from_application_server() {
	return this->disconnectFromApplicationServer();
}
	
  
bool Smartflow::startClient(const std::string clientpath, 
                           const std::string server,
                           const std::string arguments,
                           const std::string parameters)
{
	ACE_TRACE(ACE_TEXT("Smartflow::startClient"));
	if ( server.empty() || clientpath.empty())
	{ 	setError(SMARTFLOW_NOT_READY);
		return false; 
	}
	if ( readyToRun() && (connectionmanager_ != 0))
		return connectionmanager_->startClient(clientpath,server,arguments,parameters);
	setError(SMARTFLOW_NOT_READY);
	return false;
	
}

bool Smartflow::start_client(const std::string clientpath, 
							const std::string server,
							const std::string arguments,
							const std::string parameters)
{
	return this->startClient(clientpath, server, arguments, parameters);
}
	
bool Smartflow::stopClient(const std::string clientname,const std::string clientgroup)
{
	ACE_TRACE(ACE_TEXT("Smartflow::stopClient"));
	if (clientname.empty() || clientgroup.empty()) {
		setError(SMARTFLOW_NOT_READY);
		return false;
	}
	if ( readyToRun() && (connectionmanager_ != 0))
		return connectionmanager_->stopClient(clientname,clientgroup);
	setError(SMARTFLOW_NOT_READY);
	return false;
}

bool Smartflow::stop_client(const std::string clientname,const std::string clientgroup) {
	return this->stopClient(clientname, clientgroup);
}
	
bool Smartflow::getApplicationDescription(std::string &flowlist)
{
	ACE_TRACE(ACE_TEXT("Smartflow::getApplicationDescription"));
	if ( readyToRun() && (connectionmanager_ != 0))
		return connectionmanager_->getApplicationDescription(flowlist);
	setError(SMARTFLOW_NOT_READY);
	return false;
}

bool Smartflow::get_application_description(std::string &flowlist) {
	return this->getApplicationDescription(flowlist);
}

bool Smartflow::getServerList(std::string &serverlist)
{
	ACE_TRACE(ACE_TEXT("Smartflow::getServerList"));
	if ( readyToRun() && (connectionmanager_ != 0))
		return connectionmanager_->getServerList(serverlist);
	setError(SMARTFLOW_NOT_READY);
	return false;
}

bool Smartflow::get_server_list(std::string &serverlist) {
	return this->getServerList(serverlist);
}

bool Smartflow::getLastModifications(std::string &lastmodifications)
{
	ACE_TRACE(ACE_TEXT("Smartflow::getLastModifications"));
	if ( readyToRun() && (connectionmanager_ != 0))
		return connectionmanager_->getLastModifications(lastmodifications);
	setError(SMARTFLOW_NOT_READY);
	return false;
}

bool Smartflow::get_last_modifications(std::string &lastmodifications) {
	return this->getLastModifications(lastmodifications);
}

Flow * Smartflow::makeInputFlow(const std::string type,const std::string name,const std::string args,const std::string group, flow_policy_t policy)
{
  ACE_TRACE(ACE_TEXT("Smartflow::makeInputFlow"));
  if ( readyToRun() ){
    return flowmanager_->makeFlow(type,name,group,args,FLOW_INPUT,policy,0);
  }
    setError(SMARTFLOW_NOT_READY);
    return NULL;
}

Flow * Smartflow::make_input_flow(const std::string type,const std::string name,const std::string args,const std::string group, flow_policy_t policy) {
	return this->makeInputFlow(type, name, args, group, policy);	
}

	
Flow * Smartflow::makeOutputFlow(const std::string type,const std::string name,const std::string args,const std::string group, flow_policy_t policy)
{
  ACE_TRACE(ACE_TEXT("Smartflow::makeOutputFlow"));
  if ( readyToRun() ) {
    return flowmanager_->makeFlow(type,name,group,args,FLOW_OUTPUT,policy,0);
  }
    setError(SMARTFLOW_NOT_READY);
    return NULL;
}

Flow * Smartflow::make_output_flow(const std::string type,const std::string name,const std::string args,const std::string group, flow_policy_t policy) {
	return this->makeOutputFlow(type, name, args, group, policy);
}
	
Flow * Smartflow::makeSynchronizedOutputFlow(const std::string type, const std::string name, unsigned int numberOfConsumersExpected, const std::string args, const std::string group)
{
  ACE_TRACE(ACE_TEXT("Smartflow::makeSynchronizedOutputFlow"));
  if ( readyToRun() ) {
	  return flowmanager_->makeFlow(type,name,group,args,FLOW_OUTPUT, DO_NOT_DROP_BUFFER, numberOfConsumersExpected);
  }
  setError(SMARTFLOW_NOT_READY);
  return NULL;
}

Flow * Smartflow::make_synchronized_output_flow(const std::string type, const std::string name, unsigned int numberOfConsumersExpected, const std::string args, const std::string group) {
	return this->makeSynchronizedOutputFlow(type, name, numberOfConsumersExpected, args, group);
}
	
bool Smartflow::destroyFlow( Flow * flow )
{
	ACE_TRACE(ACE_TEXT("Smartflow::destroyFlow"));
	if ( flow == 0 ) { 
		setError(INVALID_FLOW); return false;
	}
	return flowmanager_->destroyFlow(flow);
}

bool Smartflow::destroy_flow( Flow * flow ) {
	return this->destroyFlow(flow);
}

void Smartflow::setError(long int error)
{	
	ACE_TRACE(ACE_TEXT("Smartflow::setError"));
	if ( errormanager_ )
		errormanager_->set_error(error);
}

void Smartflow::set_error(long int error) {
	this->setError(error);
}
	
long int Smartflow::error(void)
{	
	ACE_TRACE(ACE_TEXT("Smartflow::error"));
	if ( errormanager_ )
		return errormanager_->error();
        return -1;
}

std::string Smartflow::strError(void)
{	
	ACE_TRACE(ACE_TEXT("Smartflow::strError"));
	if (errormanager_)
		return errormanager_->strError();
    return DEFAULT_EMPTY;
}

std::string Smartflow::str_error(void) {
	return this->strError();
}


Smartflow * Smartflow::instance(const std::string application)
{
	ACE_TRACE(ACE_TEXT("Smartflow::instance"));
	std::map<std::string,Smartflow*>::iterator i = smartflow_map_.find(application);
	if ( i == smartflow_map_.end() ) {
	    return NULL;
	}
	return i->second;
}

PreferenceManager * Smartflow::getPreferenceManager()
{ 
	ACE_TRACE(ACE_TEXT("Smartflow::getPreferenceManager"));
	return preferencemanager_ ;
}

PreferenceManager * Smartflow::get_preference_manager() {
	return this->getPreferenceManager();
}

FlowManager          * Smartflow::getFlowManager()
{ 
	ACE_TRACE(ACE_TEXT("Smartflow::getFlowManager"));
	return flowmanager_; 
}

FlowManager          * Smartflow::get_flow_manager() {
	return this->getFlowManager();
}

MemoryManager    *  Smartflow::getMemoryManager()
{ 
	ACE_TRACE(ACE_TEXT("Smartflow::getMemoryManager"));
	return memorymanager_;
}
	
MemoryManager    *  Smartflow::get_memory_manager() {
	return this->getMemoryManager();
}

ConnectionManager * Smartflow::getConnectionManager()
{ 
	ACE_TRACE(ACE_TEXT("Smartflow::getConnectionManager"));
	return connectionmanager_;
}
	
ConnectionManager * Smartflow::get_connection_manager() {
	return this->getConnectionManager();
}

ErrorManager          * Smartflow::getErrorManager()
{ 
	ACE_TRACE(ACE_TEXT("Smartflow::getErrorManager"));
	return errormanager_; 
}
	
ErrorManager          * Smartflow::get_error_manager() {
	return this->getErrorManager();
}

Log_Manager * Smartflow::getLogManager() 
{
	ACE_TRACE(ACE_TEXT("Smartflow::getLogManager()"));
	return logmanager_;
}

Log_Manager * Smartflow::get_log_manager() {
	return this->getLogManager() ;
}

}//namespace
