#include "Thread_Client.h"
#include "Client_Handler.h"
#include "Log_Manager_ACE.h"
#include "ConnectionManager_ACE.h"
#include "FlowManager_STL.h"
#include "ace/Reactor.h"
#include <ace/SOCK_Connector.h>
#include <ace/INET_Addr.h>

namespace smartflow {

	
Thread_Stopper::Thread_Stopper(ACE_Reactor *reactor) : ACE_Event_Handler(reactor) {
	this->reactor(reactor);
	this->reactor()->register_handler(this, ACE_Event_Handler::READ_MASK);
}

Thread_Stopper::~Thread_Stopper() {}	
	
int Thread_Stopper::handle_input(ACE_HANDLE handle_) {
	ACE_UNUSED_ARG(handle_);
	this->reactor()->end_reactor_event_loop();
	return 0;
}	

Thread_Client::Thread_Client(ACE_TString name, ACE_TString group, ConnectionManager_ACE *cm_) : ACE_Task_Base(), clientname(name), clientgroup(group), cm(cm_) {
	Thread_Client();
}

Thread_Client::Thread_Client() {
	reactor = NULL;
	theLogHandler = NULL;
	theClientHandler = NULL;
	theThreadStopper = NULL;
	setNeedToInvokeExitMethod(false);
}
	
Thread_Client::~Thread_Client() {
}

void Thread_Client::setNameAndGroup(ACE_TString &name, ACE_TString &group) {
	clientname = name;
	clientgroup = group;
}
	
void Thread_Client::setConnectionManger(ConnectionManager_ACE *cm_) {
	cm = cm_;
}
	
int Thread_Client::svc(void) {
	ACE_TRACE(ACE_TEXT("Thread_Client::svc"));
	ACE_DEBUG((DEBUG_PREFIX ACE_TEXT("[ Thread_Client::svc ] BEGIN \n" )));
	
	//we don't want to use the singleton provided by ACE here.
	//So we create a new reactor
	reactor = new ACE_Reactor();
	
	//Creating the handler used to stop the reactor, therefore initiating the procee
	//to stop the thread
	theThreadStopper = new Thread_Stopper(reactor);
	
	//Creating the client handler.
	theClientHandler = new Client_Handler(reactor, this);
    
	//Creating the log handler.
	//Registration of the handler to the reactor is made inside the method for convenience
	theLogHandler = new Log_Handler(reactor, cm->getSmartflow());
	
    ACE_SOCK_Connector connector;
	//AF_INET to force IPv4
    ACE_INET_Addr local_server_addr (1716, "localhost", AF_INET);
   	
	//connecting to the server
    int err = connector.connect(theClientHandler->peer(),local_server_addr);	
	while (err == -1)
	{
		ACE_Time_Value tmp(0,500000);
		ACE_OS::sleep(tmp);		
		err = connector.connect(theClientHandler->peer(),local_server_addr);
		ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Thread_Client::run_svc ] Error Connect .. trying again... \n" )));
	}
	
	//telling the server who we are
	theClientHandler->peer().send_n(clientname.c_str(), 512*sizeof(ACE_TCHAR));
	theClientHandler->peer().send_n(clientgroup.c_str(), 512*sizeof(ACE_TCHAR));
	
	//registering 
	if (reactor->register_handler (theClientHandler, ACE_Event_Handler::READ_MASK) == -1) {
		ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Thread_Client::run_svc ] Error register_handler\n")));
		ACE_Thread::exit();
	}
	
	//And starting the event loop
	if (reactor->run_reactor_event_loop() == -1) {
    	ACE_DEBUG((ERROR_PREFIX ACE_TEXT("[ Thread_Client::run_svc ] Error run_reactor_event_loop()\n")));
		ACE_Thread::exit();
	}
	//if we are here, this is because the event loop has been stopped
	//it could have for different reasons:
	//	- the server broke the connection (so the Client_Handler stopped the reactor)
	//	- the disconnect method has been invoked. 
	//In both case we need to gracefully shutdown everything
	
	//if we are here we don't need the emission socket anymore
	//depending where we came from (disconnect method or client handler stopped reactor),
	//the socket is not closed yet. So we taking care of that to avoid error invoking send method
	//from an other thread.
	close_input_stream();
	
	theThreadStopper = NULL;
	theClientHandler = NULL;
	theLogHandler = NULL;
	
	//We don't need the reactor, we delete it. 
	//This call will invoke the handle_close() callback of every handler registered to the reactor.
	//Each of this handler is deleting itself within their handle_close() method. 
	//We do things this way because the WFMO Reactor (Windows implementation) behaves
	//differently from the the Select Reactor (Linux, Mac).
	//We can't manually remove the handler manually and then delete the handlers before destroying the reactor on Windows
	//because the handler removal is not registered right away by the reactor.
	//The reactor then still believes it has handlers that are already destroyed... => segfault
	if ( reactor ) 
		delete reactor;
	reactor = NULL;

	//This Event_Handler doesn't have any handle, nor a handle_close() callback.
	//So there is no 'delete this' on the object in handle_close(). 
	//As a result, the reactor doesn't destroy automatically this handler. 
	//We do it manually.
	if ( theThreadStopper )
		delete theThreadStopper;

	//All the handlers have been destroyed now...
	theThreadStopper = NULL;
	theClientHandler = NULL;
	theLogHandler = NULL;
	
	//In order to be consistent, we need to stop all the flows when we are disconnected
	Smartflow *tmp_sf =  cm->getSmartflow();
	if ( tmp_sf ) {
			FlowManager *tmp_fm = tmp_sf->getFlowManager();
			if ( tmp_fm ) {
				tmp_fm->stopAllFlows();
			}
	}

	//We test if we need to invoke the exit method, this happens in two cases:
	//	- the server broke the connection, i.e. it is gone
	//  - the server instructs the client to quit.
	//Note: this test will return true in both cases above even if the user
	//has not declared an exit method. 
	//If so, we exit
	if ( needToInvokeExitMethod() ) {
		
		//Now that we are properly disconnected we can invoke user exit method because the connectio has been lost
		//otherwise if we are in this method, it is at user request, therefore it is not necessary to invoke the exit
		//method specified by the user
		if ( tmp_sf != NULL ) {
			if (tmp_sf->user_exit_function != NULL) {
				tmp_sf->user_exit_function();
			}
			else if (tmp_sf->user_exit_function_with_parameters != NULL && tmp_sf->user_exit_func_argument != NULL) {
				tmp_sf->user_exit_function_with_parameters(tmp_sf->user_exit_func_argument);
			}
			else {
				exit(0);
			}
		}
	}
	
	//put back the boolean as it is set by the constructor
	setNeedToInvokeExitMethod(false);
	return 0;   
}

void Thread_Client::stop_thread() {
	if ( theThreadStopper && reactor ) {
		reactor->notify(theThreadStopper, ACE_Event_Handler::READ_MASK);
	}
}
	
bool Thread_Client::needToInvokeExitMethod() {
	return invoke;
}

void Thread_Client::setNeedToInvokeExitMethod(bool invoke) {
	this->invoke = invoke;
}

void Thread_Client::close_input_stream() {
	cm->close_socket();
}

	
}//closing namespace
