#include "FlowGen.h"

#include <QtCore/QFile>
#include <QtCore/QRegExp>
#include <QtGui/QFileDialog>
#include <QtGui/QMessageBox>
#include <QtGui/QAction>
#include <QtGui/QValidator>
#include <QtGui/QLineEdit>
#include <QtCore/QStringList>

#include <iostream>
using namespace std;

FlowGen::FlowGen( QWidget* parent, const char* name, Qt::WFlags fl) : flowgenBase(parent, name, fl) {
    classNameOK = descOK = bufferSizeOK = historySizeOK = needSave = isGenerated =  false;
	hasDir = false;
    model = 0;
    flowFileName = "";
    
    QRegExp rxBufferSize( "^[1-9]\\d*$" );
    QValidator* validatorBufferSize = new QRegExpValidator( rxBufferSize, this );
    bufferSize->setValidator( validatorBufferSize );
    
    QRegExp rxHistorySize( "^[1-9]\\d*$" );
    QValidator* validatorHistorySize = new QRegExpValidator( rxHistorySize, this );
    historySize->setValidator( validatorHistorySize );
    
    QRegExp rxFlowDesc( "^(\\w|\\s)+$" );
    QValidator* validatorFlowDesc = new QRegExpValidator( rxFlowDesc, this );
    flowDesc->setValidator( validatorFlowDesc );
    
    QRegExp rxFlowClassName( "^Flow_\\w+$" );
    QValidator* validatorFlowClassName = new QRegExpValidator( rxFlowClassName, this );
    flowClassName->setValidator( validatorFlowClassName );
    
    fileNew();
}

FlowGen::~FlowGen() {}

void FlowGen::loadFile(QString &s) {
    model->readFlowFile(s);
    initAllFields();
    flowFileName = s;
    needSave = false;
    isGenerated = false;
}

void FlowGen::initAll() {
	needSave = false;
    fileNew();
}

void FlowGen::fileNew() {
    if ( needSave ) {
		if ( flowFileName.isEmpty() ) {
			int ret = QMessageBox::warning(this, tr("Attention"), 
										   tr("The Flow Description File is not saved.\n"
											  "Do you want to save it?"),	
										   QMessageBox::Yes | QMessageBox::Default,
										   QMessageBox::No,
										   QMessageBox::Cancel | QMessageBox::Escape);
			if ( ret == QMessageBox::Yes )
				fileSave();
			else if ( ret == QMessageBox::Cancel )
				return;
		}	
		else {	
			int ret = QMessageBox::warning(this, tr("Attention"), 
										   tr("The Flow Description File %1 is not saved.\n"
											  "Do you want to save it?")
										   .arg(QDir::convertSeparators(flowFileName)),
										   QMessageBox::Yes | QMessageBox::Default,
										   QMessageBox::No, 
										   QMessageBox::Cancel | QMessageBox::Escape);
			if (ret == QMessageBox::Yes)
				fileSave();
			else if (ret == QMessageBox::Cancel )
				return;
		}
    }
    flowFileName = "";
    if ( model )
		delete model;
    model =  new FlowTypeModel();
    clearAllFields();
}

void FlowGen::fileOpen(){
	QFileDialog dialog(this, "Open Flow Description File", "", "Flow files (*.flow);;All (*)");
    if ( hasDir )
		dialog.setDirectory( currentDir );
    QString filename;
    if ( dialog.exec() == QDialog::Accepted ) {
		filename = dialog.selectedFile();
		currentDir = dialog.directory();
		hasDir = true;
		if ( !filename.isEmpty() ) {
			if ( !model->readFlowFile(filename))
				QMessageBox:: critical(this, tr("Attention"), 
									   tr("Error while reading the flow description file."),	
									   tr("Ok"));
			else {
				flowFileName = filename;
				QString tmp_basename = currentDir.absFilePath(model->getClassName());
				model->setBaseName(tmp_basename);
				initAllFields();
				needSave = false;
			}
		}
    }
}

void FlowGen::fileSave(){
    if ( flowFileName.isEmpty() )
		fileSaveAs();
    else {	
		QString tmp_basename = currentDir.absFilePath(model->getClassName());
		model->setBaseName(tmp_basename);
		if ( !model->writeFlowFile(flowFileName) ) {
			QMessageBox:: critical(this, tr("Attention"), 
								   tr("Error while saving the flow description file."),	
								   tr("Ok"));
		}
		else
			needSave = false;
	}	
}	

void FlowGen::fileSaveAs(){
	QFileDialog dialog(this, "Save Flow Description File", "", "Flow files (*.flow);;All (*)");
    if ( hasDir )
		dialog.setDirectory ( currentDir );
	dialog.setFileMode( QFileDialog::AnyFile );
	dialog.setAcceptMode ( QFileDialog::AcceptSave );
    QString filename;
    if ( dialog.exec() == QDialog::Accepted ) {
		filename = dialog.selectedFile();
		currentDir = dialog.directory();
		hasDir = true;
		if ( !filename.isEmpty() ) {
			QRegExp rxp( "\\.flow", true, false );
			bool nb = filename.contains( rxp );
			if ( nb == false )
				filename = filename.append(".flow");
			if (QFile::exists(filename)) {
				int ret = QMessageBox::warning(this, tr("Attention"), 
											   tr("File %1 already exists.\n"
												  "Do you want to overwrite it?")
											   .arg(QDir::convertSeparators(filename)),
											   QMessageBox::Yes | QMessageBox::Default,
											   QMessageBox::No | QMessageBox::Escape);
				if (ret == QMessageBox::Yes) {
					flowFileName = filename;
					if ( !model->writeFlowFile(flowFileName) ) {
						QMessageBox:: critical(this, tr("Attention"), 
											   tr("Error while saving the flow description file."),	
											   tr("Ok"));
					}
					else
						needSave = false;
				}
			}
			else {
				flowFileName = filename;
				QString tmp_basename = currentDir.absFilePath(model->getClassName());
				model->setBaseName(tmp_basename);
				if ( !model->writeFlowFile(flowFileName) ) {	
					QMessageBox:: critical(this, tr("Attention"), 
										   tr("Error while saving the flow description file."),	
										   tr("Ok"));
				}
				else
					needSave = false;
			}
		}
    }    
}

void FlowGen::fileExit() {
    if ( needSave ) {
		if ( flowFileName.isEmpty() ) {
			int ret = QMessageBox::warning(this, tr("Attention"), 
										   tr("The Flow Description File is not saved.\n"
											  "Do you want to save it?"),	
										   QMessageBox::Yes | QMessageBox::Default,
										   QMessageBox::No,
										   QMessageBox::Cancel | QMessageBox::Escape);
			if ( ret == QMessageBox::Yes )
				fileSave();
			else if ( ret == QMessageBox::Cancel )
				return;
		}	
		else {	
			int ret = QMessageBox::warning(this, tr("Attention"), 
										   tr("The Flow Description File %1 is not saved.\n"
											  "Do you want to save it?")
										   .arg(QDir::convertSeparators(flowFileName)),
										   QMessageBox::Yes | QMessageBox::Default,
										   QMessageBox::No, 
										   QMessageBox::Cancel | QMessageBox::Escape);
			if (ret == QMessageBox::Yes)
				fileSaveAs();
			else if (ret == QMessageBox::Cancel )
				return;
		}
    }
    hide();//exit( 0 );
}

void FlowGen::tools_Build_Flow_Library() {}

void FlowGen::tools_Generate_Code() {
    if ( needSave )
		QMessageBox::warning(this, QMessageBox::tr("Attention"), QMessageBox::tr("You need to save your work before generating the code!") );
    else
		if ( !model->generateCode() )
			QMessageBox::warning(this, QMessageBox::tr("Attention"), QMessageBox::tr("An error occured during the generation") );
	else
		QMessageBox::warning(this, QMessageBox::tr(""), QMessageBox::tr("Flow source code properly generated") );
	
}

void FlowGen::helpAbout() {
    QMessageBox::about (this, QMessageBox::tr("About the Flow Generator"), QMessageBox::tr("NSDFS Flow Generator.\nSmart Space Group.\nInformation Access Division.\nNIST.") );
}

void FlowGen::checkClassName() {
    needSave = true;
    if ( (flowClassName->text()).length() < 5 ) {
		flowClassName->setText("Flow_");
		model->setClassName(flowClassName->text());
		classNameOK = false;
    }
    else {
		model->setClassName(flowClassName->text());
		classNameOK = true;
    }
    enableGeneration();
}

void FlowGen::checkDesc() {
    needSave = true;
    if ( (flowDesc->text()).isEmpty() ) {
		model->setDescription("");
		descOK = false;
    }
    else {
		model->setDescription(flowDesc->text());
		descOK = true;
    }
    enableGeneration();
}

void FlowGen::checkBufferSize() {
    needSave = true;
    if ( (bufferSize->text()).isEmpty() ) {
		model->setBufferSize("");
		bufferSizeOK = false;
    }
    else {
		model->setBufferSize(bufferSize->text());
		bufferSizeOK = true;
    }
    enableGeneration();	
}

void FlowGen::checkHistorySize() {
    needSave = true;
    if ( (historySize->text()).isEmpty() ) {
		model->setHistorySize("");
		historySizeOK = false;
    }
    else {
		model->setHistorySize(historySize->text());
		historySizeOK = true;
    }
    enableGeneration();	
}

void FlowGen::checkSpecificMethods() {
    needSave = true;
    if ( (methodTextEdit->text()).isEmpty() )
		model->setSpecificMethods("");
    else 
		model->setSpecificMethods(methodTextEdit->text());
}

void FlowGen::checkIncludes() {
	needSave = true;
    if ( (includeTextEdit->text()).isEmpty() )
		model->setIncludes("");
    else 
		model->setIncludes(includeTextEdit->text()); 
}

void FlowGen::closeEvent(QCloseEvent *e) {
	Q_UNUSED(e);
    fileExit();
	// e->accept();
}

void FlowGen::enableGeneration() {
    if ( classNameOK && descOK && bufferSizeOK && historySizeOK )
		toolsGenerate_CodeAction->setEnabled(true);
    else
		toolsGenerate_CodeAction->setEnabled(false);
}

void FlowGen::clearAllFields() {
    flowDesc->setText("");
    flowClassName->setText("Flow_");
    bufferSize->setText("");
    historySize->setText("");
    includeTextEdit->setText("");
    methodTextEdit->setText("");
}

void FlowGen::initAllFields() {
    flowClassName->setText(model->getClassName());
    flowDesc->setText(model->getDescription());
    historySize->setText(model->getHistorySize());
    bufferSize->setText(model->getBufferSize());
    includeTextEdit->setText(((model->getIncludes()).join("\n")));
    methodTextEdit->setText( ((model->getSpecificMethods()).join("\n")) );	   
}
