#ifndef __MAPVIEW_H
#define __MAPVIEW_H

#include <QtCore/QObject>
#include <QtCore/QPoint>
#include <QtGui/QDragEnterEvent>
#include <QtGui/QDropEvent>
#include <QtGui/QMenu>
#include <QtGui/QMouseEvent>
#include <QtGui/QColor>
#include <QtGui/QPainter>
#include <QtGui/QToolTip>
#include <QtGui/QKeyEvent>
#include <Qt3Support/Q3ValueList>
#include <Qt3Support/Q3PointArray>
#include <Qt3Support/Q3Canvas>

#include <set>
#include <vector>
#include <map>

class AbstractGraphicObject;
class FlowGraphicOutlet;
class FlowGraphicOutletIn;
class FlowGraphicOutletOut;
class FlowGraphicObject;
class ClientGraphicObject;
class SubFlowGraphicObject;
class MapView;
class MapRoute;
class AddClientTool;

class PathFinding;
class Client;
class ClientApp;
class ClientConf;
class Flow;
class FlowOutlet;
class FlowOutletIn;
class FlowOutletOut;

#define TILE_SX 10
#define TILE_SY 10
#define SIZEMAPX 400
#define SIZEMAPY 200

static const int ON = 0;
static const int ON_S = 1;
static const int ON_LOG = 2;
static const int ON_LOG_S = 3;
static const int IDLE = 4;
static const int IDLE_S = 5;
static const int IDLE_LOG = 6;
static const int IDLE_LOG_S = 7;
static const int START = 8;
static const int START_S = 9;
static const int START_LOG = 10;
static const int START_LOG_S = 11;
static const int DISABLE = 12;
static const int DISABLE_S = 13;
static const int DISABLE_LOG = 14;
static const int DISABLE_LOG_S = 15;
static const int ERR = 16;
static const int ERR_S = 17;
static const int ERR_LOG = 18;
static const int ERR_LOG_S = 19;



class AbstractGraphicObject : public QObject 
{
	Q_OBJECT
public:
	AbstractGraphicObject(class MapView *mv);
	virtual ~AbstractGraphicObject();
	
	virtual void showPopupMenu(const QPoint &);
	
	virtual unsigned int getId() = 0;
	
	virtual void setSelected(bool s);
	bool isSelected() const { return selected; }
	
	MapView* getMapView() { return mapView; }
	
protected:
	bool selected;
	MapView *mapView;
	QMenu popupMenu;
};

/*********************************************************/

static const int FlowGraphicOutletRTTI = 150283;

class FlowGraphicOutlet : public AbstractGraphicObject, public Q3CanvasRectangle {
    Q_OBJECT
public:
	FlowGraphicOutlet(FlowOutlet *fo, MapView *mv);
	virtual ~FlowGraphicOutlet();
	
	void drawInMap();
	void drawFlowInMap();
	void drawOutletInMap();
	void drawFlowAndOutlet();
	void erase();
	void eraseFlow();
	void eraseFlowAndOutlet();
	
	int rtti () const { return FlowGraphicOutletRTTI; }
	
	//new
	Q3PointArray areaPoints () const;
	
	unsigned int getId();
	unsigned int getPrivateId();
	
	//new
	void setPosition(int x, int y);
	int getX() { return (int) x()/TILE_SX; }
	int getY() { return (int) y()/TILE_SY; }
	bool isSelected() const { return AbstractGraphicObject::isSelected(); }
	
	//old
	//void setPosition(int x, int y);
	//int getX() { return _x; }
	//int getY() { return _y; }
	
	void refresh();
	virtual void setSelected(bool s);
	
	FlowOutlet *getModel() { return flowOutlet; }
	
	virtual void showPopupMenu(const QPoint &);
	
	bool hasDupLog();
	void switchToLogTab();
	
public slots:
	void infoFlowAction();
	void showLogAction();
	void viewFlowAction();
	
protected:
		
	//new
	virtual void drawShape ( QPainter & p );
	
	int _x, _y;
	//stay here
	FlowOutlet *flowOutlet;
	QMenu popupMenu;
};

class FlowGraphicOutletIn : public FlowGraphicOutlet {
    Q_OBJECT
public:
	FlowGraphicOutletIn(FlowOutlet *f, MapView *mv);
	virtual ~FlowGraphicOutletIn();
};

class FlowGraphicOutletOut : public FlowGraphicOutlet {
    Q_OBJECT
public:
	FlowGraphicOutletOut(FlowOutlet *f, MapView *mv);
	virtual ~FlowGraphicOutletOut();
};

/*********************************************************/

class FlowGraphicObject : public AbstractGraphicObject {
	Q_OBJECT
public:
	FlowGraphicObject(MapView *mv, Flow *f);
	virtual ~FlowGraphicObject();
	
	//stay here
	bool drawPathInMap();
	void drawOutletsInMap();
	void erasePath();
	void erasePathWithoutOutlet();
	
	//new 
	void draw(QPainter &p);
	void show();
	void hide();
	
	void createSubFlow(unsigned int _id);
	bool subFlowExists(unsigned int _id);
	void addSubFlow(unsigned int _id, SubFlowGraphicObject *_subflow);
	void removeSubFlow(unsigned int _id);
	SubFlowGraphicObject *getSubFlow(unsigned int _id);
	unsigned int countSubFlows();
	
	void setSubFlowToBreak(SubFlowGraphicObject *sub);
	void unsetSubFlowToBreak();
	
	void setSubFlowToPause(SubFlowGraphicObject *sub);
	void unsetSubFlowToPause();
	
	//stay here
	virtual void setSelected(bool s);
	
	//stay here
	void refresh();
	
	unsigned int getId();
	
	Flow *getModel() { return flow; }
	
	virtual void showPopupMenu(const QPoint &);
	void emitDeleteFlowSignal(Flow *);
	
public slots:
	void breakLinkAction();
	void infoFlowAction();
	void viewFlowAction();
	void pauseFlowAction();
	void restartFlowAction();
	
signals:
	void deleteTheFlow(Flow *);
	
private:  
	Flow *flow;
	QMenu popupMenu;
	SubFlowGraphicObject *subFlowToBreak;
	SubFlowGraphicObject *subFlowToPause;
	//new
	std::map<unsigned int, SubFlowGraphicObject *> subflowList;
};

/**********************************************************/
static const int SubFlowRTTI = 150282;

class SubFlowGraphicObject : public AbstractGraphicObject, public Q3CanvasPolygon {
	
public:
	SubFlowGraphicObject (MapView *_mapView, FlowGraphicObject *_parent );
	virtual ~SubFlowGraphicObject();
	Q3PointArray areaPoints () const;
	void setPointList(Q3ValueList<QPoint> ptlst);
	
	int rtti () const { return SubFlowRTTI; }
	
	bool needToShow();
	
	//new
	unsigned int getId();
	
	unsigned int getSubId();
	void setSubId(unsigned int _subid);
	
	void drawSubFlow1(QPainter &p);
	void drawSubFlow2(QPainter &p, bool isPaused = false);
	
	void refresh();
	void hide();
	
protected:
	void drawShape ( QPainter & p );
	
private:
		
	static const int right = 1;
	static const int left = 2;
	static const int up = 3;
	static const int down = 4;
	
	int subid;
	
	void clearPointList();
	int direction (QPoint pt1, QPoint pt2) const;
	
	Q3ValueList<QPoint> pointList;
	FlowGraphicObject *parent;
	
};

/**********************************************************/
static const int ClientRTTI = 150281;

class ClientGraphicObject : public AbstractGraphicObject, public Q3CanvasPolygonalItem {
	Q_OBJECT
public:
	ClientGraphicObject(MapView *mv, ClientApp *ca);
	
	virtual ~ClientGraphicObject();
	
	//stay here
	virtual void setSelected(bool s);
	bool isSelected() const { return AbstractGraphicObject::isSelected(); }
	virtual QPoint pos();
	virtual void moveRequest(int _x, int _y);
	void move();
	void setLabel(const QString &l);
	
	//stay here
	int rtti () const { return ClientRTTI; }
	
	//stay here
	std::set<unsigned int> drawInMap();
	void setSize(int, int);
	int width() { return _width; }
	int height() { return _height; }
	QPoint getSize();
	void setPos(int, int);
	Q3PointArray areaPoints () const;
	void drawClient();
	void eraseInMap();
	void drawClientWithOutlets();
	void drawClientFlows();
	
	//stay here
	void hide();
	void show();
	bool isShown() { return shown; }
	virtual void draw ( QPainter & painter );
	
	//stay here
	virtual void updateClient();
	void refresh();
	
	//stay here
	unsigned int getId();
	
	ClientApp *getModel() const { return clientApp; }
	
	void setLabelColor(QColor c);
	void refreshLabel();
	
	virtual void showPopupMenu(const QPoint &);
	
	bool isOkayToEdit();
	bool isOkayToRemove();
	void emitDeleteClientSignal(ClientApp* );
	
	public slots:
	void startClientAction();
	void stopClientAction();
	void editClientAction();
	void enableDisableClientAction();
	void removeClientAction();
	void showClientLogAction();
	void clearClientLogAction();
	void setOutletsPositionAction();
	
signals:
	void deleteTheClient(ClientApp *);
	
protected:
	//stay here
	bool shown;
	unsigned int size_type;
	ClientApp *clientApp;
	
	//stay here
	class Q3CanvasText *label;
	int _width, _height;
	std::set<unsigned int> coli;
	std::set<FlowGraphicObject *> undrawFlows;
	Q3CanvasSprite *sprite;
	
	//stay here
	void resetFlows();
	void resetInputFlows();
	void clearFlows();
	void setImage();
	unsigned int getSizeType();
	void drawShape(QPainter&);
};

/*********************************************************/

class MapView : public Q3CanvasView {
	Q_OBJECT
public:
	~MapView();
	MapView ( QWidget* parent = 0, const char* name = 0, Qt::WFlags f = 0);
	
	//void setTile(int x, int y);
	void addClient(unsigned int id, ClientGraphicObject *obj);
	void remClient(unsigned int id);
	void removeAllClients();
	void addFlow(unsigned int id, FlowGraphicObject *obj);
	void remFlow(unsigned int id);
	void addOutlet(unsigned int id, FlowGraphicOutlet *obj);
	void remOutlet(unsigned int id);
	
	void update();
	bool findObject(QPoint pos, ClientGraphicObject *&co, FlowGraphicObject *&fo, FlowGraphicOutlet *&oo);
	FlowGraphicObject* getFlow(unsigned int id); 
	ClientGraphicObject* getClient(unsigned int id);
	FlowGraphicOutlet* getFlowGraphicOutlet(unsigned int id);
	PathFinding* getPf() { return pf; }
	std::set<int> flowsCol;
	Q3CanvasPixmapArray* getPixmap(int i, unsigned int size);
	void setClientsPosition();
	void removeAllSelected();
	void addSelected(AbstractGraphicObject *obj);
	void removeSelected(AbstractGraphicObject *obj);
	QString& tip( const QPoint & p );
	int getNewId();
	QString &getInfo(QPoint p);
	
	void refreshFlows();
	void refreshClientsWithoutFlows();
	double getZoomLevel() { return zoom; }
	void trashGraphicObject(Q3CanvasItem *item);
	
protected:	
	bool event(QEvent *event);
	void contentsMouseMoveEvent ( QMouseEvent * e );
	void contentsMousePressEvent ( QMouseEvent * e );
	void contentsMouseReleaseEvent ( QMouseEvent * e );
	void contentsMouseDoubleClickEvent ( QMouseEvent * e );
	void dragEnterEvent( QDragEnterEvent *event);
	void dropEvent( QDropEvent *event);
	void keyReleaseEvent(QKeyEvent * e);
	void createClientFromAddClientTool(QPoint &position);
	
	std::map<unsigned int, FlowGraphicObject *> flowObjects;
	std::map<unsigned int, ClientGraphicObject *> clientObjects;
	std::map<unsigned int, FlowGraphicOutlet *> flowOutlets;
	
	virtual void showPopupMenu(const QPoint &p);
	
public slots:
	void startAllClientsAction();
	void stopAllClientsAction();
	void setClientsPositionAction();
	void addClientAction();
	void deleteTheFlow(Flow*);
	void deleteTheClient(ClientApp *);
	void createClient();
	void deleteAddClientTool();
	void zoomIn();
	void zoomOut();
	
private:
		
	struct pos { int x1, x2, y1, y2;
		void set(int _x1, int _x2, int _y1, int _y2) { x1 = _x1; x2 = _x2; y1 = _y1; y2 = _y2;}
		void clear() { x1 =  y1 = -1; x2 = y2 = 0;}
		void test() { if ( x1 < 0 ) x1 = 0;  if ( y1 < 0 ) y1 = 0; }
	}; 
	
	pos position;
	double zoom;
	
	std::set<AbstractGraphicObject *> selected_objects;
	bool flowExists(unsigned int id);
	bool clientExists(unsigned int id);
	bool outletExists(unsigned int id);
	PathFinding* pf;
	bool isMoving;
	bool isBoxing;
	bool connecting;
	ClientGraphicObject *mo;
	FlowGraphicOutlet *flo;
	QPoint moving_start;
	std::set<int> clientsCol;
	int d_x, d_y;
	int widthItem, heightItem;
	Q3CanvasPixmapArray *pixs[60];
	void clear();
	QMenu popupMenu;
	QPoint dropPoint;
	ClientConf *clientDrop;
	AddClientTool *addClientTool;
	int newId;
	bool hasAddClientDialog;
	
	Q3Canvas *trashBin;
};

/*******************************************************************************************/

class MapCanvas : public Q3Canvas {
    Q_OBJECT
public:
	MapCanvas();
	virtual ~MapCanvas();
	
	void setZoneRect(const QRect &r);
	void drawTheLine(const QPoint &pt1, const QPoint &pt2, const QRect &r);
	
protected:
	virtual void drawForeground(QPainter &painter, const QRect &clip);
	
private:
	QRect zone_rect, zone_rect2;
	QPoint p1, p2;
};

/*******************************************************************************************/

typedef std::vector< std::vector<int> > levels;

class MapRoute {
public:
	MapRoute();
	MapRoute(int **m, int s, std::map<unsigned int, unsigned int> &_realId, MapView *map, std::vector<int> &sClient);
	~MapRoute();
	
	void initMatrix();
	bool hasCircuit(int **);
	bool organize();
	void setMatrix();
	int **delCircuit(); 
	void displayLevels();
	unsigned int getMaxLevel();
	const std::vector<QPoint>& getPositions();
	void displayPos();
	void displayMatrix();
	
private:
	int size;
	int **mat;
	std::vector<QPoint> pos;
	std::map<unsigned int, unsigned int> realId;
	MapView *mapView;
	levels lev;
	std::vector<int> sizeClients;
};

#endif
