#ifndef __PATHFINDING_H
#define __PATHFINDING_H

#include "MapView.h"

#include <Qt3Support/Q3Canvas>
#include <Qt3Support/Q3ValueList>
#include <QtCore/QPoint>

#include <vector>
#include <set>
using namespace std;

/*! \class Graph
* \brief Graph class.
*
* This class is used in the process to find a path. Basically, it constructs a tree reprensenting the tiles explored in the research for the better path. Once the goal tile is reached, this class allow to give the list of tiles representing the path. This class should never been used directly. Only the PathFinding class uses it 
*/
class Graph {
	
public:
	
	Graph();
	~Graph();
	
	void clearGraph();
	
	static const char M_NODE_STATUS = 0x30; // 00110000
	
	static const char OPEN = 0x20; //0b00100000
	static const char CLOSE = 0x10; //0b00010000
	
	static const char CLEAR_OC = 0x0F;
	static const char CLEAR_DIR = 0xF0;
	static const char NOTHING = 0x00; //0b00000000
	
	static const char RIGHT = 0x01; //0b00000001
	static const char LEFT = 0x02; //0b00000010
	static const char DOWN = 0x04; //0b00000100
	static const char UP = 0x08; //0b00001000
	
	void setTile( int x, int y, char _val, int _cost, int _hcost );
	
	
	void setCost(int x, int y, int _cost ) { map[x][y].cost = _cost; }
	int getCost(int x, int y ) { return map[x][y].cost; }
	
	void setHcost( int x, int y, int _hcost ) { map[x][y].hcost = _hcost; }
	int getHcost( int x, int y ) { return  map[x][y].hcost; }
	
	
	void setVal( int x, int y, char _val ) { map[x][y].val = _val; }
	char getVal( int x, int y ) { return map[x][y].val; }
	
	bool isOpen( int x, int y ) { return ( ( map[x][y].val & M_NODE_STATUS) == OPEN); }
	bool isClosed( int x, int y ) { return ( ( map[x][y].val & M_NODE_STATUS) == CLOSE); }
	
	bool fromRight( int x, int y ) { return ( ( map[x][y].val & RIGHT) == RIGHT); }
	bool fromLeft( int x, int y ){ return ( ( map[x][y].val & LEFT) == LEFT); }
	bool fromUp( int x, int y ){ return ( ( map[x][y].val & UP) == UP); }
	bool fromDown( int x, int y ){ return ( ( map[x][y].val & DOWN) == DOWN); }
    
	
	void addRight( int x, int y ) { map[x][y].val |= RIGHT; }
	void addLeft( int x, int y ) { map[x][y].val |= LEFT; }
	void addUp( int x, int y ) { map[x][y].val |= UP; }
	void addDown( int x, int y ) { map[x][y].val |= DOWN; }
	
	void setOpen( int x, int y ) { map[x][y].val &= CLEAR_OC; map[x][y].val |= OPEN; }
	void setClose( int x, int y ) { map[x][y].val &= CLEAR_OC; map[x][y].val |= CLOSE; }
	
	void clearDirection( int x,  int y ) { map[x][y].val &= CLEAR_DIR; }
	void setDirection(int x, int y, int px, int py);
	
private:
		
	struct tile{
		char val; char nothing; int cost; int hcost;
		tile(){}
		tile( char _val, int _cost, int _hcost ) { val = _val; cost = _cost; hcost = _hcost; }
	};
	
	tile map[SIZEMAPX][SIZEMAPY];
	
};


/*! \class PathFinding
* \brief PathFinding class.
*
* This class is used to find the best path in a graph, avoiding obstacles. The algorithm used is an adaptation of A*. Paths are only vertical and horizontal. The heuristic used is the length of Manhattan.
*/
class PathFinding 
{
public:
	/*! \fn PathFinding()
	* \brief The default constructor
	*/
	PathFinding();
	
	/*! \fn ~PathFinding()
	* \brief The default destructor
	*/
	~PathFinding();
	
	//void initMap(int val);
	
	/*! \fn void clearMap()
		* \brief Clear the Map. A map is an array with 2 dimensions composed by tiles. Tiles are representated by integers.  
		*/
	void clearMap();
	
	
	/*! \fn void setMapCell(int x, int y, int val)
		* \brief Set a tile with the coordinate "x" and "y" of the map to the value "val". The value is representing a tile, for example a value can represent a vertical flow in the tile.
		* \param x an integer representing the X-coordinate. Must be between 0 and (SIZEMAPX-1)
		* \param y an integer representing the Y-coordinate. Must be between 0 and (SIZEMAPY-1)
		* \param val an integer representing what is in the tile.
		*/
	void setMapCell(int x, int y, int val);
	
	/*! \fn int getMapCell(int x, int y)
		* \brief return the value of the tile located in the map at the coordinate x and y.
		* \param x an integer representing the X-coordinate. Must be between 0 and (SIZEMAPX-1).
		* \param y an integer representing the Y-coordinate. Must be between 0 and (SIZEMAPY-1).
		* \return an integer representing the value of the tile in the map.
		*/
	int getMapCell(int x, int y);
	
	/*! \fn void setMapCellId(int x, int y, unsigned int _id)
		* \brief Set an Id to the tile. Each tile belongs to an object a client or a flow. For example, it allows to know which flow is in the tile if we have several flows.
		* \param x an integer representing the X-coordinate. Must be between 0 and (SIZEMAPX-1)
		* \param y an integer representing the Y-coordinate. Must be between 0 and (SIZEMAPY-1)
		* \param _id an integer representing the id of the object whose belongs the tile.
		*/
	void setMapCellId(int x, int y, unsigned int _id);
	
	/*! \fn void setMapUnderCellId(int x, int y, unsigned int _id)
		* \brief Set an other Id to the tile. A tile can belongs to several objects in some cases. Example: 2 flows are crossing each others.
		* \param x an integer representing the X-coordinate. Must be between 0 and (SIZEMAPX-1)
		* \param y an integer representing the Y-coordinate. Must be between 0 and (SIZEMAPY-1)
		* \param _id an integer representing the id of the other object whose belongs the tile.
		*/ 
	void setMapCellIdUnder(int x, int y, unsigned int _id);
	
	/*! \fn unsigned int getMapCellId(int x, int y)
		* \brief Return the ID of the object (flow or client) whose belongs the tile.
		* \param x an integer representing the X-coordinate. Must be between 0 and (SIZEMAPX-1)
		* \param y an integer representing the Y-coordinate. Must be between 0 and (SIZEMAPY-1)
		* \return the id of the object whose belongs the tile. 0 means the tile doesn't belong to any object. 
		*/
	unsigned int getMapCellId(int x, int y);
	
	/*! \fn unsigned int getMapCellIdUnder(int x, int y)
		* \brief Return the ID of the other object (flow or client) whose belongs the tile.
		* \param x an integer representing the X-coordinate. Must be between 0 and (SIZEMAPX-1)
		* \param y an integer representing the Y-coordinate. Must be between 0 and (SIZEMAPY-1)
		* \return the id of the object whose belongs the tile. 0 means the tile doesn't belong to any object.
		*/
	unsigned int getMapCellIdUnder(int x, int y);
	
	
	/*! \fn void printHeap()
		* \brief Print the heap, this method is used for debugging only.
		*/
	void printHeap();
	
	/*! \fn bool getPath(int x1, int y1, int x2, int y2, unsigned int idFlow)
		* \brief This method looks for the best path between the tile (x1,y1) and the tile (x2,y2) if any. If a path is found, this one is written in the map.
		* \param x1 an integer representing the X-coordinate of the starting tile. Must be between 0 and (SIZEMAPX-1)
		* \param y1 an integer representing the Y-coordinate of the starting tile. Must be between 0 and (SIZEMAPY-1)
		* \param x2 an integer representing the X-coordinate of the ending tile. Must be between 0 and (SIZEMAPX-1)
		* \param y2 an integer representing the Y-coordinate of the ending tile. Must be between 0 and (SIZEMAPY-1)
		* \param idFlow an unsigned int representing the ID of the flow.
		* \return true if a path has been found, false otherwise.
		*/
	Q3ValueList<QPoint> getPath(int x1, int y1, int x2, int y2, unsigned int idFlow);
	
	
	/*! \fn void removeObstacle(unsigned int id)
		* \brief This method removes an obstacle (ie a client) from the map
		* \param id an integer representing the id of the obstacle (ie a client ) we are removing from the map.
		*/
	void removeObstacle(unsigned int id);
	
	/*! \fn void removePath(unsigned int idFlow)
		* \brief This method removes a path (ie a flow) from the map
		* \param idFlow an integer representing the id of the flow we remove from the map
		*/
	void removePath(unsigned int idFlow);
	
	/*! \fn void printIdMap()
		* \brief Display the id maps on stdout. For debugging.
		*/
	void printIdMap();
	
	/*! \fn void printMap()
		* \brief Display the map on stdout. For debugging.
		*/
	void printMap();
	
	/*! \fn void setMapView(MapView *m)
		* \brief Set the MapView. The MapView is an object inherited from a Qt Object to draw the Map. Not necessary for the algorithm.
		* \param m a MapView*
		*/ 
	void setMapView(MapView *m);
	
	/*! \fn set<unsigned int> createObstacle(int x, int y, int width, int height, unsigned int id)
		* \brief Create a client in the map at the position (x,y) with width "width", height "height" and id "id"
		* \param x an integer representing the X-coordinate. Must be between 0 and (SIZEMAPX-1).
		* \param y an integer representing the Y-coordinate. Must be between 0 and (SIZEMAPY-1).
		* \param width an integer representing the width (in tile) of the client.
		* \param height an integer representing the height (in tile) of the client.
		* \param id an unsigned int representing the id of the client.
		*/
	set<unsigned int> createObstacle(int x, int y, int width, int height, unsigned int id);
	
	//this constants represent the different value a tile can have in the map
	static const int NO = 0; //there is nothing in the tile
	static const int NE = 255; //there is client near this tile
	static const int H = 10; //there is an horizontal flow in this tile
	static const int V = 5; //there is a vertical flow in this tile
	static const int VonH = 4; //there is a vertical flow on an horizontal flow in this tile
	static const int HonV = 8; //there is an horizontal flow on a vertical flow in this tile
	static const int C = 1; //there is a client in this tile
	static const int BEF = 64; //the beginning or the end of a flow
	static const int BEF_U = 65; //the beginning or the end of a flow
	static const int BEF_R = 66; //the beginning or the end of a flow
	static const int BEF_L = 67; //the beginning or the end of a flow
	static const int BEF_D = 68; //the beginning or the end of a flow
	
	static const int AngleUL = 9; //there is an angle of 2 pieces of flow in this tile : one from Up, one from Left
	static const int AngleUR = 3; //there is an angle of 2 pieces of flow in this tile : one from Up, one from Right
	static const int AngleDL = 12; //there is an angle of 2 pieces of flow in this tile : one from Down, one from Left
	static const int AngleDR = 6; //there is an angle of 2 pieces of flow in this tile : one from Down, one from Right
	static const int UpT = 11; //the flow splits in 2 branch, it looks like a T, with the vertical line of the T pointing up 
	static const int DownT = 14; //the flow splits in 2 branch, it looks like a T, with the vertical line of the T pointing down 
	static const int LeftT = 13; //the flow splits in 2 branch, it looks like a T, with the vertical line of the T pointing left
	static const int RightT = 7; //the flow splits in 2 branch, it looks like a T, with the vertical line of the T pointing right
	static const int Cross = 15; //the flow split in 3 branchs, it looks like a cross
	
private:
		
	/*! \fn void clearHeap()
	* \brief Clear the heap. The heap is used each time a path is calculated.  
	*/
	void clearHeap();
	
	/*! \fn bool exist(int x, int y)
		* \brief Return wether the coordinate (x,y) are in the map or not.
		* \param x an integer representing the X-coordinate.
		* \param y an integer representing the Y-coordinate.
		* \return a boolean representing wether the coordinate are in the map.
		*/
	bool exist(int x, int y);
	
	/*! \fn int getMin(unsigned int num1, unsigned int num2)
		* \brief Return the minimum heuristic value between the leaves num1 and num2 of the heap
		* \param num1 an unsigned int giving the number of a leaf in the heap
		* \param num2 an unsigned int giving the number of a leaf in the heap
		* \return an integer, num1 or num2 or -1 if num1 and num2 or not in the heap
		*/
	int getMin(unsigned int num1, unsigned int num2);
	
	/*! \fn bool graphIsEmpty()
		* \brief Return wether the graph is empty or not.
		* \return a boolean, true if the graph is empty, false otherwise
		*/
	bool graphIsEmpty();
	
	/*! \fn void insertLeaf(int _x, int _y, int _hcost)
		* \brief Insert a leaf in the heap. The leaf represents the heuristic cost of the tile (x,y)
		* \param _x an integer representing the X-coordinate.
		* \param _y an integer representing the Y-coordinate.
		* \param _hcost an integer representing the estimated cost to go to this tile from the starting tile + the estimated cost to go from this tile to the ending tile.
		*/
	void insertLeaf(int _x, int _y, int _hcost);
	
	/*! \fn void removeMinLeaf()
		* \brief Remove the leaf with the minimum estimated cost from the heap.
		*/
	void removeMinLeaf();
	
	/*! \fn void printPath(int startx, int starty, int endx, int endy, unsigned int idFlow)
		* \brief This method is used to do the reverse parsing of the graph once the path has been found. It called internally the method updateMap() in order to write the info in the map. This method is called internally.
		* \param startx an integer representing the X-coordinate of the starting tile. Must be between 0 and (SIZEMAPX-1)
		* \param starty an integer representing the Y-coordinate of the starting tile. Must be between 0 and (SIZEMAPY-1)
		* \param endx an integer representing the X-coordinate of the ending tile. Must be between 0 and (SIZEMAPX-1)
		* \param endy an integer representing the Y-coordinate of the ending tile. Must be between 0 and (SIZEMAPY-1)
		* \param idFlow an unsigned int representing the ID of the flow.
		*/
	Q3ValueList<QPoint> printPath(int startx, int starty, int endx, int endy, unsigned int idFlow);
	
	/*! \fn void updateMap(int x, int y, int dir, int startx, int starty, int endx, int endy, unsigned int id)
		* \brief This method is used to write a new tile in the map regarding the current state of the tile. It is called internally by the method printPath(). This method is called internally.
		* \param x an integer representing the X-coordinate of the current tile. Must be between 0 and (SIZEMAPX-1)
		* \param y an integer representing the Y-coordinate of the current tile. Must be between 0 and (SIZEMAPY-1)
		* \param dir an integer representing the current direction of the path.
		* \param startx an integer representing the X-coordinate of the starting tile. Must be between 0 and (SIZEMAPX-1)
		* \param starty an integer representing the Y-coordinate of the starting tile. Must be between 0 and (SIZEMAPY-1)
		* \param endx an integer representing the X-coordinate of the ending tile. Must be between 0 and (SIZEMAPX-1)
		* \param endy an integer representing the Y-coordinate of the ending tile. Must be between 0 and (SIZEMAPY-1)
		* \param idFlow an unsigned int representing the ID of the flow.
		*/
	void updateMap(int x, int y, int dir, int startx, int starty, int endx, int endy, unsigned int id, Q3ValueList<QPoint> &pt_list);
	
	// an integer giving the previous direction for the reverse parsing. It can have 4 values corresponding to right, left, up or down 
	int previous;
	
	struct leaf {
		int x; int y;
		int hcost;
		leaf(int _x, int _y, int _hcost) { x = _x; y = _y; hcost = _hcost; }
	};
	
	//a Qt object. Basically this is a drawing area
	//QCanvas *can;
	//An object allowing us to display and manage the QCanvas*
	MapView *mv;
	
	//the map of tile. Each tile has a value representing the graph of the client and the flow. This map is mapped with the QCanvas*
	int map[SIZEMAPX][SIZEMAPY];
	//the map of IDS.
	unsigned int idMap[2][SIZEMAPX][SIZEMAPY];
	
	//the graph used to calculate the path.
	Graph *graph;
	
	//the heap used to calculate the path
	vector<leaf> heap;
	
	/*! \fn leaf getMinLeaf()
		* \brief This method return the leaf with the mininum cost.
		* \return the leaf of the heap with the minimum heuristic cost
		*/
	leaf getMinLeaf();
	
	// use in the process in order to put angle of flow between 1/3 and 2/3 of the distance of the flow.
	int startXGlobal, startYGlobal;
	
	/*! \fn int getHeuristicDistance(int startx, int starty, int endx, int endy)
		* \brief Return an integer representing the estimated lenght between the starting tile (startx, starty) and the ending tile (endx, endy). The heuristic used is the lenght of manhattan.
		* \param startx an integer representing the X-coordinate of the starting tile. Must be between 0 and (SIZEMAPX-1)
		* \param starty an integer representing the Y-coordinate of the starting tile. Must be between 0 and (SIZEMAPY-1)
		* \param endx an integer representing the X-coordinateof the ending tile. Must be between 0 and (SIZEMAPX-1)
		* \param endy an integer representing the Y-coordinateof the ending tile. Must be between 0 and (SIZEMAPY-1)
		* \return an integer representing the estimated distance between the two tiles.
		*/
	int getHeuristicDistance(int startx, int starty, int endx, int endy); 
	
	/*! \fn void explore(int x, int y, int px, int py, int end_x, int end_y, unsigned int idFlow)
		* \brief This method explores new tiles, to check if it is possible to uses them is the best path.
		* \param x an integer representing the X-coordinate of the current explorated tile. Must be between 0 and (SIZEMAPX-1)
		* \param y an integer representing the Y-coordinate of the current explorated tile. Must be between 0 and (SIZEMAPY-1)
		* \param px an integer representing the X-coordinate of the tile form where the exploration was initiated. Must be between 0 and (SIZEMAPX-1)
		* \param py an integer representing the Y-coordinate of the tile form where the exploration was initiated. Must be between 0 and (SIZEMAPY-1)
		* \param end_x an integer representing the X-coordinateof the ending tile. Must be between 0 and (SIZEMAPX-1)
		* \param end_y an integer representing the Y-coordinateof the ending tile. Must be between 0 and (SIZEMAPY-1)
		* \param idFlow an unsigned int representing the ID of the flow.
		*/
	void explore(int x, int y, int px, int py, int end_x, int end_y, unsigned int idFlow);
	
};

typedef struct {
	int x; int y;
} point;


#endif
