#include "SfControl.h"
#include "CheckSmartflowModifications.h"
#include "SfControlThread.h"
#include "mainwindow.h"
#include "GlobalCC.h"
#include "Client.h"
#include "Flow.h"

#include <QtCore/QString>

#include <string>
#include <iostream>

using namespace std;

SfControl::SfControl(GlobalCC *_global, SfControlThread *_sfct) {//: QObject() {
    global = _global;
	sfct = _sfct;
	connectedToServer = false;
    sf = new smartflow::Smartflow();
    if ( sf ) {
		sf->setUserExitFunction(disconnectFromServerHandler, this);
		mutex = new QMutex();
		smartflowModificationsChecker = new CheckSmartflowModifications(sf, mutex, global, this);
		int tmp_argc = 0;
		char **tmp_argv = NULL;
		sf->init(tmp_argc, tmp_argv);
    }	
    else {
		std::cout << "Error creation smartflow object" << std::endl;
		abort();
    }
	
}

bool SfControl::connectToServer() {
	//cout << "SfControl::connectToServer() : TRY LOCK" << endl;
	mutex->lock();
	//cout << "[SfControl::connectToServer()] trying to connect" << endl;
	//cout << "SfControl::connectToServer() : LOCK" << endl;
	bool res = sf->connectToApplicationServer("NDFSCC", "admin");
	//cout << "SfControl : TRY UNLOCK" << endl;
	mutex->unlock();
	//cout << "SfControl::connectToServer() : UNLOCK" << endl;
	if ( !res ) {
		connectedToServer = false;
		//return false;
		//EMIT SIGNAL WITH RETURN VALUE
		MainWindow::instance()->emitConnectToServerSignal();
	}
	else {
		connectedToServer = true;
		//we set the exit method which should be called if the server send a sigint signal
		//sf->set_user_exit_function(disconnectFromServerHandler, this);
		launchSmartflowModificationChecker();
		//return true;
		//EMIT SIGNAL WITH RETURN VALUE
		MainWindow::instance()->emitConnectToServerSignal();
	}
	return res;
}

//this is a callback methods that is invoked when the user press CTRL-C
//this method actively disconnect from the server (at user request)
void SfControl::disconnectFromServer() {
	if ( sf ) {
		if ( connectedToServer ) {
			connectedToServer = false;
			sf->disconnectFromApplicationServer();
			MainWindow::instance()->emitDisconnectFromServerSignal();
		}
	}
}

//this is the handler given to the Smartflow oobject so this method is called when the server quits instead of killing the process (default behavior)
void SfControl::disconnectFromServerHandler(void *arg) {
	//There is no point to disconnect from the server, because we already are: the server just quits.
	SfControl *sfc_tmp = NULL;
	sfc_tmp = (SfControl*)(arg);
	if ( sfc_tmp != NULL ) {
		sfc_tmp->emitDisconnectFromServerSignal();
	} 
}


SfControl::~SfControl() {
	if ( smartflowModificationsChecker ) {
		stopSmartflowModificationChecker();
		delete smartflowModificationsChecker;
	}
	if ( !sf->disconnectFromApplicationServer() )
		std::cout << "Disconnection error: " << sf->strError() <<std::endl;
	if ( sf )
		delete sf;
	if ( mutex )
		delete mutex;
}

void SfControl::getServerList() {
	//cout << "SfControl : getServerList()" << endl;
	string serverList_tmp;
	//if ( !connectedToServer ) 
	//connectToServer();
	if ( connectedToServer ) {
		//cout << "SfControl::getServerList() : TRY LOCK" << endl;
		mutex->lock();
		//cout << "SfControl::getServerList() : LOCK" << endl;
		bool res = sf->getServerList(serverList_tmp);
		//cout << "SfControl : TRY UNLOCK" << endl;
		mutex->unlock();
		//cout << "SfControl::getServerList() : UNLOCK" << endl;
		if ( res ) {
			QString serverList(QString::fromStdString(serverList_tmp));
			//cout << "ServerList\n" << serverList.toStdString() << endl;
			MainWindow::instance()->emitGetServerListSignal(res, serverList);
		}
		//else
		//	std::cout << "SfControl::getServerList() : " << sf->strerror() <<std::endl;
	}
	//else
	//	std::cout << "SfControl::getServerList() : Not connected to server" << endl;
}

void SfControl::startClient(unsigned int id) {
	ApplicationCC *app = global->getApplication();
	if ( app ) {
		ClientApp *client = app->getClient(id);
		startClient(client);
	}  
}

void SfControl::startClient(ClientApp *client) {
	if ( isConnectedToServer() ) {
		if ( client && !client->isStarting() && !client->isRunning()) {
			if ( client->getServer()->getAddress() != "No Server" ) {
				//client->addLogLine("Client starting...");
				//cout << "Starting:" << client->getName() << " ConnectionId: " << client->getGroup() << endl;
				//here we use the smartflow object to launch the client
				QString serverName(client->getServer()->getAddress());
				QString clientPath(client->getPath());
				QString clientParam(client->getParam());
				QString clientArguments = "";
				QString flowInfo = " --sf flowinfo=";
				
				clientArguments.append("--sf \"clientname=");
				clientArguments.append(client->getName());
				clientArguments.append("\"");
				clientArguments.append(" --sf clientgroup=");
				clientArguments.append(client->getGroup());
				clientArguments.append(" --sf clientpath=");
				clientArguments.append(client->getPath());
				clientArguments.append(" ");

				std::map<unsigned int, FlowOutletIn*>* inputFlowList =  client->getInputFlowList();
				std::map<unsigned int, FlowOutletIn*>::iterator iter_flowIn;
				for ( iter_flowIn = inputFlowList->begin(); iter_flowIn != inputFlowList->end(); iter_flowIn++ ) {
					FlowOutletIn *fin = dynamic_cast<FlowOutletIn *>(iter_flowIn->second);
					if ( fin ) {
						clientArguments.append(" --sf flow-");
						clientArguments.append(fin->getName());
						clientArguments.append("=");
						clientArguments.append(fin->getGroup());
						
						if ( flowInfo.length() > 15 )
							flowInfo.append("%%");
						flowInfo.append("I");
						flowInfo.append("%");
						flowInfo.append(fin->getName());
						flowInfo.append("%");
						flowInfo.append(fin->getDataType());
					}
				}
				std::map<unsigned int, FlowOutletOut*>* outputFlowList =  client->getOutputFlowList();
				std::map<unsigned int, FlowOutletOut*>::iterator iter_flowOut;
				for ( iter_flowOut = outputFlowList->begin(); iter_flowOut != outputFlowList->end(); iter_flowOut++ ) {
					FlowOutletOut *fout = dynamic_cast<FlowOutletOut *>(iter_flowOut->second);
					if ( fout ) {
						clientArguments.append(" --sf flow-");
						clientArguments.append(fout->getName());
						clientArguments.append("=");
						clientArguments.append(fout->getGroup());
						
						if ( flowInfo.length() > 15 )
							flowInfo.append("%%");
						flowInfo.append("O");
						flowInfo.append("%");
						flowInfo.append(fout->getName());
						flowInfo.append("%");
						flowInfo.append(fout->getDataType());
					}
				}
				//cout << "Before start" << clientPath << endl;
				//cout << "SfControl : TRY LOCK" << endl;
				mutex->lock();
				//cout << "SfControl : LOCK" << endl;
				clientArguments.append(flowInfo);
				//cout << clientArguments << endl;
				//cout << "Client Path= " << clientPath.ascii()  << " Servername= " << serverName.ascii()  << " args= " <<  clientArguments.ascii() <<" parameters= " <<  clientParam.ascii() << endl;
				bool res = sf->startClient(clientPath.ascii(), serverName.ascii(), clientArguments.ascii(), clientParam.ascii());
				QString mess;
				if ( !res ) {
					cout << "there is an error: " << sf->strError() << endl;
					mess.append("Error while starting the client: ");
					mess.append(sf->strError().c_str());
				}
				//EMIT SIGNAL
				MainWindow::instance()->emitStartClientSignal(res, mess, client);
			
				//cout << "After start" << clientPath << endl;
				//cout << "SfControl : TRY UNLOCK" << endl;
				mutex->unlock();
				//cout << "SfControl : UNLOCK" << endl;
			}
		}
	}
}

void SfControl::stopClient(unsigned int id) {
	ApplicationCC *app = global->getApplication();
	if ( app ) {
		ClientApp *client = app->getClient(id);
		stopClient(client);
	}  
}

void SfControl::stopClient(ClientApp* client) {
	if ( isConnectedToServer() ) {
		if ( client && ( client->isRunning() || client->isStarting() )) {
			//client->addLogLine("Client stopping...");
			//cout << "Stopping:" << client->getName() << " ConnectionId: " << client->getGroup() << endl;
			//here we use the smartflow object to launch the client
			QString clientName(client->getName());
			QString clientGroup(client->getGroup());
			//cout << "SfControl : TRY LOCK" << endl;
			mutex->lock();
			//cout << "SfControl : LOCK" << endl;
			bool res = sf->stopClient(clientName.ascii(), clientGroup.ascii());
			//EMIT SIGNAL
			MainWindow::instance()->emitStopClientSignal(res, client);
			
			//cout << "SfControl : TRY UNLOCK" << endl;
			mutex->unlock();
			//cout << "SfControl : UNLOCK" << endl;
		}
	}
}

void SfControl::startAllClients() {
	//cout << "SfControl::startAllClients()" << endl;
	if ( isConnectedToServer() ) {
		ApplicationCC *app = global->getApplication();
		if ( app ) {
			std::map<unsigned int, ClientApp *> *clientList = app->getClientList();
			std::map<unsigned int, ClientApp *>::iterator iter;
			for ( iter = clientList->begin(); iter != clientList->end(); iter++ ) {
				ClientApp *client = (ClientApp *) (iter->second);
				if ( client ) 
					if ( !client->isStarting() && !client->isRunning()  && !client->isDisable() ) {
						sfct->doSleep();
						startClient(client);
					}
			}
			//cout << "Application is started" << endl;
			//applicationStarted = true;
		}
	}
}

void SfControl::stopAllClients() {
	//cout << "SfControl::stopAllClients()" << endl;
	if ( isConnectedToServer() ) {
		ApplicationCC *app = global->getApplication();
		if ( app ) {
			std::map<unsigned int, ClientApp *> *clientList = app->getClientList();
			std::map<unsigned int, ClientApp *>::iterator iter;
			for ( iter = clientList->begin(); iter != clientList->end(); iter++ ) {
				ClientApp *client = (ClientApp *) (iter->second);
				if ( client ) 
					if ( client->isRunning() || client->isStarting() ) {
						sfct->doSleep();
						stopClient(client);
					}
			}
			//applicationStarted = false;
		}
	}
}

void SfControl::discoverApplication() {
	if ( isConnectedToServer() ) {
		ApplicationCC *app = global->getApplication();
		if ( app ) {
			string appliDescription_tmp;
			mutex->lock();
			bool res = sf->getApplicationDescription(appliDescription_tmp);
			mutex->unlock();
			if ( res ) {
				QString appliDescription(QString::fromStdString(appliDescription_tmp));
				MainWindow::instance()->emitDiscoverApplicationSignal(appliDescription);
				//EMIT SIGNAL
				//if ( !global->getApplication()->readApplicationDescription(appliDescription) ) {
				//	cout << "Error During the discovery process" << endl;
				//} 
			}
			else {
				cout << "Error During the discovery process" << endl;
			}
		}
	}
}

void SfControl::pauseFlow(Flow *f, ClientApp *cl) {
	if ( f != NULL && cl != NULL ) {
		if ( isConnectedToServer() ) {
			ApplicationCC *app = global->getApplication();
			if ( app ) {
				QString flowGroup = f->getOutletOut()->getGroup();
				QString flowType = f->getDataType();
				QString clientGroup = cl->getGroup();
				QString clientName = cl->getName();
				//cout << "PAUSE ClientName=" << clientName << " ClientGroup=" << clientGroup << " FlowType=" << flowType << " FlowGroup=" << flowGroup << endl;
				sf->pauseFlow(clientName.ascii(), clientGroup.ascii(), flowType.ascii(), flowGroup.ascii());
			}
		}
	}
}

void SfControl::restartFlow(Flow *f, ClientApp *cl) {
	if ( f != NULL && cl != NULL ) {
		if ( isConnectedToServer() ) {
			ApplicationCC *app = global->getApplication();
			if ( app ) {
				QString flowGroup = f->getOutletOut()->getGroup();
				QString flowType = f->getDataType();
				QString clientGroup = cl->getGroup();
				QString clientName = cl->getName();
				//cout << "RESTART ClientName=" << clientName << " ClientGroup=" << clientGroup << " FlowType=" << flowType << " FlowGroup=" << flowGroup << endl;
				sf->restartFlow(clientName.ascii(), clientGroup.ascii(), flowType.ascii(), flowGroup.ascii());
			}
		}
	}
}

void SfControl::setGlobalCC(GlobalCC *_global) {
    if ( _global )
		global = _global;
}

void SfControl::unsetGlobalCC() {
    global = NULL;
}

GlobalCC* SfControl::getGlobalCC() {
    return global;
}

bool SfControl::isConnectedToServer() { return connectedToServer; }

//bool SfControl::isApplicationStarted() { return applicationStarted; }

void SfControl::launchSmartflowModificationChecker() {
	if ( smartflowModificationsChecker ) {
		//cout << "Starting last modif thread" << endl;
		smartflowModificationsChecker->start();
	}
}

void SfControl::stopSmartflowModificationChecker() {
	if ( smartflowModificationsChecker ) {
		smartflowModificationsChecker->stop_work();
		smartflowModificationsChecker->wait();
	}
}

void SfControl::emitConnectToServerSignal() {
	emit connectToServerSignal();
}

void SfControl::emitDisconnectFromServerSignal() {
	emit disconnectFromServerSignal();
}

void SfControl::emitGetServerListSignal() {
	emit getServerListSignal();
}

void SfControl::emitStartClientSignal(unsigned int id) {
	emit startClientSignal(id);
}

void SfControl::emitStartClientSignal(ClientApp *client) {
	emit startClientSignal(client);
}

void SfControl::emitStopClientSignal(unsigned int id) {
	emit stopClientSignal(id);
}

void SfControl::emitStopClientSignal(ClientApp* client) {
	emit stopClientSignal(client);
}

void SfControl::emitStartAllClientsSignal() {
	emit startAllClientsSignal();
}

void SfControl::emitStopAllClientsSignal() {
	emit stopAllClientsSignal();
}

void SfControl::emitDiscoverApplicationSignal() {
	emit discoverApplicationSignal();
}

void SfControl::emitPauseFlowSignal(Flow *f, ClientApp *cl) {
	emit pauseFlowSignal(f, cl);
}

void SfControl::emitRestartFlowSignal(Flow *f, ClientApp *cl) {
	emit restartFlowSignal(f, cl);
}

